/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiAnnouncementEventLogTable.c 5418 2005-11-14 16:45:18Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiAnnouncementEventLogTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiAnnouncementEventLogTable_oid[] = { saHpiAnnouncementEventLogTable_TABLE_OID };
size_t saHpiAnnouncementEventLogTable_oid_len = OID_LENGTH(saHpiAnnouncementEventLogTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long announcement_event_log_entry_count = 0;

static oid saHpiAnnouncementEventLogActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,30 };

int handle_saHpiAnnouncementEventLogActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);

int initialize_table_saHpiAnnouncementEventLogActiveEntries(void);

/**
 * 
 * @sessionid:
 * @rdr_entry:
 * @rpt_entry:
 * @full_oid:
 * @full_oid_len:
 * @child_oid:
 * @child_oid_len:
 * 
 * @return 
 */
 
SaErrorT populate_saHpiAnnouncementEventLogTable(SaHpiSessionIdT sessionid, 
                                         SaHpiRdrT *rdr_entry,
                                         SaHpiRptEntryT *rpt_entry,
                                         oid *full_oid, size_t full_oid_len,
                                         oid *child_oid, size_t *child_oid_len)
{					 
	SaErrorT rv = SA_OK;
        SaErrorT rc = SA_OK;

	oid announcement_oid[ANNOUNCEMENT_EVENT_INDEX_NR];
	netsnmp_index announcement_index;
	saHpiAnnouncementEventLogTable_context *announcement_ctx;
	
	oid column[2];
	int column_len = 2;
	
	SaHpiAnnouncementT announcement;
	
	announcement.EntryId = SAHPI_FIRST_ENTRY;
		
	/* Used for decoding */
	oh_big_textbuffer bigbuf;
	SaHpiTextBufferT buf;
		
        DEBUGMSGTL ((AGENT, "populate_saHpiAnnouncementEventLogTable, called\n"));		
	
	
	if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_saHpiAnnouncementEventLogTable() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_saHpiAnnouncementEventLogTable() passed NULL rpt_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
		
        rv = saHpiAnnunciatorGetNext(sessionid, 
	                         rpt_entry->ResourceId,
				 rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum,
				 SAHPI_ALL_SEVERITIES,
				 SAHPI_FALSE,         //Get all announcements.
				 &announcement);

	if ((rv != SA_OK) && (rv != SA_ERR_HPI_NOT_PRESENT)) {
		DEBUGMSGTL ((AGENT, 
		"populate_saHpiAnnouncementEventLogTable: saHpiAnnunciatorGetNext Failed: rv = %d\n",
		rv));
		return AGENT_ERR_INTERNAL_ERROR;
	}
	else if (rv == SA_ERR_HPI_NOT_PRESENT) {
	        return SA_OK; //first call returned nothing
	}
	
	while ((rv != SA_ERR_HPI_NOT_PRESENT) && (rv == SA_OK)) {
                /* BUILD oid for new row */
                /* assign the number of indices */	
	        announcement_index.len = ANNOUNCEMENT_EVENT_INDEX_NR;
                /** Index saHpiDomainId is external */	
                announcement_oid[0] = announcement.StatusCond.DomainId;
                /** Index saHpiResourceId is external */	
                announcement_oid[1] = announcement.StatusCond.ResourceId;
                /** Index saHpiAnnouncementEntryId is internal */	
	        announcement_oid[2] = announcement.EntryId;
                /* assign the indices to the index */	
	        announcement_index.oids = (oid *) & announcement_oid;
 
                /* See if Row exists. */
                announcement_ctx = NULL;
                announcement_ctx = CONTAINER_FIND(cb.container, &announcement_index);

                if (!announcement_ctx) {
                       // New entry. Add it
                        announcement_ctx = 
                             saHpiAnnouncementEventLogTable_create_row(&announcement_index);
                }
                if (!announcement_ctx) {
                        snmp_log (LOG_ERR, "Not enough memory for an Announcement row!");
                        return AGENT_ERR_INTERNAL_ERROR;
                }
	 
	        /** SaHpiEntryId = ASN_UNSIGNED */
                announcement_ctx->saHpiAnnouncementEventLogEntryId = announcement.EntryId;
	 	
                /** SaHpiTime = ASN_OCTET_STR */
		hpitime_to_snmptime(announcement.Timestamp, announcement_ctx->saHpiAnnouncementEventLogTimestamp);
		announcement_ctx->saHpiAnnouncementEventLogTimestamp_len = sizeof(SaHpiTimeT);
 
                /** TruthValue = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogAddedByUser = announcement.AddedByUser;
 
                /** SaHpiSeverity = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogSeverity = announcement.Severity + 1;
 
                /** TruthValue = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogAcknowledged = announcement.Acknowledged;
 
                /** INTEGER = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogStatusCondType = announcement.StatusCond.Type + 1;

                /** SaHpiEntityPath = ASN_OCTET_STR */
	        rc = oh_decode_entitypath(&announcement.StatusCond.Entity, &bigbuf);
		
	        if (rc != SA_OK) {
		        DEBUGMSGTL ((AGENT, 
		                "populate_saHpiAnnouncementEventLogTable: oh_decode_entitypath Failed: rc = %d\n",
		                 rc));
	        }
	        else {
		        memcpy(announcement_ctx->saHpiAnnouncementEventLogEntityPath,
		                                       bigbuf.Data, bigbuf.DataLength);
	                announcement_ctx->saHpiAnnouncementEventLogEntityPath_len = bigbuf.DataLength;
	        }

               
	
	        if (announcement.StatusCond.Type == SAHPI_STATUS_COND_TYPE_SENSOR) {
		        /** UNSIGNED32 = ASN_UNSIGNED */
               	        announcement_ctx->saHpiAnnouncementEventLogSensorNum = announcement.StatusCond.SensorNum;

	                /** SaHpiEventState = ASN_OCTET_STR */
	                rc = oh_decode_eventstate(announcement.StatusCond.EventState, SAHPI_EC_SENSOR_SPECIFIC, &buf);
			
		        if (rc != SA_OK) {
	                        DEBUGMSGTL ((AGENT, 
  		                "populate_saHpiAnnouncementEventLogTable: oh_decode_eventstate Failed: rc = %d\n",
		                 rc));
		        }	         	
		        else {
		               memcpy(announcement_ctx->saHpiAnnouncementEventLogEventState, buf.Data, buf.DataLength); 
               	               announcement_ctx->saHpiAnnouncementEventLogEventState_len = buf.DataLength;
		        }	
                }
	
                /** OCTETSTR = ASN_OCTET_STR */
                memcpy(announcement_ctx->saHpiAnnouncementEventLogName, announcement.StatusCond.Name.Value, 
	                                                announcement.StatusCond.Name.Length);
                announcement_ctx->saHpiAnnouncementEventLogName_len = announcement.StatusCond.Name.Length;
 
                /** SaHpiManufacturerId = ASN_UNSIGNED */
                announcement_ctx->saHpiAnnouncementEventLogMid = announcement.StatusCond.Mid;
 
                /** SaHpiTextType = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogTextType = announcement.StatusCond.Data.DataType + 1;
 
                /** SaHpiTextLanguage = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogTextLanguage = announcement.StatusCond.Data.Language + 1;
 
                /** SaHpiText = ASN_OCTET_STR */
                memcpy(announcement_ctx->saHpiAnnouncementEventLogText, announcement.StatusCond.Data.Data, 
	                                                 announcement.StatusCond.Data.DataLength);
                announcement_ctx->saHpiAnnouncementEventLogText_len = announcement.StatusCond.Data.DataLength;
 
                /** RowStatus = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementEventLogDelete = SAHPIANNOUNCEMENTEVENTLOGDELETE_ACTIVE;

	        CONTAINER_INSERT (cb.container, announcement_ctx);
			
                /* create full oid on This row for parent RowPointer */
                column[0] = 1;
                column[1] = COLUMN_SAHPIANNOUNCEMENTEVENTLOGTIMESTAMP;
                memset(child_oid, 0, MAX_OID_LEN);
                build_full_oid(saHpiAnnouncementEventLogTable_oid, saHpiAnnouncementEventLogTable_oid_len,
                              column, column_len,
                              &announcement_index,
                              child_oid, MAX_OID_LEN, child_oid_len);
			      
                rv = saHpiAnnunciatorGetNext(sessionid, 
	                         rpt_entry->ResourceId,
				 rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum,
				 SAHPI_ALL_SEVERITIES,
				 SAHPI_FALSE,         //Get all announcements.
				 &announcement);			      	
	
	}
	
	
	announcement_event_log_entry_count = CONTAINER_SIZE (cb.container);
	
		       
        return SA_OK; 		       
}
/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiAnnouncementEventLogActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiAnnouncementEventLogActiveEntries, called\n"));

        announcement_event_log_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			(u_char *) &announcement_event_log_entry_count,
        			sizeof(announcement_event_log_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @return: 
 */
int initialize_table_saHpiAnnouncementEventLogActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiAnnouncementEventLogActiveEntries, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiAnnouncementEventLogActiveEntries", 
					handle_saHpiAnnouncementEventLogActiveEntries,
                                        saHpiAnnouncementEventLogActiveEntries_oid, 
			                OID_LENGTH(saHpiAnnouncementEventLogActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}

 

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiAnnouncementEventLogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiAnnouncementEventLogTable_cmp( const void *lhs, const void *rhs )
{
        saHpiAnnouncementEventLogTable_context *context_l =
        (saHpiAnnouncementEventLogTable_context *)lhs;
        saHpiAnnouncementEventLogTable_context *context_r =
        (saHpiAnnouncementEventLogTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiAnnouncementEventLogTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiAnnouncementEventLogTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiAnnouncementEntryId */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                return 0;
                        }
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiAnnouncementEventLogTable module
 */
void
init_saHpiAnnouncementEventLogTable(void)
{

        DEBUGMSGTL ((AGENT, "init_saHpiAnnouncementEventLogTable, called\n"));        
	
	initialize_table_saHpiAnnouncementEventLogTable();
     
        initialize_table_saHpiAnnouncementEventLogActiveEntries();
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiAnnouncementEventLogTable_row_copy(saHpiAnnouncementEventLogTable_context * dst,
                         saHpiAnnouncementEventLogTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiAnnouncementEventLogEntryId = src->saHpiAnnouncementEventLogEntryId;

    memcpy(dst->saHpiAnnouncementEventLogTimestamp,
           src->saHpiAnnouncementEventLogTimestamp,
	   src->saHpiAnnouncementEventLogTimestamp_len);
	   
    dst->saHpiAnnouncementEventLogTimestamp_len = src->saHpiAnnouncementEventLogTimestamp_len;   

    dst->saHpiAnnouncementEventLogAddedByUser = src->saHpiAnnouncementEventLogAddedByUser;

    dst->saHpiAnnouncementEventLogSeverity = src->saHpiAnnouncementEventLogSeverity;

    dst->saHpiAnnouncementEventLogAcknowledged = src->saHpiAnnouncementEventLogAcknowledged;

    dst->saHpiAnnouncementEventLogStatusCondType = src->saHpiAnnouncementEventLogStatusCondType;

    memcpy( dst->saHpiAnnouncementEventLogEntityPath, src->saHpiAnnouncementEventLogEntityPath, src->saHpiAnnouncementEventLogEntityPath_len );
    dst->saHpiAnnouncementEventLogEntityPath_len = src->saHpiAnnouncementEventLogEntityPath_len;

    dst->saHpiAnnouncementEventLogSensorNum = src->saHpiAnnouncementEventLogSensorNum;

    memcpy( dst->saHpiAnnouncementEventLogEventState, src->saHpiAnnouncementEventLogEventState, src->saHpiAnnouncementEventLogEventState_len );
    dst->saHpiAnnouncementEventLogEventState_len = src->saHpiAnnouncementEventLogEventState_len;

    memcpy( dst->saHpiAnnouncementEventLogName, src->saHpiAnnouncementEventLogName, src->saHpiAnnouncementEventLogName_len );
    dst->saHpiAnnouncementEventLogName_len = src->saHpiAnnouncementEventLogName_len;

    dst->saHpiAnnouncementEventLogMid = src->saHpiAnnouncementEventLogMid;

    dst->saHpiAnnouncementEventLogTextType = src->saHpiAnnouncementEventLogTextType;

    dst->saHpiAnnouncementEventLogTextLanguage = src->saHpiAnnouncementEventLogTextLanguage;

    memcpy( dst->saHpiAnnouncementEventLogText, src->saHpiAnnouncementEventLogText, src->saHpiAnnouncementEventLogText_len );
    dst->saHpiAnnouncementEventLogText_len = src->saHpiAnnouncementEventLogText_len;

    dst->saHpiAnnouncementEventLogDelete = src->saHpiAnnouncementEventLogDelete;

    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiAnnouncementEventLogTable_extract_index( saHpiAnnouncementEventLogTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiAnnouncementEntryId;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementEventLogTable_extract_index, called\n"));

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiAnnouncementEntryId;

        memset( &var_saHpiAnnouncementEntryId, 0x00, sizeof(var_saHpiAnnouncementEntryId) );
        var_saHpiAnnouncementEntryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiAnnouncementEntryId.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiResourceId */
   
                /** skipping external index saHpiAnnouncementEntryId */
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
		err = saHpiAnnouncementEntryId_check_index(*var_saHpiAnnouncementEntryId.val.integer);   

        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiAnnouncementEventLogTable_can_activate(saHpiAnnouncementEventLogTable_context *undo_ctx,
                      saHpiAnnouncementEventLogTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiAnnouncementEventLogTable_can_deactivate(saHpiAnnouncementEventLogTable_context *undo_ctx,
                        saHpiAnnouncementEventLogTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiAnnouncementEventLogTable_can_delete(saHpiAnnouncementEventLogTable_context *undo_ctx,
                    saHpiAnnouncementEventLogTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiAnnouncementEventLogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiAnnouncementEventLogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiAnnouncementEventLogTable_context *
saHpiAnnouncementEventLogTable_create_row( netsnmp_index* hdr)
{
    saHpiAnnouncementEventLogTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiAnnouncementEventLogTable_context);
    if(!ctx)
        return NULL;
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiAnnouncementEventLogTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiAnnouncementEventLogSeverity = 0;
     ctx->saHpiAnnouncementEventLogAcknowledged = 0;
     ctx->saHpiAnnouncementEventLogStatusCondType = 0;
     ctx->saHpiAnnouncementEventLogEntityPath = 0;
     ctx->saHpiAnnouncementEventLogSensorNum = 0;
     ctx->saHpiAnnouncementEventLogEventState = 0;
     ctx->saHpiAnnouncementEventLogName = 0;
     ctx->saHpiAnnouncementEventLogMid = 0;
     ctx->saHpiAnnouncementEventLogTextType = 0;
     ctx->saHpiAnnouncementEventLogTextLanguage = 0;
     ctx->saHpiAnnouncementEventLogText = 0;
     ctx->saHpiAnnouncementEventLogDelete = 0;
    */

    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiAnnouncementEventLogTable_context *
saHpiAnnouncementEventLogTable_duplicate_row( saHpiAnnouncementEventLogTable_context * row_ctx)
{
    saHpiAnnouncementEventLogTable_context * dup;

    if(!row_ctx)
        return NULL;

    dup = SNMP_MALLOC_TYPEDEF(saHpiAnnouncementEventLogTable_context);
    if(!dup)
        return NULL;
        
    if(saHpiAnnouncementEventLogTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiAnnouncementEventLogTable_delete_row( saHpiAnnouncementEventLogTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiAnnouncementEventLogTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiAnnouncementEventLogTable_context *row_ctx =
            (saHpiAnnouncementEventLogTable_context *)rg->existing_row;
//    saHpiAnnouncementEventLogTable_context *undo_ctx =
//            (saHpiAnnouncementEventLogTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogSeverity));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogAcknowledged));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogStatusCondType));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */	
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > sizeof(row_ctx->saHpiAnnouncementEventLogEntityPath)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogSensorNum));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */	    
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > sizeof(row_ctx->saHpiAnnouncementEventLogEventState)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }		    
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > sizeof(row_ctx->saHpiAnnouncementEventLogName)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	    
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogMid));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogTextType));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogTextLanguage));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > sizeof(row_ctx->saHpiAnnouncementEventLogText)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	    
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiAnnouncementEventLogDelete));
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiAnnouncementEventLogTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiAnnouncementEventLogTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiAnnouncementEventLogTable_context *row_ctx = (saHpiAnnouncementEventLogTable_context *)rg->existing_row;
//    saHpiAnnouncementEventLogTable_context *undo_ctx = (saHpiAnnouncementEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiAnnouncementEventLogTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiAnnouncementEventLogTable_context *row_ctx = (saHpiAnnouncementEventLogTable_context *)rg->existing_row;
//    saHpiAnnouncementEventLogTable_context *undo_ctx = (saHpiAnnouncementEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;

    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            row_ctx->saHpiAnnouncementEventLogSeverity = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiAnnouncementEventLogAcknowledged = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiAnnouncementEventLogStatusCondType = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementEventLogEntityPath,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementEventLogEntityPath_len = var->val_len;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            row_ctx->saHpiAnnouncementEventLogSensorNum = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementEventLogEventState,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementEventLogEventState_len = var->val_len;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementEventLogName,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementEventLogName_len = var->val_len;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            row_ctx->saHpiAnnouncementEventLogMid = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            row_ctx->saHpiAnnouncementEventLogTextType = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            row_ctx->saHpiAnnouncementEventLogTextLanguage = *var->val.integer;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementEventLogText,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementEventLogText_len = var->val_len;
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
            row_ctx->saHpiAnnouncementEventLogDelete = *var->val.integer;
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }

    subagent_unlock(&hpi_lock_data);

    return;


}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiAnnouncementEventLogTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiAnnouncementEventLogTable_context *row_ctx = (saHpiAnnouncementEventLogTable_context *)rg->existing_row;
//    saHpiAnnouncementEventLogTable_context *undo_ctx = (saHpiAnnouncementEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiAnnouncementEventLogTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiAnnouncementEventLogTable_context *row_ctx = (saHpiAnnouncementEventLogTable_context *)rg->existing_row;
//    saHpiAnnouncementEventLogTable_context *undo_ctx = (saHpiAnnouncementEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
		break;
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiAnnouncementEventLogTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiAnnouncementEventLogTable_context *row_ctx = (saHpiAnnouncementEventLogTable_context *)rg->existing_row;
//    saHpiAnnouncementEventLogTable_context *undo_ctx = (saHpiAnnouncementEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}


/************************************************************
 *
 * Initialize the saHpiAnnouncementEventLogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiAnnouncementEventLogTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiAnnouncementEventLogTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiAnnouncementEventLogTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiAnnouncementEventLogTable_oid,
                                             saHpiAnnouncementEventLogTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiAnnouncementEventLogTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiAnnouncementEntryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiAnnouncementEventLogTable_COL_MIN;
    table_info->max_column = saHpiAnnouncementEventLogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiAnnouncementEventLogTable_get_value;
    cb.container = netsnmp_container_find("saHpiAnnouncementEventLogTable_primary:"
                                          "saHpiAnnouncementEventLogTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiAnnouncementEventLogTable_secondary:"
                                                       "saHpiAnnouncementEventLogTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiAnnouncementEventLogTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiAnnouncementEventLogTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiAnnouncementEventLogTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiAnnouncementEventLogTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiAnnouncementEventLogTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiAnnouncementEventLogTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiAnnouncementEventLogTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiAnnouncementEventLogTable_can_delete;

    cb.set_reserve1 = saHpiAnnouncementEventLogTable_set_reserve1;
    cb.set_reserve2 = saHpiAnnouncementEventLogTable_set_reserve2;
    cb.set_action = saHpiAnnouncementEventLogTable_set_action;
    cb.set_commit = saHpiAnnouncementEventLogTable_set_commit;
    cb.set_free = saHpiAnnouncementEventLogTable_set_free;
    cb.set_undo = saHpiAnnouncementEventLogTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiAnnouncementEventLogTable",
                "Registering table saHpiAnnouncementEventLogTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiAnnouncementEventLogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiAnnouncementEventLogTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiAnnouncementEventLogTable_context *context = (saHpiAnnouncementEventLogTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTRYID:
            /** SaHpiEntryId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementEventLogEntryId,
                         sizeof(context->saHpiAnnouncementEventLogEntryId) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEventLogTimestamp,
                         sizeof(context->saHpiAnnouncementEventLogTimestamp) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGADDEDBYUSER:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogAddedByUser,
                         sizeof(context->saHpiAnnouncementEventLogAddedByUser) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogSeverity,
                         sizeof(context->saHpiAnnouncementEventLogSeverity) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogAcknowledged,
                         sizeof(context->saHpiAnnouncementEventLogAcknowledged) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogStatusCondType,
                         sizeof(context->saHpiAnnouncementEventLogStatusCondType) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEventLogEntityPath,
                         context->saHpiAnnouncementEventLogEntityPath_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementEventLogSensorNum,
                         sizeof(context->saHpiAnnouncementEventLogSensorNum) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEventLogEventState,
                         context->saHpiAnnouncementEventLogEventState_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGNAME:
            /** OCTETSTR = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEventLogName,
                         context->saHpiAnnouncementEventLogName_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementEventLogMid,
                         sizeof(context->saHpiAnnouncementEventLogMid) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogTextType,
                         sizeof(context->saHpiAnnouncementEventLogTextType) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogTextLanguage,
                         sizeof(context->saHpiAnnouncementEventLogTextLanguage) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEventLogText,
                         context->saHpiAnnouncementEventLogText_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTLOGDELETE:
            /** RowStatus = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementEventLogDelete,
                         sizeof(context->saHpiAnnouncementEventLogDelete) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiAnnouncementEventLogTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiAnnouncementEventLogTable_get_by_idx
 */
const saHpiAnnouncementEventLogTable_context *
saHpiAnnouncementEventLogTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiAnnouncementEventLogTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


