/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.hones18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiAnnouncementTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiAnnouncementTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>

#include <hpiAnnunciatorMapping.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiAnnouncementTable_oid[] = { saHpiAnnouncementTable_TABLE_OID };
size_t saHpiAnnouncementTable_oid_len = OID_LENGTH(saHpiAnnouncementTable_oid);


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int ann_initialized = FALSE;               
static GHashTable *dre_table;


/*************************************************************
 * oid and fucntion declarations scalars
 */
 
static u_long announcement_entry_count = 0;

static oid saHpiAnnouncementActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,1,31 };

int handle_saHpiAnnouncementActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);

int initialize_table_saHpiAnnouncementActiveEntries(void);


/**
 * 
 * @sessionid:
 * @rdr_entry:
 * @rpt_entry:
 * @full_oid:
 * @full_oid_len:
 * @child_oid:
 * @child_oid_len:
 * 
 * @return 
 */
SaErrorT populate_saHpiAnnouncementTable(SaHpiSessionIdT sessionid, 
                                         SaHpiRdrT *rdr_entry,
                                         SaHpiRptEntryT *rpt_entry)
{
	SaErrorT rv = SA_OK;
        SaErrorT rc = SA_OK;
        int new_row = MIB_FALSE;

	oid announcement_oid[ANNOUNCEMENT_INDEX_NR];
	netsnmp_index announcement_index;
	saHpiAnnouncementTable_context *announcement_ctx;
	
	SaHpiAnnouncementT announcement;

        DRE_XREF *dre_entry;
        key_tuple dre_tuple;

	announcement.EntryId = SAHPI_FIRST_ENTRY;
		
	/* Used for decoding */
	oh_big_textbuffer bigbuf;
	SaHpiTextBufferT buf;
		
        DEBUGMSGTL ((AGENT, "populate_saHpiAnnouncementTable, called\n"));		
	
	
	if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_saHpiAnnouncementTable() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_saHpiAnnouncementTable() passed NULL rpt_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        
	subagent_lock(&hpi_lock_data);
		
        rv = saHpiAnnunciatorGetNext(sessionid, 
	                         rpt_entry->ResourceId,
				 rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum,
				 SAHPI_ALL_SEVERITIES,
				 SAHPI_FALSE,         //Get all announcements.
				 &announcement);

	if ((rv != SA_OK) && (rv != SA_ERR_HPI_NOT_PRESENT)) {
		DEBUGMSGTL ((AGENT, 
		"populate_saHpiAnnouncementTable: saHpiAnnunciatorGetNext Failed: rv = %d\n",
		rv));
		
		subagent_unlock(&hpi_lock_data);
		
		return AGENT_ERR_INTERNAL_ERROR;
	}
	else if (rv == SA_ERR_HPI_NOT_PRESENT) {
          	
		subagent_unlock(&hpi_lock_data);
	        
		return SA_OK; //first call returned nothing
	}
	
	
	while ((rv != SA_ERR_HPI_NOT_PRESENT) && (rv == SA_OK)) {

                /* Store sa value                                       */
                /* this is the same as the EntryId obtained from HPI    */
                /* BUT in the case of a user defined Announcement they  */
                /* will be different                                    */
                dre_tuple.key_tuple_array[0] = announcement.StatusCond.DomainId;
                dre_tuple.key_tuple_array[1] = announcement.StatusCond.ResourceId;
                dre_tuple.key_tuple_array[2] = announcement.EntryId;

                /* domain_resource_entry_get() generates unique keys based on */
                /* domainid, resourceid, and entryid tuples.                  */
                dre_entry = domain_resource_entry_get(&dre_tuple, &dre_table); 

                if (dre_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: populate_saHpiAnnouncementTable() "
                                     "domain_resource_entry_get returned NULL\n"));
				     
			subagent_unlock(&hpi_lock_data);
				     
                        return AGENT_ERR_INTERNAL_ERROR;
                }
                /* here is where the mapped id is the same as the key value     */
                /* remember when the user creates an announcement the user      */
                /* specifies the entryId and will be different from the         */
                /* entryId returned from the HPI API that generates the         */
                /* new announcement                                             */    
                dre_entry->hpi_entry_id = announcement.EntryId;

                /* BUILD oid for new row */
                /* assign the number of indices */	
	        announcement_index.len = ANNOUNCEMENT_INDEX_NR;
                /** Index saHpiDomainId is external */	
                announcement_oid[0] = announcement.StatusCond.DomainId;
                /** Index saHpiResourceId is external */	
                announcement_oid[1] = announcement.StatusCond.ResourceId;
                /** Index saHpiAnnouncementEntryId is internal */	
	        announcement_oid[2] = announcement.EntryId;
                /* assign the indices to the index */	
	        announcement_index.oids = (oid *) & announcement_oid;
 
                /* See if Row exists. */
                announcement_ctx = NULL;
                announcement_ctx = CONTAINER_FIND(cb.container, &announcement_index);

                if (!announcement_ctx) {
                       // New entry. Add it
                        announcement_ctx = 
                             saHpiAnnouncementTable_create_row(&announcement_index);
                        new_row = MIB_TRUE;
                }
		else {
			new_row = MIB_FALSE;
		}	
		
                if (!announcement_ctx) {
                        snmp_log (LOG_ERR, "Not enough memory for an Announcement row!");
			
			subagent_unlock(&hpi_lock_data);
			
                        return AGENT_ERR_INTERNAL_ERROR;
                }
	 
	        /** SaHpiEntryId = ASN_UNSIGNED */
                announcement_ctx->saHpiAnnouncementEntryId = 
                        announcement.EntryId;

                /** SaHpiInstrumentId = ASN_UNSIGNED */
                announcement_ctx->saHpiAnnouncementAnnunciatorNum = 
                        rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum;
	 	
                /** SaHpiTime = ASN_OCTET_STR */
		hpitime_to_snmptime(announcement.Timestamp, announcement_ctx->saHpiAnnouncementTimestamp);
                announcement_ctx->saHpiAnnouncementTimestamp_len = sizeof(SaHpiTimeT);
                /** TruthValue = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementAddedByUser = 
                        (announcement.AddedByUser == SAHPI_TRUE) 
                        ? MIB_TRUE : MIB_FALSE;

                /** SaHpiSeverity = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementSeverity = 
                        announcement.Severity + 1;
 
                /** TruthValue = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementAcknowledged = 
                        (announcement.Acknowledged == SAHPI_TRUE) 
                        ? MIB_TRUE : MIB_FALSE;

                /** SaHpiSeverity = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementAckBySeverity =
                        0;
 
                /** INTEGER = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementStatusCondType = 
                        announcement.StatusCond.Type + 1;

                /** SaHpiEntityPath = ASN_OCTET_STR */
	        rc = oh_decode_entitypath(&announcement.StatusCond.Entity, 
                                          &bigbuf);
		
	        if (rc != SA_OK) {
		        DEBUGMSGTL ((AGENT, 
		                "populate_saHpiAnnouncementTable: "
                                "oh_decode_entitypath Failed: rc = %d\n",
		                 rc));
	        }
	        else {
		        memcpy(announcement_ctx->saHpiAnnouncementEntityPath,
		               bigbuf.Data, bigbuf.DataLength);
	                announcement_ctx->saHpiAnnouncementEntityPath_len = 
                                bigbuf.DataLength;
	        }

               
	
	        if (announcement.StatusCond.Type == 
                    SAHPI_STATUS_COND_TYPE_SENSOR) {
		        /** UNSIGNED32 = ASN_UNSIGNED */
               	        announcement_ctx->saHpiAnnouncementSensorNum = 
                                announcement.StatusCond.SensorNum;

	                /** SaHpiEventState = ASN_OCTET_STR */
	                rc = oh_decode_eventstate(
                                announcement.StatusCond.EventState, 
                                SAHPI_EC_SENSOR_SPECIFIC, &buf);
			
		        if (rc != SA_OK) {
	                        DEBUGMSGTL ((AGENT, 
  		                "populate_saHpiAnnouncementTable: "
                                "oh_decode_eventstate Failed: rc = %d\n",
		                 rc));
		        }	         	
		        else {
		               memcpy(announcement_ctx->saHpiAnnouncementEventState, 
                                      buf.Data, buf.DataLength); 
               	               announcement_ctx->saHpiAnnouncementEventState_len = 
                                       buf.DataLength;
		        }	
                }
	
                /** saHpiAnnouncementName = ASN_OCTET_STR */
                memcpy(announcement_ctx->saHpiAnnouncementName, 
                       announcement.StatusCond.Name.Value, 
	               announcement.StatusCond.Name.Length);
                announcement_ctx->saHpiAnnouncementName_len = 
                        announcement.StatusCond.Name.Length;
 
                /** SaHpiManufacturerId = ASN_UNSIGNED */
                announcement_ctx->saHpiAnnouncementMid = 
                        announcement.StatusCond.Mid;
 
                /** SaHpiTextType = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementTextType = 
                        announcement.StatusCond.Data.DataType + 1;
 
                /** SaHpiTextLanguage = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementTextLanguage = 
                        announcement.StatusCond.Data.Language + 1;
 
                /** SaHpiText = ASN_OCTET_STR */
                memcpy(announcement_ctx->saHpiAnnouncementText, 
                       announcement.StatusCond.Data.Data, 
	               announcement.StatusCond.Data.DataLength);
                announcement_ctx->saHpiAnnouncementText_len = 
                        announcement.StatusCond.Data.DataLength;
 
                /** RowStatus = ASN_INTEGER */
                announcement_ctx->saHpiAnnouncementDelete = 
                        SAHPIANNOUNCEMENTDELETE_ACTIVE;

                if (new_row == MIB_TRUE) 
                        CONTAINER_INSERT (cb.container, announcement_ctx);
			      
                rv = saHpiAnnunciatorGetNext(sessionid, 
	                         rpt_entry->ResourceId,
				 rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum,
				 SAHPI_ALL_SEVERITIES,
				 SAHPI_FALSE,         //Get all announcements.
				 &announcement);			      	
	
	}
	
	announcement_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);
		       
        return SA_OK; 		       

}

/**
 * 
 * @param row_ctx
 * 
 * @return 
 */
int announcement_add (saHpiAnnouncementTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "announcement_add, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiAnnunciatorNumT annunciator_num;
        SaHpiAnnouncementT  announcement;

        DRE_XREF *dre_entry;
        key_tuple dre_tuple;
        
        if ((row_ctx->sahpi_announcement_annunciator_num_set == MIB_TRUE) &&
            (row_ctx->sahpi_announcement_severity_set == MIB_TRUE) &&
            (row_ctx->sahpi_announcement_status_cond_type_set == MIB_TRUE) &&
            (row_ctx->sahpi_announcement_entitypath_set == MIB_TRUE) &&      
            (row_ctx->sahpi_announcement_sensornum_set == MIB_TRUE) &&       
            (row_ctx->sahpi_announcement_event_state_set == MIB_TRUE) &&     
            (row_ctx->sahpi_announcement_name_set == MIB_TRUE) &&            
            (row_ctx->sahpi_announcement_mid == MIB_TRUE) &&                 
            (row_ctx->sahpi_announcement_text_type_set == MIB_TRUE) &&   
            (row_ctx->sahpi_announcement_text_language_set == MIB_TRUE) &&
            (row_ctx->sahpi_announcement_text == MIB_TRUE) &&
            (row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX] !=
             SAHPI_ENTRY_UNSPECIFIED)) {

                /* Acknowledged */
                announcement.Acknowledged = 
                        (row_ctx->saHpiAnnouncementAcknowledged == MIB_TRUE) ? 
                        SAHPI_TRUE : SAHPI_FALSE;

                /* AddedByUser */
                announcement.AddedByUser = 
                        (row_ctx->saHpiAnnouncementAddedByUser == MIB_TRUE) ? 
                        SAHPI_TRUE : SAHPI_FALSE;                        

                /* EntryId */
                announcement.EntryId = row_ctx->saHpiAnnouncementEntryId;

                /*StatusCond */
                announcement.StatusCond.Data.DataLength =
                        row_ctx->saHpiAnnouncementText_len;
                memcpy(announcement.StatusCond.Data.Data, 
                       row_ctx->saHpiAnnouncementText, 
                       row_ctx->saHpiAnnouncementText_len);
                announcement.StatusCond.Type = 
                        row_ctx->saHpiAnnouncementStatusCondType - 1;
                announcement.StatusCond.Data.DataType = 
                        row_ctx->saHpiAnnouncementTextType;
                announcement.StatusCond.Data.Language =
                        row_ctx->saHpiAnnouncementTextLanguage;
                announcement.StatusCond.DomainId =
                        row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX];

                /* EP */
                rc = oh_encode_entitypath((gchar*)row_ctx->saHpiAnnouncementEntityPath, 
                                          &announcement.StatusCond.Entity);
                if (rc != SA_OK) {
                        DEBUGMSGTL ((AGENT, 
                        "announcement_add: oh_encode_entitypath Failed:"
                        " rc = %s\n",
                        oh_lookup_error(rc)));
                }


                /* EventState */
                SaHpiTextBufferT        buffer;
                SaHpiEventCategoryT     event_cat;
                memcpy(buffer.Data, 
                       row_ctx->saHpiAnnouncementEventState, 
                       row_ctx->saHpiAnnouncementEventState_len);
                buffer.DataLength = row_ctx->saHpiAnnouncementEventState_len;

                if ((row_ctx->saHpiAnnouncementStatusCondType - 1) == 
                    SAHPI_STATUS_COND_TYPE_SENSOR) {
                        rc = oh_encode_eventstate(&buffer,
                                                  &announcement.StatusCond.EventState,
                                                  &event_cat);
                        if (rc != SA_OK) {
                                DEBUGMSGTL ((AGENT, 
                                "announcement_add: oh_encode_eventstate Failed:"
                                " rc = %s\n",
                                 oh_lookup_error(rc)));
                        }	         	
                }

                /* Mid */
                announcement.StatusCond.Mid = 
                        row_ctx->saHpiAnnouncementMid;

                /* Name */
                announcement.StatusCond.Name.Length = 
                        row_ctx->saHpiAnnouncementName_len;
                memcpy(announcement.StatusCond.Name.Value,
                       row_ctx->saHpiAnnouncementName,
                       row_ctx->saHpiAnnouncementName_len);

                /* ResourceId */
                announcement.StatusCond.ResourceId =
                        row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];

                /* SensorNum */
                announcement.StatusCond.SensorNum =
                        row_ctx->saHpiAnnouncementSensorNum;

                /* Type */
                announcement.StatusCond.Type = 
                        row_ctx->saHpiAnnouncementStatusCondType - 1;

                

                session_id = get_session_id(row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX]);
                resource_id = row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];
                annunciator_num = row_ctx->saHpiAnnouncementAnnunciatorNum;

                rc = saHpiAnnunciatorAdd(session_id, 
                                         resource_id, 
                                         annunciator_num, 
                                         &announcement);

                if (rc != SA_OK) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: saHpiAnnunciatorAdd() "
                                     "Failed to add Announcement rc: [%s]\n", 
                                     oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }
		
		/* Timestamp */
		hpitime_to_snmptime(announcement.Timestamp, row_ctx->saHpiAnnouncementTimestamp);
                row_ctx->saHpiAnnouncementTimestamp_len = sizeof(SaHpiTimeT);
		
                dre_tuple.key_tuple_array[0] = 
                        row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX];
                dre_tuple.key_tuple_array[1] = 
                        row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];
                dre_tuple.key_tuple_array[2] = 
                        row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX];

                /* domain_resource_entry_get() generates unique keys based on */
                /* domainid, resourceid, and entryid tuples.                  */
                dre_entry = domain_resource_entry_get(&dre_tuple, &dre_table); 

                if (dre_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: populate_saHpiAnnouncementTable() "
                                     "domain_resource_entry_get returned NULL\n"));
                        return AGENT_ERR_INTERNAL_ERROR;
                }

                /* we have retrieved the HPI Annunciator EntryId */
                dre_entry->hpi_entry_id = announcement.EntryId;

        } 
        
        return SNMP_ERR_NOERROR; 
}


/**
 * 
 * @param row_ctx
 * 
 * @return 
 */
int announcement_ack (saHpiAnnouncementTable_context *row_ctx)
{

        DEBUGMSGTL ((AGENT, "announcement_ack, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiEntryIdT       entry_id;

        DRE_XREF *dre_entry;
        key_tuple dre_tuple;

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX]);
        resource_id = row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];

        /* see if this is an attempt to delete based on severtiy */
        if (row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX] == 0) {
                /* we are deleting based on Severity */
                entry_id = SAHPI_ENTRY_UNSPECIFIED;
        } else {

                dre_tuple.key_tuple_array[0] = row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX];
                dre_tuple.key_tuple_array[1] = row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];
                dre_tuple.key_tuple_array[2] = row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX];

                /* domain_resource_entry_get() generates unique keys based on */
                /* domainid, resourceid, and entryid tuples.                  */
                dre_entry = domain_resource_entry_get(&dre_tuple, &dre_table); 

                if (dre_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: populate_saHpiAnnouncementTable() "
                                     "domain_resource_entry_get returned NULL\n"));
                        return AGENT_ERR_INTERNAL_ERROR;
                }

                /* we have retrieved the HPI Annunciator EntryId */
                entry_id = dre_entry->hpi_entry_id;
        }

        rc = saHpiAnnunciatorAcknowledge(session_id, 
                                         resource_id, 
                                         row_ctx->saHpiAnnouncementAnnunciatorNum,
                                         entry_id, 
                                         row_ctx->saHpiAnnouncementSeverity);

        if (rc != SA_OK) {
                snmp_log (LOG_ERR,
                "announcement_ack: Call to "
                "saHpiAnnunciatorAcknowledge() failed rc: %s.\n",
                oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                "announcement_ack: Call to "
                "saHpiAnnunciatorAcknowledge() failed rc: %s.\n",
                oh_lookup_error(rc)));
                return get_snmp_error(rc);
        } 

        DEBUGMSGTL ((AGENT, "announcement_ack: Call to "
                   "saHpiAnnunciatorAcknowledge() succeeded rc: %s.\n",
                   oh_lookup_error(rc)));

        /* if we ack'd based on severity then we need to set ack to true   */
        /* for all rows for the given SessionId <==> DomainId, ResourceId, */
        /* AnnunciatorNum and Severity                                     */
        if (entry_id == SAHPI_ENTRY_UNSPECIFIED) {

                netsnmp_index *row_idx;
                saHpiAnnouncementTable_context *announcement_ctx;

                row_idx = CONTAINER_FIRST(cb.container);
                if (row_idx) //At least one entry was found.
                {
                        do {
                                announcement_ctx = CONTAINER_FIND(cb.container, row_idx);

                                if ((announcement_ctx->index.oids[saHpiAnnouncementDomainId_INDEX] ==
                                     row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX]) &&

                                    (announcement_ctx->index.oids[saHpiAnnouncementResourceId_INDEX] ==
                                     row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX]) &&

                                    (announcement_ctx->saHpiAnnouncementAnnunciatorNum == 
                                     row_ctx->saHpiAnnouncementAnnunciatorNum) &&

                                    ((announcement_ctx->saHpiAnnouncementSeverity == 
                                      row_ctx->saHpiAnnouncementSeverity) ||
                                    (row_ctx->saHpiAnnouncementAckBySeverity == 
                                     SAHPI_ALL_SEVERITIES)))

                                     row_ctx->saHpiAnnouncementAcknowledged = MIB_TRUE;
                                     DEBUGMSGTL ((AGENT, "announcement_ack: found row for "
                                                         "setting ack based on severity\n"));

                                row_idx = CONTAINER_NEXT(cb.container, row_idx);
                        } while (row_idx);
                }
        } /* end check for all rows deleted based on Severity */

        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @param row_ctx
 * 
 * @return 
 */
int announcement_delete (saHpiAnnouncementTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "announcement_delete, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiAnnouncementT  announcement;

        DRE_XREF *dre_entry;
        key_tuple dre_tuple;

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX]);
        resource_id = row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];

        dre_tuple.key_tuple_array[0] = row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX];
        dre_tuple.key_tuple_array[1] = row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX];
        dre_tuple.key_tuple_array[2] = row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX];

        /* domain_resource_entry_get() generates unique keys based on */
        /* domainid, resourceid, and entryid tuples.                  */
        dre_entry = domain_resource_entry_get(&dre_tuple, &dre_table); 

        if (dre_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_saHpiAnnouncementTable() "
                             "domain_resource_entry_get returned NULL\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        /* see if this is an attempt to delete based on severtiy */
        if (row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX] == 0) {
                /* we are deleting based on Severity */
                announcement.EntryId = SAHPI_ENTRY_UNSPECIFIED;
        } else {
                /* we have retrieved the HPI Annunciator EntryId */
                /* now we can call the delete function           */
                announcement.EntryId = dre_entry->hpi_entry_id;
        }

        rc = saHpiAnnunciatorDelete(session_id, 
                                    resource_id, 
                                    row_ctx->saHpiAnnouncementAnnunciatorNum,
                                    announcement.EntryId,
                                    row_ctx->saHpiAnnouncementSeverity - 1);
        if (rc != SA_OK) {
                snmp_log (LOG_ERR,
                "announcement_delete: Call to "
                "saHpiAnnouncementDelete() failed rc: %s.\n",
                oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                "announcement_delete: Call to "
                "saHpiAnnouncementDelete() failed rc: %s.\n",
                oh_lookup_error(rc)));
                return get_snmp_error(rc);
        } 

        DEBUGMSGTL ((AGENT, "announcement_delete: Call to "
                   "saHpiAnnouncementDelete() succeeded rc: %s.\n",
                   oh_lookup_error(rc)));

        /* if we deleted based on severity then we need to remove      */
        /* all rows for the given SessionId <==> DomainId, ResourceId, */
        /* AnnunciatorNum and Severity                                 */
        if (announcement.EntryId == SAHPI_ENTRY_UNSPECIFIED) {

                netsnmp_index *row_idx;
                saHpiAnnouncementTable_context *announcement_ctx;

                row_idx = CONTAINER_FIRST(cb.container);
                if (row_idx) //At least one entry was found.
                {
                        do {
                                announcement_ctx = CONTAINER_FIND(cb.container, row_idx);

                                if ((announcement_ctx->index.oids[saHpiAnnouncementDomainId_INDEX] ==
                                     row_ctx->index.oids[saHpiAnnouncementDomainId_INDEX]) &&

                                    (announcement_ctx->index.oids[saHpiAnnouncementResourceId_INDEX] ==
                                     row_ctx->index.oids[saHpiAnnouncementResourceId_INDEX]) &&

                                    (announcement_ctx->saHpiAnnouncementAnnunciatorNum == 
                                     row_ctx->saHpiAnnouncementAnnunciatorNum) &&

                                    ((announcement_ctx->saHpiAnnouncementSeverity == 
                                      row_ctx->saHpiAnnouncementSeverity) ||
                                    (row_ctx->saHpiAnnouncementAckBySeverity == 
                                     SAHPI_ALL_SEVERITIES)) &&

                                    (announcement_ctx->index.oids[saHpiAnnouncementEntryId_INDEX] !=
                                     row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX])) {

                                        row_idx = CONTAINER_NEXT(cb.container, row_idx);
					/* all conditions met remove row */
                                        CONTAINER_REMOVE (cb.container, announcement_ctx);
                                        saHpiAnnouncementTable_delete_row (announcement_ctx);
                                        announcement_entry_count = CONTAINER_SIZE (cb.container);
                                        DEBUGMSGTL ((AGENT, "announcement_delete: found row for "
                                                     "deletion based on severity\n"));
                                }
				else {
				     row_idx = CONTAINER_NEXT(cb.container, row_idx);
				} 
                                
                        } while (row_idx);
                }
        } /* end check for all rows deleted based on Severity */

        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @SessionId: current session
 *
 * @return:
 */
 
void update_announcements (SaHpiSessionIdT sessionid)
{     
	SaErrorT                rv; 
	
	// FOR RPT
	SaHpiEntryIdT           rpt_entry_id;
	SaHpiRptEntryT          rpt_entry;
	
	// FOR RDR
	SaHpiEntryIdT    	rdr_entry_id;
	SaHpiRdrT		rdr_entry;


	DEBUGMSGTL ((AGENT, "update_announcements() Called\n"));
		
	rv = clear_announcements();
	
	rpt_entry_id = SAHPI_FIRST_ENTRY;
        do {
		rv = saHpiRptEntryGet(sessionid, rpt_entry_id, &rpt_entry_id, &rpt_entry);
		rdr_entry_id = SAHPI_FIRST_ENTRY;
		do {
			rv =  saHpiRdrGet(sessionid, 
				          rpt_entry.ResourceId, 
				          rdr_entry_id, 
				          &rdr_entry_id, 
				          &rdr_entry);
					  
			if (rdr_entry.RdrType == SAHPI_ANNUNCIATOR_RDR) {
				
				populate_saHpiAnnouncementTable(sessionid,
				                                &rdr_entry,
								&rpt_entry);
				
			}			  
					  
	        }while (rdr_entry_id !=  SAHPI_LAST_ENTRY);
		
	}while(rpt_entry_id != SAHPI_LAST_ENTRY);

}

int clear_announcements (void)
{
	netsnmp_index *ann_idx;
	saHpiAnnouncementTable_context *ann_ctx;
	int count = 0;
	
	DEBUGMSGTL ((AGENT, "clear_announcements() called\n"));
	DEBUGMSGTL ((AGENT, "Wiping out announcement table!!!\n"));
	
	ann_idx = CONTAINER_FIRST(cb.container);
	
	if (ann_idx) //At least one found
	{
		do{
			ann_ctx = CONTAINER_FIND(cb.container, ann_idx);
			ann_idx = CONTAINER_NEXT(cb.container, ann_idx);
			
			if(ann_ctx)
			{
							   
			   count++;	
	                   CONTAINER_REMOVE (cb.container, ann_ctx);
                           saHpiAnnouncementTable_delete_row (ann_ctx);
			   announcement_entry_count = CONTAINER_SIZE (cb.container);
			   
			}
			
		} while(ann_idx);
	}
	
	DEBUGMSGTL ((AGENT, "clear_announcements removed %d rows\n", count));		   
	
	return SA_OK;	
}	

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiAnnouncementActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiAnnouncementActiveEntries, called\n"));

        announcement_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			        (u_char *) &announcement_entry_count,
        			        sizeof(announcement_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

 
/**
 * 
 * @return: 
 */ 
int initialize_table_saHpiAnnouncementActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiAnnouncementActiveEntries, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiAnnouncementActiveEntries", 
					handle_saHpiAnnouncementActiveEntries,
                                        saHpiAnnouncementActiveEntries_oid, 
					OID_LENGTH(saHpiAnnouncementActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
 

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiAnnouncementTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiAnnouncementTable_cmp( const void *lhs, const void *rhs )
{
        saHpiAnnouncementTable_context *context_l =
        (saHpiAnnouncementTable_context *)lhs;
        saHpiAnnouncementTable_context *context_r =
        (saHpiAnnouncementTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiAnnouncementTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiAnnouncementEntryId */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                return 0;
                        }
                }
        }

        return 0;
}


/************************************************************
 * Initializes the saHpiAnnouncementTable module
 */
void
init_saHpiAnnouncementTable(void)
{
        
        DEBUGMSGTL ((AGENT, "init_saHpiAnnouncementTable, called\n"));
	
	initialize_table_saHpiAnnouncementTable();

        initialize_table_saHpiAnnouncementActiveEntries();

        domain_resource_entry_initialize(&ann_initialized, &dre_table);
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiAnnouncementTable_row_copy(saHpiAnnouncementTable_context * dst,
                         saHpiAnnouncementTable_context * src)
{
    if(!dst||!src)
        return 1;

    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_row_copy, called\n"));

    /*
     * copy index, if provided
     */
     
    subagent_lock(&hpi_lock_data); 
     
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiAnnouncementEntryId = src->saHpiAnnouncementEntryId;

    dst->saHpiAnnouncementAnnunciatorNum = src->saHpiAnnouncementAnnunciatorNum;

    memcpy( dst->saHpiAnnouncementTimestamp, 
            src->saHpiAnnouncementTimestamp, 
	    src->saHpiAnnouncementTimestamp_len );
	     
    dst->saHpiAnnouncementTimestamp_len = src->saHpiAnnouncementTimestamp_len;	     

    dst->saHpiAnnouncementAddedByUser = src->saHpiAnnouncementAddedByUser;

    dst->saHpiAnnouncementSeverity = src->saHpiAnnouncementSeverity;

    dst->saHpiAnnouncementAcknowledged = src->saHpiAnnouncementAcknowledged;

    dst->saHpiAnnouncementAckBySeverity = src->saHpiAnnouncementAckBySeverity;

    dst->saHpiAnnouncementStatusCondType = src->saHpiAnnouncementStatusCondType;

    memcpy( dst->saHpiAnnouncementEntityPath, src->saHpiAnnouncementEntityPath, src->saHpiAnnouncementEntityPath_len );
    dst->saHpiAnnouncementEntityPath_len = src->saHpiAnnouncementEntityPath_len;

    dst->saHpiAnnouncementSensorNum = src->saHpiAnnouncementSensorNum;

    memcpy( dst->saHpiAnnouncementEventState, src->saHpiAnnouncementEventState, src->saHpiAnnouncementEventState_len );
    dst->saHpiAnnouncementEventState_len = src->saHpiAnnouncementEventState_len;

    memcpy( dst->saHpiAnnouncementName, src->saHpiAnnouncementName, src->saHpiAnnouncementName_len );
    dst->saHpiAnnouncementName_len = src->saHpiAnnouncementName_len;

    dst->saHpiAnnouncementMid = src->saHpiAnnouncementMid;

    dst->saHpiAnnouncementTextType = src->saHpiAnnouncementTextType;

    dst->saHpiAnnouncementTextLanguage = src->saHpiAnnouncementTextLanguage;

    memcpy( dst->saHpiAnnouncementText, src->saHpiAnnouncementText, src->saHpiAnnouncementText_len );
    dst->saHpiAnnouncementText_len = src->saHpiAnnouncementText_len;

    dst->saHpiAnnouncementDelete = src->saHpiAnnouncementDelete;
 
    subagent_unlock(&hpi_lock_data);

    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiAnnouncementTable_extract_index( saHpiAnnouncementTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiAnnouncementEntryId;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_extract_index, called\n"));

        subagent_lock(&hpi_lock_data);
	
	/*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        
			subagent_unlock(&hpi_lock_data);
			return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiAnnouncementEntryId;

        memset( &var_saHpiAnnouncementEntryId, 0x00, sizeof(var_saHpiAnnouncementEntryId) );
        var_saHpiAnnouncementEntryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiAnnouncementEntryId.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiResourceId */
   
                ctx->saHpiAnnouncementEntryId = *var_saHpiAnnouncementEntryId.val.integer;
                
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
		err = saHpiAnnouncementEntryId_check_index(*var_saHpiAnnouncementEntryId.val.integer);
   
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
	
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiAnnouncementTable_can_activate(saHpiAnnouncementTable_context *undo_ctx,
                      saHpiAnnouncementTable_context *row_ctx,
                      netsnmp_request_group * rg)
{

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_can_activate, called\n"));

    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiAnnouncementTable_can_deactivate(saHpiAnnouncementTable_context *undo_ctx,
                        saHpiAnnouncementTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiAnnouncementTable_can_delete(saHpiAnnouncementTable_context *undo_ctx,
                    saHpiAnnouncementTable_context *row_ctx,
                    netsnmp_request_group * rg)
{

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_can_delete, called\n"));

    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiAnnouncementTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiAnnouncementTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiAnnouncementTable_context *
saHpiAnnouncementTable_create_row( netsnmp_index* hdr)
{

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_create_row, called\n"));

    
    subagent_lock(&hpi_lock_data);
    
    saHpiAnnouncementTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiAnnouncementTable_context);
    if(!ctx)
        return NULL;
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiAnnouncementTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiAnnouncementAnnunciatorNum = 0;
     ctx->saHpiAnnouncementSeverity = 0;
     ctx->saHpiAnnouncementAcknowledged = 0;
     ctx->saHpiAnnouncementAckBySeverity = 0;
     ctx->saHpiAnnouncementStatusCondType = 0;
     ctx->saHpiAnnouncementEntityPath = 0;
     ctx->saHpiAnnouncementSensorNum = 0;
     ctx->saHpiAnnouncementEventState = 0;
     ctx->saHpiAnnouncementName = 0;
     ctx->saHpiAnnouncementMid = 0;
     ctx->saHpiAnnouncementTextType = 0;
     ctx->saHpiAnnouncementTextLanguage = 0;
     ctx->saHpiAnnouncementText = 0;
     ctx->saHpiAnnouncementDelete = 0;
    */

    ctx->sahpi_announcement_annunciator_num_set = MIB_FALSE; 
    ctx->sahpi_announcement_severity_set =  MIB_FALSE;
    ctx->sahpi_announcement_status_cond_type_set =  MIB_FALSE;
    ctx->sahpi_announcement_entitypath_set =  MIB_FALSE;
    ctx->sahpi_announcement_sensornum_set =  MIB_FALSE;
    ctx->sahpi_announcement_event_state_set =  MIB_FALSE;
    ctx->sahpi_announcement_name_set =  MIB_FALSE;
    ctx->sahpi_announcement_mid =  MIB_FALSE;
    ctx->sahpi_announcement_text_type_set =  MIB_FALSE;
    ctx->sahpi_announcement_text_language_set =  MIB_FALSE;
    ctx->sahpi_announcement_text =  MIB_FALSE;


    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiAnnouncementTable_context *
saHpiAnnouncementTable_duplicate_row( saHpiAnnouncementTable_context * row_ctx)
{

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_duplicate_row, called\n"));

    saHpiAnnouncementTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
    
    dup = SNMP_MALLOC_TYPEDEF(saHpiAnnouncementTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL;
    }	
        
    if(saHpiAnnouncementTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiAnnouncementTable_delete_row( saHpiAnnouncementTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_delete_row, called\n"));

    subagent_lock(&hpi_lock_data);
    
    
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiAnnouncementTable_set_reserve1( netsnmp_request_group *rg )
{

        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_set_reserve1, called\n"));

    saHpiAnnouncementTable_context *row_ctx =
            (saHpiAnnouncementTable_context *)rg->existing_row;
//    saHpiAnnouncementTable_context *undo_ctx =
//            (saHpiAnnouncementTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                          sizeof(row_ctx->saHpiAnnouncementAnnunciatorNum));
        break;

        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                      sizeof(row_ctx->saHpiAnnouncementSeverity));
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                      sizeof(row_ctx->saHpiAnnouncementAcknowledged));
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                      sizeof(row_ctx->saHpiAnnouncementAckBySeverity));
        break;

        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                      sizeof(row_ctx->saHpiAnnouncementStatusCondType));
        break;

        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiAnnouncementEntityPath)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	    
        break;

        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                        sizeof(row_ctx->saHpiAnnouncementSensorNum));
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiAnnouncementEventState)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	    
        break;

        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiAnnouncementName)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	    
        break;

        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                        sizeof(row_ctx->saHpiAnnouncementMid));
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                        sizeof(row_ctx->saHpiAnnouncementTextType));
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                        sizeof(row_ctx->saHpiAnnouncementTextLanguage));
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiAnnouncementText)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }		    
        break;

        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                        sizeof(row_ctx->saHpiAnnouncementDelete));
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiAnnouncementTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiAnnouncementTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiAnnouncementTable_context *row_ctx = (saHpiAnnouncementTable_context *)rg->existing_row;
//    saHpiAnnouncementTable_context *undo_ctx = (saHpiAnnouncementTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_set_reserve2, called\n"));

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
                if (oh_lookup_severity(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
                if (oh_lookup_severity(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
                if (oh_lookup_statuscondtype(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */

                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
                if (oh_lookup_texttype(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
                if (oh_lookup_language(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
                /*
                 * TODO: routine to check valid values
                 *
                 * EXAMPLE:
                 *
                * if ( XXX_check_value( var->val.string, XXX ) ) {
            *    rc = SNMP_ERR_INCONSISTENTVALUE;
            *    rc = SNMP_ERR_BADVALUE;
            * }
            */
        break;

        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
                if (*var->val.integer != SAHPIANNOUNCEMENTDELETE_DESTROY)
                        rc = SNMP_ERR_BADVALUE;
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiAnnouncementTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiAnnouncementTable_context *row_ctx = (saHpiAnnouncementTable_context *)rg->existing_row;
//    saHpiAnnouncementTable_context *undo_ctx = (saHpiAnnouncementTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;

    subagent_lock(&hpi_lock_data);

    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_set_action, called\n"));

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
            row_ctx->saHpiAnnouncementAnnunciatorNum = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_annunciator_num_set = MIB_TRUE; 
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            row_ctx->saHpiAnnouncementSeverity = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete == 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
            } 
            row_ctx->sahpi_announcement_severity_set = MIB_TRUE;
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiAnnouncementAcknowledged = *var->val.integer;
      
            if (rg->row_created == 1) {                                 /* createAndWait */
                    row_ctx->saHpiAnnouncementDelete = SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; 

            } else if ((rg->row_created != 1) &&                       /* active row, this is an ACK */
                       (row_ctx->saHpiAnnouncementDelete == SAHPIANNOUNCEMENTDELETE_ACTIVE)) {  
                    if (*var->val.integer == MIB_TRUE) {
                            row_err = announcement_ack(row_ctx);
                    }
            } 

        break;

        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            row_ctx->saHpiAnnouncementAckBySeverity = *var->val.integer;

            if (rg->row_created == 1) {                                 /* createAndWait */
                    row_ctx->saHpiAnnouncementDelete = SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; 

            } else if ((rg->row_created != 1) &&                       /* severity set, time to ACK */
                       (row_ctx->sahpi_announcement_severity_set == MIB_TRUE) &&
                       (row_ctx->saHpiAnnouncementDelete == SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT) &&
                       (row_ctx->index.oids[saHpiAnnouncementEntryId_INDEX] ==
                        SAHPI_ENTRY_UNSPECIFIED)) { 
                    if (*var->val.integer == MIB_TRUE) {
                            row_err = announcement_ack(row_ctx);
                    }
            }
        break;

        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiAnnouncementStatusCondType = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_status_cond_type_set = MIB_TRUE; 
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementEntityPath,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementEntityPath_len = var->val_len;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_entitypath_set = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            row_ctx->saHpiAnnouncementSensorNum = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_sensornum_set = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementEventState,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementEventState_len = var->val_len;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_event_state_set = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementName,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementName_len = var->val_len;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_name_set = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            row_ctx->saHpiAnnouncementMid = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_mid = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            row_ctx->saHpiAnnouncementTextType = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_text_type_set = MIB_TRUE; 
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            row_ctx->saHpiAnnouncementTextLanguage = *var->val.integer;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_text_language_set = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiAnnouncementText,var->val.string,var->val_len);
            row_ctx->saHpiAnnouncementText_len = var->val_len;
            if ( (rg->row_created == 1) || (row_ctx->saHpiAnnouncementDelete ==
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT)) {
                    row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT; /* createAndWait */
                    row_ctx->sahpi_announcement_text = MIB_TRUE;
            }
            row_err = announcement_add(row_ctx);
        break;

        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
                if (((row_ctx->sahpi_announcement_severity_set == MIB_TRUE) && 
		     (*var->val.integer == SAHPIANNOUNCEMENTDELETE_DESTROY) &&
                     (row_ctx->sahpi_announcement_annunciator_num_set == MIB_TRUE)) ||
                    (row_ctx->saHpiAnnouncementDelete == SAHPIANNOUNCEMENTDELETE_ACTIVE)) {

                        row_ctx->saHpiAnnouncementDelete = *var->val.integer;
                        row_err = announcement_delete(row_ctx);
                        /* delete succeeded remove row */
                        if (row_err == SNMP_ERR_NOERROR) {
                                rg->row_deleted = 1;
                        }

                } else if ( (rg->row_created == 1)) {
                        
			row_ctx->saHpiAnnouncementDelete = 
                            SAHPIANNOUNCEMENTDELETE_CREATEANDWAIT;
                        
                }
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }

    subagent_unlock(&hpi_lock_data);
    return;

}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiAnnouncementTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiAnnouncementTable_context *row_ctx = (saHpiAnnouncementTable_context *)rg->existing_row;
//    saHpiAnnouncementTable_context *undo_ctx = (saHpiAnnouncementTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;


    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_set_commit, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiAnnouncementTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiAnnouncementTable_context *row_ctx = (saHpiAnnouncementTable_context *)rg->existing_row;
//    saHpiAnnouncementTable_context *undo_ctx = (saHpiAnnouncementTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_set_free, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
	        break;
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiAnnouncementTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiAnnouncementTable_context *row_ctx = (saHpiAnnouncementTable_context *)rg->existing_row;
//    saHpiAnnouncementTable_context *undo_ctx = (saHpiAnnouncementTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_set_undo, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}


/************************************************************
 *
 * Initialize the saHpiAnnouncementTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiAnnouncementTable(void)
{
    netsnmp_table_registration_info *table_info;

    DEBUGMSGTL ((AGENT, "initialize_table_saHpiAnnouncementTable, called\n"));

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiAnnouncementTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiAnnouncementTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiAnnouncementTable_oid,
                                             saHpiAnnouncementTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiAnnouncementTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiAnnouncementEntryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiAnnouncementTable_COL_MIN;
    table_info->max_column = saHpiAnnouncementTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiAnnouncementTable_get_value;
    cb.container = netsnmp_container_find("saHpiAnnouncementTable_primary:"
                                          "saHpiAnnouncementTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiAnnouncementTable_secondary:"
                                                       "saHpiAnnouncementTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiAnnouncementTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiAnnouncementTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiAnnouncementTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiAnnouncementTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiAnnouncementTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiAnnouncementTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiAnnouncementTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiAnnouncementTable_can_delete;

    cb.set_reserve1 = saHpiAnnouncementTable_set_reserve1;
    cb.set_reserve2 = saHpiAnnouncementTable_set_reserve2;
    cb.set_action = saHpiAnnouncementTable_set_action;
    cb.set_commit = saHpiAnnouncementTable_set_commit;
    cb.set_free = saHpiAnnouncementTable_set_free;
    cb.set_undo = saHpiAnnouncementTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiAnnouncementTable",
                "Registering table saHpiAnnouncementTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiAnnouncementTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiAnnouncementTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiAnnouncementTable_context *context = (saHpiAnnouncementTable_context *)item;

    DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_get_value, called\n"));

    switch(table_info->colnum) {

        case COLUMN_SAHPIANNOUNCEMENTENTRYID:
            /** SaHpiEntryId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementEntryId,
                         sizeof(context->saHpiAnnouncementEntryId) );
        break;

        case COLUMN_SAHPIANNOUNCEMENTANNUNCIATORNUM:
             /** SaHpiInstrumentId = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementAnnunciatorNum,
                         sizeof(context->saHpiAnnouncementAnnunciatorNum) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementTimestamp,
                         sizeof(context->saHpiAnnouncementTimestamp) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTADDEDBYUSER:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementAddedByUser,
                         sizeof(context->saHpiAnnouncementAddedByUser) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementSeverity,
                         sizeof(context->saHpiAnnouncementSeverity) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementAcknowledged,
                         sizeof(context->saHpiAnnouncementAcknowledged) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementAckBySeverity,
                         sizeof(context->saHpiAnnouncementAckBySeverity) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementStatusCondType,
                         sizeof(context->saHpiAnnouncementStatusCondType) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEntityPath,
                         context->saHpiAnnouncementEntityPath_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementSensorNum,
                         sizeof(context->saHpiAnnouncementSensorNum) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementEventState,
                         context->saHpiAnnouncementEventState_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTNAME:
            /** OCTETSTR = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementName,
                         context->saHpiAnnouncementName_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiAnnouncementMid,
                         sizeof(context->saHpiAnnouncementMid) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementTextType,
                         sizeof(context->saHpiAnnouncementTextType) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementTextLanguage,
                         sizeof(context->saHpiAnnouncementTextLanguage) );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiAnnouncementText,
                         context->saHpiAnnouncementText_len );
        break;
    
        case COLUMN_SAHPIANNOUNCEMENTDELETE:
            /** RowStatus = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiAnnouncementDelete,
                         sizeof(context->saHpiAnnouncementDelete) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiAnnouncementTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiAnnouncementTable_get_by_idx
 */
const saHpiAnnouncementTable_context *
saHpiAnnouncementTable_get_by_idx(netsnmp_index * hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiAnnouncementTable_get_by_idx, called\n"));

    return (const saHpiAnnouncementTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


