/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiAnnunciatorTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiAnnunciatorTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceTable.h>
#include <session_info.h>
#include <oh_utils.h>
#include <saHpiAnnouncementTable.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiAnnunciatorTable_oid[] = { saHpiAnnunciatorTable_TABLE_OID};
size_t saHpiAnnunciatorTable_oid_len = OID_LENGTH(saHpiAnnunciatorTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long annunciator_entry_count = 0;
static oid saHpiAnnunciatorActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,4,5};
int handle_saHpiAnnunciatorActiveEntries(netsnmp_mib_handler *handler,
                                      netsnmp_handler_registration *reginfo,
                                      netsnmp_agent_request_info   *reqinfo,
                                      netsnmp_request_info         *requests);
int initialize_table_saHpiAnnunciatorActiveEntries(void);

/**
 * 
 * @sessionid:
 * @rdr_entry:
 * @rpt_entry:
 * @full_oid:
 * @full_oid_len:
 * @child_oid:
 * @child_oid_len:
 * 
 * @return 
 */
SaErrorT populate_annunciator(SaHpiSessionIdT sessionid, 
                              SaHpiRdrT *rdr_entry,
                              SaHpiRptEntryT *rpt_entry,
                              oid *full_oid, size_t full_oid_len,
                              oid *child_oid, size_t *child_oid_len)
{

        DEBUGMSGTL ((AGENT, "populate_annunciator, called\n"));

        SaErrorT rv = SA_OK; 
	SaErrorT rc = SA_OK; 
        int new_row = MIB_FALSE;
        SaHpiAnnunciatorModeT mode;

        oid annunciator_oid[ANNUNCIATOR_INDEX_NR];
        netsnmp_index annunciator_index;
        saHpiAnnunciatorTable_context *annunciator_context;

        oid column[2];
        int column_len = 2;

        DEBUGMSGTL ((AGENT, "SAHPI_ANNUNCIATOR_RDR populate_annunciator() called\n"));

        /* check for NULL pointers */
        if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_annunciator() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_annunciator() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        subagent_lock(&hpi_lock_data);

        /* BUILD oid for new row */
        /* assign the number of indices */
        annunciator_index.len = ANNUNCIATOR_INDEX_NR;
        /** Index saHpiDomainId is external */
        annunciator_oid[0] = get_domain_id(sessionid);
        /** Index saHpiResourceId is external */
        annunciator_oid[1] = rpt_entry->ResourceId;
        /** Index saHpiSensorNum */
        annunciator_oid[2] = rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum;
        /* assign the indices to the index */
        annunciator_index.oids = (oid *) & annunciator_oid;

        /* create full oid on This row for parent RowPointer */
        column[0] = 1;
        column[1] = COLUMN_SAHPIANNUNCIATORTYPE;
        memset(child_oid, 0, MAX_OID_LEN);
        build_full_oid(saHpiAnnunciatorTable_oid, saHpiAnnunciatorTable_oid_len,
                       column, column_len,
                       &annunciator_index,
                       child_oid, MAX_OID_LEN, child_oid_len);

        /* See if Row exists. */
        annunciator_context = NULL;
        annunciator_context = CONTAINER_FIND(cb.container, &annunciator_index);

        if (!annunciator_context) {
                // New entry. Add it
                annunciator_context = 
                saHpiAnnunciatorTable_create_row(&annunciator_index);
                new_row = MIB_TRUE;
        }
        if (!annunciator_context) {
                snmp_log (LOG_ERR, "Not enough memory for a Annunciator row!");
		
		subagent_unlock(&hpi_lock_data);
		
                return AGENT_ERR_INTERNAL_ERROR;
        }

        /** SaHpiInstrumentId = ASN_UNSIGNED */
        annunciator_context->saHpiAnnunciatorNum =
        rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum;

        /** INTEGER = ASN_INTEGER */
        annunciator_context->saHpiAnnunciatorType =
        rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorType + 1;

        /** TruthValue = ASN_INTEGER */
        annunciator_context->saHpiAnnunciatorModeReadOnly = 
        (rdr_entry->RdrTypeUnion.AnnunciatorRec.ModeReadOnly == SAHPI_TRUE)
        ? MIB_TRUE : MIB_FALSE;

        /** UNSIGNED32 = ASN_UNSIGNED */
        annunciator_context->saHpiAnnunciatorMaxConditions =
        rdr_entry->RdrTypeUnion.AnnunciatorRec.MaxConditions;

        /** INTEGER = ASN_INTEGER */
        rv = saHpiAnnunciatorModeGet(sessionid, rpt_entry->ResourceId,
                                     rdr_entry->RdrTypeUnion.AnnunciatorRec.AnnunciatorNum,
                                     &mode);
        if (rv == SA_OK) {
                annunciator_context->saHpiAnnunciatorMode = mode + 1;
        } else {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_annunciator() saHpiAnnunciatorModeGet() ERRORED out\n"));
                saHpiAnnunciatorTable_delete_row( annunciator_context );
		
		subagent_unlock(&hpi_lock_data);
		
                return AGENT_ERR_INTERNAL_ERROR;
        }

        /** UNSIGNED32 = ASN_UNSIGNED */
        annunciator_context->saHpiAnnunciatorOem =
        rdr_entry->RdrTypeUnion.AnnunciatorRec.Oem;

        /** RowPointer = ASN_OBJECT_ID */
        memset(annunciator_context->saHpiAnnunciatorRDR, 
               0, 
               sizeof(annunciator_context->saHpiAnnunciatorRDR));
        annunciator_context->saHpiAnnunciatorRDR_len = 
        full_oid_len * sizeof(oid);
        memcpy(annunciator_context->saHpiAnnunciatorRDR, 
               full_oid, 
               annunciator_context->saHpiAnnunciatorRDR_len);

        if (new_row == MIB_TRUE) 
                CONTAINER_INSERT (cb.container, annunciator_context);

        annunciator_entry_count = CONTAINER_SIZE (cb.container);

	rc = populate_saHpiAnnouncementTable(sessionid, rdr_entry, rpt_entry);
	
	subagent_unlock(&hpi_lock_data);
			      
        DEBUGMSGTL ((AGENT, "populate_annunciator: populate_saHpiAnnouncementTable returned %d\n", rc));			      

        return rv;
} 

/**
 * 
 * @domainId
 * @resourceId
 * 
 * @return 
 */
SaErrorT clear_annunciator(SaHpiDomainIdT domainId, 
                           SaHpiResourceIdT resourceId)

{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiAnnunciatorTable_context *ctx;

	DEBUGMSGTL ((AGENT, "clear_annunciator, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domainId));	
	DEBUGMSGTL ((AGENT, "           resourceId [%d]\n", resourceId));

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        /* based on the found row_idx get the pointer   */
                        /* to its context (row data)                    */
                        ctx = CONTAINER_FIND(cb.container, row_idx);

                        /* before we delete the context we should get the  */
                        /* next row (context) if any before we delete this */ 
                        /* one.                                            */
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if ((ctx->index.oids[saHpiDomainId_INDEX] ==
                             domainId) &&

                            (ctx->index.oids[saHpiResourceEntryId_INDEX] ==
                             resourceId)) {

                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctx);
                                saHpiAnnunciatorTable_delete_row (ctx);
                                annunciator_entry_count = 
                                        CONTAINER_SIZE (cb.container);
                                DEBUGMSGTL ((AGENT, "clear_annunciator: "
                                                    "found row: removing\n"));

                        }

                } while (row_idx);
        } 

        return rv;
}

int set_table_annun_mode (saHpiAnnunciatorTable_context *row_ctx)
{

        DEBUGMSGTL ((AGENT, "set_table_annun_mode, called\n"));

        SaErrorT                rc = SA_OK;
        SaHpiSessionIdT         session_id;
        SaHpiResourceIdT        resource_id;
        SaHpiAnnunciatorModeT   mode;   

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
        resource_id = row_ctx->index.oids[saHpiResourceEntryId_INDEX];
        mode =  row_ctx->saHpiAnnunciatorMode - 1;

        rc = saHpiAnnunciatorModeSet(session_id, 
                                     resource_id,
                                     row_ctx->saHpiAnnunciatorNum,
                                     mode);

        if (rc != SA_OK) {
                snmp_log (LOG_ERR,
                          "SAHPI_ANNUNCIATOR_RDR: Call to saHpiAnnunciatorModeSet failed to set Mode rc: %s.\n",
                          oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                             "SAHPI_ANNUNCIATOR_RDR: Call to saHpiAnnunciatorModeSet failed to set Mode rc: %s.\n",
                             oh_lookup_error(rc)));
                return get_snmp_error(rc);
        }

        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiAnnunciatorActiveEntries(netsnmp_mib_handler *handler,
                                  netsnmp_handler_registration *reginfo,
                                  netsnmp_agent_request_info   *reqinfo,
                                  netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiAnnunciatorActiveEntries, called\n"));

	annunciator_entry_count = CONTAINER_SIZE (cb.container);

        switch (reqinfo->mode) {
        
        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
                                         (u_char *) &annunciator_entry_count,
                                         sizeof(annunciator_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}


/**
 * 
 * @return: 
 */
int initialize_table_saHpiAnnunciatorActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiSensorActiveEntries, called\n"));

        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
                                       "saHpiAnnunciatorActiveEntries", 
                                       handle_saHpiAnnunciatorActiveEntries,
                                       saHpiAnnunciatorActiveEntries_oid, 
                                       OID_LENGTH(saHpiAnnunciatorActiveEntries_oid),
                                       HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiAnnunciatorTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiAnnunciatorTable_cmp( const void *lhs, const void *rhs )
{
        saHpiAnnunciatorTable_context *context_l =
        (saHpiAnnunciatorTable_context *)lhs;
        saHpiAnnunciatorTable_context *context_r =
        (saHpiAnnunciatorTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiAnnunciatorTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiAnnunciatorNum */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                return 0;
                        }
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiAnnunciatorTable module
 */
void
init_saHpiAnnunciatorTable(void)
{

        DEBUGMSGTL ((AGENT, "init_saHpiAnnunciatorTable, called\n"));

        initialize_table_saHpiAnnunciatorTable();

        initialize_table_saHpiAnnunciatorActiveEntries();

}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiAnnunciatorTable_row_copy(saHpiAnnunciatorTable_context * dst,
                                          saHpiAnnunciatorTable_context * src)
{
        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_row_copy, called\n"));

        if (!dst||!src)
                return 1;

        
	subagent_lock(&hpi_lock_data);
	/*
         * copy index, if provided
         */
        if (dst->index.oids)
                free(dst->index.oids);
        if (snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                            src->index.len * sizeof(oid) )) {
                dst->index.oids = NULL;
                subagent_unlock(&hpi_lock_data);
		return 1;
        }
        dst->index.len = src->index.len;


        /*
         * copy components into the context structure
         */
        /** TODO: add code for external index(s)! */
        dst->saHpiAnnunciatorNum = src->saHpiAnnunciatorNum;

        dst->saHpiAnnunciatorType = src->saHpiAnnunciatorType;

        dst->saHpiAnnunciatorModeReadOnly = src->saHpiAnnunciatorModeReadOnly;

        dst->saHpiAnnunciatorMaxConditions = src->saHpiAnnunciatorMaxConditions;

        dst->saHpiAnnunciatorMode = src->saHpiAnnunciatorMode;

        dst->saHpiAnnunciatorOem = src->saHpiAnnunciatorOem;

        memcpy( dst->saHpiAnnunciatorRDR, src->saHpiAnnunciatorRDR, src->saHpiAnnunciatorRDR_len );
        dst->saHpiAnnunciatorRDR_len = src->saHpiAnnunciatorRDR_len;

        subagent_unlock(&hpi_lock_data);
	return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiAnnunciatorTable_extract_index( saHpiAnnunciatorTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiAnnunciatorNum;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_extract_index, called\n"));

        
	subagent_lock(&hpi_lock_data);
	
	/*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
			return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiAnnunciatorNum;

        memset( &var_saHpiAnnunciatorNum, 0x00, sizeof(var_saHpiAnnunciatorNum) );
        var_saHpiAnnunciatorNum.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiAnnunciatorNum.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */

                /** skipping external index saHpiResourceId */


                ctx->saHpiAnnunciatorNum = *var_saHpiAnnunciatorNum.val.integer;

                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
                err = saHpiAnnunciatorNum_check_index(*var_saHpiAnnunciatorNum.val.integer);    
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
	
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiAnnunciatorTable_can_activate(saHpiAnnunciatorTable_context *undo_ctx,
                                       saHpiAnnunciatorTable_context *row_ctx,
                                       netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_can_activate, called\n"));

        /*
         * TODO: check for activation requirements here
         */


        /*
         * be optimistic.
         */
        return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiAnnunciatorTable_can_deactivate(saHpiAnnunciatorTable_context *undo_ctx,
                                         saHpiAnnunciatorTable_context *row_ctx,
                                         netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_can_deactivate, called\n"));
        /*
         * TODO: check for deactivation requirements here
         */
        return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiAnnunciatorTable_can_delete(saHpiAnnunciatorTable_context *undo_ctx,
                                     saHpiAnnunciatorTable_context *row_ctx,
                                     netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_can_delete, called\n"));
        /*
         * probably shouldn't delete a row that we can't
         * deactivate.
         */
        if (saHpiAnnunciatorTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
                return 0;

        /*
         * TODO: check for other deletion requirements here
         */
        return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiAnnunciatorTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiAnnunciatorTable_context *
saHpiAnnunciatorTable_create_row( netsnmp_index* hdr)
{
        
	subagent_lock(&hpi_lock_data);

	saHpiAnnunciatorTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiAnnunciatorTable_context);

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_create_row, called\n"));

        if (!ctx) {
                subagent_unlock(&hpi_lock_data);
	        return NULL;
	}	

        /*
         * TODO: check indexes, if necessary.
         */
	
	if (saHpiAnnunciatorTable_extract_index( ctx, hdr )) {
                free(ctx->index.oids);
                free(ctx);
                subagent_unlock(&hpi_lock_data);
		return NULL;
        }

        /* netsnmp_mutex_init(ctx->lock);
           netsnmp_mutex_lock(ctx->lock); */

        /*
         * TODO: initialize any default values here. This is also
         * first place you really should allocate any memory for
         * yourself to use.  If you allocated memory earlier,
         * make sure you free it for earlier error cases!
         */
        /**
         ctx->saHpiAnnunciatorMode = 0;
        */

        
	subagent_unlock(&hpi_lock_data);
	return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiAnnunciatorTable_context *
saHpiAnnunciatorTable_duplicate_row( saHpiAnnunciatorTable_context * row_ctx)
{
        saHpiAnnunciatorTable_context * dup;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_duplicate_row, called\n"));

        if (!row_ctx)
                return NULL;

        subagent_lock(&hpi_lock_data);
	
	dup = SNMP_MALLOC_TYPEDEF(saHpiAnnunciatorTable_context);
        if (!dup) {
		subagent_unlock(&hpi_lock_data);
                return NULL;
	}	

        if (saHpiAnnunciatorTable_row_copy(dup,row_ctx)) {
                free(dup);
                dup = NULL;
        }

        subagent_unlock(&hpi_lock_data);
	
	return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiAnnunciatorTable_delete_row( saHpiAnnunciatorTable_context * ctx )
{
        /* netsnmp_mutex_destroy(ctx->lock); */

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_delete_row, called\n"));

        
	subagent_lock(&hpi_lock_data);
	
	if (ctx->index.oids)
                free(ctx->index.oids);

        /*
         * TODO: release any memory you allocated here...
         */

        /*
         * release header
         */
        free( ctx );

	subagent_unlock(&hpi_lock_data);
	
	return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiAnnunciatorTable_set_reserve1( netsnmp_request_group *rg )
{
        saHpiAnnunciatorTable_context *row_ctx =
        (saHpiAnnunciatorTable_context *)rg->existing_row;
//        saHpiAnnunciatorTable_context *undo_ctx =
//        (saHpiAnnunciatorTable_context *)rg->undo_info;
        netsnmp_variable_list *var;
        netsnmp_request_group_item *current;
        int rc;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_set_reserve1, called\n"));

        /*
         * TODO: loop through columns, check syntax and lengths. For
         * columns which have no dependencies, you could also move
         * the value/range checking here to attempt to catch error
         * cases as early as possible.
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;
                rc = SNMP_ERR_NOERROR;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIANNUNCIATORMODE:
                        /** INTEGER = ASN_INTEGER */
                        rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                            sizeof(row_ctx->saHpiAnnunciatorMode));
                        break;

                default: /** We shouldn't get here */
                        rc = SNMP_ERR_GENERR;
                        snmp_log(LOG_ERR, "unknown column in "
                                 "saHpiAnnunciatorTable_set_reserve1\n");
                }

                if (rc)
                        netsnmp_request_set_error( current->ri, rc );
                rg->status = SNMP_MAX( rg->status, current->ri->status );
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

void saHpiAnnunciatorTable_set_reserve2( netsnmp_request_group *rg )
{
//      saHpiAnnunciatorTable_context *row_ctx = (saHpiAnnunciatorTable_context *)rg->existing_row;
//      saHpiAnnunciatorTable_context *undo_ctx = (saHpiAnnunciatorTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;
        netsnmp_variable_list *var;
        int rc;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_set_reserve2, called\n"));
        rg->rg_void = rg->list->ri;

        /*
         * TODO: loop through columns, check for valid
         * values and any range constraints.
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;
                rc = SNMP_ERR_NOERROR;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIANNUNCIATORMODE:
                	if ( ((*var->val.integer - 1) != SAHPI_ANNUNCIATOR_MODE_AUTO) &&
                     	     ((*var->val.integer - 1) != SAHPI_ANNUNCIATOR_MODE_USER) &&
			     ((*var->val.integer - 1) != SAHPI_ANNUNCIATOR_MODE_SHARED) ) {
                        	DEBUGMSGTL ((AGENT, "COLUMN_SAHPIANNUNCIATORMODE"
                                             " saHpiAnnunciatorTable_set_reserve2"
                                             " Only modes of "
					     " SAHPI_ANNUNCIATOR_MODE_AUTO and"
					     " SAHPI_ANNUNCIATOR_MODE_USER and"
					     " SAHPI_ANNUNCIATOR_MODE_SHARED"
                                             " are allowed\n"
                                            ));
                                rc = SNMP_ERR_BADVALUE;
			}	
                break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }

                if (rc)
                        netsnmp_request_set_error( current->ri, rc);
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiAnnunciatorTable_set_action( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
        saHpiAnnunciatorTable_context *row_ctx = (saHpiAnnunciatorTable_context *)rg->existing_row;
//        saHpiAnnunciatorTable_context *undo_ctx = (saHpiAnnunciatorTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        int            row_err = 0;

        subagent_lock(&hpi_lock_data);

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_set_action, called\n"));
        /*
         * TODO: loop through columns, copy varbind values
         * to context structure for the row.
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIANNUNCIATORMODE:
                        /** INTEGER = ASN_INTEGER */
                        row_ctx->saHpiAnnunciatorMode = *var->val.integer;
                        row_err = set_table_annun_mode (row_ctx);
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }
        }

        if (row_err) {
                netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                               row_err);
        }

        subagent_unlock(&hpi_lock_data);

        return;

        /*
         * TODO: if you have dependencies on other tables, this would be
         * a good place to check those, too.
         */
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiAnnunciatorTable_set_commit( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
//        saHpiAnnunciatorTable_context *row_ctx = (saHpiAnnunciatorTable_context *)rg->existing_row;
//        saHpiAnnunciatorTable_context *undo_ctx = (saHpiAnnunciatorTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_set_commit, called\n"));
        /*
         * loop through columns
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIANNUNCIATORMODE:
                        /** INTEGER = ASN_INTEGER */
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiAnnunciatorTable_set_free( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
//        saHpiAnnunciatorTable_context *row_ctx = (saHpiAnnunciatorTable_context *)rg->existing_row;
//        saHpiAnnunciatorTable_context *undo_ctx = (saHpiAnnunciatorTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_set_free, called\n"));
        /*
         * loop through columns
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIANNUNCIATORMODE:
                        /** INTEGER = ASN_INTEGER */
                        break;

                default: 
                        break;
                        /** We shouldn't get here */
                        /** should have been logged in reserve1 */
                }
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiAnnunciatorTable_set_undo( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
//        saHpiAnnunciatorTable_context *row_ctx = (saHpiAnnunciatorTable_context *)rg->existing_row;
//        saHpiAnnunciatorTable_context *undo_ctx = (saHpiAnnunciatorTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_set_undo, called\n"));
        /*
         * loop through columns
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIANNUNCIATORMODE:
                        /** INTEGER = ASN_INTEGER */
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}


/************************************************************
 *
 * Initialize the saHpiAnnunciatorTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiAnnunciatorTable(void)
{
        netsnmp_table_registration_info *table_info;

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiAnnunciatorTable, called\n"));
        if (my_handler) {
                snmp_log(LOG_ERR, "initialize_table_saHpiAnnunciatorTable_handler called again\n");
                return;
        }

        memset(&cb, 0x00, sizeof(cb));

        /** create the table structure itself */
        table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

        /* if your table is read only, it's easiest to change the
           HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
        my_handler = netsnmp_create_handler_registration("saHpiAnnunciatorTable",
                                                         netsnmp_table_array_helper_handler,
                                                         saHpiAnnunciatorTable_oid,
                                                         saHpiAnnunciatorTable_oid_len,
                                                         HANDLER_CAN_RWRITE);

        if (!my_handler || !table_info) {
                snmp_log(LOG_ERR, "malloc failed in "
                         "initialize_table_saHpiAnnunciatorTable_handler\n");
                return; /** mallocs failed */
        }

        /***************************************************
         * Setting up the table's definition
         */
        /*
         * TODO: add any external indexes here.
         */
        /** TODO: add code for external index(s)! */

        /*
         * internal indexes
         */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiAnnunciatorNum */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

        table_info->min_column = saHpiAnnunciatorTable_COL_MIN;
        table_info->max_column = saHpiAnnunciatorTable_COL_MAX;

        /***************************************************
         * registering the table with the master agent
         */
        cb.get_value = saHpiAnnunciatorTable_get_value;
        cb.container = netsnmp_container_find("saHpiAnnunciatorTable_primary:"
                                              "saHpiAnnunciatorTable:"
                                              "table_container");

        netsnmp_container_add_index(cb.container,
                                    netsnmp_container_find("saHpiAnnunciatorTable_secondary:"
                                                           "saHpiAnnunciatorTable:"
                                                           "table_container"));
        cb.container->next->compare = saHpiAnnunciatorTable_cmp;


        cb.can_set = 1;

        cb.create_row = (UserRowMethod*)saHpiAnnunciatorTable_create_row;

        cb.duplicate_row = (UserRowMethod*)saHpiAnnunciatorTable_duplicate_row;
        cb.delete_row = (UserRowMethod*)saHpiAnnunciatorTable_delete_row;
        cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiAnnunciatorTable_row_copy;

        cb.can_activate = (Netsnmp_User_Row_Action *)saHpiAnnunciatorTable_can_activate;
        cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiAnnunciatorTable_can_deactivate;
        cb.can_delete = (Netsnmp_User_Row_Action *)saHpiAnnunciatorTable_can_delete;

        cb.set_reserve1 = saHpiAnnunciatorTable_set_reserve1;
        cb.set_reserve2 = saHpiAnnunciatorTable_set_reserve2;
        cb.set_action = saHpiAnnunciatorTable_set_action;
        cb.set_commit = saHpiAnnunciatorTable_set_commit;
        cb.set_free = saHpiAnnunciatorTable_set_free;
        cb.set_undo = saHpiAnnunciatorTable_set_undo;

        DEBUGMSGTL(("initialize_table_saHpiAnnunciatorTable",
                    "Registering table saHpiAnnunciatorTable "
                    "as a table array\n"));
        netsnmp_table_container_register(my_handler, table_info, &cb,
                                         cb.container, 1);
}

/************************************************************
 * saHpiAnnunciatorTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiAnnunciatorTable_get_value(
                                   netsnmp_request_info *request,
                                   netsnmp_index *item,
                                   netsnmp_table_request_info *table_info )
{
        netsnmp_variable_list *var = request->requestvb;
        saHpiAnnunciatorTable_context *context = (saHpiAnnunciatorTable_context *)item;

        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_get_value, called\n"));
        switch (table_info->colnum) {
        
        case COLUMN_SAHPIANNUNCIATORNUM:
                /** SaHpiInstrumentId = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                                         (u_char*)&context->saHpiAnnunciatorNum,
                                         sizeof(context->saHpiAnnunciatorNum) );
                break;

        case COLUMN_SAHPIANNUNCIATORTYPE:
                /** INTEGER = ASN_INTEGER */
                snmp_set_var_typed_value(var, ASN_INTEGER,
                                         (u_char*)&context->saHpiAnnunciatorType,
                                         sizeof(context->saHpiAnnunciatorType) );
                break;

        case COLUMN_SAHPIANNUNCIATORMODEREADONLY:
                /** TruthValue = ASN_INTEGER */
                snmp_set_var_typed_value(var, ASN_INTEGER,
                                         (u_char*)&context->saHpiAnnunciatorModeReadOnly,
                                         sizeof(context->saHpiAnnunciatorModeReadOnly) );
                break;

        case COLUMN_SAHPIANNUNCIATORMAXCONDITIONS:
                /** UNSIGNED32 = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                                         (u_char*)&context->saHpiAnnunciatorMaxConditions,
                                         sizeof(context->saHpiAnnunciatorMaxConditions) );
                break;

        case COLUMN_SAHPIANNUNCIATORMODE:
                /** INTEGER = ASN_INTEGER */
                snmp_set_var_typed_value(var, ASN_INTEGER,
                                         (u_char*)&context->saHpiAnnunciatorMode,
                                         sizeof(context->saHpiAnnunciatorMode) );
                break;

        case COLUMN_SAHPIANNUNCIATOROem:
                /** UNSIGNED32 = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                                         (u_char*)&context->saHpiAnnunciatorOem,
                                         sizeof(context->saHpiAnnunciatorOem) );
                break;

        case COLUMN_SAHPIANNUNCIATORRDR:
                /** RowPointer = ASN_OBJECT_ID */
                snmp_set_var_typed_value(var, ASN_OBJECT_ID,
                                         (u_char*)&context->saHpiAnnunciatorRDR,
                                         context->saHpiAnnunciatorRDR_len );
                break;

        default: /** We shouldn't get here */
                snmp_log(LOG_ERR, "unknown column in "
                         "saHpiAnnunciatorTable_get_value\n");
                return SNMP_ERR_GENERR;
        }
        return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiAnnunciatorTable_get_by_idx
 */
const saHpiAnnunciatorTable_context *
saHpiAnnunciatorTable_get_by_idx(netsnmp_index * hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiAnnunciatorTable_get_by_idx, called\n"));
        return(const saHpiAnnunciatorTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


