/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiAreaTable.c 5419 2005-11-14 20:37:28Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiAreaTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceTable.h>
#include <saHpiFieldTable.h>
#include <saHpiInventoryTable.h>
#include <session_info.h>
#include <hash_utils.h>
#include <oh_utils.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiAreaTable_oid[] = { saHpiAreaTable_TABLE_OID};
size_t saHpiAreaTable_oid_len = OID_LENGTH(saHpiAreaTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;               
static GHashTable *dri_table;


/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long area_entry_count = 0;
static oid saHpiAreaActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,4,8,3};
int handle_saHpiAreaActiveEntries(netsnmp_mib_handler *handler,
                               netsnmp_handler_registration *reginfo,
                               netsnmp_agent_request_info   *reqinfo,
                               netsnmp_request_info         *requests);
int initialize_table_saHpiAreaActiveEntries(void);

/**
 * 
 * @sessionid:
 * @rdr_entry:
 * @rpt_entry:
 * @full_oid:
 * @full_oid_len:
 * @child_oid:
 * @child_oid_len:
 * 
 * @return 
 */
SaErrorT populate_area (SaHpiSessionIdT sessionid, 
                        SaHpiRdrT *rdr_entry,
                        SaHpiRptEntryT *rpt_entry)
{

        DEBUGMSGTL ((AGENT, "populate_area, called\n"));

        SaErrorT rv = SA_OK;
        SaHpiEntryIdT area_id;
        SaHpiIdrAreaHeaderT header;

        DRI_XREF *dri_entry;
        SaHpiDomainIdResourceIdInventoryIdArrayT dri_tuple;

        oid area_oid[AREA_INDEX_NR];
        netsnmp_index area_index;
        saHpiAreaTable_context *area_context;

        SaHpiEntryIdT subagent_area_id;

        DEBUGMSGTL ((AGENT, "SAHPI_ANNUNCIATOR_RDR populate_area() called\n"));

        /* check for NULL pointers */
        if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_area() passed NULL rdr_entry pointer\n"));
			     		 
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_area() passed NULL rdr_entry pointer\n"));
			     			     
                return AGENT_ERR_INTERNAL_ERROR;
        }


        area_id = SAHPI_FIRST_ENTRY;
	
	subagent_lock(&hpi_lock_data);
	
        do {
                rv = saHpiIdrAreaHeaderGet(sessionid, 
                                           rpt_entry->ResourceId, 
                                           rdr_entry->RdrTypeUnion.InventoryRec.IdrId,
                                           SAHPI_IDR_AREATYPE_UNSPECIFIED,
                                           area_id,
                                           &area_id,
                                           &header);
                if (rv != SA_OK) {
                        DEBUGMSGTL ((AGENT, "saHpiIdrAreaHeaderGet Failed: rv = %d\n",rv));
                        rv =  AGENT_ERR_INTERNAL_ERROR;
                        break;
                }

                /* Create MIB AreaId value                              */
                /* this is not the same as the AreaId obtained from HPI */
                /* it is maintained in the subagent                     */
                dri_tuple.domainId_resourceId_idr_arry[0] = get_domain_id(sessionid);
                dri_tuple.domainId_resourceId_idr_arry[1] = rpt_entry->ResourceId;
                dri_tuple.domainId_resourceId_idr_arry[2] = rdr_entry->RdrTypeUnion.InventoryRec.IdrId;

                /* domain_resource_idr_get() generates unique keys based on */
                /* domainid, resourceid, and IdrId tuples.                  */
                dri_entry = domain_resource_idr_get(&dri_tuple, &dri_table); 

                if (dri_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: populate_area() domain_resource_idr_get returned NULL\n"));
				     
		  	subagent_unlock(&hpi_lock_data);
			
                        return AGENT_ERR_INTERNAL_ERROR;
                }
                subagent_area_id = dri_entry->entry_id++;

                /* BUILD oid for new row */
                /* assign the number of indices */
                area_index.len = AREA_INDEX_NR;
                /** Index saHpiDomainId is external */
                area_oid[0] = get_domain_id(sessionid);
                /** Index saHpiResourceId is external */
                area_oid[1] = rpt_entry->ResourceId;
                /** Index saHpiResourceIsHistorical is external */
                area_oid[2] = MIB_FALSE;
                /** Index saHpiInventoryId */
                area_oid[3] = rdr_entry->RdrTypeUnion.InventoryRec.IdrId;
                /** Index saHpiAreaId */
                area_oid[4] = subagent_area_id;
                /* assign the indices to the index */
                area_index.oids = (oid *) & area_oid;

                /* See if Row exists. */
                area_context = NULL;
                area_context = CONTAINER_FIND(cb.container, &area_index);

                if (!area_context) {
                        // New entry. Add it
                        area_context = 
                        saHpiAreaTable_create_row(&area_index);
                }
                if (!area_context) {
                        snmp_log (LOG_ERR, "Not enough memory for a Area row!");
			
			subagent_unlock(&hpi_lock_data);
			
                        return AGENT_ERR_INTERNAL_ERROR;
                }

                /** SaHpiInstrumentId = ASN_UNSIGNED */
                area_context->saHpiAreaId = subagent_area_id;

                /** SaHpiInstrumentId = ASN_UNSIGNED */
                area_context->saHpiAreaIdIndex = header.AreaId;

                /** INTEGER = ASN_INTEGER */
                area_context->saHpiAreaType = header.Type + 1;

                /** TruthValue = ASN_INTEGER */
                area_context->saHpiAreaIsReadOnly =
                (header.ReadOnly == SAHPI_TRUE) ? 
                MIB_TRUE : MIB_FALSE;

                /** RowStatus = ASN_INTEGER */
                area_context->saHpiAreaRowStatus = SNMP_ROW_ACTIVE;

                /** UNSIGNED32 = ASN_UNSIGNED */
                area_context->saHpiAreaNumDataFields = header.NumFields;

                populate_field (sessionid, rdr_entry, rpt_entry, area_context);

                CONTAINER_INSERT (cb.container, area_context);

        } while (area_id !=  SAHPI_LAST_ENTRY );

        area_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);
        
	return rv;
}

/**
 * 
 * @domainId
 * @resourceId
 * @idrId
 * 
 * @return 
 */
SaErrorT clear_area(SaHpiDomainIdT   domainId, 
                    SaHpiResourceIdT resourceId,
                    SaHpiIdrIdT      idrId)

{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiAreaTable_context *ctx;

        DRI_XREF *dri_entry;
        SaHpiDomainIdResourceIdInventoryIdArrayT dri_tuple;

	DEBUGMSGTL ((AGENT, "area_inventory, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domainId));	
	DEBUGMSGTL ((AGENT, "           resourceId [%d]\n", resourceId));
	DEBUGMSGTL ((AGENT, "           idrId      [%d]\n", idrId));

        /***************************/
        /* build the key to lookup */
        /***************************/
        dri_tuple.domainId_resourceId_idr_arry[0] = domainId;
        dri_tuple.domainId_resourceId_idr_arry[1] = resourceId;
        dri_tuple.domainId_resourceId_idr_arry[2] = idrId;

        /* domain_resource_idr_get() generates unique keys based on */
        /* domainid, resourceid, and IdrId tuples.                  */
        dri_entry = domain_resoruce_idr_lookup(&dri_tuple, &dri_table);

        if (dri_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                             "INFO: area_inventory() domain_resource_idr_get"
                             " returned NULL\n"));
                rv = SA_ERR_HPI_NOT_PRESENT;
        } else {
                dri_entry->entry_id = 0;
        }

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        /* based on the found row_idx get the pointer   */
                        /* to its context (row data)                    */
                        ctx = CONTAINER_FIND(cb.container, row_idx);

                        /* before we delete the context we should get the  */
                        /* next row (context) if any before we delete this */ 
                        /* one.                                            */
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if ((ctx->index.oids[saHpiDomainId_INDEX] ==
                             domainId) &&

                            (ctx->index.oids[saHpiResourceEntryId_INDEX] ==
                             resourceId)) {                     

                                /********************************************/
                                /* Clear all the under lying tables as well */
                                /********************************************/
                                rv = clear_field(domainId, 
                                                 resourceId, 
                                                 idrId,
                                                 ctx->saHpiAreaId);

                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctx);
                                saHpiAreaTable_delete_row (ctx);
                                area_entry_count = 
                                        CONTAINER_SIZE (cb.container);
                                DEBUGMSGTL ((AGENT, "area_inventory: "
                                                    "found row: removing\n"));
                        }

                } while (row_idx);
        } 

        return rv;
}

/**
 * 
 * @field_oids:
 * @num_fields:
 * 
 * @return:
 */
int update_num_data_fields ( oid *field_oids,  SaHpiUint32T num_fields)
{
        oid area_oid[AREA_INDEX_NR];
        netsnmp_index area_index;
        saHpiAreaTable_context *area_context;

        DEBUGMSGTL ((AGENT, "update_num_data_fields, called\n"));

        /* BUILD oid for new row */
        /* assign the number of indices */
        area_index.len = AREA_INDEX_NR;
        /** Index saHpiDomainId is external */
        area_oid[0] = field_oids[saHpiDomainId_field_INDEX];
        /** Index saHpiResourceId is external */
        area_oid[1] = field_oids[saHpiResourceEntryId_field_INDEX];
        /** Index saHpiResourceIsHistorical is external */
        area_oid[2] = field_oids[saHpiResourceIsHistorical_field_INDEX];
        /** Index saHpiInventoryId */                        
        area_oid[3] = field_oids[saHpiInventoryId_field_INDEX];
        /** Index saHpiAreaId */
        area_oid[4] = field_oids[saHpiAreaId_field_INDEX];
        /* assign the indices to the index */
        area_index.oids = (oid *) & area_oid;

        /* See if Row exists. */
        area_context = NULL;
        area_context = CONTAINER_FIND(cb.container, &area_index); 
        if (!area_context) {
                return SA_ERR_HPI_UNKNOWN;
        }

        area_context->saHpiAreaNumDataFields = num_fields;

        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int set_table_area_delete (saHpiAreaTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "********************************************\n"));
        DEBUGMSGTL ((AGENT, "*** NEED TO DELETE ALL ASSOCIATED FIELDS ***\n"));
        DEBUGMSGTL ((AGENT, "********************************************\n"));

        DEBUGMSGTL ((AGENT, "set_table_area_delete, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiIdrIdT         idr_id;    
        SaHpiEntryIdT       area_id;    

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
        resource_id = row_ctx->index.oids[saHpiResourceEntryId_INDEX];
        idr_id = row_ctx->index.oids[saHpiInventoryId_INDEX];
        area_id = row_ctx->saHpiAreaIdIndex;

        rc = saHpiIdrAreaDelete(session_id, resource_id, idr_id, area_id);
	
        if (rc == SA_OK) {
	
		rc = decrement_area_num(session_id, resource_id, idr_id);
		
                rc = delete_fields(session_id, resource_id, idr_id, row_ctx->saHpiAreaId);

                return SNMP_ERR_NOERROR; 
		
        } else {
                snmp_log (LOG_ERR,
                          "set_table_area_delete: Call to saHpiIdrAreaDelete failed %s.\n",
                          oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                             "set_table_area_delete: Call to saHpiIdrAreaDelete failed %s.\n",
                             oh_lookup_error(rc)));
                return get_snmp_error(rc);
        }

        return SNMP_ERR_NOERROR; 
}


/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int set_table_area_type (saHpiAreaTable_context *row_ctx)
{

        DEBUGMSGTL ((AGENT, "set_table_area_type, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiIdrIdT         idr_id;    
        SaHpiIdrAreaTypeT   area_type;    
        SaHpiEntryIdT       area_id;
        SaHpiIdrInfoT       idr_info;    

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
        resource_id = row_ctx->index.oids[saHpiResourceEntryId_INDEX];
        idr_id = row_ctx->index.oids[saHpiInventoryId_INDEX];
        area_id = row_ctx->saHpiAreaIdIndex;
        area_type = row_ctx->saHpiAreaType - 1;

        rc = saHpiIdrAreaAdd(session_id, resource_id, idr_id, area_type, &area_id);

        if (rc == SA_OK) {
                row_ctx->saHpiAreaIdIndex = area_id;

                /* Get Idr Info to set the AreaHeaderReadOnly Flag */
                rc = saHpiIdrInfoGet(session_id, resource_id, idr_id, &idr_info);
                if (rc == SA_OK) {
                        row_ctx->saHpiAreaIsReadOnly = 
                                (idr_info.ReadOnly == SAHPI_TRUE) ? 
                                        MIB_TRUE : MIB_FALSE;
                        return SNMP_ERR_NOERROR; 
                } else if (rc != SA_OK) {
                        snmp_log (LOG_ERR,
                        "set_table_area_type: Call to saHpiIdrInfoGet() failed to set Mode rc: %s.\n",
                        oh_lookup_error(rc));
                        DEBUGMSGTL ((AGENT,
                        "set_table_area_type: Call to saHpiIdrInfoGet() failed to set Mode rc: %s.\n",
                        oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }

        } else if (rc != SA_OK) {
                snmp_log (LOG_ERR,
                          "set_table_area_type: Call to saHpiIdrAreaAdd() failed to set Mode rc: %s.\n",
                          oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                             "set_table_area_type: Call to saHpiIdrAreaAdd() failed to set Mode rc: %s.\n",
                             oh_lookup_error(rc)));
                return get_snmp_error(rc);
        }

        return SNMP_ERR_NOERROR; 
}


/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int handle_saHpiAreaActiveEntries(netsnmp_mib_handler *handler,
                               netsnmp_handler_registration *reginfo,
                               netsnmp_agent_request_info   *reqinfo,
                               netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */

        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiAreaActiveEntries, called\n"));

	area_entry_count = CONTAINER_SIZE (cb.container);

        switch (reqinfo->mode) {
        
        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
                                         (u_char *) &area_entry_count,
                                         sizeof(area_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @return: 
 */
int initialize_table_saHpiAreaActiveEntries(void)
{
        DEBUGMSGTL ((AGENT, "initialize_table_saHpiAreaActiveEntries, called\n"));

        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
                                                                  "saHpiAreaActiveEntries", 
                                                                  handle_saHpiAreaActiveEntries,
                                                                  saHpiAreaActiveEntries_oid, 
                                                                  OID_LENGTH(saHpiAreaActiveEntries_oid),
                                                                  HANDLER_CAN_RONLY ));

        return 0;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/



/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiAreaTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiAreaTable_cmp( const void *lhs, const void *rhs )
{
        saHpiAreaTable_context *context_l =
        (saHpiAreaTable_context *)lhs;
        saHpiAreaTable_context *context_r =
        (saHpiAreaTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiAreaTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiAreaTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiResourceIsHistorical */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                /* If saHpiResourceIsHistorical index is equal sort by forth index */
                                /* CHECK FORTH INDEX,  saHpiInventoryId */
                                if ( context_l->index.oids[3] < context_r->index.oids[3])
                                        return -1;

                                if ( context_l->index.oids[3] > context_r->index.oids[3])
                                        return 1;

                                if ( context_l->index.oids[3] == context_r->index.oids[3]) {
                                        /* If saHpiInventoryId index is equal sort by forth index */
                                        /* CHECK FIFTH INDEX,  saHpiAreaId */
                                        if ( context_l->index.oids[4] < context_r->index.oids[4])
                                                return -1;

                                        if ( context_l->index.oids[4] > context_r->index.oids[4])
                                                return 1;

                                        if ( context_l->index.oids[4] == context_r->index.oids[4])
                                                return 0;

                                }
                        }
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiAreaTable module
 */
void
init_saHpiAreaTable(void)
{
        DEBUGMSGTL ((AGENT, "init_saHpiAreaTable, called\n"));

        initialize_table_saHpiAreaTable();

        initialize_table_saHpiAreaActiveEntries();

        domain_resource_idr_initialize(&initialized, &dri_table);

}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiAreaTable_row_copy(saHpiAreaTable_context * dst,
                                   saHpiAreaTable_context * src)
{
        if (!dst||!src)
                return 1;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_row_copy, called\n"));
        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (dst->index.oids)
                free(dst->index.oids);
        if (snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                            src->index.len * sizeof(oid) )) {
                dst->index.oids = NULL;
                subagent_unlock(&hpi_lock_data);
                return 1;
        }
        dst->index.len = src->index.len;


        /*
         * copy components into the context structure
         */
        /** TODO: add code for external index(s)! */
        dst->saHpiAreaId = src->saHpiAreaId;

        dst->saHpiAreaIdIndex = src->saHpiAreaIdIndex;

        dst->saHpiAreaType = src->saHpiAreaType;

        dst->saHpiAreaIsReadOnly = src->saHpiAreaIsReadOnly;

        dst->saHpiAreaRowStatus = src->saHpiAreaRowStatus;

        dst->saHpiAreaNumDataFields = src->saHpiAreaNumDataFields;

        subagent_unlock(&hpi_lock_data);
        return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiAreaTable_extract_index( saHpiAreaTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiResourceIsHistorical;
        netsnmp_variable_list var_saHpiInventoryId;
        netsnmp_variable_list var_saHpiAreaId;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_extract_index, called\n"));
        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
                        return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiResourceIsHistorical;

        memset( &var_saHpiResourceIsHistorical, 0x00, sizeof(var_saHpiResourceIsHistorical) );
        var_saHpiResourceIsHistorical.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceIsHistorical.next_variable = &var_saHpiInventoryId;

        memset( &var_saHpiInventoryId, 0x00, sizeof(var_saHpiInventoryId) );
        var_saHpiInventoryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiInventoryId.next_variable = &var_saHpiAreaId;

        memset( &var_saHpiAreaId, 0x00, sizeof(var_saHpiAreaId) );
        var_saHpiAreaId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiAreaId.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */

                /** skipping external index saHpiResourceId */

                /** skipping external index saHpiResourceIsHistorical */

                /** skipping external index saHpiInventoryId */

                ctx->saHpiAreaId = *var_saHpiAreaId.val.integer;


                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
                err = saHpiResourceIsHistorical_check_index(*var_saHpiResourceIsHistorical.val.integer);
                err = saHpiInventoryId_check_index(*var_saHpiInventoryId.val.integer);
                err = saHpiAreaId_check_index(*var_saHpiAreaId.val.integer);
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );
        subagent_unlock(&hpi_lock_data);

        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiAreaTable_can_activate(saHpiAreaTable_context *undo_ctx,
                                saHpiAreaTable_context *row_ctx,
                                netsnmp_request_group * rg)
{

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_can_activate, called\n"));

        /*
         * TODO: check for activation requirements here
         */


        /*
         * be optimistic.
         */
        return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiAreaTable_can_deactivate(saHpiAreaTable_context *undo_ctx,
                                  saHpiAreaTable_context *row_ctx,
                                  netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiAreaTable_can_deactivate, called\n"));

        /*
         * TODO: check for deactivation requirements here
         */
        return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiAreaTable_can_delete(saHpiAreaTable_context *undo_ctx,
                              saHpiAreaTable_context *row_ctx,
                              netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiAreaTable_can_delete, called\n"));

        /*
         * probably shouldn't delete a row that we can't
         * deactivate.
         */
        if (saHpiAreaTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
                return 0;

        /*
         * TODO: check for other deletion requirements here
         */
        return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiAreaTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiAreaTable_context *
saHpiAreaTable_create_row( netsnmp_index* hdr)
{
        saHpiAreaTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiAreaTable_context);

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_create_row, called\n"));
        subagent_lock(&hpi_lock_data);

        if (!ctx) {
                subagent_unlock(&hpi_lock_data);
                return NULL;
        }

        /*
         * TODO: check indexes, if necessary.
         */
        if (saHpiAreaTable_extract_index( ctx, hdr )) {
                free(ctx->index.oids);
                free(ctx);
                subagent_unlock(&hpi_lock_data);
                return NULL;
        }

        /* netsnmp_mutex_init(ctx->lock);
           netsnmp_mutex_lock(ctx->lock); */

        /*
         * TODO: initialize any default values here. This is also
         * first place you really should allocate any memory for
         * yourself to use.  If you allocated memory earlier,
         * make sure you free it for earlier error cases!
         */
        /**
         ctx->saHpiAreaType = 0;
         ctx->saHpiAreaRowStatus = 0;
        */

        subagent_unlock(&hpi_lock_data);
        return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiAreaTable_context *
saHpiAreaTable_duplicate_row( saHpiAreaTable_context * row_ctx)
{
        DEBUGMSGTL ((AGENT, "saHpiAreaTable_duplicate_row, called\n"));
        subagent_lock(&hpi_lock_data);

        saHpiAreaTable_context * dup;

        if (!row_ctx) {
                subagent_unlock(&hpi_lock_data);
                return NULL;
        }

        dup = SNMP_MALLOC_TYPEDEF(saHpiAreaTable_context);
        if (!dup) {
                subagent_unlock(&hpi_lock_data);
                return NULL;
        }

        if (saHpiAreaTable_row_copy(dup,row_ctx)) {
                free(dup);
                dup = NULL;
        }

        subagent_unlock(&hpi_lock_data);
        return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiAreaTable_delete_row( saHpiAreaTable_context * ctx )
{
        /* netsnmp_mutex_destroy(ctx->lock); */

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_delete_row, called\n"));
        subagent_lock(&hpi_lock_data);

        if (ctx->index.oids)
                free(ctx->index.oids);

        /*
         * TODO: release any memory you allocated here...
         */

        /*
         * release header
         */
        free( ctx );
        subagent_unlock(&hpi_lock_data);

        return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiAreaTable_set_reserve1( netsnmp_request_group *rg )
{
        saHpiAreaTable_context *row_ctx =
        (saHpiAreaTable_context *)rg->existing_row;
//    saHpiAreaTable_context *undo_ctx =
//            (saHpiAreaTable_context *)rg->undo_info;
        netsnmp_variable_list *var;
        netsnmp_request_group_item *current;
        int rc;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_set_reserve1, called\n"));

        DRI_XREF *dri_entry;
        SaHpiDomainIdResourceIdInventoryIdArrayT dri_tuple;

        /*
         * TODO: loop through columns, check syntax and lengths. For
         * columns which have no dependencies, you could also move
         * the value/range checking here to attempt to catch error
         * cases as early as possible.
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;
                rc = SNMP_ERR_NOERROR;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIAREATYPE:
                        /** INTEGER = ASN_INTEGER */
                        rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                            sizeof(row_ctx->saHpiAreaType));
                        /* check correct range for types */
                        if (rc == SNMP_ERR_NOERROR) {
                                if (!oh_lookup_idrareatype(*var->val.integer - 1)) {
                                        DEBUGMSGTL ((AGENT, "COLUMN_SA_HPI_AREA_TYPE, TYPE invalid\n"));
                                        rc = SNMP_ERR_WRONGVALUE;
                                }
                        }

                        /* rowstatus state transition */
                        if (row_ctx->saHpiAreaRowStatus != SNMP_ROW_CREATEANDWAIT) {
                                DEBUGMSGTL ((AGENT, "COLUMN_SA_HPI_AREA_TYPE, RowStatus invalid\n"));
                                rc = SNMP_ERR_WRONGVALUE;
                        }

                        break;

                case COLUMN_SAHPIAREAROWSTATUS:
                        /** RowStatus = ASN_INTEGER */
                        rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                            sizeof(row_ctx->saHpiAreaRowStatus));
                        /* check for valid status and state transitions*/
                        if ((rc == SNMP_ERR_NOERROR ) && 
                            (rg->row_created) && 
                            (*var->val.integer != SNMP_ROW_CREATEANDWAIT)) {
                                DEBUGMSGTL ((AGENT, 
                                             "AREA_RS: COLUMN_SAHPIAREAROWSTATUS, should be SNMP_ROW_CREATEANDWAIT on new row\n"));
                                rc = SNMP_ERR_WRONGVALUE;
                        } else if ((rc == SNMP_ERR_NOERROR ) && (!rg->row_created)) {
                                if ((row_ctx->saHpiAreaRowStatus != SNMP_ROW_DESTROY) &&
                                    (*var->val.integer == SNMP_ROW_DESTROY)) {
                                        rc = SNMP_ERR_NOERROR;
                                } else {
                                        DEBUGMSGTL ((AGENT, 
                                                     "COLUMN_SA_HPI_AREA_ROW_STATUS, SNMP_ERR_INCONSISTENTVALUE\n"));
                                        rc = SNMP_ERR_INCONSISTENTVALUE;
                                }
                        }

                        /* if this is a new row check the validity of the user specified AreaId */
                        if (((rg->row_created)) && (rc == SNMP_ERR_NOERROR)) {
                                /* check for valid areaId specified */
                                /* this is not the same as the AreaId obtained from HPI */
                                /* it is maintained in the subagent                     */
                                dri_tuple.domainId_resourceId_idr_arry[0] = 
                                row_ctx->index.oids[saHpiDomainId_INDEX];
                                dri_tuple.domainId_resourceId_idr_arry[1] = 
                                row_ctx->index.oids[saHpiResourceEntryId_INDEX];
                                dri_tuple.domainId_resourceId_idr_arry[2] = 
                                row_ctx->index.oids[saHpiInventoryId_INDEX];

                                /* domain_resource_idr_get() generates unique keys based on */
                                /* domainid, resourceid, and IdrId tuples.                  */
                                dri_entry = domain_resource_idr_get(&dri_tuple, &dri_table); 

                                if (dri_entry == NULL) {
                                        DEBUGMSGTL ((AGENT, 
                                                     "ERROR: saHpiAreaTable_set_reserve1() domain_resource_idr_get returned NULL\n"));
                                        rc = SNMP_ERR_GENERR;
                                }

                                if ( row_ctx->index.oids[saHpiAreaId_INDEX] >= dri_entry->entry_id ) {
                                        dri_entry->entry_id = row_ctx->index.oids[saHpiAreaId_INDEX];
                                        dri_entry->entry_id++;
                                } else {
                                        DEBUGMSGTL ((AGENT,"dri_entry->entry_id [%d]\n", dri_entry->entry_id));
                                        DEBUGMSGTL ((AGENT,"row_ctx->index.oids[saHpiAreaId_INDEX] [%ld]\n", row_ctx->index.oids[saHpiAreaId_INDEX]));
                                        DEBUGMSGTL ((AGENT, 
                                                     "ERROR: saHpiAreaTable_set_reserve1() User specified AreaId invalid!!!\n"));
                                        rc = SNMP_ERR_WRONGVALUE;
                                }
                        }
                        break;

                default: /** We shouldn't get here */
                        rc = SNMP_ERR_GENERR;
                        snmp_log(LOG_ERR, "unknown column in "
                                 "saHpiAreaTable_set_reserve1\n");
                }

                if (rc)
                        netsnmp_request_set_error( current->ri, rc );
                rg->status = SNMP_MAX( rg->status, current->ri->status );
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

void saHpiAreaTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiAreaTable_context *row_ctx = (saHpiAreaTable_context *)rg->existing_row;
//    saHpiAreaTable_context *undo_ctx = (saHpiAreaTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;
        netsnmp_variable_list *var;
        int rc;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_set_reserve2, called\n"));

        rg->rg_void = rg->list->ri;

        /*
         * TODO: loop through columns, check for valid
         * values and any range constraints.
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;
                rc = SNMP_ERR_NOERROR;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIAREATYPE:
                        /** INTEGER = ASN_INTEGER */
                        /*
                         * TODO: routine to check valid values
                         *
                         * EXAMPLE:
                         *
                        * if ( *var->val.integer != XXX ) {
                    *    rc = SNMP_ERR_INCONSISTENTVALUE;
                    *    rc = SNMP_ERR_BADVALUE;
                    * }
                    */
                        break;

                case COLUMN_SAHPIAREAROWSTATUS:
                        /** RowStatus = ASN_INTEGER */
                        /*
                         * TODO: routine to check valid values
                         *
                         * EXAMPLE:
                         *
                        * if ( *var->val.integer != XXX ) {
                    *    rc = SNMP_ERR_INCONSISTENTVALUE;
                    *    rc = SNMP_ERR_BADVALUE;
                    * }
                    */
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }

                if (rc)
                        netsnmp_request_set_error( current->ri, rc);
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiAreaTable_set_action( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
        saHpiAreaTable_context *row_ctx = (saHpiAreaTable_context *)rg->existing_row;
//    saHpiAreaTable_context *undo_ctx = (saHpiAreaTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        int            row_err = 0;

        subagent_lock(&hpi_lock_data);

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_set_action, called\n"));

        /*
         * TODO: loop through columns, copy varbind values
         * to context structure for the row.
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIAREATYPE:
                        /** INTEGER = ASN_INTEGER */
                        if (row_ctx->saHpiAreaRowStatus == SNMP_ROW_CREATEANDWAIT) {
                                row_ctx->saHpiAreaType = *var->val.integer;
                                row_ctx->saHpiAreaRowStatus = SNMP_ROW_ACTIVE;
                                row_err = set_table_area_type (row_ctx);
                        }
                        break;

                case COLUMN_SAHPIAREAROWSTATUS:
                        /** RowStatus = ASN_INTEGER */
                        if (*var->val.integer == SNMP_ROW_DESTROY) {
                                if (row_ctx->saHpiAreaRowStatus == SNMP_ROW_ACTIVE) {
                                        row_err = set_table_area_delete (row_ctx);
                                }
                                rg->row_deleted = 1;
                        }
                        row_ctx->saHpiAreaRowStatus = *var->val.integer;
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }
        }

        if (row_err) {
                netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                               row_err);
        }

        subagent_unlock(&hpi_lock_data);

        return;

}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiAreaTable_set_commit( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
//    saHpiAreaTable_context *row_ctx = (saHpiAreaTable_context *)rg->existing_row;
//    saHpiAreaTable_context *undo_ctx = (saHpiAreaTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_set_commit, called\n"));

        /*
         * loop through columns
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIAREATYPE:
                        /** INTEGER = ASN_INTEGER */
                        break;

                case COLUMN_SAHPIAREAROWSTATUS:
                        /** RowStatus = ASN_INTEGER */
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiAreaTable_set_free( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
//    saHpiAreaTable_context *row_ctx = (saHpiAreaTable_context *)rg->existing_row;
//    saHpiAreaTable_context *undo_ctx = (saHpiAreaTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_set_free, called\n"));

        /*
         * loop through columns
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIAREATYPE:
                        /** INTEGER = ASN_INTEGER */
                        break;

                case COLUMN_SAHPIAREAROWSTATUS:
                        /** RowStatus = ASN_INTEGER */
                        break;

                default: 
                        break;
                        /** We shouldn't get here */
                        /** should have been logged in reserve1 */
                }
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiAreaTable_set_undo( netsnmp_request_group *rg )
{
        netsnmp_variable_list *var;
//    saHpiAreaTable_context *row_ctx = (saHpiAreaTable_context *)rg->existing_row;
//    saHpiAreaTable_context *undo_ctx = (saHpiAreaTable_context *)rg->undo_info;
        netsnmp_request_group_item *current;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_set_undo, called\n"));

        /*
         * loop through columns
         */
        for ( current = rg->list; current; current = current->next ) {

                var = current->ri->requestvb;

                switch (current->tri->colnum) {
                
                case COLUMN_SAHPIAREATYPE:
                        /** INTEGER = ASN_INTEGER */
                        break;

                case COLUMN_SAHPIAREAROWSTATUS:
                        /** RowStatus = ASN_INTEGER */
                        break;

                default: /** We shouldn't get here */
                        netsnmp_assert(0); /** why wasn't this caught in reserve1? */
                }
        }

        /*
         * done with all the columns. Could check row related
         * requirements here.
         */
}


/************************************************************
 *
 * Initialize the saHpiAreaTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiAreaTable(void)
{
        netsnmp_table_registration_info *table_info;

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiAreaTable, called\n"));

        if (my_handler) {
                snmp_log(LOG_ERR, "initialize_table_saHpiAreaTable_handler called again\n");
                return;
        }

        memset(&cb, 0x00, sizeof(cb));

        /** create the table structure itself */
        table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

        /* if your table is read only, it's easiest to change the
           HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
        my_handler = netsnmp_create_handler_registration("saHpiAreaTable",
                                                         netsnmp_table_array_helper_handler,
                                                         saHpiAreaTable_oid,
                                                         saHpiAreaTable_oid_len,
                                                         HANDLER_CAN_RWRITE);

        if (!my_handler || !table_info) {
                snmp_log(LOG_ERR, "malloc failed in "
                         "initialize_table_saHpiAreaTable_handler\n");
                return; /** mallocs failed */
        }

        /***************************************************
         * Setting up the table's definition
         */
        /*
         * TODO: add any external indexes here.
         */
        /** TODO: add code for external index(s)! */

        /*
         * internal indexes
         */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceIsHistorical */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiInventoryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiAreaId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

        table_info->min_column = saHpiAreaTable_COL_MIN;
        table_info->max_column = saHpiAreaTable_COL_MAX;

        /***************************************************
         * registering the table with the master agent
         */
        cb.get_value = saHpiAreaTable_get_value;
        cb.container = netsnmp_container_find("saHpiAreaTable_primary:"
                                              "saHpiAreaTable:"
                                              "table_container");

        netsnmp_container_add_index(cb.container,
                                    netsnmp_container_find("saHpiAreaTable_secondary:"
                                                           "saHpiAreaTable:"
                                                           "table_container"));
        cb.container->next->compare = saHpiAreaTable_cmp;


        cb.can_set = 1;

        cb.create_row = (UserRowMethod*)saHpiAreaTable_create_row;

        cb.duplicate_row = (UserRowMethod*)saHpiAreaTable_duplicate_row;
        cb.delete_row = (UserRowMethod*)saHpiAreaTable_delete_row;
        cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiAreaTable_row_copy;

        cb.can_activate = (Netsnmp_User_Row_Action *)saHpiAreaTable_can_activate;
        cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiAreaTable_can_deactivate;
        cb.can_delete = (Netsnmp_User_Row_Action *)saHpiAreaTable_can_delete;

        cb.set_reserve1 = saHpiAreaTable_set_reserve1;
        cb.set_reserve2 = saHpiAreaTable_set_reserve2;
        cb.set_action = saHpiAreaTable_set_action;
        cb.set_commit = saHpiAreaTable_set_commit;
        cb.set_free = saHpiAreaTable_set_free;
        cb.set_undo = saHpiAreaTable_set_undo;

        DEBUGMSGTL(("initialize_table_saHpiAreaTable",
                    "Registering table saHpiAreaTable "
                    "as a table array\n"));
        netsnmp_table_container_register(my_handler, table_info, &cb,
                                         cb.container, 1);
}

/************************************************************
 * saHpiAreaTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiAreaTable_get_value(
                            netsnmp_request_info *request,
                            netsnmp_index *item,
                            netsnmp_table_request_info *table_info )
{
        netsnmp_variable_list *var = request->requestvb;
        saHpiAreaTable_context *context = (saHpiAreaTable_context *)item;

        DEBUGMSGTL ((AGENT, "saHpiAreaTable_get_value, called\n"));

        switch (table_info->colnum) {
        
        case COLUMN_SAHPIAREAID:
                /** SaHpiInstrumentId = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                                         (u_char*)&context->saHpiAreaId,
                                         sizeof(context->saHpiAreaId) );
                break;

        case COLUMN_SAHPIAREAIDINDEX:
                /** SaHpiInstrumentId = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                                         (u_char*)&context->saHpiAreaIdIndex,
                                         sizeof(context->saHpiAreaIdIndex) );
                break;

        case COLUMN_SAHPIAREATYPE:
                /** INTEGER = ASN_INTEGER */
                snmp_set_var_typed_value(var, ASN_INTEGER,
                                         (u_char*)&context->saHpiAreaType,
                                         sizeof(context->saHpiAreaType) );
                break;

        case COLUMN_SAHPIAREAISREADONLY:
                /** TruthValue = ASN_INTEGER */
                snmp_set_var_typed_value(var, ASN_INTEGER,
                                         (u_char*)&context->saHpiAreaIsReadOnly,
                                         sizeof(context->saHpiAreaIsReadOnly) );
                break;

        case COLUMN_SAHPIAREAROWSTATUS:
                /** RowStatus = ASN_INTEGER */
                snmp_set_var_typed_value(var, ASN_INTEGER,
                                         (u_char*)&context->saHpiAreaRowStatus,
                                         sizeof(context->saHpiAreaRowStatus) );
                break;

        case COLUMN_SAHPIAREANUMDATAFIELDS:
                /** UNSIGNED32 = ASN_UNSIGNED */
                snmp_set_var_typed_value(var, ASN_UNSIGNED,
                                         (u_char*)&context->saHpiAreaNumDataFields,
                                         sizeof(context->saHpiAreaNumDataFields) );
                break;

        default: /** We shouldn't get here */
                snmp_log(LOG_ERR, "unknown column in "
                         "saHpiAreaTable_get_value\n");
                return SNMP_ERR_GENERR;
        }
        return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiAreaTable_get_by_idx
 */
const saHpiAreaTable_context *
saHpiAreaTable_get_by_idx(netsnmp_index * hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiAreaTable_get_by_idx, called\n"));

        return(const saHpiAreaTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


