/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiCtrlAnalogTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiCtrlAnalogTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceTable.h>
#include <session_info.h>

#include <oh_utils.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiCtrlAnalogTable_oid[] = { saHpiCtrlAnalogTable_TABLE_OID};
size_t saHpiCtrlAnalogTable_oid_len = OID_LENGTH(saHpiCtrlAnalogTable_oid);


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;		      
static GHashTable *dr_table;

/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long ctrl_analog_entry_count = 0;
static oid saHpiCtrlAnalogActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,4,7,6 };
int handle_saHpiCtrlAnalogActiveEntries( netsnmp_mib_handler *handler, 
			       netsnmp_handler_registration *reginfo,
			       netsnmp_agent_request_info   *reqinfo, 
			       netsnmp_request_info *requests);
int initialize_table_saHpiCtrlAnalogActiveEntries(void);

/*
 * SaErrorT populate_ctrl_analog()
 */
SaErrorT populate_ctrl_analog(SaHpiSessionIdT sessionid, 
			      SaHpiRdrT *rdr_entry,
			      SaHpiRptEntryT *rpt_entry,
			      oid *full_oid, size_t full_oid_len,
			      oid *child_oid, size_t *child_oid_len)
{

 	DEBUGMSGTL ((AGENT, "populate_ctrl_analog, called\n"));
	SaErrorT rv = SA_OK;
        int new_row = MIB_FALSE;

	oid ctrl_analog_oid[CTRL_ANALOG_INDEX_NR];
	netsnmp_index ctrl_analog_index;
	saHpiCtrlAnalogTable_context *ctrl_analog_context;

	DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

	oid column[2];
	int column_len = 2;

	DEBUGMSGTL ((AGENT, "SAHPI_CTRL_TYPE_DISCRETE populate_ctrl_digital() called\n"));

	/* check for NULL pointers */
	if (!rdr_entry) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_ctrl_analog() passed NULL rdr_entry pointer\n"));

		return AGENT_ERR_INTERNAL_ERROR;
	}    
	if (!rpt_entry) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_ctrl_analog() passed NULL rdr_entry pointer\n"));
		
		return AGENT_ERR_INTERNAL_ERROR;
	}

        subagent_lock(&hpi_lock_data);

	/* BUILD oid for new row */
		/* assign the number of indices */
	ctrl_analog_index.len = CTRL_ANALOG_INDEX_NR;
		/** Index saHpiDomainId is external */
	ctrl_analog_oid[0] = get_domain_id(sessionid);
		/** Index saHpiResourceId is external */
	ctrl_analog_oid[1] = rpt_entry->ResourceId;
		/** Index saHpiResourceIsHistorical is external */
	ctrl_analog_oid[2] = MIB_FALSE;
		/** Index saHpiCtrlDigitalEntryId is internal */
	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
	dr_pair.domainId_resourceId_arry[1] = rpt_entry->ResourceId;
	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
	if (dr_entry == NULL) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_ctrl_analog() domain_resource_pair_get returned NULL\n"));
		
		subagent_unlock(&hpi_lock_data);
		
		return AGENT_ERR_INTERNAL_ERROR;
	}
	ctrl_analog_oid[3] = dr_entry->entry_id++;
		/* assign the indices to the index */
	ctrl_analog_index.oids = (oid *) & ctrl_analog_oid;

	/* create full oid on This row for parent RowPointer */
	column[0] = 1;
	column[1] = COLUMN_SAHPICTRLANALOGNUM;
	memset(child_oid, 0, MAX_OID_LEN);
	build_full_oid(saHpiCtrlAnalogTable_oid, saHpiCtrlAnalogTable_oid_len,
			column, column_len,
			&ctrl_analog_index,
			child_oid, MAX_OID_LEN, child_oid_len);

	/* See if Row exists. */
	ctrl_analog_context = NULL;
	ctrl_analog_context = CONTAINER_FIND(cb.container, &ctrl_analog_index);

	if (!ctrl_analog_context) { 
		// New entry. Add it
		ctrl_analog_context = 
			saHpiCtrlAnalogTable_create_row(&ctrl_analog_index);
                new_row = MIB_TRUE;
	}
	if (!ctrl_analog_context) {
		snmp_log (LOG_ERR, "Not enough memory for a Ctrl Analog row!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	} 

	/** SaHpiEntryId = ASN_UNSIGNED */
        ctrl_analog_context->saHpiCtrlAnalogEntryId =
		ctrl_analog_oid[3];

        /** SaHpiInstrumentId = ASN_UNSIGNED */
        ctrl_analog_context->saHpiCtrlAnalogNum =
		rdr_entry->RdrTypeUnion.CtrlRec.Num;

        /** SaHpiCtrlOutputType = ASN_INTEGER */
        ctrl_analog_context->saHpiCtrlAnalogOutputType = 
		rdr_entry->RdrTypeUnion.CtrlRec.OutputType + 1;

        /** SaHpiCtrlMode = ASN_INTEGER */
        ctrl_analog_context->saHpiCtrlAnalogDefaultMode = 
		rdr_entry->RdrTypeUnion.CtrlRec.DefaultMode.Mode + 1;

        /** SaHpiCtrlMode = ASN_INTEGER */
        ctrl_analog_context->saHpiCtrlAnalogDefaultMode = 
		rdr_entry->RdrTypeUnion.CtrlRec.DefaultMode.Mode + 1;

        /** TruthValue = ASN_INTEGER */
        ctrl_analog_context->saHpiCtrlAnalogIsReadOnly =
            (rdr_entry->RdrTypeUnion.CtrlRec.DefaultMode.ReadOnly == SAHPI_TRUE)
		? MIB_TRUE : MIB_FALSE;

        /** TruthValue = ASN_INTEGER */
        ctrl_analog_context->saHpiCtrlAnalogIsWriteOnly =
	    (rdr_entry->RdrTypeUnion.CtrlRec.WriteOnly == SAHPI_TRUE) 
		? MIB_TRUE : MIB_FALSE;

	/** INTEGER = ASN_INTEGER */
	ctrl_analog_context->saHpiCtrlAnalogDefaultMinState =
		rdr_entry->RdrTypeUnion.CtrlRec.TypeUnion.Analog.Min;

	/** INTEGER = ASN_INTEGER */
	ctrl_analog_context->saHpiCtrlAnalogDefaultMaxState =
		rdr_entry->RdrTypeUnion.CtrlRec.TypeUnion.Analog.Max;

	/** INTEGER = ASN_INTEGER */
	ctrl_analog_context->saHpiCtrlAnalogDefaultState =
		rdr_entry->RdrTypeUnion.CtrlRec.TypeUnion.Analog.Default;

        /** saHpiCtrlDiscreteState = UNSIGNED32 */
        ctrl_analog_context->saHpiCtrlAnalogState = 
		rdr_entry->RdrTypeUnion.CtrlRec.TypeUnion.Analog.Default;

        /** UNSIGNED32 = ASN_UNSIGNED */
        ctrl_analog_context->saHpiCtrlAnalogOem = 
		rdr_entry->RdrTypeUnion.CtrlRec.Oem;

        /** RowPointer = ASN_OBJECT_ID */
	memset(ctrl_analog_context->saHpiCtrlAnalogRDR, 
	       0, 
	       sizeof(ctrl_analog_context->saHpiCtrlAnalogRDR));

	ctrl_analog_context->saHpiCtrlAnalogRDR_len = full_oid_len * sizeof(oid);
	memcpy(ctrl_analog_context->saHpiCtrlAnalogRDR, 
	       full_oid, 
	       ctrl_analog_context->saHpiCtrlAnalogRDR_len);


	if (new_row == MIB_TRUE)
                 CONTAINER_INSERT (cb.container, ctrl_analog_context);
		
	ctrl_analog_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);

	return rv;
}

/**
 * 
 * @domainId
 * @resourceId
 * 
 * @return 
 */
SaErrorT clear_ctrl_analog(SaHpiDomainIdT domainId,
                           SaHpiResourceIdT resourceId)

{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiCtrlAnalogTable_context *ctrl_ctx;

	DEBUGMSGTL ((AGENT, "clear_ctrl_analog, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domainId));	
	DEBUGMSGTL ((AGENT, "           resourceId [%d]\n", resourceId));

        DR_XREF *dr_entry;
        SaHpiDomainIdResourceIdArrayT dr_pair;

        /* reset any existing indexEntry value */
        dr_pair.domainId_resourceId_arry[0] = domainId;
        dr_pair.domainId_resourceId_arry[1] = resourceId;
        dr_entry = domain_resoruce_pair_lookup(&dr_pair, &dr_table); 
        if (dr_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                "INFO: clear_ctrl_analog() domain_resource_pair_get returned NULL\n"));
                return SA_ERR_HPI_NOT_PRESENT;
        } else {
                dr_entry->entry_id = 0;
        }

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        /* based on the found row_idx get the pointer   */
                        /* to its context (row data)                    */
                        ctrl_ctx = CONTAINER_FIND(cb.container, row_idx);

                        /* before we delete the context we should get the  */
                        /* next row (context) if any before we delete this */ 
                        /* one.                                            */
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if ((ctrl_ctx->index.oids[saHpiCtrlAnalogDomainId_INDEX] ==
                             domainId) &&

                            (ctrl_ctx->index.oids[saHpiCtrlAnalogResourceEntryId_INDEX] ==
                             resourceId)) {

                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctrl_ctx);
                                saHpiCtrlAnalogTable_delete_row (ctrl_ctx);
                                ctrl_analog_entry_count = 
                                        CONTAINER_SIZE (cb.container);
                                DEBUGMSGTL ((AGENT, "clear_ctrl_analog: "
                                                    "found row: removing\n"));

                        }

                } while (row_idx);
        } 

        return rv;
}

/*
 * int set_table_ctrl_analog_mode()
 */
int set_table_ctrl_analog_mode (saHpiCtrlAnalogTable_context *row_ctx)
{

 	DEBUGMSGTL ((AGENT, "set_table_ctrl_analog_mode, called\n"));

	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiCtrlStateT	    ctrl_state;	

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiCtrlAnalogDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiCtrlAnalogResourceEntryId_INDEX];

	ctrl_state.StateUnion.Analog = row_ctx->saHpiCtrlAnalogState;
	ctrl_state.Type = SAHPI_CTRL_TYPE_ANALOG;

	rc = saHpiControlSet(session_id, resource_id, 
			     row_ctx->saHpiCtrlAnalogNum, 
			     row_ctx->saHpiCtrlAnalogMode - 1, 
			     &ctrl_state);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "SAHPI_CTRL_TYPE_ANALOG: Call to saHpiControlSet failed to set Mode rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "SAHPI_CTRL_TYPE_ANALOG: Call to saHpiControlSet failed to set Mode rc: %s.\n",
			     oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/*
 * int set_table_ctrl_analog_mode()
 */

int set_table_ctrl_analog_state (saHpiCtrlAnalogTable_context *row_ctx)
{

 	DEBUGMSGTL ((AGENT, "set_table_ctrl_analog_state, called\n"));

	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiCtrlStateT	    ctrl_state;	

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiCtrlAnalogDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiCtrlAnalogResourceEntryId_INDEX];

	ctrl_state.StateUnion.Analog = row_ctx->saHpiCtrlAnalogState;
	ctrl_state.Type = SAHPI_CTRL_TYPE_ANALOG; 

	rc = saHpiControlSet(session_id, resource_id, 
			     row_ctx->saHpiCtrlAnalogNum, 
			     row_ctx->saHpiCtrlAnalogMode - 1, 
			     &ctrl_state); 

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "SAHPI_CTRL_TYPE_ANALOG: Call to saHpiControlSet failed to set State rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "SAHPI_CTRL_TYPE_ANALOG: Call to saHpiControlSet failed to set State rc: %s.\n",
			     oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/*
 * int handle_saHpiCtrlAnalogActiveEntries()
 */
int handle_saHpiCtrlAnalogActiveEntries(netsnmp_mib_handler *handler,
				     netsnmp_handler_registration *reginfo,
				     netsnmp_agent_request_info   *reqinfo,
				     netsnmp_request_info         *requests)
{
    /* We are never called for a GETNEXT if it's registered as a
       "instance", as it's "magically" handled for us.  */

    /* a instance handler also only hands us one request at a time, so
       we don't need to loop over a list of requests; we'll only get one. */


 	DEBUGMSGTL ((AGENT, "handle_saHpiCtrlAnalogActiveEntries, called\n"));

        ctrl_analog_entry_count = CONTAINER_SIZE (cb.container);

	switch(reqinfo->mode) {
	
	    case MODE_GET:
		snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
					 (u_char *) &ctrl_analog_entry_count,
					 sizeof(ctrl_analog_entry_count));
		break;
	
	
	    default:
		/* we should never get here, so this is a really bad error */
		return SNMP_ERR_GENERR;
	}
	
	return SNMP_ERR_NOERROR;
}

/*
 * int initialize_table_saHpiCtrlAnalogActiveEntries(void)
 */
int initialize_table_saHpiCtrlAnalogActiveEntries(void)
{

 	DEBUGMSGTL ((AGENT, "initialize_table_saHpiCtrlAnalogActiveEntries, called\n"));

	netsnmp_register_scalar(
		netsnmp_create_handler_registration(
			"saHpiCtrlAnalogActiveEntries", 
			handle_saHpiCtrlAnalogActiveEntries,
                        saHpiCtrlAnalogActiveEntries_oid, 
			OID_LENGTH(saHpiCtrlAnalogActiveEntries_oid),
                        HANDLER_CAN_RONLY));
	return 0;
}


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiCtrlAnalogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiCtrlAnalogTable_cmp( const void *lhs, const void *rhs )
{
	saHpiCtrlAnalogTable_context *context_l =
	(saHpiCtrlAnalogTable_context *)lhs;
	saHpiCtrlAnalogTable_context *context_r =
	(saHpiCtrlAnalogTable_context *)rhs;

	/*
	 * check primary key, then secondary. Add your own code if
	 * there are more than 2 indexes
	 */
 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_cmp, called\n"));

       /* check for NULL pointers */
	if (lhs == NULL || rhs == NULL ) {
		DEBUGMSGTL((AGENT,"saHpiCtrlDigitalTable_cmp() NULL pointer ERROR\n" ));
		return 0;
	}
	/* CHECK FIRST INDEX,  saHpiDomainId */
	if ( context_l->index.oids[0] < context_r->index.oids[0])
		return -1;

	if ( context_l->index.oids[0] > context_r->index.oids[0])
		return 1;

	if ( context_l->index.oids[0] == context_r->index.oids[0]) {
		/* If saHpiDomainId index is equal sort by second index */
		/* CHECK SECOND INDEX,  saHpiResourceEntryId */
		if ( context_l->index.oids[1] < context_r->index.oids[1])
			return -1;

		if ( context_l->index.oids[1] > context_r->index.oids[1])
			return 1;

		if ( context_l->index.oids[1] == context_r->index.oids[1]) {
			/* If saHpiResourceEntryId index is equal sort by third index */
			/* CHECK THIRD INDEX,  saHpiResourceIsHistorical */
			if ( context_l->index.oids[2] < context_r->index.oids[2])
				return -1;

			if ( context_l->index.oids[2] > context_r->index.oids[2])
				return 1;

			if ( context_l->index.oids[2] == context_r->index.oids[2]) {
				/* If saHpiResourceIsHistorical index is equal sort by forth index */
				/* CHECK FORTH INDEX,  saHpiCtrlAnalogEntryId */
				if ( context_l->index.oids[3] < context_r->index.oids[3])
					return -1;

				if ( context_l->index.oids[3] > context_r->index.oids[3])
					return 1;

				if ( context_l->index.oids[3] == context_r->index.oids[3])
					return 0;
			}
		}
	}

	return 0;
}

/************************************************************
 * Initializes the saHpiCtrlAnalogTable module
 */
void
init_saHpiCtrlAnalogTable(void)
{

 	DEBUGMSGTL ((AGENT, "init_saHpiCtrlAnalogTable, called\n"));

	initialize_table_saHpiCtrlAnalogTable();

	initialize_table_saHpiCtrlAnalogActiveEntries();

	domain_resource_pair_initialize(&initialized, &dr_table);
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiCtrlAnalogTable_row_copy(saHpiCtrlAnalogTable_context * dst,
					 saHpiCtrlAnalogTable_context * src)
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_row_copy, called\n"));

	if (!dst||!src)
		return 1;

	
	subagent_lock(&hpi_lock_data);
	/*
	 * copy index, if provided
	 */
	if (dst->index.oids)
		free(dst->index.oids);
	if (snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
			    src->index.len * sizeof(oid) )) {
		dst->index.oids = NULL;
		subagent_unlock(&hpi_lock_data);
		return 1;
	}
	dst->index.len = src->index.len;


	/*
	 * copy components into the context structure
	 */
	/** TODO: add code for external index(s)! */
	dst->saHpiCtrlAnalogEntryId = src->saHpiCtrlAnalogEntryId;

	dst->saHpiCtrlAnalogNum = src->saHpiCtrlAnalogNum;

	dst->saHpiCtrlAnalogOutputType = src->saHpiCtrlAnalogOutputType;

	dst->saHpiCtrlAnalogDefaultMode = src->saHpiCtrlAnalogDefaultMode;

	dst->saHpiCtrlAnalogMode = src->saHpiCtrlAnalogMode;

	dst->saHpiCtrlAnalogIsReadOnly = src->saHpiCtrlAnalogIsReadOnly;

	dst->saHpiCtrlAnalogIsWriteOnly = src->saHpiCtrlAnalogIsWriteOnly;

	dst->saHpiCtrlAnalogDefaultMinState = src->saHpiCtrlAnalogDefaultMinState;

	dst->saHpiCtrlAnalogDefaultMaxState = src->saHpiCtrlAnalogDefaultMaxState;

	dst->saHpiCtrlAnalogDefaultState = src->saHpiCtrlAnalogDefaultState;

	dst->saHpiCtrlAnalogState = src->saHpiCtrlAnalogState;

	dst->saHpiCtrlAnalogOem = src->saHpiCtrlAnalogOem;

	memcpy( dst->saHpiCtrlAnalogRDR, src->saHpiCtrlAnalogRDR, src->saHpiCtrlAnalogRDR_len );
	dst->saHpiCtrlAnalogRDR_len = src->saHpiCtrlAnalogRDR_len;

	subagent_unlock(&hpi_lock_data);
	return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiCtrlAnalogTable_extract_index( saHpiCtrlAnalogTable_context * ctx, netsnmp_index * hdr )
{
	/*
	 * temporary local storage for extracting oid index
	 *
	 * extract index uses varbinds (netsnmp_variable_list) to parse
	 * the index OID into the individual components for each index part.
	 */
	/** TODO: add storage for external index(s)! */
	netsnmp_variable_list var_saHpiDomainId;
	netsnmp_variable_list var_saHpiResourceId;
	netsnmp_variable_list var_saHpiResourceIsHistorical;
	netsnmp_variable_list var_saHpiCtrlAnalogEntryId;
	int err;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_extract_index, called\n"));

	
	subagent_lock(&hpi_lock_data);
	/*
	 * copy index, if provided
	 */
	if (hdr) {
		netsnmp_assert(ctx->index.oids == NULL);
		if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
				    hdr->len * sizeof(oid) )) {
			subagent_unlock(&hpi_lock_data);
			return -1;
		}
		ctx->index.len = hdr->len;
	}

	/*
	 * initialize variable that will hold each component of the index.
	 * If there are multiple indexes for the table, the variable_lists
	 * need to be linked together, in order.
	 */
	/** TODO: add code for external index(s)! */
	memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
	var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiDomainId.next_variable = &var_saHpiResourceId;

	memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
	var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiResourceId.next_variable = &var_saHpiResourceIsHistorical;

	memset( &var_saHpiResourceIsHistorical, 0x00, sizeof(var_saHpiResourceIsHistorical) );
	var_saHpiResourceIsHistorical.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiResourceIsHistorical.next_variable = &var_saHpiCtrlAnalogEntryId;

	memset( &var_saHpiCtrlAnalogEntryId, 0x00, sizeof(var_saHpiCtrlAnalogEntryId) );
	var_saHpiCtrlAnalogEntryId.type = ASN_UNSIGNED;	/* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiCtrlAnalogEntryId.next_variable = NULL;


	/*
	 * parse the oid into the individual index components
	 */
	err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
	if (err == SNMP_ERR_NOERROR) {
		/*
		 * copy index components into the context structure
		 */
		/** skipping external index saHpiDomainId */

		/** skipping external index saHpiResourceId */

		/** skipping external index saHpiResourceIsHistorical */

		ctx->saHpiCtrlAnalogEntryId = *var_saHpiCtrlAnalogEntryId.val.integer;

		err = saHpiDomainId_check_index(
				*var_saHpiDomainId.val.integer);
		err = saHpiResourceEntryId_check_index(
				*var_saHpiResourceId.val.integer);  
		err = saHpiResourceIsHistorical_check_index(
				*var_saHpiResourceIsHistorical.val.integer);
		err = saHpiCtrlAnalogEntryId_check_index(
				*var_saHpiCtrlAnalogEntryId.val.integer);
	}

	/*
	 * parsing may have allocated memory. free it.
	 */
	snmp_reset_var_buffers( &var_saHpiDomainId );

	
	subagent_unlock(&hpi_lock_data);
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiCtrlAnalogTable_can_activate(saHpiCtrlAnalogTable_context *undo_ctx,
				      saHpiCtrlAnalogTable_context *row_ctx,
				      netsnmp_request_group * rg)
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_can_activate, called\n"));
	/*
	 * TODO: check for activation requirements here
	 */


	/*
	 * be optimistic.
	 */
	return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiCtrlAnalogTable_can_deactivate(saHpiCtrlAnalogTable_context *undo_ctx,
					saHpiCtrlAnalogTable_context *row_ctx,
					netsnmp_request_group * rg)
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_can_deactivate, called\n"));
	/*
	 * TODO: check for deactivation requirements here
	 */
	return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiCtrlAnalogTable_can_delete(saHpiCtrlAnalogTable_context *undo_ctx,
				    saHpiCtrlAnalogTable_context *row_ctx,
				    netsnmp_request_group * rg)
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_can_delete, called\n"));
	/*
	 * probably shouldn't delete a row that we can't
	 * deactivate.
	 */
	if (saHpiCtrlAnalogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
		return 0;

	/*
	 * TODO: check for other deletion requirements here
	 */
	return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiCtrlAnalogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiCtrlAnalogTable_context *
saHpiCtrlAnalogTable_create_row( netsnmp_index* hdr)
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_create_row, called\n"));

	
	subagent_lock(&hpi_lock_data);
	
	saHpiCtrlAnalogTable_context * ctx =
	SNMP_MALLOC_TYPEDEF(saHpiCtrlAnalogTable_context);
	if (!ctx) {
		subagent_unlock(&hpi_lock_data);
		return NULL;
         }
	/*
	 * TODO: check indexes, if necessary.
	 */
	if (saHpiCtrlAnalogTable_extract_index( ctx, hdr )) {
		free(ctx->index.oids);
		free(ctx);
		subagent_unlock(&hpi_lock_data);
		return NULL;
	}

	/* netsnmp_mutex_init(ctx->lock);
	   netsnmp_mutex_lock(ctx->lock); */

	/*
	 * TODO: initialize any default values here. This is also
	 * first place you really should allocate any memory for
	 * yourself to use.  If you allocated memory earlier,
	 * make sure you free it for earlier error cases!
	 */
	/**
	 ctx->saHpiCtrlAnalogMode = 0;
	 ctx->saHpiCtrlAnalogState = 0;
	*/

	subagent_unlock(&hpi_lock_data);
	return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiCtrlAnalogTable_context *
saHpiCtrlAnalogTable_duplicate_row( saHpiCtrlAnalogTable_context * row_ctx)
{
	saHpiCtrlAnalogTable_context * dup;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_duplicate_row, called\n"));

	if (!row_ctx)
		return NULL;

	subagent_lock(&hpi_lock_data);
	
	dup = SNMP_MALLOC_TYPEDEF(saHpiCtrlAnalogTable_context);
	if (!dup) {
		subagent_unlock(&hpi_lock_data);
		return NULL;
	}
	if (saHpiCtrlAnalogTable_row_copy(dup,row_ctx)) {
		free(dup);
		dup = NULL;
	}

	subagent_unlock(&hpi_lock_data);
	return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiCtrlAnalogTable_delete_row( saHpiCtrlAnalogTable_context * ctx )
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_delete_row, called\n"));

	/* netsnmp_mutex_destroy(ctx->lock); */

	subagent_lock(&hpi_lock_data);
	
	if (ctx->index.oids)
		free(ctx->index.oids);

	/*
	 * TODO: release any memory you allocated here...
	 */

	/*
	 * release header
	 */
	free( ctx );

	subagent_unlock(&hpi_lock_data);
	
	return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiCtrlAnalogTable_set_reserve1( netsnmp_request_group *rg )
{
	saHpiCtrlAnalogTable_context *row_ctx =
	(saHpiCtrlAnalogTable_context *)rg->existing_row;
//	saHpiCtrlAnalogTable_context *undo_ctx =
//	(saHpiCtrlAnalogTable_context *)rg->undo_info;
	netsnmp_variable_list *var;
	netsnmp_request_group_item *current;
	int rc;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_set_reserve1, called\n"));

	/*
	 * TODO: loop through columns, check syntax and lengths. For
	 * columns which have no dependencies, you could also move
	 * the value/range checking here to attempt to catch error
	 * cases as early as possible.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;
		rc = SNMP_ERR_NOERROR;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICTRLANALOGMODE:
			/** SaHpiCtrlMode = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiCtrlAnalogMode));
			break;

		case COLUMN_SAHPICTRLANALOGSTATE:
			/** INTEGER = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiCtrlAnalogState));
			break;

		default: /** We shouldn't get here */
			rc = SNMP_ERR_GENERR;
			snmp_log(LOG_ERR, "unknown column in "
				 "saHpiCtrlAnalogTable_set_reserve1\n");
		}

		if (rc)
			netsnmp_request_set_error( current->ri, rc );
		rg->status = SNMP_MAX( rg->status, current->ri->status );
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

void saHpiCtrlAnalogTable_set_reserve2( netsnmp_request_group *rg )
{
	saHpiCtrlAnalogTable_context *row_ctx = (saHpiCtrlAnalogTable_context *)rg->existing_row;
//	saHpiCtrlAnalogTable_context *undo_ctx = (saHpiCtrlAnalogTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;
	netsnmp_variable_list *var;
	int rc = SNMP_ERR_NOERROR;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_set_reserve2, called\n"));

	rg->rg_void = rg->list->ri;

	/*
	 * TODO: loop through columns, check for valid
	 * values and any range constraints.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;
		rc = SNMP_ERR_NOERROR;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICTRLANALOGMODE:
			if (row_ctx->saHpiCtrlAnalogIsReadOnly == MIB_TRUE) {
				       snmp_log(LOG_ERR, "COLUMN_SAHPICTRLANALOGMODE mode is ReadOnly, Failed\n");
				       DEBUGMSGTL ((AGENT, "COLUMN_SAHPICTRLANALOGMODE mode is ReadOnly, Failed\n"));
				       rc = SNMP_ERR_READONLY;
			       }  
			       if (oh_lookup_ctrlmode(*var->val.integer - 1) == NULL) {
				       snmp_log(LOG_ERR, "COLUMN_SAHPICTRLANALOGMODE Invalid Mode, Failed\n");
				       DEBUGMSGTL ((AGENT, "COLUMN_SAHPICTRLANALOGMODE Invalid Mode, Failed\n"));
				       rc = SNMP_ERR_BADVALUE;
			       }
			break;

		case COLUMN_SAHPICTRLANALOGSTATE:
			if (row_ctx->saHpiCtrlAnalogMode == 
			    (SAHPI_CTRL_MODE_AUTO + 1) ) {
				snmp_log(LOG_ERR, "COLUMN_SAHPICTRLANALOGSTATE mode is AUTO, Failed\n");
				DEBUGMSGTL ((AGENT, "COLUMN_SAHPICTRLANALOGSTATE mode is AUTO, Failed\n"));
				rc = SNMP_ERR_GENERR;
			} 
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}

		if (rc)
			netsnmp_request_set_error( current->ri, rc);
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiCtrlAnalogTable_set_action( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
	saHpiCtrlAnalogTable_context *row_ctx = (saHpiCtrlAnalogTable_context *)rg->existing_row;
//	saHpiCtrlAnalogTable_context *undo_ctx = (saHpiCtrlAnalogTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	int	row_err = 0;

        subagent_lock(&hpi_lock_data);

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_set_action, called\n"));

	/*
	 * TODO: loop through columns, copy varbind values
	 * to context structure for the row.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICTRLANALOGMODE:
			/** SaHpiCtrlMode = ASN_INTEGER */
			row_ctx->saHpiCtrlAnalogMode = *var->val.integer;
			row_err = set_table_ctrl_analog_mode (row_ctx);
			break;

		case COLUMN_SAHPICTRLANALOGSTATE:
			/** INTEGER = ASN_INTEGER */
			row_ctx->saHpiCtrlAnalogState = *var->val.integer;
			row_err = set_table_ctrl_analog_state (row_ctx);
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	if (row_err) {
		netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
					       row_err);
	}

        subagent_unlock(&hpi_lock_data);

        return;

	/*
	 * TODO: if you have dependencies on other tables, this would be
	 * a good place to check those, too.
	 */
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiCtrlAnalogTable_set_commit( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiCtrlAnalogTable_context *row_ctx = (saHpiCtrlAnalogTable_context *)rg->existing_row;
//	saHpiCtrlAnalogTable_context *undo_ctx = (saHpiCtrlAnalogTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_set_commit, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICTRLANALOGMODE:
			/** SaHpiCtrlMode = ASN_INTEGER */
			break;

		case COLUMN_SAHPICTRLANALOGSTATE:
			/** INTEGER = ASN_INTEGER */
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiCtrlAnalogTable_set_free( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiCtrlAnalogTable_context *row_ctx = (saHpiCtrlAnalogTable_context *)rg->existing_row;
//	saHpiCtrlAnalogTable_context *undo_ctx = (saHpiCtrlAnalogTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_set_free, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICTRLANALOGMODE:
			/** SaHpiCtrlMode = ASN_INTEGER */
			break;

		case COLUMN_SAHPICTRLANALOGSTATE:
			/** INTEGER = ASN_INTEGER */
			break;

		default: 
			break;
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiCtrlAnalogTable_set_undo( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiCtrlAnalogTable_context *row_ctx = (saHpiCtrlAnalogTable_context *)rg->existing_row;
//	saHpiCtrlAnalogTable_context *undo_ctx = (saHpiCtrlAnalogTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_set_undo, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICTRLANALOGMODE:
			/** SaHpiCtrlMode = ASN_INTEGER */
			break;

		case COLUMN_SAHPICTRLANALOGSTATE:
			/** INTEGER = ASN_INTEGER */
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}


/************************************************************
 *
 * Initialize the saHpiCtrlAnalogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiCtrlAnalogTable(void)
{
	netsnmp_table_registration_info *table_info;

 	DEBUGMSGTL ((AGENT, "initialize_table_saHpiCtrlAnalogTable, called\n"));

	if (my_handler) {
		snmp_log(LOG_ERR, "initialize_table_saHpiCtrlAnalogTable_handler called again\n");
		return;
	}

	memset(&cb, 0x00, sizeof(cb));

	/** create the table structure itself */
	table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

	/* if your table is read only, it's easiest to change the
	   HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
	my_handler = netsnmp_create_handler_registration("saHpiCtrlAnalogTable",
							 netsnmp_table_array_helper_handler,
							 saHpiCtrlAnalogTable_oid,
							 saHpiCtrlAnalogTable_oid_len,
							 HANDLER_CAN_RWRITE);

	if (!my_handler || !table_info) {
		snmp_log(LOG_ERR, "malloc failed in "
			 "initialize_table_saHpiCtrlAnalogTable_handler\n");
		return;	/** mallocs failed */
	}

	/***************************************************
	 * Setting up the table's definition
	 */
	/*
	 * TODO: add any external indexes here.
	 */
	/** TODO: add code for external index(s)! */

	/*
	 * internal indexes
	 */
	/** index: saHpiDomainId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
	/** index: saHpiResourceId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
	/** index: saHpiResourceIsHistorical */
	netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
	/** index: saHpiCtrlAnalogEntryId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

	table_info->min_column = saHpiCtrlAnalogTable_COL_MIN;
	table_info->max_column = saHpiCtrlAnalogTable_COL_MAX;

	/***************************************************
	 * registering the table with the master agent
	 */
	cb.get_value = saHpiCtrlAnalogTable_get_value;
	cb.container = netsnmp_container_find("saHpiCtrlAnalogTable_primary:"
					      "saHpiCtrlAnalogTable:"
					      "table_container");

	netsnmp_container_add_index(cb.container,
				    netsnmp_container_find("saHpiCtrlAnalogTable_secondary:"
							   "saHpiCtrlAnalogTable:"
							   "table_container"));
	cb.container->next->compare = saHpiCtrlAnalogTable_cmp;


	cb.can_set = 1;

	cb.create_row = (UserRowMethod*)saHpiCtrlAnalogTable_create_row;

	cb.duplicate_row = (UserRowMethod*)saHpiCtrlAnalogTable_duplicate_row;
	cb.delete_row = (UserRowMethod*)saHpiCtrlAnalogTable_delete_row;
	cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiCtrlAnalogTable_row_copy;

	cb.can_activate = (Netsnmp_User_Row_Action *)saHpiCtrlAnalogTable_can_activate;
	cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiCtrlAnalogTable_can_deactivate;
	cb.can_delete = (Netsnmp_User_Row_Action *)saHpiCtrlAnalogTable_can_delete;

	cb.set_reserve1 = saHpiCtrlAnalogTable_set_reserve1;
	cb.set_reserve2 = saHpiCtrlAnalogTable_set_reserve2;
	cb.set_action = saHpiCtrlAnalogTable_set_action;
	cb.set_commit = saHpiCtrlAnalogTable_set_commit;
	cb.set_free = saHpiCtrlAnalogTable_set_free;
	cb.set_undo = saHpiCtrlAnalogTable_set_undo;

	DEBUGMSGTL(("initialize_table_saHpiCtrlAnalogTable",
		    "Registering table saHpiCtrlAnalogTable "
		    "as a table array\n"));
	netsnmp_table_container_register(my_handler, table_info, &cb,
					 cb.container, 1);
}

/************************************************************
 * saHpiCtrlAnalogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiCtrlAnalogTable_get_value(
				  netsnmp_request_info *request,
				  netsnmp_index *item,
				  netsnmp_table_request_info *table_info )
{
	netsnmp_variable_list *var = request->requestvb;
	saHpiCtrlAnalogTable_context *context = (saHpiCtrlAnalogTable_context *)item;

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_get_value, called\n"));

	switch (table_info->colnum) {
	
	case COLUMN_SAHPICTRLANALOGENTRYID:
		/** SaHpiEntryId = ASN_UNSIGNED */
		snmp_set_var_typed_value(var, ASN_UNSIGNED,
					 (u_char*)&context->saHpiCtrlAnalogEntryId,
					 sizeof(context->saHpiCtrlAnalogEntryId) );
		break;

	case COLUMN_SAHPICTRLANALOGNUM:
		/** SaHpiInstrumentId = ASN_UNSIGNED */
		snmp_set_var_typed_value(var, ASN_UNSIGNED,
					 (u_char*)&context->saHpiCtrlAnalogNum,
					 sizeof(context->saHpiCtrlAnalogNum) );
		break;

	case COLUMN_SAHPICTRLANALOGOUTPUTTYPE:
		/** SaHpiCtrlOutputType = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogOutputType,
					 sizeof(context->saHpiCtrlAnalogOutputType) );
		break;

	case COLUMN_SAHPICTRLANALOGDEFAULTMODE:
		/** SaHpiCtrlMode = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogDefaultMode,
					 sizeof(context->saHpiCtrlAnalogDefaultMode) );
		break;

	case COLUMN_SAHPICTRLANALOGMODE:
		/** SaHpiCtrlMode = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogMode,
					 sizeof(context->saHpiCtrlAnalogMode) );
		break;

	case COLUMN_SAHPICTRLANALOGISREADONLY:
		/** TruthValue = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogIsReadOnly,
					 sizeof(context->saHpiCtrlAnalogIsReadOnly) );
		break;

	case COLUMN_SAHPICTRLANALOGISWRITEONLY:
		/** TruthValue = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogIsWriteOnly,
					 sizeof(context->saHpiCtrlAnalogIsWriteOnly) );
		break;

	case COLUMN_SAHPICTRLANALOGDEFAULTMINSTATE:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogDefaultMinState,
					 sizeof(context->saHpiCtrlAnalogDefaultMinState) );
		break;

	case COLUMN_SAHPICTRLANALOGDEFAULTMAXSTATE:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogDefaultMaxState,
					 sizeof(context->saHpiCtrlAnalogDefaultMaxState) );
		break;

	case COLUMN_SAHPICTRLANALOGDEFAULTSTATE:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogDefaultState,
					 sizeof(context->saHpiCtrlAnalogDefaultState) );
		break;

	case COLUMN_SAHPICTRLANALOGSTATE:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCtrlAnalogState,
					 sizeof(context->saHpiCtrlAnalogState) );
		break;

	case COLUMN_SAHPICTRLANALOGOem:
		/** UNSIGNED32 = ASN_UNSIGNED */
		snmp_set_var_typed_value(var, ASN_UNSIGNED,
					 (u_char*)&context->saHpiCtrlAnalogOem,
					 sizeof(context->saHpiCtrlAnalogOem) );
		break;

	case COLUMN_SAHPICTRLANALOGRDR:
		/** RowPointer = ASN_OBJECT_ID */
		snmp_set_var_typed_value(var, ASN_OBJECT_ID,
					 (u_char*)&context->saHpiCtrlAnalogRDR,
					 context->saHpiCtrlAnalogRDR_len );
		break;

	default: /** We shouldn't get here */
		snmp_log(LOG_ERR, "unknown column in "
			 "saHpiCtrlAnalogTable_get_value\n");
		return SNMP_ERR_GENERR;
	}
	return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiCtrlAnalogTable_get_by_idx
 */
const saHpiCtrlAnalogTable_context *
saHpiCtrlAnalogTable_get_by_idx(netsnmp_index * hdr)
{

 	DEBUGMSGTL ((AGENT, "saHpiCtrlAnalogTable_get_by_idx, called\n"));

	return(const saHpiCtrlAnalogTable_context *)
	CONTAINER_FIND(cb.container, hdr );
}


