/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiCurrentSensorStateTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiCurrentSensorStateTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceTable.h>
#include <session_info.h>

#include <oh_utils.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiCurrentSensorStateTable_oid[] = { saHpiCurrentSensorStateTable_TABLE_OID};
size_t saHpiCurrentSensorStateTable_oid_len = OID_LENGTH(saHpiCurrentSensorStateTable_oid);

/**
 * 
 * @sessionid
 * @rdr_entry
 * @rpt_entry
 * @full_oid
 * @full_oid_len
 * 
 * @return 
 */
SaErrorT populate_current_sensor_state(SaHpiSessionIdT sessionid, 
				       SaHpiRdrT *rdr_entry,
				       SaHpiRptEntryT *rpt_entry,
				       oid *full_oid, size_t full_oid_len)
{
	DEBUGMSGTL ((AGENT, "populate_current_sensor_state, called\n"));

	SaErrorT rv = SA_OK;
        int new_row = MIB_FALSE;

	oid             current_sensor_state_oid[CURRENT_SENSOR_STATE_INDEX_NR];
	netsnmp_index   current_sensor_state_index;
	saHpiCurrentSensorStateTable_context *current_sensor_state_context;

	SaHpiTextBufferT        buffer;

	SaHpiSensorReadingT     reading;
	SaHpiEventStateT        event_state;

	SaHpiBoolT              sensor_enabled;
	SaHpiBoolT              sensor_events_enabled;

	SaHpiEventStateT        assert_event_mask;
	SaHpiEventStateT        deassert_event_mask;

	DEBUGMSGTL ((AGENT, "SAHPI_SENSOR_RDR populate_current_sensor_state() called\n"));

	/* check for NULL pointers */
	if (!rdr_entry) {
		DEBUGMSGTL ((AGENT, 
			     "ERROR: populate_current_sensor_state() passed NULL rdr_entry pointer\n"));

		return AGENT_ERR_INTERNAL_ERROR;
	}
	if (!rpt_entry) {
		DEBUGMSGTL ((AGENT, 
			     "ERROR: populate_current_sensor_state() passed NULL rdr_entry pointer\n"));

		return AGENT_ERR_INTERNAL_ERROR;
	}

        subagent_lock(&hpi_lock_data);

	/* BUILD oid for new row */
	/* assign the number of indices */
	current_sensor_state_index.len = CURRENT_SENSOR_STATE_INDEX_NR;
	/** Index saHpiDomainId is external */
	current_sensor_state_oid[0] = get_domain_id(sessionid);
	/** Index saHpiResourceId is external */
	current_sensor_state_oid[1] = rpt_entry->ResourceId;
	/** Index saHpiSensorNum */
	current_sensor_state_oid[2] = rdr_entry->RdrTypeUnion.SensorRec.Num;
	/* assign the indices to the index */
	current_sensor_state_index.oids = (oid *) & current_sensor_state_oid;

	/* See if Row exists. */
	current_sensor_state_context = NULL;
	current_sensor_state_context = CONTAINER_FIND(cb.container, &current_sensor_state_index);

	if (!current_sensor_state_context) {
		// New entry. Add it
		current_sensor_state_context = 
		saHpiCurrentSensorStateTable_create_row(&current_sensor_state_index);
                new_row = MIB_FALSE;
	}
	if (!current_sensor_state_context) {
		snmp_log (LOG_ERR, "Not enough memory for a Ctrl Text row!");
		
		subagent_unlock(&hpi_lock_data);
		
		return AGENT_ERR_INTERNAL_ERROR;
	}

	/** INTEGER = ASN_INTEGER */
	if (!rdr_entry->RdrTypeUnion.SensorRec.Events && 
	    !rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported) {
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		4; /* noData(4) */
	} else if (rdr_entry->RdrTypeUnion.SensorRec.Events && 
		   rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported) {
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		3; /* readingAndEventState(3) */
	} else if (rdr_entry->RdrTypeUnion.SensorRec.Events) {
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		2; /* eventState(2) */
	} else { /* (rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported) */
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		1; /* reading(1) */
	}

	/* if state or event support exist get */
	if (current_sensor_state_context->saHpiCurrentSensorStateDataPresent) {
		rv = saHpiSensorReadingGet (sessionid, 
					    rpt_entry->ResourceId, 
					    rdr_entry->RdrTypeUnion.SensorRec.Num,
					    &reading,
					    &event_state);
		if (rv != SA_OK) {
			DEBUGMSGTL ((AGENT, 
				     "ERROR: populate_current_sensor_state() "
                                     "oh_decode_eventstate() ERRORED out\n"));
			saHpiCurrentSensorStateTable_delete_row( current_sensor_state_context );
			
			subagent_unlock(&hpi_lock_data);
			
			return AGENT_ERR_INTERNAL_ERROR;
		}
	}

	/* If reading supported populate the following two elements */
	if (rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported == SAHPI_TRUE) {
		/** SaHpiSensorReadingType = ASN_INTEGER */
		current_sensor_state_context->saHpiCurrentSensorStateType =
		rdr_entry->RdrTypeUnion.SensorRec.DataFormat.ReadingType + 1;

		/** SaHpiSensorReadingValue = ASN_OCTET_STR */
		current_sensor_state_context->saHpiCurrentSensorStateValue_len = 
		set_sensor_reading_value(&reading,
					 current_sensor_state_context->saHpiCurrentSensorStateEventState);
	}

	/** SaHpiEventState = ASN_OCTET_STR */
	if (rdr_entry->RdrTypeUnion.SensorRec.Events) {
		rv = oh_decode_eventstate(event_state,
					  rdr_entry->RdrTypeUnion.SensorRec.Category,
					  &buffer);
		oh_decode_char(&buffer);
		memcpy(current_sensor_state_context->saHpiCurrentSensorStateEventState,
		       buffer.Data, buffer.DataLength);
		current_sensor_state_context->saHpiCurrentSensorStateEventState_len =
		buffer.DataLength;
	}

	/** TruthValue = ASN_INTEGER */
	rv = saHpiSensorEnableGet(sessionid, rpt_entry->ResourceId, 
				  rdr_entry->RdrTypeUnion.SensorRec.Num, 
				  &sensor_enabled);
	current_sensor_state_context->saHpiCurrentSensorStateSensorEnable =
	(sensor_enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;    

	/** TruthValue = ASN_INTEGER */
	rv = saHpiSensorEventEnableGet(sessionid, rpt_entry->ResourceId, 
				       rdr_entry->RdrTypeUnion.SensorRec.Num,
				       &sensor_events_enabled);
	current_sensor_state_context->saHpiCurrentSensorStateEventEnable =
	(sensor_events_enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;



	/*******************************************************/
	/* get EventMasks and convert to string representation */
	/*******************************************************/
	rv = saHpiSensorEventMasksGet (sessionid,
				       rpt_entry->ResourceId,
				       rdr_entry->RdrTypeUnion.SensorRec.Num,
				       &assert_event_mask,
				       &deassert_event_mask);

	/*******************************************************/
	/* decode assert values  			       */
	/*******************************************************/
	rv = oh_decode_eventstate(assert_event_mask,
				  rdr_entry->RdrTypeUnion.SensorRec.Category,
				  &buffer);
	oh_decode_char(&buffer);

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateAssertAddEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateAssertAddEventMask_len =
	buffer.DataLength;

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateAssertRemoveEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateAssertRemoveEventMask_len =
	buffer.DataLength;

	/*******************************************************/
	/* decode deassert values  			       */
	/*******************************************************/
	rv = oh_decode_eventstate(deassert_event_mask,
				  rdr_entry->RdrTypeUnion.SensorRec.Category,
				  &buffer);
	oh_decode_char(&buffer);

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateDeassertAddEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateDeassertAddEventMask_len =
	buffer.DataLength;                    

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateDeassertRemoveEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateDeassertRemoveEventMask_len =
	buffer.DataLength;  

	if (new_row == MIB_TRUE)
                CONTAINER_INSERT (cb.container, current_sensor_state_context);

        subagent_unlock(&hpi_lock_data);

	return rv;
}  

SaErrorT async_sensor_enable_change_add(SaHpiSessionIdT sessionid,
                                        SaHpiEventT *event, 
                                        SaHpiRdrT *rdr_entry, 
                                        SaHpiRptEntryT *rpt_entry)
{
	SaErrorT rv = SA_OK;

	oid             current_sensor_state_oid[CURRENT_SENSOR_STATE_INDEX_NR];
	netsnmp_index   current_sensor_state_index;
	saHpiCurrentSensorStateTable_context *current_sensor_state_context;

	SaHpiTextBufferT        buffer;

	SaHpiSensorReadingT     reading;
	SaHpiEventStateT        event_state;

	SaHpiBoolT              sensor_enabled;
	SaHpiBoolT              sensor_events_enabled;

	SaHpiEventStateT        assert_event_mask;
	SaHpiEventStateT        deassert_event_mask;

	DEBUGMSGTL ((AGENT, "async_sensor_enable_change_add, called\n"));

	/* check for NULL pointers */
	if (!rdr_entry) {
		DEBUGMSGTL ((AGENT, 
			     "ERROR: async_sensor_enable_change_add passed "
                             "NULL rdr_entry pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}
	if (!rpt_entry) {
		DEBUGMSGTL ((AGENT, 
			     "ERROR: async_sensor_enable_change_add passed "
                             "NULL rdr_entry pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}

	/* BUILD oid for new row */
	/* assign the number of indices */
	current_sensor_state_index.len = CURRENT_SENSOR_STATE_INDEX_NR;
	/** Index saHpiDomainId is external */
	current_sensor_state_oid[0] = get_domain_id(sessionid);
	/** Index saHpiResourceId is external */
	current_sensor_state_oid[1] = event->Source;
	/** Index saHpiSensorNum */
	current_sensor_state_oid[2] = event->EventDataUnion.SensorEnableChangeEvent.SensorNum;
	/* assign the indices to the index */
	current_sensor_state_index.oids = (oid *) & current_sensor_state_oid;

	/* See if Row exists. */
	current_sensor_state_context = NULL;
	current_sensor_state_context = CONTAINER_FIND(cb.container, 
                                                      &current_sensor_state_index);

	if (!current_sensor_state_context) {
		DEBUGMSGTL ((AGENT, "async_sensor_enable_change_add: "
                          "row not found!\n"));
		return AGENT_ERR_NOT_FOUND;
	}

	/** INTEGER = ASN_INTEGER */  
	if (!rdr_entry->RdrTypeUnion.SensorRec.Events && 
	    !rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported) {
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		4; /* noData(4) */
	} else if (rdr_entry->RdrTypeUnion.SensorRec.Events && 
		   rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported) {
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		3; /* readingAndEventState(3) */
	} else if (rdr_entry->RdrTypeUnion.SensorRec.Events) {
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		2; /* eventState(2) */
	} else { /* (rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported) */
		current_sensor_state_context->saHpiCurrentSensorStateDataPresent =
		1; /* reading(1) */
	}

	/* if state or event support exist get */
	if (current_sensor_state_context->saHpiCurrentSensorStateDataPresent) {
		rv = saHpiSensorReadingGet (sessionid, 
					    rpt_entry->ResourceId, 
					    rdr_entry->RdrTypeUnion.SensorRec.Num,
					    &reading,
					    &event_state);
		if (rv != SA_OK) {
			DEBUGMSGTL ((AGENT, 
				     "ERROR: async_sensor_enable_change_add"
                                     " oh_decode_eventstate() ERRORED out\n"));
			return AGENT_ERR_INTERNAL_ERROR;
		}
	}

	/* If reading supported populate the following two elements */
	if (rdr_entry->RdrTypeUnion.SensorRec.DataFormat.IsSupported == SAHPI_TRUE) {
		/** SaHpiSensorReadingType = ASN_INTEGER */
		current_sensor_state_context->saHpiCurrentSensorStateType =
		rdr_entry->RdrTypeUnion.SensorRec.DataFormat.ReadingType + 1;

		/** SaHpiSensorReadingValue = ASN_OCTET_STR */
		current_sensor_state_context->saHpiCurrentSensorStateValue_len = 
		set_sensor_reading_value(&reading,
                    current_sensor_state_context->saHpiCurrentSensorStateEventState);
	}

	/** SaHpiEventState = ASN_OCTET_STR */
	if (rdr_entry->RdrTypeUnion.SensorRec.Events) {
		rv = oh_decode_eventstate(event_state,
					  rdr_entry->RdrTypeUnion.SensorRec.Category,
					  &buffer);
		oh_decode_char(&buffer);
		memcpy(current_sensor_state_context->saHpiCurrentSensorStateEventState,
		       buffer.Data, buffer.DataLength);
		current_sensor_state_context->saHpiCurrentSensorStateEventState_len =
		buffer.DataLength;
	}

	/** TruthValue = ASN_INTEGER */
	rv = saHpiSensorEnableGet(sessionid, rpt_entry->ResourceId, 
				  rdr_entry->RdrTypeUnion.SensorRec.Num, 
				  &sensor_enabled);
	current_sensor_state_context->saHpiCurrentSensorStateSensorEnable =
	(sensor_enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;    

	/** TruthValue = ASN_INTEGER */
	rv = saHpiSensorEventEnableGet(sessionid, rpt_entry->ResourceId, 
				       rdr_entry->RdrTypeUnion.SensorRec.Num,
				       &sensor_events_enabled);
	current_sensor_state_context->saHpiCurrentSensorStateEventEnable =
	(sensor_events_enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;



	/*******************************************************/
	/* get EventMasks and convert to string representation */
	/*******************************************************/
	rv = saHpiSensorEventMasksGet (sessionid,
				       rpt_entry->ResourceId,
				       rdr_entry->RdrTypeUnion.SensorRec.Num,
				       &assert_event_mask,
				       &deassert_event_mask);

	/*******************************************************/
	/* decode assert values  			       */
	/*******************************************************/
	rv = oh_decode_eventstate(assert_event_mask,
				  rdr_entry->RdrTypeUnion.SensorRec.Category,
				  &buffer);
	oh_decode_char(&buffer);

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateAssertAddEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateAssertAddEventMask_len =
	buffer.DataLength;

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateAssertRemoveEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateAssertRemoveEventMask_len =
	buffer.DataLength;

	/*******************************************************/
	/* decode deassert values  			       */
	/*******************************************************/
	rv = oh_decode_eventstate(deassert_event_mask,
				  rdr_entry->RdrTypeUnion.SensorRec.Category,
				  &buffer);
	oh_decode_char(&buffer);

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateDeassertAddEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateDeassertAddEventMask_len =
	buffer.DataLength;                    

	/** SaHpiEventState = ASN_OCTET_STR */
	memcpy(current_sensor_state_context->saHpiCurrentSensorStateDeassertRemoveEventMask,
	       buffer.Data, buffer.DataLength);
	current_sensor_state_context->saHpiCurrentSensorStateDeassertRemoveEventMask_len =
	buffer.DataLength;  

	return rv;
}


/**
 * 
 * @domainId
 * @resourceId
 * 
 * @return 
 */
SaErrorT clear_current_sensor_state(SaHpiDomainIdT domainId, 
                                    SaHpiResourceIdT resourceId)

{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiCurrentSensorStateTable_context *ctx;

	DEBUGMSGTL ((AGENT, "clear_current_sensor_state, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domainId));	
	DEBUGMSGTL ((AGENT, "           resourceId [%d]\n", resourceId));

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        /* based on the found row_idx get the pointer   */
                        /* to its context (row data)                    */
                        ctx = CONTAINER_FIND(cb.container, row_idx);

                        /* before we delete the context we should get the  */
                        /* next row (context) if any before we delete this */ 
                        /* one.                                            */
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if ((ctx->index.oids[saHpiDomainId_INDEX] ==
                             domainId) &&

                            (ctx->index.oids[saHpiResourceEntryId_INDEX] ==
                             resourceId)) {

                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctx);
                                saHpiCurrentSensorStateTable_delete_row (ctx);
                                DEBUGMSGTL ((AGENT, "clear_current_sensor_state: "
                                                    "found row: removing\n"));

                        }

                } while (row_idx);
        } 

        return rv;
}


/*
 * int set_table_sensor_enable (saHpiCurrentSensorStateTable_context *row_ctx) 
 */ 
int set_table_sensor_enable (saHpiCurrentSensorStateTable_context *row_ctx) 
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiSensorNumT     sensor_num;
	SaHpiBoolT          sensor_enabled;


	DEBUGMSGTL ((AGENT, "set_table_sensor_enable, called\n"));  

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceEntryId_INDEX];
	sensor_num = row_ctx->index.oids[saHpiSensorNumCurrent_INDEX]; 
	sensor_enabled = 
	(row_ctx->saHpiCurrentSensorStateSensorEnable == MIB_TRUE ) ? 
	SAHPI_TRUE : SAHPI_FALSE;

	rc = saHpiSensorEnableSet(session_id, resource_id, 
				  sensor_num, sensor_enabled);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "saHpiSensorEnableSet failed to set Mode rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "saHpiSensorEnableSet failed to set Mode rc: %s.\n",
			     oh_lookup_error(rc)));
		return get_snmp_error(rc);
	}

	return SNMP_ERR_NOERROR;
}

/*
 * int set_table_sensor_enable (saHpiCurrentSensorStateTable_context *row_ctx) 
 */ 
int set_table_sensor_event_enable (saHpiCurrentSensorStateTable_context *row_ctx) 
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiSensorNumT     sensor_num;
	SaHpiBoolT          sensor_event_enabled;


	DEBUGMSGTL ((AGENT, "set_table_sensor_enable, called\n"));  

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceEntryId_INDEX];
	sensor_num = row_ctx->index.oids[saHpiSensorNumCurrent_INDEX]; 
	sensor_event_enabled = 
	(row_ctx->saHpiCurrentSensorStateEventEnable == MIB_TRUE ) ? 
	SAHPI_TRUE : SAHPI_FALSE;

	rc = saHpiSensorEventEnableSet(session_id, resource_id, 
				       sensor_num, sensor_event_enabled);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "saHpiSensorEventEnableSet failed to set Mode rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "saHpiSensorEventEnableSet failed to set Mode rc: %s.\n",
			     oh_lookup_error(rc)));
		return get_snmp_error(rc);
	}

	return SNMP_ERR_NOERROR;
}

/*
 * int set_table_sensor_assert (saHpiCurrentSensorStateTable_context *row_ctx) 
 */ 
int set_table_sensor_masks (saHpiCurrentSensorStateTable_context *row_ctx, 
			    int column) 
{
	SaErrorT            rc = SA_OK;
	SaHpiTextBufferT    buffer;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiSensorNumT     sensor_num;
	SaHpiEventStateT    event_mask;
	SaHpiEventCategoryT event_cat;
	int rv = SNMP_ERR_NOERROR;

	DEBUGMSGTL ((AGENT, "set_table_sensor_enable, called\n"));  

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceEntryId_INDEX];
	sensor_num = row_ctx->index.oids[saHpiSensorNumCurrent_INDEX]; 

	switch (column) {
	case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
		memcpy(buffer.Data, 
		       row_ctx->saHpiCurrentSensorStateAssertAddEventMask, 
		       row_ctx->saHpiCurrentSensorStateAssertAddEventMask_len);
		buffer.DataLength = 
		row_ctx->saHpiCurrentSensorStateAssertAddEventMask_len;
		oh_encode_char(&buffer);
		rc = oh_encode_eventstate(&buffer, &event_mask, &event_cat);
		if (rc != SA_OK) {
			DEBUGMSGTL((AGENT, 
				    "set_table_sensor_masks: ASSERT_ADD_EVENT_MASK: FALIED\n"));
			break;
		}
		rc = saHpiSensorEventMasksSet (session_id, 
					       resource_id, 
					       sensor_num, 
					       SAHPI_SENS_ADD_EVENTS_TO_MASKS, 
					       event_mask, 
					       0);
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
		memcpy(buffer.Data, 
		       row_ctx->saHpiCurrentSensorStateAssertRemoveEventMask, 
		       row_ctx->saHpiCurrentSensorStateAssertRemoveEventMask_len);
		buffer.DataLength = 
		row_ctx->saHpiCurrentSensorStateAssertRemoveEventMask_len;
		oh_encode_char(&buffer);
		rc = oh_encode_eventstate(&buffer, &event_mask, &event_cat);
		if (rc != SA_OK) {
			DEBUGMSGTL((AGENT, 
				    "set_table_sensor_masks: ASSERT_REMOVE_EVENT_MASK: FALIED\n"));
			break;
		}
		rc = saHpiSensorEventMasksSet (session_id, 
					       resource_id, 
					       sensor_num, 
					       SAHPI_SENS_REMOVE_EVENTS_FROM_MASKS, 
					       event_mask, 
					       0);
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
		memcpy(buffer.Data, 
		       row_ctx->saHpiCurrentSensorStateDeassertAddEventMask, 
		       row_ctx->saHpiCurrentSensorStateDeassertAddEventMask_len);
		buffer.DataLength = 
		row_ctx->saHpiCurrentSensorStateDeassertAddEventMask_len;
		oh_encode_char(&buffer);
		rc = oh_encode_eventstate(&buffer, &event_mask, &event_cat);
		if (rc != SA_OK) {
			DEBUGMSGTL((AGENT, 
				    "set_table_sensor_masks: DEASSERT_ADD_EVENT_MASK: FALIED\n"));
			break;
		}
		rc = saHpiSensorEventMasksSet (session_id, 
					       resource_id, 
					       sensor_num, 
					       SAHPI_SENS_ADD_EVENTS_TO_MASKS, 
					       0, 
					       event_mask);
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
		memcpy(buffer.Data, 
		       row_ctx->saHpiCurrentSensorStateDeassertRemoveEventMask, 
		       row_ctx->saHpiCurrentSensorStateDeassertRemoveEventMask_len);
		buffer.DataLength = 
		row_ctx->saHpiCurrentSensorStateDeassertRemoveEventMask_len;
		oh_encode_char(&buffer);
		rc = oh_encode_eventstate(&buffer, &event_mask, &event_cat);
		if (rc != SA_OK) {
			DEBUGMSGTL((AGENT, 
				    "set_table_sensor_masks: DEASSERT_REMOVE_EVENT_MASK: FALIED\n"));
			break;
		}
		rc = saHpiSensorEventMasksSet (session_id, 
					       resource_id, 
					       sensor_num, 
					       SAHPI_SENS_REMOVE_EVENTS_FROM_MASKS, 
					       0, 
					       event_mask);
		break;

	default:
		DEBUGMSGTL ((AGENT, "set_table_sensor_masks: ERROR: \n"));  
		rv = SNMP_ERR_BADVALUE;
	}

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "saHpiSensorEventEnableSet failed to set Mode rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "saHpiSensorEventEnableSet failed to set Mode rc: %s.\n",
			     oh_lookup_error(rc)));
		rv =  get_snmp_error(rc);
	}

	return rv;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiCurrentSensorStateTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiCurrentSensorStateTable_cmp( const void *lhs, const void *rhs )
{
	saHpiCurrentSensorStateTable_context *context_l =
	(saHpiCurrentSensorStateTable_context *)lhs;
	saHpiCurrentSensorStateTable_context *context_r =
	(saHpiCurrentSensorStateTable_context *)rhs;

	/*
	 * check primary key, then secondary. Add your own code if
	 * there are more than 2 indexes
	 */
	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_cmp, called\n"));

	/* check for NULL pointers */
	if (lhs == NULL || rhs == NULL ) {
		DEBUGMSGTL((AGENT,"saHpiCurrentSensorStateTable_cmp() NULL pointer ERROR\n" ));
		return 0;
	}
	/* CHECK FIRST INDEX,  saHpiDomainId */
	if ( context_l->index.oids[0] < context_r->index.oids[0])
		return -1;

	if ( context_l->index.oids[0] > context_r->index.oids[0])
		return 1;

	if ( context_l->index.oids[0] == context_r->index.oids[0]) {
		/* If saHpiDomainId index is equal sort by second index */
		/* CHECK SECOND INDEX,  saHpiResourceEntryId */
		if ( context_l->index.oids[1] < context_r->index.oids[1])
			return -1;

		if ( context_l->index.oids[1] > context_r->index.oids[1])
			return 1;

		if ( context_l->index.oids[1] == context_r->index.oids[1]) {
			/* If saHpiResourceEntryId index is equal sort by third index */
			/* CHECK THIRD INDEX,  saHpiSensorNum */
			if ( context_l->index.oids[2] < context_r->index.oids[2])
				return -1;

			if ( context_l->index.oids[2] > context_r->index.oids[2])
				return 1;

			if ( context_l->index.oids[2] == context_r->index.oids[2]) {
				return 0;
			}
		}
	}

	return 0;
}

/************************************************************
 * Initializes the saHpiCurrentSensorStateTable module
 */
void
init_saHpiCurrentSensorStateTable(void)
{
	DEBUGMSGTL((AGENT, "Called init_saHpiCurrentSensorStateTable()\n"));

	initialize_table_saHpiCurrentSensorStateTable();

}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiCurrentSensorStateTable_row_copy(saHpiCurrentSensorStateTable_context * dst,
						 saHpiCurrentSensorStateTable_context * src)
{
	if (!dst||!src)
		return 1;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_row_copy, called\n"));

	subagent_lock(&hpi_lock_data);
	
	/*
	 * copy index, if provided
	 */
	if (dst->index.oids)
		free(dst->index.oids);
	if (snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
			    src->index.len * sizeof(oid) )) {
		dst->index.oids = NULL;
		subagent_unlock(&hpi_lock_data);
		return 1;
	}
	dst->index.len = src->index.len;


	/*
	 * copy components into the context structure
	 */
	/** TODO: add code for external index(s)! */
	dst->saHpiCurrentSensorStateDataPresent = src->saHpiCurrentSensorStateDataPresent;

	dst->saHpiCurrentSensorStateType = src->saHpiCurrentSensorStateType;

	memcpy( dst->saHpiCurrentSensorStateValue, src->saHpiCurrentSensorStateValue, src->saHpiCurrentSensorStateValue_len );
	dst->saHpiCurrentSensorStateValue_len = src->saHpiCurrentSensorStateValue_len;

	memcpy( dst->saHpiCurrentSensorStateEventState, src->saHpiCurrentSensorStateEventState, src->saHpiCurrentSensorStateEventState_len );
	dst->saHpiCurrentSensorStateEventState_len = src->saHpiCurrentSensorStateEventState_len;

	dst->saHpiCurrentSensorStateSensorEnable = src->saHpiCurrentSensorStateSensorEnable;

	dst->saHpiCurrentSensorStateEventEnable = src->saHpiCurrentSensorStateEventEnable;

	memcpy( dst->saHpiCurrentSensorStateAssertAddEventMask, src->saHpiCurrentSensorStateAssertAddEventMask, src->saHpiCurrentSensorStateAssertAddEventMask_len );
	dst->saHpiCurrentSensorStateAssertAddEventMask_len = src->saHpiCurrentSensorStateAssertAddEventMask_len;

	memcpy( dst->saHpiCurrentSensorStateAssertRemoveEventMask, src->saHpiCurrentSensorStateAssertRemoveEventMask, src->saHpiCurrentSensorStateAssertRemoveEventMask_len );
	dst->saHpiCurrentSensorStateAssertRemoveEventMask_len = src->saHpiCurrentSensorStateAssertRemoveEventMask_len;

	memcpy( dst->saHpiCurrentSensorStateDeassertAddEventMask, src->saHpiCurrentSensorStateDeassertAddEventMask, src->saHpiCurrentSensorStateDeassertAddEventMask_len );
	dst->saHpiCurrentSensorStateDeassertAddEventMask_len = src->saHpiCurrentSensorStateDeassertAddEventMask_len;

	memcpy( dst->saHpiCurrentSensorStateDeassertRemoveEventMask, src->saHpiCurrentSensorStateDeassertRemoveEventMask, src->saHpiCurrentSensorStateDeassertRemoveEventMask_len );
	dst->saHpiCurrentSensorStateDeassertRemoveEventMask_len = src->saHpiCurrentSensorStateDeassertRemoveEventMask_len;


	subagent_unlock(&hpi_lock_data);
	return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiCurrentSensorStateTable_extract_index( saHpiCurrentSensorStateTable_context * ctx, netsnmp_index * hdr )
{
	/*
	 * temporary local storage for extracting oid index
	 *
	 * extract index uses varbinds (netsnmp_variable_list) to parse
	 * the index OID into the individual components for each index part.
	 */
	/** TODO: add storage for external index(s)! */
	netsnmp_variable_list var_saHpiDomainId;
	netsnmp_variable_list var_saHpiResourceId;
	netsnmp_variable_list var_saHpiSensorNum;
	int err;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_extract_index, called\n"));

	
	subagent_lock(&hpi_lock_data);
	/*
	 * copy index, if provided
	 */
	if (hdr) {
		netsnmp_assert(ctx->index.oids == NULL);
		if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
				    hdr->len * sizeof(oid) )) {
			subagent_unlock(&hpi_lock_data);
			return -1;
		}
		ctx->index.len = hdr->len;
	}

	/*
	 * initialize variable that will hold each component of the index.
	 * If there are multiple indexes for the table, the variable_lists
	 * need to be linked together, in order.
	 */
	/** TODO: add code for external index(s)! */
	memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
	var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiDomainId.next_variable = &var_saHpiResourceId;

	memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
	var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiResourceId.next_variable = &var_saHpiSensorNum;

	memset( &var_saHpiSensorNum, 0x00, sizeof(var_saHpiSensorNum) );
	var_saHpiSensorNum.type = ASN_UNSIGNED;	/* type hint for parse_oid_indexes */
	/** TODO: link this index to the next, or NULL for the last one */
	var_saHpiSensorNum.next_variable = NULL;


	/*
	 * parse the oid into the individual index components
	 */
	err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
	if (err == SNMP_ERR_NOERROR) {
		err = saHpiDomainId_check_index(
					       *var_saHpiDomainId.val.integer);
		err = saHpiResourceEntryId_check_index(
						      *var_saHpiResourceId.val.integer);  
		err = saHpiSensorNum_check_index(
						*var_saHpiSensorNum.val.integer);
	}

	/*
	 * parsing may have allocated memory. free it.
	 */
	snmp_reset_var_buffers( &var_saHpiDomainId );

	
	subagent_unlock(&hpi_lock_data);
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiCurrentSensorStateTable_can_activate(saHpiCurrentSensorStateTable_context *undo_ctx,
					      saHpiCurrentSensorStateTable_context *row_ctx,
					      netsnmp_request_group * rg)
{

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_can_activate, called\n"));
	/*
	 * TODO: check for activation requirements here
	 */


	/*
	 * be optimistic.
	 */
	return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiCurrentSensorStateTable_can_deactivate(saHpiCurrentSensorStateTable_context *undo_ctx,
						saHpiCurrentSensorStateTable_context *row_ctx,
						netsnmp_request_group * rg)
{

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_can_deactivate, called\n"));
	/*
	 * TODO: check for deactivation requirements here
	 */
	return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiCurrentSensorStateTable_can_delete(saHpiCurrentSensorStateTable_context *undo_ctx,
					    saHpiCurrentSensorStateTable_context *row_ctx,
					    netsnmp_request_group * rg)
{

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_can_delete, called\n"));

	/*
	 * probably shouldn't delete a row that we can't
	 * deactivate.
	 */
	if (saHpiCurrentSensorStateTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
		return 0;

	/*
	 * TODO: check for other deletion requirements here
	 */
	return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiCurrentSensorStateTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiCurrentSensorStateTable_context *
saHpiCurrentSensorStateTable_create_row( netsnmp_index* hdr)
{

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_create_row, called\n"));

	subagent_lock(&hpi_lock_data);
	
	saHpiCurrentSensorStateTable_context * ctx =
	SNMP_MALLOC_TYPEDEF(saHpiCurrentSensorStateTable_context);
	if (!ctx) {
		subagent_unlock(&hpi_lock_data);
		return NULL;
	}

	/*
	 * TODO: check indexes, if necessary.
	 */
	if (saHpiCurrentSensorStateTable_extract_index( ctx, hdr )) {
		free(ctx->index.oids);
		free(ctx);
		subagent_unlock(&hpi_lock_data);
		return NULL;
	}

	/* netsnmp_mutex_init(ctx->lock);
	   netsnmp_mutex_lock(ctx->lock); */

	/*
	 * TODO: initialize any default values here. This is also
	 * first place you really should allocate any memory for
	 * yourself to use.  If you allocated memory earlier,
	 * make sure you free it for earlier error cases!
	 */
	/**
	 ctx->saHpiCurrentSensorStateSensorEnable = 0;
	 ctx->saHpiCurrentSensorStateEventEnable = 0;
	 ctx->saHpiCurrentSensorStateAssertAddEventMask = 0;
	 ctx->saHpiCurrentSensorStateAssertRemoveEventMask = 0;
	 ctx->saHpiCurrentSensorStateDeassertAddEventMask = 0;
	 ctx->saHpiCurrentSensorStateDeassertRemoveEventMask = 0;
	*/

	subagent_unlock(&hpi_lock_data);
	return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiCurrentSensorStateTable_context *
saHpiCurrentSensorStateTable_duplicate_row( saHpiCurrentSensorStateTable_context * row_ctx)
{
	saHpiCurrentSensorStateTable_context * dup;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_duplicate_row, called\n"));

	if (!row_ctx)
		return NULL;

	subagent_lock(&hpi_lock_data);
	
	dup = SNMP_MALLOC_TYPEDEF(saHpiCurrentSensorStateTable_context);
	if (!dup) {
		subagent_unlock(&hpi_lock_data);
		return NULL;
	}
	if (saHpiCurrentSensorStateTable_row_copy(dup,row_ctx)) {
		free(dup);
		dup = NULL;
	}

	subagent_unlock(&hpi_lock_data);
	return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiCurrentSensorStateTable_delete_row( saHpiCurrentSensorStateTable_context * ctx )
{

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_delete_row, called\n"));

	/* netsnmp_mutex_destroy(ctx->lock); */

	subagent_lock(&hpi_lock_data);
	
	if (ctx->index.oids)
		free(ctx->index.oids);

	/*
	 * TODO: release any memory you allocated here...
	 */

	/*
	 * release header
	 */
	free( ctx );

	subagent_unlock(&hpi_lock_data);
	
	return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiCurrentSensorStateTable_set_reserve1( netsnmp_request_group *rg )
{
	saHpiCurrentSensorStateTable_context *row_ctx =
	(saHpiCurrentSensorStateTable_context *)rg->existing_row;
//	saHpiCurrentSensorStateTable_context *undo_ctx =
//	(saHpiCurrentSensorStateTable_context *)rg->undo_info;
	netsnmp_variable_list *var;
	netsnmp_request_group_item *current;
	int rc;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_set_reserve1, called\n"));

	/*
	 * TODO: loop through columns, check syntax and lengths. For
	 * columns which have no dependencies, you could also move
	 * the value/range checking here to attempt to catch error
	 * cases as early as possible.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;
		rc = SNMP_ERR_NOERROR;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
			/** TruthValue = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
				sizeof(row_ctx->saHpiCurrentSensorStateSensorEnable));
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
			/** TruthValue = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
				sizeof(row_ctx->saHpiCurrentSensorStateEventEnable));
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
			if (rc == SNMP_ERR_NOERROR ) {
				if (var->val_len > 
				    sizeof(row_ctx->saHpiCurrentSensorStateAssertAddEventMask)) {
					DEBUGMSGTL ((AGENT, 
					"COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK ERROR: %d\n", 
					rc));
					rc = SNMP_ERR_WRONGLENGTH;
				}
			}
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
			if (rc == SNMP_ERR_NOERROR ) {
				if (var->val_len > 
				    sizeof(row_ctx->saHpiCurrentSensorStateAssertRemoveEventMask)) {
					DEBUGMSGTL ((AGENT, 
					"COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK ERROR: %d\n", rc));
					rc = SNMP_ERR_WRONGLENGTH;
				}
			}
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
			if (rc == SNMP_ERR_NOERROR ) {
				if (var->val_len > 
				    sizeof(row_ctx->saHpiCurrentSensorStateDeassertAddEventMask)) {
					DEBUGMSGTL ((AGENT, 
					"COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK ERROR: %d\n", rc));
					rc = SNMP_ERR_WRONGLENGTH;
				}
			}
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
			if (rc == SNMP_ERR_NOERROR ) {
				if (var->val_len > 
				    sizeof(row_ctx->saHpiCurrentSensorStateDeassertRemoveEventMask)) {
					DEBUGMSGTL ((AGENT, 
					"COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK ERROR: %d\n", rc));
					rc = SNMP_ERR_WRONGLENGTH;
				}
			}
			break;

		default: /** We shouldn't get here */
			rc = SNMP_ERR_GENERR;
			snmp_log(LOG_ERR, "unknown column in "
				 "saHpiCurrentSensorStateTable_set_reserve1\n");
		}

		if (rc)
			netsnmp_request_set_error( current->ri, rc );
		rg->status = SNMP_MAX( rg->status, current->ri->status );
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

void saHpiCurrentSensorStateTable_set_reserve2( netsnmp_request_group *rg )
{
//	saHpiCurrentSensorStateTable_context *row_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->existing_row;
//	saHpiCurrentSensorStateTable_context *undo_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;
	netsnmp_variable_list *var;
	int rc;

	SaHpiTextBufferT    buffer;
	SaHpiEventStateT    event_state;
	SaHpiEventCategoryT event_cat;
	SaErrorT            rv;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_set_reserve2, called\n"));

	rg->rg_void = rg->list->ri;

	/*
	 * TODO: loop through columns, check for valid
	 * values and any range constraints.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;
		rc = SNMP_ERR_NOERROR;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
			/** TruthValue = ASN_INTEGER */
			rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
			/** TruthValue = ASN_INTEGER */
			rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(buffer.Data, var->val.string, var->val_len);
			buffer.DataLength = var->val_len;
			oh_encode_char(&buffer);
			rv = oh_encode_eventstate(&buffer,
						  &event_state,
						  &event_cat);
			if (rv != SA_OK) {
				DEBUGMSGTL((AGENT, 
					    "reserve2: ASSERT_ADD_EVENT_MASK: FALIED\n"));
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(buffer.Data, var->val.string, var->val_len);
			buffer.DataLength = var->val_len;
			oh_encode_char(&buffer);
			rv = oh_encode_eventstate(&buffer,
						  &event_state,
						  &event_cat);
			if (rv != SA_OK) {
				DEBUGMSGTL((AGENT, 
					    "reserve2: ASSERT_REMOVE_EVENT_MASK: FALIED\n"));
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(buffer.Data, var->val.string, var->val_len);
			buffer.DataLength = var->val_len;
			oh_encode_char(&buffer);
			rv = oh_encode_eventstate(&buffer,
						  &event_state,
						  &event_cat);
			if (rv != SA_OK) {
				DEBUGMSGTL((AGENT, 
					    "reserve2: DEASSERT_ADD_EVENT_MASK: FALIED\n"));
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(buffer.Data, var->val.string, var->val_len);
			buffer.DataLength = var->val_len;
			oh_encode_char(&buffer);
			rv = oh_encode_eventstate(&buffer,
						  &event_state,
						  &event_cat);
			if (rv != SA_OK) {
				DEBUGMSGTL((AGENT, 
					    "reserve2: DEASSERT_REMOVE_EVENT_MASK: FALIED\n"));
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}

		if (rc)
			netsnmp_request_set_error( current->ri, rc);
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiCurrentSensorStateTable_set_action( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
	saHpiCurrentSensorStateTable_context *row_ctx = (saHpiCurrentSensorStateTable_context *)rg->existing_row;
//	saHpiCurrentSensorStateTable_context *undo_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	int            row_err = 0;

        subagent_lock(&hpi_lock_data);

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_set_action, called\n"));

	/*
	 * TODO: loop through columns, copy varbind values
	 * to context structure for the row.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
			/** TruthValue = ASN_INTEGER */
			row_ctx->saHpiCurrentSensorStateSensorEnable = *var->val.integer;
			row_err = set_table_sensor_enable (row_ctx);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
			/** TruthValue = ASN_INTEGER */
			row_ctx->saHpiCurrentSensorStateEventEnable = *var->val.integer;
			row_err = set_table_sensor_event_enable (row_ctx);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(row_ctx->saHpiCurrentSensorStateAssertAddEventMask,
			       var->val.string, var->val_len);
			row_ctx->saHpiCurrentSensorStateAssertAddEventMask_len = 
			var->val_len;
			row_err = set_table_sensor_masks (row_ctx, 
							  COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(row_ctx->saHpiCurrentSensorStateAssertRemoveEventMask,
			       var->val.string, var->val_len);
			row_ctx->saHpiCurrentSensorStateAssertRemoveEventMask_len = 
			var->val_len;
			row_err = set_table_sensor_masks (row_ctx, 
							  COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(row_ctx->saHpiCurrentSensorStateDeassertAddEventMask,
			       var->val.string, var->val_len);
			row_ctx->saHpiCurrentSensorStateDeassertAddEventMask_len = 
			var->val_len;
			row_err = set_table_sensor_masks (row_ctx, 
							  COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK);
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			memcpy(row_ctx->saHpiCurrentSensorStateDeassertRemoveEventMask,
			       var->val.string, var->val_len);
			row_ctx->saHpiCurrentSensorStateDeassertRemoveEventMask_len = 
			var->val_len;
			row_err = set_table_sensor_masks (row_ctx, 
							  COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK);
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	if (row_err) {
		netsnmp_request_set_error(
					       (netsnmp_request_info*)rg->rg_void,
					       row_err);
	}

        subagent_unlock(&hpi_lock_data);

        return;

}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiCurrentSensorStateTable_set_commit( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiCurrentSensorStateTable_context *row_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->existing_row;
//	saHpiCurrentSensorStateTable_context *undo_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_set_commit, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
			/** TruthValue = ASN_INTEGER */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
			/** TruthValue = ASN_INTEGER */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiCurrentSensorStateTable_set_free( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiCurrentSensorStateTable_context *row_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->existing_row;
//	saHpiCurrentSensorStateTable_context *undo_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_set_free, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
			/** TruthValue = ASN_INTEGER */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
			/** TruthValue = ASN_INTEGER */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		default: 
			break;
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiCurrentSensorStateTable_set_undo( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiCurrentSensorStateTable_context *row_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->existing_row;
//	saHpiCurrentSensorStateTable_context *undo_ctx = 
// 		(saHpiCurrentSensorStateTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_set_undo, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
			/** TruthValue = ASN_INTEGER */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
			/** TruthValue = ASN_INTEGER */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
			/** SaHpiEventState = ASN_OCTET_STR */
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 *
 * Initialize the saHpiCurrentSensorStateTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiCurrentSensorStateTable(void)
{
	netsnmp_table_registration_info *table_info;

	DEBUGMSGTL ((AGENT, "initialize_table_saHpiCurrentSensorStateTable, called\n"));

	if (my_handler) {
		snmp_log(LOG_ERR, "initialize_table_saHpiCurrentSensorStateTable_handler called again\n");
		return;
	}

	memset(&cb, 0x00, sizeof(cb));

	/** create the table structure itself */
	table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

	/* if your table is read only, it's easiest to change the
	   HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
	my_handler = netsnmp_create_handler_registration("saHpiCurrentSensorStateTable",
							 netsnmp_table_array_helper_handler,
							 saHpiCurrentSensorStateTable_oid,
							 saHpiCurrentSensorStateTable_oid_len,
							 HANDLER_CAN_RWRITE);

	if (!my_handler || !table_info) {
		snmp_log(LOG_ERR, "malloc failed in "
			 "initialize_table_saHpiCurrentSensorStateTable_handler\n");
		return;	/** mallocs failed */
	}

	/***************************************************
	 * Setting up the table's definition
	 */
	/*
	 * TODO: add any external indexes here.
	 */
	/** TODO: add code for external index(s)! */

	/*
	 * internal indexes
	 */
	/** index: saHpiDomainId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
	/** index: saHpiResourceId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
	/** index: saHpiSensorNum */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

	table_info->min_column = saHpiCurrentSensorStateTable_COL_MIN;
	table_info->max_column = saHpiCurrentSensorStateTable_COL_MAX;

	/***************************************************
	 * registering the table with the master agent
	 */
	cb.get_value = saHpiCurrentSensorStateTable_get_value;
	cb.container = netsnmp_container_find("saHpiCurrentSensorStateTable_primary:"
					      "saHpiCurrentSensorStateTable:"
					      "table_container");

	netsnmp_container_add_index(cb.container,
				    netsnmp_container_find("saHpiCurrentSensorStateTable_secondary:"
							   "saHpiCurrentSensorStateTable:"
							   "table_container"));
	cb.container->next->compare = saHpiCurrentSensorStateTable_cmp;


	cb.can_set = 1;

	cb.create_row = (UserRowMethod*)saHpiCurrentSensorStateTable_create_row;

	cb.duplicate_row = (UserRowMethod*)saHpiCurrentSensorStateTable_duplicate_row;
	cb.delete_row = (UserRowMethod*)saHpiCurrentSensorStateTable_delete_row;
	cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiCurrentSensorStateTable_row_copy;

	cb.can_activate = (Netsnmp_User_Row_Action *)saHpiCurrentSensorStateTable_can_activate;
	cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiCurrentSensorStateTable_can_deactivate;
	cb.can_delete = (Netsnmp_User_Row_Action *)saHpiCurrentSensorStateTable_can_delete;

	cb.set_reserve1 = saHpiCurrentSensorStateTable_set_reserve1;
	cb.set_reserve2 = saHpiCurrentSensorStateTable_set_reserve2;
	cb.set_action = saHpiCurrentSensorStateTable_set_action;
	cb.set_commit = saHpiCurrentSensorStateTable_set_commit;
	cb.set_free = saHpiCurrentSensorStateTable_set_free;
	cb.set_undo = saHpiCurrentSensorStateTable_set_undo;

	DEBUGMSGTL(("initialize_table_saHpiCurrentSensorStateTable",
		    "Registering table saHpiCurrentSensorStateTable "
		    "as a table array\n"));
	netsnmp_table_container_register(my_handler, table_info, &cb,
					 cb.container, 1);
}

/************************************************************
 * saHpiCurrentSensorStateTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiCurrentSensorStateTable_get_value(
					  netsnmp_request_info *request,
					  netsnmp_index *item,
					  netsnmp_table_request_info *table_info )
{
	netsnmp_variable_list *var = request->requestvb;
	saHpiCurrentSensorStateTable_context *context = (saHpiCurrentSensorStateTable_context *)item;

	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_get_value, called\n"));

	switch (table_info->colnum) {
	
	case COLUMN_SAHPICURRENTSENSORSTATEDATAPRESENT:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCurrentSensorStateDataPresent,
					 sizeof(context->saHpiCurrentSensorStateDataPresent) );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATETYPE:
		/** SaHpiSensorReadingType = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCurrentSensorStateType,
					 sizeof(context->saHpiCurrentSensorStateType) );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEVALUE:
		/** SaHpiSensorReadingValue = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (u_char*)&context->saHpiCurrentSensorStateValue,
					 context->saHpiCurrentSensorStateValue_len );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEEVENTSTATE:
		/** SaHpiEventState = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (u_char*)&context->saHpiCurrentSensorStateEventState,
					 context->saHpiCurrentSensorStateEventState_len );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATESENSORENABLE:
		/** TruthValue = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCurrentSensorStateSensorEnable,
					 sizeof(context->saHpiCurrentSensorStateSensorEnable) );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEEVENTENABLE:
		/** TruthValue = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (u_char*)&context->saHpiCurrentSensorStateEventEnable,
					 sizeof(context->saHpiCurrentSensorStateEventEnable) );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEASSERTADDEVENTMASK:
		/** SaHpiEventState = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (u_char*)&context->saHpiCurrentSensorStateAssertAddEventMask,
					 context->saHpiCurrentSensorStateAssertAddEventMask_len );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEASSERTREMOVEEVENTMASK:
		/** SaHpiEventState = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (u_char*)&context->saHpiCurrentSensorStateAssertRemoveEventMask,
					 context->saHpiCurrentSensorStateAssertRemoveEventMask_len );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTADDEVENTMASK:
		/** SaHpiEventState = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (u_char*)&context->saHpiCurrentSensorStateDeassertAddEventMask,
					 context->saHpiCurrentSensorStateDeassertAddEventMask_len );
		break;

	case COLUMN_SAHPICURRENTSENSORSTATEDEASSERTREMOVEEVENTMASK:
		/** SaHpiEventState = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (u_char*)&context->saHpiCurrentSensorStateDeassertRemoveEventMask,
					 context->saHpiCurrentSensorStateDeassertRemoveEventMask_len );
		break;

	default: /** We shouldn't get here */
		snmp_log(LOG_ERR, "unknown column in "
			 "saHpiCurrentSensorStateTable_get_value\n");
		return SNMP_ERR_GENERR;
	}
	return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiCurrentSensorStateTable_get_by_idx
 */
const saHpiCurrentSensorStateTable_context *
saHpiCurrentSensorStateTable_get_by_idx(netsnmp_index * hdr)
{
	DEBUGMSGTL ((AGENT, "saHpiCurrentSensorStateTable_get_by_idx, called\n"));

	return(const saHpiCurrentSensorStateTable_context *)
	CONTAINER_FIND(cb.container, hdr );
}
