/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiDomainAlarmTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiDomainAlarmTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>

#include <hpiDomainAlarmMapping.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiDomainAlarmTable_oid[] = { saHpiDomainAlarmTable_TABLE_OID };
size_t saHpiDomainAlarmTable_oid_len = OID_LENGTH(saHpiDomainAlarmTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int doma_initialized = FALSE;               
static GHashTable *doma_de_table;


/*************************************************************
 * oid and fucntion declarations scalars
 */

static u_long domain_alarm_entry_count = 0;

static oid saHpiDomainAlarmActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,2,5 };

int  handle_saHpiDomainAlarmActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);
					
int initialize_table_saHpiDomainAlarmActiveEntries(void);

/**
 * 
 * @sessionid:
 * 
 * @return:
 */
SaErrorT populate_saHpiDomainAlarmTable(SaHpiSessionIdT sessionid)
{
	SaErrorT rv = SA_OK;
        SaErrorT rc = SA_OK;
        int new_row = MIB_FALSE;

	oid domain_alarm_oid[DOMAIN_ALARM_INDEX_NR];
	netsnmp_index domain_alarm_idx;
	saHpiDomainAlarmTable_context *domain_alarm_ctx;
	
	/* Used for saHpiDomainAlarmGetNext call */
        SaHpiSeverityT Severity;
	SaHpiAlarmT Alarm;
	
	/* Used for decoding */
	oh_big_textbuffer bigbuf;
	SaHpiTextBufferT buf;
	
	/* Used for mapping */
        DE_XREF *de_entry;
        doma_keys keys;
	
        DEBUGMSGTL ((AGENT, "populate_saHpiDomainAlarmTable, called\n"));		
		
       	/* Get the first DAT Entry for this Session */	
	Alarm.AlarmId = SAHPI_FIRST_ENTRY;
	/* Get all the severities */
	Severity = SAHPI_ALL_SEVERITIES;
	
        subagent_lock(&hpi_lock_data);
	
	rv = saHpiAlarmGetNext( sessionid,
		                Severity,
                                SAHPI_FALSE, //Get ALL Alarms
				&Alarm);
	
	if ((rv != SA_OK) && (rv != SA_ERR_HPI_NOT_PRESENT)) {
		DEBUGMSGTL ((AGENT, 
		"populate_saHpiDomainAlarmTable: saHpiDomainAlarmGetNext Failed: rv = %d\n",
		rv));
		
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}
	else if (rv == SA_ERR_HPI_NOT_PRESENT) {
	        
		subagent_unlock(&hpi_lock_data);
		return SA_OK; //first call returned nothing
	}
	
	while ((rv != SA_ERR_HPI_NOT_PRESENT) && (rv == SA_OK)) {

                /* Store sa value                                       */
                /* this is the same as the AlarmId obtained from HPI    */
                /* BUT in the case of a user defined Alarm they  */
                /* will be different                                    */
                keys.key_array[0] = Alarm.AlarmCond.DomainId;
                keys.key_array[1] = Alarm.AlarmId;

                /* domain_alarm_entry_get() generates unique keys based on */
                /* domainid and alarmid values.                            */
                de_entry = domain_alarm_entry_get(&keys, &doma_de_table); 

                if (de_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: populate_saHpiDomainAlarmTable() "
                                     "domain_alarm_entry_get returned NULL\n"));
                        
			subagent_unlock(&hpi_lock_data);
			return AGENT_ERR_INTERNAL_ERROR;
                }
                /* here is where the mapped id is the same as the key value     */
                /* remember when the user creates an announcement the user      */
                /* specifies the entryId and will be different from the         */
                /* entryId returned from the HPI API that generates the         */
                /* new announcement                                             */    
                de_entry->hpi_alarm_id = Alarm.AlarmId;

	        
		/* assign the number of indices */
	        domain_alarm_idx.len = DOMAIN_ALARM_INDEX_NR;
		/** Index saHpiDomainId is external */
	        domain_alarm_oid[0] = Alarm.AlarmCond.DomainId;
	        /** Index saHpiDomainAlarmId is internal */
                domain_alarm_oid[1] = Alarm.AlarmId;
	        ///** Index saHpiDomainAlarmSeverity is internal */
	        //domain_alarm_oid[2] = Alarm.Severity + 1;
		
	        domain_alarm_idx.oids = (oid *) & domain_alarm_oid;
	   
	        /* See if Row exists. */
	        domain_alarm_ctx = NULL;
	        domain_alarm_ctx = CONTAINER_FIND(cb.container, &domain_alarm_idx);

	        if (!domain_alarm_ctx) { 
		        // New entry. Add it
		        domain_alarm_ctx = 
			        saHpiDomainAlarmTable_create_row(&domain_alarm_idx);
                        new_row = MIB_TRUE;
	        }
	        if (!domain_alarm_ctx) {
		        snmp_log (LOG_ERR, "Not enough memory for a Domain Alarm row!");
		        
			subagent_unlock(&hpi_lock_data);
			return AGENT_ERR_INTERNAL_ERROR;
	        }

                /** SaHpiEntryId = ASN_UNSIGNED */
                domain_alarm_ctx->saHpiDomainAlarmId = Alarm.AlarmId;
	
                /** SaHpiTime = ASN_OCTET_STR */
		hpitime_to_snmptime(Alarm.Timestamp, domain_alarm_ctx->saHpiDomainAlarmTimestamp);
	        domain_alarm_ctx->saHpiDomainAlarmTimestamp_len = sizeof(SaHpiTimeT);
		
                /** TruthValue = ASN_INTEGER */
	        domain_alarm_ctx->saHpiDomainAlarmAcknowledged = 
                        (Alarm.Acknowledged == SAHPI_TRUE)
                        ? MIB_TRUE : MIB_FALSE; 

                /** TruthValue = ASN_INTEGER */
	        domain_alarm_ctx->saHpiDomainAlarmAckBySeverity = 0;
	
	        /** INTEGER = ASN_INTEGER */
        	domain_alarm_ctx->saHpiDomainAlarmCondStatusCondType = Alarm.AlarmCond.Type + 1;
		
		/** SaHpiEntityPath = ASN_OCTET_STR */
                rc = oh_decode_entitypath(&Alarm.AlarmCond.Entity, &bigbuf);
		
		if (rc != SA_OK) {
		        DEBUGMSGTL ((AGENT, 
		          "populate_saHpiDomainAlarmTable: oh_decode_entitypath Failed: rc = %d\n",
		           rc));
		}
		else {
		       memcpy(domain_alarm_ctx->saHpiDomainAlarmCondEntityPath,
		              bigbuf.Data, bigbuf.DataLength);
	               domain_alarm_ctx->saHpiDomainAlarmCondEntityPath_len = bigbuf.DataLength;
		}
	
		
		if (Alarm.AlarmCond.Type == SAHPI_STATUS_COND_TYPE_SENSOR) {
		
		        /** UNSIGNED32 = ASN_UNSIGNED */
			domain_alarm_ctx->saHpiDomainAlarmCondSensorNum = Alarm.AlarmCond.SensorNum;
			
		        /** SaHpiEventState = ASN_OCTET_STR */
		        rc = oh_decode_eventstate(Alarm.AlarmCond.EventState, SAHPI_EC_SENSOR_SPECIFIC, &buf);
			
			if (rc != SA_OK) {
		                DEBUGMSGTL ((AGENT, 
		                "populate_saHpiDomainAlarmTable: oh_decode_eventstate Failed: rc = %d\n",
		                 rc));
			}	 
			else {	 		        
		                memcpy(domain_alarm_ctx->saHpiDomainAlarmCondEventState, 
				                                buf.Data, buf.DataLength); 
								
				domain_alarm_ctx->saHpiDomainAlarmCondEventState_len = buf.DataLength;
			}					
		}

	        /** OCTETSTR = ASN_OCTET_STR */
		memcpy(domain_alarm_ctx->saHpiDomainAlarmCondNameValue, 
		                Alarm.AlarmCond.Name.Value, Alarm.AlarmCond.Name.Length);
		
	        domain_alarm_ctx->saHpiDomainAlarmCondNameValue_len = Alarm.AlarmCond.Name.Length;
		
		/** SaHpiManufacturerId = ASN_UNSIGNED */
		domain_alarm_ctx->saHpiDomainAlarmCondMid = Alarm.AlarmCond.Mid;
		
	        /** SaHpiTextType = ASN_INTEGER */
                domain_alarm_ctx->saHpiDomainAlarmCondTextType = Alarm.AlarmCond.Data.DataType + 1;
	        
		/** SaHpiTextLanguage = ASN_INTEGER */
		domain_alarm_ctx->saHpiDomainAlarmCondTextLanguage = Alarm.AlarmCond.Data.Language + 1;
        
	        /** SaHpiText = ASN_OCTET_STR */
                domain_alarm_ctx->saHpiDomainAlarmCondText_len = Alarm.AlarmCond.Data.DataLength;
		
		memcpy(domain_alarm_ctx->saHpiDomainAlarmCondText, 
		                        Alarm.AlarmCond.Data.Data, Alarm.AlarmCond.Data.DataLength);
					
		domain_alarm_ctx->saHpiDomainAlarmRowStatus = SAHPIDOMAINALARMROWSTATUS_ACTIVE;

                if (new_row == MIB_TRUE) {
                        CONTAINER_INSERT (cb.container, domain_alarm_ctx);
			new_row = MIB_FALSE;
		}	
	
	        //Get the next Alarm
        	rv = saHpiAlarmGetNext( sessionid,
		                        Severity,
                                        SAHPI_FALSE, //Get Unacknowledged Alarms
				        &Alarm);
		
        }	
	
	domain_alarm_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);

        return SA_OK;   								    
}


/**
 * 
 * @param row_ctx
 * 
 * @return 
 */
int domain_alarm_delete (saHpiDomainAlarmTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "domain_alarm_delete, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiAlarmIdT       alarmId;

        DE_XREF *de_entry;
        doma_keys keys;

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX]);

        keys.key_array[0] = row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX];
        keys.key_array[1] = row_ctx->index.oids[saHpiDomainAlarmEntryId_INDEX];

        /* domain_alarm_entry_get() generates unique keys based on */
        /* domainid and entryid.                                   */
        de_entry = domain_alarm_entry_get(&keys, &doma_de_table); 

        if (de_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: domain_alarm_delete() "
                             "domain_alarm_entry_get returned NULL\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        /* see if this is an attempt to delete based on severtiy */
        if (row_ctx->index.oids[saHpiDomainAlarmEntryId_INDEX] == 0) {
                /* we are deleting based on Severity */
                alarmId = SAHPI_ENTRY_UNSPECIFIED;
        } else {
                /* we have retrieved the HPI Alarm ID  */
                /* now we can call the delete function */
                alarmId = de_entry->hpi_alarm_id;
        }

        rc = saHpiAlarmDelete(session_id, 
                              alarmId,
                              row_ctx->saHpiDomainAlarmSeverity - 1);
        if (rc != SA_OK) {
                snmp_log (LOG_ERR,
                "domain_alarm_delete: Call to "
                "saHpiAlarmDelete() failed rc: %s.\n",
                oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                "domain_alarm_delete: Call to "
                "saHpiAlarmDelete() failed rc: %s.\n",
                oh_lookup_error(rc)));
                return get_snmp_error(rc);
        } 

        DEBUGMSGTL ((AGENT, "domain_alarm_delete: Call to "
                   "saHpiAlarmDelete() succeeded rc: %s.\n",
                   oh_lookup_error(rc)));

        /* if we deleted based on severity then we need to remove      */
        /* all rows for the given SessionId <==> DomainId, ResourceId, */
        /* AnnunciatorNum and Severity                                 */
        if (alarmId == SAHPI_ENTRY_UNSPECIFIED) {

                netsnmp_index *row_idx;
                saHpiDomainAlarmTable_context *da_ctx;

                row_idx = CONTAINER_FIRST(cb.container);
                if (row_idx) //At least one entry was found.
                {
                        do {
                                da_ctx = CONTAINER_FIND(cb.container, row_idx);

                                row_idx = CONTAINER_NEXT(cb.container, row_idx);
                                
				if ((da_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX] ==
                                     row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX]) &&
                                    
				    ((da_ctx->saHpiDomainAlarmSeverity == 
                                     row_ctx->saHpiDomainAlarmSeverity) || 
				    (row_ctx->saHpiDomainAlarmSeverity == 
                                     SAHPI_ALL_SEVERITIES)) &&
                                    
				    (da_ctx->index.oids[saHpiDomainAlarmEntryId_INDEX] !=
                                     row_ctx->index.oids[saHpiDomainAlarmEntryId_INDEX])) {
                                        /* all conditions met remove row */
                                        CONTAINER_REMOVE (cb.container, da_ctx);
                                        saHpiDomainAlarmTable_delete_row (da_ctx);
                                        domain_alarm_entry_count = CONTAINER_SIZE (cb.container);
                                        DEBUGMSGTL ((AGENT, "domain_alarm_delete: found row for "
                                                     "deletion based on severity\n"));
                                }
                                
                        } while (row_idx);
                }
        } /* end check for all rows deleted based on Severity */

        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @param row_ctx
 * 
 * @return 
 */
int domain_alarm_add (saHpiDomainAlarmTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "domain_alarm_add, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiAlarmT         alarm;

        DE_XREF *de_entry;
        doma_keys keys;
        
        if ((row_ctx->sahpi_domain_alarm_severity_set         == MIB_TRUE) &&
            (row_ctx->sahpi_domain_alarm_status_cond_type_set == MIB_TRUE) &&
            (row_ctx->sahpi_domain_alarm_entitypath_set       == MIB_TRUE) &&                 
            (row_ctx->sahpi_domain_alarm_name_set             == MIB_TRUE) &&                           
            (row_ctx->sahpi_domain_alarm_text_type_set        == MIB_TRUE) &&   
            (row_ctx->sahpi_domain_alarm_text_language_set    == MIB_TRUE) &&
            (row_ctx->sahpi_domain_alarm_text                 == MIB_TRUE)) {

                /* EntryId */
                alarm.AlarmId = row_ctx->saHpiDomainAlarmId;

                
		/* Severity */
		alarm.Severity =
			row_ctx->saHpiDomainAlarmSeverity;
			
                /* Acknowledged */
                alarm.Acknowledged = 
                        (row_ctx->saHpiDomainAlarmAcknowledged == MIB_TRUE) ? 
                        SAHPI_TRUE : SAHPI_FALSE;			
                
		/*StatusCond */		
                alarm.AlarmCond.Type = 
                        row_ctx->saHpiDomainAlarmCondStatusCondType - 1;	
			
                rc = oh_encode_entitypath((gchar*)row_ctx->saHpiDomainAlarmCondEntityPath, 
                                          &alarm.AlarmCond.Entity);
		/* Domain */			  
                alarm.AlarmCond.DomainId =
                        row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX];					  

                /* Resource ID */
		//Must be SAHPI_UNSPECIFIED_RESOURCE_ID, since we can only add USER alarms.			  		
		alarm.AlarmCond.ResourceId = SAHPI_UNSPECIFIED_RESOURCE_ID;												

                /* SensorNum */
		//Undefined since user cannot add Sensor Alarms. 
                alarm.AlarmCond.SensorNum = 0;
		
		/* EventState */
		//Undefined since user cannot add Sensor Alarms.
		alarm.AlarmCond.EventState = 0;

                /* Name */
                alarm.AlarmCond.Name.Length = 
                        row_ctx->saHpiDomainAlarmCondNameValue_len;
                memcpy(alarm.AlarmCond.Name.Value,
                       row_ctx->saHpiDomainAlarmCondNameValue,
                       row_ctx->saHpiDomainAlarmCondNameValue_len);

                /* Mid */
		//Undefined since user cannot add Oem Alarms.
                alarm.AlarmCond.Mid = 0;
		       		       										  					  							
                /* SaHpiTextBufferT Stuff */
		alarm.AlarmCond.Data.DataType = 
                        row_ctx->saHpiDomainAlarmCondTextType;

                alarm.AlarmCond.Data.Language =
                        row_ctx->saHpiDomainAlarmCondTextLanguage;
		
		alarm.AlarmCond.Data.DataLength =
                        row_ctx->saHpiDomainAlarmCondText_len;
			               
		memcpy(alarm.AlarmCond.Data.Data, 
                       row_ctx->saHpiDomainAlarmCondText, 
                       row_ctx->saHpiDomainAlarmCondText_len);


                session_id = get_session_id(row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX]);

                rc = saHpiAlarmAdd(session_id, &alarm);

                if (rc != SA_OK) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: saHpiAlarmAdd() "
                                     "Failed to add an Alarm rc: [%s]\n", 
                                     oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }

                keys.key_array[0] = 
                        row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX];
                keys.key_array[1] = 
                        row_ctx->index.oids[saHpiDomainAlarmEntryId_INDEX];

                /* domain_alarm_entry_get() generates unique keys based on */
                /* domainid, and entryid info.                             */
                de_entry = domain_alarm_entry_get(&keys, &doma_de_table); 

                if (de_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: domain_alarm_add() "
                                     "domain_alarm_entry_get returned NULL\n"));
                        return AGENT_ERR_INTERNAL_ERROR;
                }

                /* we have retrieved the HPI Annunciator EntryId */
                de_entry->hpi_alarm_id = alarm.AlarmId;
		
		row_ctx->saHpiDomainAlarmId =  alarm.AlarmId;
		
		hpitime_to_snmptime(alarm.Timestamp, row_ctx->saHpiDomainAlarmTimestamp);
		row_ctx->saHpiDomainAlarmTimestamp_len = sizeof(SaHpiTimeT);
		
		row_ctx->saHpiDomainAlarmRowStatus = SAHPIDOMAINALARMROWSTATUS_ACTIVE;
		
		//RESET
		row_ctx->sahpi_domain_alarm_severity_set	 = MIB_FALSE;
                row_ctx->sahpi_domain_alarm_status_cond_type_set = MIB_FALSE;
                row_ctx->sahpi_domain_alarm_entitypath_set	 = MIB_FALSE;
                row_ctx->sahpi_domain_alarm_name_set		 = MIB_FALSE;
                row_ctx->sahpi_domain_alarm_text_type_set	 = MIB_FALSE;
                row_ctx->sahpi_domain_alarm_text_language_set	 = MIB_FALSE;
                row_ctx->sahpi_domain_alarm_text		 = MIB_FALSE;
		
        }
        
        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @param row_ctx
 * 
 * @return 
 */
int domain_alarm_ack (saHpiDomainAlarmTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "domain_alarm_ack, called\n"));

        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiAlarmIdT       alarmId;

        DE_XREF *de_entry;
        doma_keys keys;

        if (!row_ctx)
                return AGENT_ERR_NULL_DATA;

        session_id = get_session_id(row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX]);

       /* see if this is an attempt to ack based on severtiy */
        if (row_ctx->saHpiDomainAlarmRowStatus == SAHPI_ENTRY_UNSPECIFIED) {
                /* we are acking based on Severity */
                alarmId = SAHPI_ENTRY_UNSPECIFIED;
        } else {

       		keys.key_array[0] = row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX];
        	keys.key_array[1] = row_ctx->index.oids[saHpiDomainAlarmEntryId_INDEX];

                /* domain_alarm_entry_get() generates unique keys based on */
                /* domainid and alarmId tuples.                            */
                de_entry = domain_alarm_entry_get(&keys, &doma_de_table); 

                if (de_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: domain_alarm_ack() "
                                     "domain_alarm_entry_get returned NULL\n"));
                        return AGENT_ERR_INTERNAL_ERROR;
                }

                /* we have retrieved the HPI Annunciator EntryId */
                alarmId = de_entry->hpi_alarm_id;
        }

        rc = saHpiAlarmAcknowledge(session_id,                             
                                   alarmId, 
                                   row_ctx->saHpiDomainAlarmSeverity - 1);

        if (rc != SA_OK) {
                snmp_log (LOG_ERR,
                "domain_alarm_ack: Call to "
                "saHpiAlarmAcknowledge() failed rc: %s.\n",
                oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
                "domain_alarm_ack: Call to "
                "saHpiAlarmAcknowledge() failed rc: %s.\n",
                oh_lookup_error(rc)));
                return get_snmp_error(rc);
        } 

        DEBUGMSGTL ((AGENT, "domain_alarm_ack: Call to "
                   "saHpiAlarmAcknowledge() succeeded rc: %s.\n",
                   oh_lookup_error(rc)));

        /* if we ack'd based on severity then we need to set ack to true   */
        /* for all rows for the given DomainId and Severity               */
        if (alarmId == SAHPI_ENTRY_UNSPECIFIED) {

                netsnmp_index *row_idx;
                saHpiDomainAlarmTable_context *doma_ctx;

                row_idx = CONTAINER_FIRST(cb.container);
                if (row_idx) //At least one entry was found.
                {
                        do {
                                doma_ctx = CONTAINER_FIND(cb.container, row_idx);

                                if ((doma_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX] ==
                                     row_ctx->index.oids[saHpiDomainAlarmDomainId_INDEX]) &&

                                    ((doma_ctx->saHpiDomainAlarmSeverity == 
                                      row_ctx->saHpiDomainAlarmSeverity) ||
                                    (row_ctx->saHpiDomainAlarmSeverity == 
                                     SAHPI_ALL_SEVERITIES))) {
                                    
                                     doma_ctx->saHpiDomainAlarmAcknowledged = MIB_TRUE;
                                     DEBUGMSGTL ((AGENT, "domain_alarm_ack: found row for "
                                                         "setting ack based on severity\n"));
                                 }
                                row_idx = CONTAINER_NEXT(cb.container, row_idx);
                        } while (row_idx);
                }
        } /* end check for all rows deleted based on Severity */

        return SNMP_ERR_NOERROR; 
}


/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int 
handle_saHpiDomainAlarmActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiDomainAlarmActiveEntries, called\n"));

        domain_alarm_entry_count = CONTAINER_SIZE (cb.container);
	       
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			        (u_char *) &domain_alarm_entry_count,
        			        sizeof(domain_alarm_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

 
/**
 * 
 * @return: 
 */
int initialize_table_saHpiDomainAlarmActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiDomainAlarmActiveEntries, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiDomainAlarmActiveEntries", 
					handle_saHpiDomainAlarmActiveEntries,
                                        saHpiDomainAlarmActiveEntries_oid, 
					OID_LENGTH(saHpiDomainAlarmActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
} 


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiDomainAlarmTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiDomainAlarmTable_cmp( const void *lhs, const void *rhs )
{
        saHpiDomainAlarmTable_context *context_l =
        (saHpiDomainAlarmTable_context *)lhs;
        saHpiDomainAlarmTable_context *context_r =
        (saHpiDomainAlarmTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiDomainAlarmTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiDomainAlarmTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiDoaminAlaramId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) 
                        return 0;

        }

        return 0;
}

/************************************************************
 * Initializes the saHpiDomainAlarmTable module
 */
void
init_saHpiDomainAlarmTable(void)
{
        
        DEBUGMSGTL ((AGENT, "init_saHpiDomainAlarmTable, called\n"));
	
        initialize_table_saHpiDomainAlarmTable();

        initialize_table_saHpiDomainAlarmActiveEntries();
	
        domain_alarm_entry_initialize(&doma_initialized, &doma_de_table);
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiDomainAlarmTable_row_copy(saHpiDomainAlarmTable_context * dst,
                         saHpiDomainAlarmTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    /*
     * copy index, if provided
     */
    subagent_lock(&hpi_lock_data);
   
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiDomainAlarmId = src->saHpiDomainAlarmId;

    memcpy( dst->saHpiDomainAlarmTimestamp, 
            src->saHpiDomainAlarmTimestamp, 
	    src->saHpiDomainAlarmTimestamp_len );
	    
    dst->saHpiDomainAlarmTimestamp_len = src->saHpiDomainAlarmTimestamp_len;

    dst->saHpiDomainAlarmSeverity = src->saHpiDomainAlarmSeverity;

    dst->saHpiDomainAlarmAcknowledged = src->saHpiDomainAlarmAcknowledged;

    dst->saHpiDomainAlarmAckBySeverity = src->saHpiDomainAlarmAckBySeverity;

    dst->saHpiDomainAlarmCondStatusCondType = src->saHpiDomainAlarmCondStatusCondType;

    memcpy( dst->saHpiDomainAlarmCondEntityPath, src->saHpiDomainAlarmCondEntityPath, src->saHpiDomainAlarmCondEntityPath_len );
    dst->saHpiDomainAlarmCondEntityPath_len = src->saHpiDomainAlarmCondEntityPath_len;

    dst->saHpiDomainAlarmCondSensorNum = src->saHpiDomainAlarmCondSensorNum;

    memcpy( dst->saHpiDomainAlarmCondEventState, src->saHpiDomainAlarmCondEventState, src->saHpiDomainAlarmCondEventState_len );
    dst->saHpiDomainAlarmCondEventState_len = src->saHpiDomainAlarmCondEventState_len;

    memcpy( dst->saHpiDomainAlarmCondNameValue, src->saHpiDomainAlarmCondNameValue, src->saHpiDomainAlarmCondNameValue_len );
    dst->saHpiDomainAlarmCondNameValue_len = src->saHpiDomainAlarmCondNameValue_len;

    dst->saHpiDomainAlarmCondMid = src->saHpiDomainAlarmCondMid;

    dst->saHpiDomainAlarmCondTextType = src->saHpiDomainAlarmCondTextType;

    dst->saHpiDomainAlarmCondTextLanguage = src->saHpiDomainAlarmCondTextLanguage;

    memcpy( dst->saHpiDomainAlarmCondText, src->saHpiDomainAlarmCondText, src->saHpiDomainAlarmCondText_len );
    dst->saHpiDomainAlarmCondText_len = src->saHpiDomainAlarmCondText_len;

    dst->saHpiDomainAlarmRowStatus = src->saHpiDomainAlarmRowStatus;

    subagent_unlock(&hpi_lock_data);
    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiDomainAlarmTable_extract_index( saHpiDomainAlarmTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiDomainAlarmId;

        int err;

        DEBUGMSGTL ((AGENT, "saHpiDomainAlarmTable_extract_index, called\n"));

        
	subagent_lock(&hpi_lock_data);
	/*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
			return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiDomainAlarmId;

        memset( &var_saHpiDomainAlarmId, 0x00, sizeof(var_saHpiDomainAlarmId) );
        var_saHpiDomainAlarmId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainAlarmId.next_variable = NULL;

        //memset( &var_saHpiDomainAlarmSeverity, 0x00, sizeof(var_saHpiDomainAlarmSeverity) );
        //var_saHpiDomainAlarmSeverity.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        //var_saHpiDomainAlarmSeverity.next_variable = NULL;



        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                ctx->saHpiDomainAlarmId = *var_saHpiDomainAlarmId.val.integer;
   
                //ctx->saHpiDomainAlarmSeverity = *var_saHpiDomainAlarmSeverity.val.integer;
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer); 
		err = saHpiDomainAlarmId_check_index(*var_saHpiDomainAlarmId.val.integer);
		//err = saHpiDomainAlarmSeverity_check_index(*var_saHpiDomainAlarmSeverity.val.integer);
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
	
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiDomainAlarmTable_can_activate(saHpiDomainAlarmTable_context *undo_ctx,
                      saHpiDomainAlarmTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiDomainAlarmTable_can_deactivate(saHpiDomainAlarmTable_context *undo_ctx,
                        saHpiDomainAlarmTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiDomainAlarmTable_can_delete(saHpiDomainAlarmTable_context *undo_ctx,
                    saHpiDomainAlarmTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiDomainAlarmTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

#ifdef saHpiDomainAlarmTable_ROW_CREATION
/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiDomainAlarmTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiDomainAlarmTable_context *
saHpiDomainAlarmTable_create_row( netsnmp_index* hdr)
{
    
    subagent_lock(&hpi_lock_data);
    
    saHpiDomainAlarmTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiDomainAlarmTable_context);
    if(!ctx) {
    	subagent_unlock(&hpi_lock_data);
        return NULL;
    }	
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiDomainAlarmTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiDomainAlarmTimestamp = 0;
     ctx->saHpiDomainAlarmSeverity = 0;
     ctx->saHpiDomainAlarmAcknowledged = 0;
     ctx->saHpiDomainAlarmAckBySeverity = 0;
     ctx->saHpiDomainAlarmCondStatusCondType = 0;
     ctx->saHpiDomainAlarmCondEntityPath = 0;
     ctx->saHpiDomainAlarmCondSensorNum = 0;
     ctx->saHpiDomainAlarmCondEventState = 0;
     ctx->saHpiDomainAlarmCondNameValue = 0;
     ctx->saHpiDomainAlarmCondMid = 0;
     ctx->saHpiDomainAlarmCondTextType = 0;
     ctx->saHpiDomainAlarmCondTextLanguage = 0;
     ctx->saHpiDomainAlarmCondText = 0;
     ctx->saHpiDomainAlarmRowStatus = 0;
    */

    ctx->sahpi_domain_alarm_severity_set         = MIB_FALSE;
    ctx->sahpi_domain_alarm_acknowledged_set     = MIB_FALSE;
    ctx->sahpi_domain_alarm_status_cond_type_set = MIB_FALSE;
    ctx->sahpi_domain_alarm_entitypath_set       = MIB_FALSE;
    ctx->sahpi_domain_alarm_name_set             = MIB_FALSE;
    ctx->sahpi_domain_alarm_text_type_set        = MIB_FALSE;
    ctx->sahpi_domain_alarm_text_language_set    = MIB_FALSE;
    ctx->sahpi_domain_alarm_text                 = MIB_FALSE;

    subagent_unlock(&hpi_lock_data);
    return ctx;
}
#endif

/************************************************************
 * the *_duplicate row routine
 */
saHpiDomainAlarmTable_context *
saHpiDomainAlarmTable_duplicate_row( saHpiDomainAlarmTable_context * row_ctx)
{
    saHpiDomainAlarmTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
    
    dup = SNMP_MALLOC_TYPEDEF(saHpiDomainAlarmTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL;
     }	
        
    if(saHpiDomainAlarmTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiDomainAlarmTable_delete_row( saHpiDomainAlarmTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

    subagent_lock(&hpi_lock_data);
    
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiDomainAlarmTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiDomainAlarmTable_context *row_ctx =
            (saHpiDomainAlarmTable_context *)rg->existing_row;
//    saHpiDomainAlarmTable_context *undo_ctx =
//            (saHpiDomainAlarmTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmSeverity));
        break;

        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmAcknowledged));
        break;

        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmAckBySeverity));
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmCondStatusCondType));
        break;

        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiDomainAlarmCondEntityPath)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	
        break;

        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiDomainAlarmCondNameValue)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }	
        break;


        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmCondTextType));
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmCondTextLanguage));
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > 
                        sizeof(row_ctx->saHpiDomainAlarmCondText)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
            	    }
            }
        break;

        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiDomainAlarmRowStatus));
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiDomainAlarmTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiDomainAlarmTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiDomainAlarmTable_context *row_ctx = (saHpiDomainAlarmTable_context *)rg->existing_row;
    saHpiDomainAlarmTable_context *undo_ctx = (saHpiDomainAlarmTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */	
                if (oh_lookup_severity(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
                if (oh_lookup_severity(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
                if (oh_lookup_statuscondtype(*var->val.integer - 1) == NULL) {
                        rc = SNMP_ERR_BADVALUE;
		}	
		if ((*var->val.integer - 1) != SAHPI_STATUS_COND_TYPE_USER) {
                        DEBUGMSGTL ((AGENT, "COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE"
                                             " saHpiDomainAlarmTable_set_reserve2:"
                                             " Only type SAHPI_STATUS_COND_TYPE_USER"
                                             " is allowed\n"));
                        rc = SNMP_ERR_BADVALUE;
		}		
					     		        	
        break;

        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;


        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
                if (oh_lookup_texttype(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
                if (oh_lookup_language(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
                rc = netsnmp_check_vb_rowstatus(current->ri->requestvb,
                                                undo_ctx ?
                                                undo_ctx->saHpiDomainAlarmRowStatus:0);
                rg->rg_void = current->ri;
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiDomainAlarmTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiDomainAlarmTable_context *row_ctx = (saHpiDomainAlarmTable_context *)rg->existing_row;
//    saHpiDomainAlarmTable_context *undo_ctx = (saHpiDomainAlarmTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    
    subagent_lock(&hpi_lock_data);

    DEBUGMSGTL ((AGENT, "saHpiDomainAlarmTable_set_action, called\n"));
    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            row_ctx->saHpiDomainAlarmSeverity = *var->val.integer;
            if (rg->row_created == SAHPIDOMAINALARMROWSTATUS_ACTIVE) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
            } 
            row_ctx->sahpi_domain_alarm_severity_set = MIB_TRUE;
            row_err = domain_alarm_add(row_ctx);
        break;

        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiDomainAlarmAcknowledged = *var->val.integer;
            if (rg->row_created == 1) {                                 /* createAndWait */
                    row_ctx->saHpiDomainAlarmRowStatus = SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; 
                    row_ctx->sahpi_domain_alarm_acknowledged_set = MIB_TRUE;
		    row_err = domain_alarm_add(row_ctx);
            } else if ((rg->row_created != 1) &&                       /* active row, this is an ACK */
                       (row_ctx->saHpiDomainAlarmRowStatus == SAHPIDOMAINALARMROWSTATUS_ACTIVE)) {  
                    if (*var->val.integer == MIB_TRUE) {
                            row_err = domain_alarm_ack(row_ctx);
                    }
            } else if ((rg->row_created != 1) &&                
                       (row_ctx->saHpiDomainAlarmRowStatus == SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT) &&
		       (row_ctx->sahpi_domain_alarm_severity_set == MIB_TRUE)) {
		       
		       if (*var->val.integer == MIB_TRUE) {
                            row_err = domain_alarm_ack(row_ctx);
                       }
            }		       
        break;

        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            row_ctx->saHpiDomainAlarmAckBySeverity = *var->val.integer;

            if (rg->row_created == 1) {                                 /* createAndWait */
                    row_ctx->saHpiDomainAlarmRowStatus = SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; 

            } else if ((rg->row_created != 1) &&                       /* severity set, time to ACK */
                       (row_ctx->sahpi_domain_alarm_severity_set == MIB_TRUE) &&
                       (row_ctx->saHpiDomainAlarmRowStatus == SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT)) { 
                    if (*var->val.integer == MIB_TRUE) {
                            row_err = domain_alarm_ack(row_ctx);
                    }
            }	    
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiDomainAlarmCondStatusCondType = *var->val.integer;
	    
            if (rg->row_created == 1) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
                     
            }
	    
	    row_ctx->sahpi_domain_alarm_status_cond_type_set = MIB_TRUE;
            row_err = domain_alarm_add(row_ctx);	    
        break;

        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiDomainAlarmCondEntityPath,var->val.string,var->val_len);
            row_ctx->saHpiDomainAlarmCondEntityPath_len = var->val_len;
	    
            if (rg->row_created == 1) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
                    
            }
	    
	    row_ctx->sahpi_domain_alarm_entitypath_set = MIB_TRUE;
            row_err = domain_alarm_add(row_ctx);	    
        break;

        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiDomainAlarmCondNameValue,var->val.string,var->val_len);
            row_ctx->saHpiDomainAlarmCondNameValue_len = var->val_len;
            if (rg->row_created == 1) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
                    
            }
	    row_ctx->sahpi_domain_alarm_name_set = MIB_TRUE;
            row_err = domain_alarm_add(row_ctx);	    
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            row_ctx->saHpiDomainAlarmCondTextType = *var->val.integer;
            if (rg->row_created == 1) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
                    
            }
	    row_ctx->sahpi_domain_alarm_text_type_set = MIB_TRUE; 
            row_err = domain_alarm_add(row_ctx);	    
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            row_ctx->saHpiDomainAlarmCondTextLanguage = *var->val.integer;
            if (rg->row_created == 1) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
                    
            }
	    row_ctx->sahpi_domain_alarm_text_language_set = MIB_TRUE;
            row_err = domain_alarm_add(row_ctx);	    
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiDomainAlarmCondText,var->val.string,var->val_len);
            row_ctx->saHpiDomainAlarmCondText_len = var->val_len;
            if (rg->row_created == 1) {
                    row_ctx->saHpiDomainAlarmRowStatus = 
                            SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT; /* createAndWait */
                    
            }
	    row_ctx->sahpi_domain_alarm_text = MIB_TRUE;
            row_err = domain_alarm_add(row_ctx);
	    	    
        break;

        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
                if (((row_ctx->sahpi_domain_alarm_severity_set == MIB_TRUE) ||
                    (row_ctx->saHpiDomainAlarmRowStatus == SAHPIDOMAINALARMROWSTATUS_ACTIVE)) &&
		     (*var->val.integer == SAHPIDOMAINALARMROWSTATUS_DESTROY)){

                        row_ctx->saHpiDomainAlarmRowStatus = *var->val.integer;
                        row_err = domain_alarm_delete(row_ctx);
                        /* delete succeeded remove row */
                        if (row_err == SNMP_ERR_NOERROR) {
                                rg->row_deleted = 1;
                        }

                } else if ((rg->row_created == 1) ){
			   
			row_ctx->saHpiDomainAlarmRowStatus = SAHPIDOMAINALARMROWSTATUS_CREATEANDWAIT;	
                        
                } else {
		    	row_err = SNMP_ERR_COMMITFAILED;
		}    
		
		 	    
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }
    subagent_unlock(&hpi_lock_data);
    return;

    /*
     * TODO: if you have dependencies on other tables, this would be
     * a good place to check those, too.
     */
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiDomainAlarmTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiDomainAlarmTable_context *row_ctx = (saHpiDomainAlarmTable_context *)rg->existing_row;
//    saHpiDomainAlarmTable_context *undo_ctx = (saHpiDomainAlarmTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINALARMTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiDomainAlarmTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiDomainAlarmTable_context *row_ctx = (saHpiDomainAlarmTable_context *)rg->existing_row;
//    saHpiDomainAlarmTable_context *undo_ctx = (saHpiDomainAlarmTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINALARMTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
	        break;
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiDomainAlarmTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiDomainAlarmTable_context *row_ctx = (saHpiDomainAlarmTable_context *)rg->existing_row;
//    saHpiDomainAlarmTable_context *undo_ctx = (saHpiDomainAlarmTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINALARMTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}


/************************************************************
 *
 * Initialize the saHpiDomainAlarmTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiDomainAlarmTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiDomainAlarmTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiDomainAlarmTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiDomainAlarmTable_oid,
                                             saHpiDomainAlarmTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiDomainAlarmTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiDomainAlarmId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiDomainAlarmSeverity */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);

    table_info->min_column = saHpiDomainAlarmTable_COL_MIN;
    table_info->max_column = saHpiDomainAlarmTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiDomainAlarmTable_get_value;
    cb.container = netsnmp_container_find("saHpiDomainAlarmTable_primary:"
                                          "saHpiDomainAlarmTable:"
                                          "table_container");
#ifdef saHpiDomainAlarmTable_IDX2
    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiDomainAlarmTable_secondary:"
                                                       "saHpiDomainAlarmTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiDomainAlarmTable_cmp;
#endif
#ifdef saHpiDomainAlarmTable_SET_HANDLING
    cb.can_set = 1;
#ifdef saHpiDomainAlarmTable_ROW_CREATION
    cb.create_row = (UserRowMethod*)saHpiDomainAlarmTable_create_row;
#endif
    cb.duplicate_row = (UserRowMethod*)saHpiDomainAlarmTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiDomainAlarmTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiDomainAlarmTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiDomainAlarmTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiDomainAlarmTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiDomainAlarmTable_can_delete;

    cb.set_reserve1 = saHpiDomainAlarmTable_set_reserve1;
    cb.set_reserve2 = saHpiDomainAlarmTable_set_reserve2;
    cb.set_action = saHpiDomainAlarmTable_set_action;
    cb.set_commit = saHpiDomainAlarmTable_set_commit;
    cb.set_free = saHpiDomainAlarmTable_set_free;
    cb.set_undo = saHpiDomainAlarmTable_set_undo;
#endif
    DEBUGMSGTL(("initialize_table_saHpiDomainAlarmTable",
                "Registering table saHpiDomainAlarmTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiDomainAlarmTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiDomainAlarmTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiDomainAlarmTable_context *context = (saHpiDomainAlarmTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIDOMAINALARMID:
            /** SaHpiEntryId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiDomainAlarmId,
                         sizeof(context->saHpiDomainAlarmId) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiDomainAlarmTimestamp,
                         sizeof(context->saHpiDomainAlarmTimestamp) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmSeverity,
                         sizeof(context->saHpiDomainAlarmSeverity) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMACKNOWLEDGED:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmAcknowledged,
                         sizeof(context->saHpiDomainAlarmAcknowledged) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMACKBYSEVERITY:
            /** SaHpiSeverity = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmAckBySeverity,
                         sizeof(context->saHpiDomainAlarmAckBySeverity) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDSTATUSCONDTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmCondStatusCondType,
                         sizeof(context->saHpiDomainAlarmCondStatusCondType) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDENTITYPATH:
            /** SaHpiEntityPath = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiDomainAlarmCondEntityPath,
                         context->saHpiDomainAlarmCondEntityPath_len );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDSENSORNUM:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiDomainAlarmCondSensorNum,
                         sizeof(context->saHpiDomainAlarmCondSensorNum) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDEVENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiDomainAlarmCondEventState,
                         context->saHpiDomainAlarmCondEventState_len );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDNAMEVALUE:
            /** OCTETSTR = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiDomainAlarmCondNameValue,
                         context->saHpiDomainAlarmCondNameValue_len );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDMID:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiDomainAlarmCondMid,
                         sizeof(context->saHpiDomainAlarmCondMid) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmCondTextType,
                         sizeof(context->saHpiDomainAlarmCondTextType) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmCondTextLanguage,
                         sizeof(context->saHpiDomainAlarmCondTextLanguage) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMCONDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiDomainAlarmCondText,
                         context->saHpiDomainAlarmCondText_len );
        break;
    
        case COLUMN_SAHPIDOMAINALARMROWSTATUS:
            /** RowStatus = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiDomainAlarmRowStatus,
                         sizeof(context->saHpiDomainAlarmRowStatus) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiDomainAlarmTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiDomainAlarmTable_get_by_idx
 */
const saHpiDomainAlarmTable_context *
saHpiDomainAlarmTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiDomainAlarmTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


