/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiDomainInfoTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiDomainInfoTable.h"
#include <saHpiDomainReferenceTable.h>
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>

#include <SaHpi.h>
#include <oh_utils.h>
#include <session_info.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiDomainInfoTable_oid[] = { saHpiDomainInfoTable_TABLE_OID };
size_t saHpiDomainInfoTable_oid_len = OID_LENGTH(saHpiDomainInfoTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * oid declarations scalars
 */
static u_long domain_info_entry_count = 0;
static oid saHpiDomainInfoActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,2,1 };
int handle_saHpiDomainInfoActiveEntries(netsnmp_mib_handler  *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests);
int initialize_table_saHpiDomainInfoActiveEntries(void);


/*
 *  int populate_saHpiDomainInfoTable(SaHpiSessionIdT sessionid)
 */
int populate_saHpiDomainInfoTable(SaHpiSessionIdT sessionid) 
{
	SaErrorT rv;
        int new_row = MIB_FALSE;
	SaHpiDomainInfoT domain_info;
	
	oid domain_info_oid[DOMAIN_INFO_INDEX_NR];
	netsnmp_index domain_info_index;
	saHpiDomainInfoTable_context *domain_info_context;

	DEBUGMSGTL ((AGENT, "populate_saHpiDomainInfoTable\n"));

        subagent_lock(&hpi_lock_data);

	rv = saHpiDomainInfoGet(sessionid, &domain_info);
	if (rv != SA_OK) {
		DEBUGMSGTL ((AGENT, "saHpiDomainInfoGet Failed: rv = %d\n",rv));
		
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}
	
	domain_info_index.len = DOMAIN_INFO_INDEX_NR;
	domain_info_oid[0] = domain_info.DomainId;
	domain_info_index.oids = (oid *) & domain_info_oid;
	
	/* See if it exists. */
	domain_info_context = NULL;
	domain_info_context = CONTAINER_FIND (cb.container, &domain_info_index);
		
	if (!domain_info_context) { 
		// New entry. Add it
		domain_info_context = 
			saHpiDomainInfoTable_create_row ( &domain_info_index);
                new_row = MIB_TRUE;
	}
	if (!domain_info_context) {
		snmp_log (LOG_ERR, "Not enough memory for a DomainInfo row!");
		
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}	

	/** SaHpiDomainId = ASN_UNSIGNED */
        domain_info_context->saHpiDomainId = domain_info.DomainId;

	/** BITS = ASN_OCTET_STR */
	memset(	domain_info_context->saHpiDomainCapabilities, 
    		0, 
    		sizeof(domain_info_context->saHpiDomainCapabilities));
	if(domain_info.DomainCapabilities == (SaHpiDomainCapabilitiesT)0X00000001) {
    	memcpy(	domain_info_context->saHpiDomainCapabilities,
    			"SAHPI_DOMAIN_CAP_AUTOINSERT_READ_ONLY",
    			sizeof("SAHPI_DOMAIN_CAP_AUTOINSERT_READ_ONLY") );
     	domain_info_context->saHpiDomainCapabilities_len = 
    		sizeof("SAHPI_DOMAIN_CAP_AUTOINSERT_READ_ONLY");   		
	} else {
		domain_info_context->saHpiDomainCapabilities_len = 0;
	}

	/** TruthValue = ASN_INTEGER */
	domain_info_context->saHpiDomainIsPeer =
		(domain_info.IsPeer == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

	/** SaHpiTextType = ASN_INTEGER */
	domain_info_context->saHpiDomainTagTextType =
		domain_info.DomainTag.DataType;

   	/** SaHpiTextLanguage = ASN_INTEGER */
	domain_info_context->saHpiDomainTagTextLanguage =
		domain_info.DomainTag.Language;

	/** SaHpiText = ASN_OCTET_STR */
	memcpy(domain_info_context->saHpiDomainTag, 
		domain_info.DomainTag.Data,
		domain_info.DomainTag.DataLength);
	domain_info_context->saHpiDomainTag_len =
		domain_info.DomainTag.DataLength;

	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainReferenceUpdateCount =
		domain_info.DrtUpdateCount;

	/** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(domain_info.DrtUpdateTimestamp, 
	                 domain_info_context->saHpiDomainReferenceUpdateTimestamp);
        domain_info_context->saHpiDomainReferenceUpdateTimestamp_len = sizeof(SaHpiTimeT);			 
    	
	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainResourcePresenceUpdateCount =
		domain_info.RptUpdateCount;
    	
	/** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(domain_info.RptUpdateTimestamp, 
	                 domain_info_context->saHpiDomainResourcePresenceUpdateTimestamp);
			 
        domain_info_context->saHpiDomainResourcePresenceUpdateTimestamp_len = sizeof(SaHpiTimeT);
	
	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainAlarmUpdateCount =
		domain_info.DatUpdateCount;

	/** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(domain_info.DatUpdateTimestamp, 
	                 domain_info_context->saHpiDomainAlarmUpdateTimestamp);
			 
	domain_info_context->saHpiDomainAlarmUpdateTimestamp_len = sizeof(SaHpiTimeT);		 

	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainActiveAlarms =
		domain_info.ActiveAlarms;

	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainCriticalAlarms =
		domain_info.CriticalAlarms;

	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainMajorAlarms = 
		domain_info.MajorAlarms;

	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainMinorAlarms =
		domain_info.MinorAlarms;

	/** UNSIGNED32 = ASN_UNSIGNED */
	domain_info_context->saHpiDomainAlarmUserLimit =
		domain_info.DatUserAlarmLimit;

	/** TruthValue = ASN_INTEGER */
	domain_info_context->saHpiDomainAlarmOverflow =
		(domain_info.DatOverflow == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

	/** SaHpiGuid = ASN_OCTET_STR */
	/* typedef SaHpiUint8T    SaHpiGuidT[16]; */
	memcpy(domain_info_context->saHpiDomainGuid, 
		domain_info.Guid,
		16);
	domain_info_context->saHpiDomainGuid_len = 16;

        if (new_row == MIB_TRUE) 
                CONTAINER_INSERT (cb.container, domain_info_context);
	
	domain_info_entry_count = CONTAINER_SIZE (cb.container);
	
	subagent_unlock(&hpi_lock_data);
		
	return rv;
}

SaErrorT async_domain_add(SaHpiSessionIdT sessionid, SaHpiEventT *event, 
                          SaHpiRdrT *rdr, SaHpiRptEntryT *rpt_entry)
{        
        SaErrorT rv = SA_OK;

        if (event->EventDataUnion.DomainEvent.Type == 
            SAHPI_DOMAIN_REF_REMOVED) {
                rv = clear_domain_info_entry(
                        event->EventDataUnion.DomainEvent.DomainId);
                if (rv != SA_OK) return rv;
                rv = clear_domain_reference_entry(
                        event->EventDataUnion.DomainEvent.DomainId);
        } else if (event->EventDataUnion.DomainEvent.Type == 
                   SAHPI_DOMAIN_REF_ADDED) {
                rv = populate_saHpiDomainInfoTable(
                        get_session_id(event->EventDataUnion.DomainEvent.DomainId));
                if (rv != SA_OK) return rv;
                rv = poplulate_saHpiDomainReferenceTable(
                        get_session_id(event->EventDataUnion.DomainEvent.DomainId)); 
        } else {
                rv =  SA_ERR_HPI_INVALID_PARAMS;
        }

        return rv;
}

/**
 * 
 * @sessionid
 * 
 * @return 
 */
SaErrorT clear_domain_info_entry(SaHpiDomainIdT domain_id)
{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiDomainInfoTable_context *ctx;

	DEBUGMSGTL ((AGENT, "clear_domain_info_entry, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domain_id));	

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        ctx = CONTAINER_FIND(cb.container, row_idx);
                        
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if (ctx->index.oids[saHpiDomainId_INDEX] == domain_id) {

                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctx);
                                saHpiDomainInfoTable_delete_row (ctx);
                                domain_info_entry_count = 
                                        CONTAINER_SIZE (cb.container);
                                DEBUGMSGTL ((AGENT, "clear_domain_info_entry:"
                                             " found row: removing\n"));
                        }

                } while (row_idx);
        } 

        return rv;
}

/*
 * int set_table_domain_tag (saHpiDomainInfoTable_context *row_ctx)
 */
int set_table_domain_tag (saHpiDomainInfoTable_context *row_ctx)
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiTextBufferT    buffer;

        DEBUGMSGTL ((AGENT, "saHpiDomainInfoTable.c -> set_table_domain_tag called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	    
	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	
	// copy the data into a TextBufferT structure
	buffer.DataType   = row_ctx->saHpiDomainTagTextType - 1;
	buffer.Language   = row_ctx->saHpiDomainTagTextLanguage - 1;
	buffer.DataLength = row_ctx->saHpiDomainTag_len;
	memcpy(buffer.Data, row_ctx->saHpiDomainTag, row_ctx->saHpiDomainTag_len);

	rc = saHpiDomainTagSet (session_id, &buffer); 

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			    "Call to saHpiDomainTagSet failed with return code: %s.\n",
			    oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "Call to saHpiDomainTagSet failed with rc: %s.\n",
			     oh_lookup_error(rc)));
		return AGENT_ERR_OPERATION;

        }
		
	return AGENT_ERR_NOERROR; 
}	



/*
 * int handle_saHpiDomainInfoActiveEntries()
 */
int handle_saHpiDomainInfoActiveEntries(netsnmp_mib_handler  *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests)
{
	DEBUGMSGTL ((AGENT, "handle_saHpiDomainInfoActiveEntries: Entry Count is %ld\n", 
					domain_info_entry_count));  	
	
    /* We are never called for a GETNEXT if it's registered as a
       "instance", as it's "magically" handled for us.  */

    /* a instance handler also only hands us one request at a time, so
       we don't need to loop over a list of requests; we'll only get one. */

	domain_info_entry_count = CONTAINER_SIZE (cb.container);
  
    switch(reqinfo->mode) {
        case MODE_GET:        
            snmp_set_var_typed_value(
            		requests->requestvb, 
            		ASN_GAUGE,
                    (u_char *) &domain_info_entry_count,
                    sizeof(domain_info_entry_count) );
            break;
        default:
            /* we should never get here, so this is a really bad error */
            return SNMP_ERR_GENERR;
    }   

    return SNMP_ERR_NOERROR;
}

/*
 * int initialize_table_saHpiDomainInfoActiveEntries(void)
 */
int initialize_table_saHpiDomainInfoActiveEntries(void)
{
	DEBUGMSGTL ((AGENT, "initialize_table_saHpiDomainInfoActiveEntries, called\n"));	
	netsnmp_register_scalar( 
        netsnmp_create_handler_registration(
        	"saHpiDomainInfoActiveEntries", 
        	handle_saHpiDomainInfoActiveEntries,
            saHpiDomainInfoActiveEntries_oid, 
            OID_LENGTH(saHpiDomainInfoActiveEntries_oid),
            HANDLER_CAN_RONLY) ); 
	return 0;
}
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiDomainInfoTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiDomainInfoTable_cmp( const void *lhs, const void *rhs )
{
    saHpiDomainInfoTable_context *context_l =
        (saHpiDomainInfoTable_context *)lhs;
    saHpiDomainInfoTable_context *context_r =
        (saHpiDomainInfoTable_context *)rhs;

	/* check for NULL pointers */
    if(lhs == NULL || rhs == NULL ) {
        DEBUGMSGTL((AGENT,"saHpiDomainInfoTable_compare NULL pointer ERROR\n" ));
    	return 0;
    }	

	if ( context_l->index.oids[0] < context_r->index.oids[0])
		return -1;
		
	if ( context_l->index.oids[0] == context_r->index.oids[0])
		return 0;
		
	if ( context_l->index.oids[0] > context_r->index.oids[0])
		return 1;			     

	return 0;
		
}


/************************************************************
 * Initializes the saHpiDomainInfoTable module
 */
void
init_saHpiDomainInfoTable(void)
{	
	DEBUGMSGTL ((AGENT, "init_saHpiDomainInfoTable\n"));

    initialize_table_saHpiDomainInfoTable();
    
    initialize_table_saHpiDomainInfoActiveEntries();
    
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiDomainInfoTable_row_copy(saHpiDomainInfoTable_context * dst,
                         saHpiDomainInfoTable_context * src)
{
    if(!dst||!src)
        return 1;
        
   
   subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiDomainId = src->saHpiDomainId;

    memcpy( dst->saHpiDomainCapabilities, src->saHpiDomainCapabilities, src->saHpiDomainCapabilities_len );
    dst->saHpiDomainCapabilities_len = src->saHpiDomainCapabilities_len;

    dst->saHpiDomainIsPeer = src->saHpiDomainIsPeer;

    dst->saHpiDomainTagTextType = src->saHpiDomainTagTextType;

    dst->saHpiDomainTagTextLanguage = src->saHpiDomainTagTextLanguage;

    memcpy( dst->saHpiDomainTag, src->saHpiDomainTag, src->saHpiDomainTag_len );
    dst->saHpiDomainTag_len = src->saHpiDomainTag_len;

    dst->saHpiDomainReferenceUpdateCount = src->saHpiDomainReferenceUpdateCount;

    memcpy( dst->saHpiDomainReferenceUpdateTimestamp, 
            src->saHpiDomainReferenceUpdateTimestamp, 
	    src->saHpiDomainReferenceUpdateTimestamp_len );
    dst->saHpiDomainReferenceUpdateTimestamp_len = src->saHpiDomainReferenceUpdateTimestamp_len;

    dst->saHpiDomainResourcePresenceUpdateCount = src->saHpiDomainResourcePresenceUpdateCount;

    memcpy( dst->saHpiDomainResourcePresenceUpdateTimestamp, src->saHpiDomainResourcePresenceUpdateTimestamp, src->saHpiDomainResourcePresenceUpdateTimestamp_len );
    dst->saHpiDomainResourcePresenceUpdateTimestamp_len = src->saHpiDomainResourcePresenceUpdateTimestamp_len;

    dst->saHpiDomainAlarmUpdateCount = src->saHpiDomainAlarmUpdateCount;

    memcpy( dst->saHpiDomainAlarmUpdateTimestamp, src->saHpiDomainAlarmUpdateTimestamp, src->saHpiDomainAlarmUpdateTimestamp_len );
    dst->saHpiDomainAlarmUpdateTimestamp_len = src->saHpiDomainAlarmUpdateTimestamp_len;

    dst->saHpiDomainActiveAlarms = src->saHpiDomainActiveAlarms;

    dst->saHpiDomainCriticalAlarms = src->saHpiDomainCriticalAlarms;

    dst->saHpiDomainMajorAlarms = src->saHpiDomainMajorAlarms;

    dst->saHpiDomainMinorAlarms = src->saHpiDomainMinorAlarms;

    dst->saHpiDomainAlarmUserLimit = src->saHpiDomainAlarmUserLimit;

    dst->saHpiDomainAlarmOverflow = src->saHpiDomainAlarmOverflow;

    memcpy( dst->saHpiDomainGuid, src->saHpiDomainGuid, src->saHpiDomainGuid_len );
    dst->saHpiDomainGuid_len = src->saHpiDomainGuid_len;

    subagent_unlock(&hpi_lock_data);
    return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiDomainInfoTable_extract_index( saHpiDomainInfoTable_context * ctx, netsnmp_index * hdr )
{
    /*
     * temporary local storage for extracting oid index
     *
     * extract index uses varbinds (netsnmp_variable_list) to parse
     * the index OID into the individual components for each index part.
     */   
	netsnmp_variable_list var_saHpiDomainId;
    int err;

    
    subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */    
 	if(hdr) {
        netsnmp_assert(ctx->index.oids == NULL);
        if(snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                           hdr->len * sizeof(oid) )) {
            subagent_unlock(&hpi_lock_data);
	    return -1;
        }
        ctx->index.len = hdr->len;
    }

    /*
     * initialize variable that will hold each component of the index.
     * If there are multiple indexes for the table, the variable_lists
     * need to be linked together, in order.
     */
       memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
       var_saHpiDomainId.type = ASN_UNSIGNED; 
       var_saHpiDomainId.next_variable = NULL;

    /*
     * parse the oid into the individual index components
     */
    err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
    if (err == SNMP_ERR_NOERROR) {
       /*
        * copy index components into the context structure
        */
 		ctx->saHpiDomainId = *var_saHpiDomainId.val.integer;
   
      	/*
   		 *Check Index 
         */
    	err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
         
    }

    /*
     * parsing may have allocated memory. free it.
     */
    snmp_reset_var_buffers( &var_saHpiDomainId );

    subagent_unlock(&hpi_lock_data);
    
    return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiDomainInfoTable_can_activate(saHpiDomainInfoTable_context *undo_ctx,
                      saHpiDomainInfoTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiDomainInfoTable_can_deactivate(saHpiDomainInfoTable_context *undo_ctx,
                        saHpiDomainInfoTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiDomainInfoTable_can_delete(saHpiDomainInfoTable_context *undo_ctx,
                    saHpiDomainInfoTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiDomainInfoTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiDomainInfoTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiDomainInfoTable_context *
saHpiDomainInfoTable_create_row( netsnmp_index* hdr)
{
    
    subagent_lock(&hpi_lock_data);
    
    saHpiDomainInfoTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiDomainInfoTable_context);
        
 	DEBUGMSGTL ((AGENT, "saHpiDomainInfoTable_create_row: Called\n"));        
        
    if(!ctx) {
        subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiDomainInfoTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
     
     
    /**
     ctx->saHpiDomainTagTextType = 0;
     ctx->saHpiDomainTagTextLanguage = 0;
     ctx->saHpiDomainTag = 0;
    */

	subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiDomainInfoTable_context *
saHpiDomainInfoTable_duplicate_row( saHpiDomainInfoTable_context * row_ctx)
{
    saHpiDomainInfoTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
    
    dup = SNMP_MALLOC_TYPEDEF(saHpiDomainInfoTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    if(saHpiDomainInfoTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiDomainInfoTable_delete_row( saHpiDomainInfoTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

    subagent_lock(&hpi_lock_data);
    
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiDomainInfoTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiDomainInfoTable_context *row_ctx =
            (saHpiDomainInfoTable_context *)rg->existing_row;
//    saHpiDomainInfoTable_context *undo_ctx =
//            (saHpiDomainInfoTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;

     DEBUGMSGTL ((AGENT, 
     	"saHpiDomainInfoTable_set_reserve1: Called\n"));

    /*
     * Loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(
            			var, 
            			ASN_INTEGER,
                        sizeof(row_ctx->saHpiDomainTagTextType));
        break;

        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(
            			var, 
            			ASN_INTEGER,
                        sizeof(row_ctx->saHpiDomainTagTextLanguage));
        break;

        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */  
			rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);            	
			if(rc == SNMP_ERR_NOERROR ) {
				if (var->val_len > SAHPI_MAX_TEXT_BUFFER_LENGTH) {
        			rc = SNMP_ERR_WRONGLENGTH;
                                }	
			}
			if (rc == SNMP_ERR_NOERROR)		              
	     		DEBUGMSGTL ((AGENT, 
     				"COLUMN_SAHPIDOMAINTAG NO ERROR: %d\n", rc));                       
			else
	     		DEBUGMSGTL ((AGENT, 
     				"COLUMN_SAHPIDOMAINTAG ERROR: %d\n", rc));				     					
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiDomainInfoTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiDomainInfoTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiDomainInfoTable_context *row_ctx = (saHpiDomainInfoTable_context *)rg->existing_row;
//    saHpiDomainInfoTable_context *undo_ctx = (saHpiDomainInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

     DEBUGMSGTL ((AGENT, 
     	"saHpiDomainInfoTable_set_reserve2: Called"));

    rg->rg_void = rg->list->ri;

    /*
     * Loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            if (oh_lookup_texttype(*var->val.integer - 1) == NULL) {
     			DEBUGMSGTL ((AGENT, 
	     			"COLUMN_SAHPIDOMAINTAGTEXTTYPE: SNMP_ERR_BADVALUE\n"));            	
                rc = SNMP_ERR_BADVALUE;
        	}
        break;

        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
			if ( oh_lookup_language(*var->val.integer - 1) == NULL ) {
     			DEBUGMSGTL ((AGENT, 
	     			"COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE: SNMP_ERR_BADVALUE\n"));				
          		rc = SNMP_ERR_BADVALUE;
         	}
        break;

        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */
       		if ( var->val_len > SAHPI_MAX_TEXT_BUFFER_LENGTH  ) {
     			DEBUGMSGTL ((AGENT, 
	     			"COLUMN_SAHPIDOMAINTAG: SNMP_ERR_BADVALUE\n"));       			
               	rc = SNMP_ERR_BADVALUE;
          	}    
        break;

        default: /** We shouldn't get here */
     		DEBUGMSGTL ((AGENT, 
	     		"saHpiDomainInfoTable_set_reserve2: netsnmp_assert\n"));        
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiDomainInfoTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiDomainInfoTable_context *row_ctx = (saHpiDomainInfoTable_context *)rg->existing_row;
//    saHpiDomainInfoTable_context *undo_ctx = (saHpiDomainInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;

    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            row_ctx->saHpiDomainTagTextType = *var->val.integer;
	    row_err = set_table_domain_tag(row_ctx);
        break;

        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            row_ctx->saHpiDomainTagTextLanguage = *var->val.integer;
    	    row_err = set_table_domain_tag(row_ctx);
        break;

        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiDomainTag,var->val.string,var->val_len);
            row_ctx->saHpiDomainTag_len = var->val_len;
	    row_err = set_table_domain_tag(row_ctx);	    
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiDomainInfoTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiDomainInfoTable_context *row_ctx = (saHpiDomainInfoTable_context *)rg->existing_row;
//    saHpiDomainInfoTable_context *undo_ctx = (saHpiDomainInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiDomainInfoTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiDomainInfoTable_context *row_ctx = (saHpiDomainInfoTable_context *)rg->existing_row;
//    saHpiDomainInfoTable_context *undo_ctx = (saHpiDomainInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        	break;

        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        	break;

        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */
        	break;

        default: 
            break;/** We shouldn't get here */
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiDomainInfoTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiDomainInfoTable_context *row_ctx = (saHpiDomainInfoTable_context *)rg->existing_row;
//    saHpiDomainInfoTable_context *undo_ctx = (saHpiDomainInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}


/************************************************************
 *
 * Initialize the saHpiDomainInfoTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiDomainInfoTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiDomainInfoTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiDomainInfoTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiDomainInfoTable_oid,
                                             saHpiDomainInfoTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiDomainInfoTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * internal indexes
     */
   	/** index: saHpiDomainId */
  	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiDomainInfoTable_COL_MIN;
    table_info->max_column = saHpiDomainInfoTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiDomainInfoTable_get_value;
    cb.container = netsnmp_container_find("saHpiDomainInfoTable_primary:"
                                          "saHpiDomainInfoTable:"
                                          "table_container");
                                          
    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiDomainInfoTable_secondary:"
                                                       "saHpiDomainInfoTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiDomainInfoTable_cmp; 

    cb.can_set = 1;  

    cb.create_row = (UserRowMethod*)saHpiDomainInfoTable_create_row; 

    cb.duplicate_row = (UserRowMethod*)saHpiDomainInfoTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiDomainInfoTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiDomainInfoTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiDomainInfoTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiDomainInfoTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiDomainInfoTable_can_delete;

    cb.set_reserve1 = saHpiDomainInfoTable_set_reserve1;
    cb.set_reserve2 = saHpiDomainInfoTable_set_reserve2;
    cb.set_action = saHpiDomainInfoTable_set_action;
    cb.set_commit = saHpiDomainInfoTable_set_commit;
    cb.set_free = saHpiDomainInfoTable_set_free;
    cb.set_undo = saHpiDomainInfoTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiDomainInfoTable",
                "Registering table saHpiDomainInfoTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiDomainInfoTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiDomainInfoTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiDomainInfoTable_context *context = (saHpiDomainInfoTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIDOMAINID:
            /** SaHpiDomainId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainId,
                         sizeof(context->saHpiDomainId) );
        break;
    
        case COLUMN_SAHPIDOMAINCAPABILITIES:
            /** BITS = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (unsigned char*)&context->saHpiDomainCapabilities,
                         context->saHpiDomainCapabilities_len );
        break;
    
        case COLUMN_SAHPIDOMAINISPEER:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (unsigned char*)&context->saHpiDomainIsPeer,
                         sizeof(context->saHpiDomainIsPeer) );
        break;
    
        case COLUMN_SAHPIDOMAINTAGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (unsigned char*)&context->saHpiDomainTagTextType,
                         sizeof(context->saHpiDomainTagTextType) );
        break;
    
        case COLUMN_SAHPIDOMAINTAGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (unsigned char*)&context->saHpiDomainTagTextLanguage,
                         sizeof(context->saHpiDomainTagTextLanguage) );
        break;
    
        case COLUMN_SAHPIDOMAINTAG:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (unsigned char*)&context->saHpiDomainTag,
                         context->saHpiDomainTag_len );
        break;
    
        case COLUMN_SAHPIDOMAINREFERENCEUPDATECOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainReferenceUpdateCount,
                         sizeof(context->saHpiDomainReferenceUpdateCount) );
        break;
    
        case COLUMN_SAHPIDOMAINREFERENCEUPDATETIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (unsigned char*)&context->saHpiDomainReferenceUpdateTimestamp,
                         sizeof(context->saHpiDomainReferenceUpdateTimestamp) );
        break;
    
        case COLUMN_SAHPIDOMAINRESOURCEPRESENCEUPDATECOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainResourcePresenceUpdateCount,
                         sizeof(context->saHpiDomainResourcePresenceUpdateCount) );
        break;
    
        case COLUMN_SAHPIDOMAINRESOURCEPRESENCEUPDATETIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (unsigned char*)&context->saHpiDomainResourcePresenceUpdateTimestamp,
                         sizeof(context->saHpiDomainResourcePresenceUpdateTimestamp) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMUPDATECOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainAlarmUpdateCount,
                         sizeof(context->saHpiDomainAlarmUpdateCount) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMUPDATETIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (unsigned char*)&context->saHpiDomainAlarmUpdateTimestamp,
                         sizeof(context->saHpiDomainAlarmUpdateTimestamp) );
        break;
    
        case COLUMN_SAHPIDOMAINACTIVEALARMS:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainActiveAlarms,
                         sizeof(context->saHpiDomainActiveAlarms) );
        break;
    
        case COLUMN_SAHPIDOMAINCRITICALALARMS:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainCriticalAlarms,
                         sizeof(context->saHpiDomainCriticalAlarms) );
        break;
    
        case COLUMN_SAHPIDOMAINMAJORALARMS:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainMajorAlarms,
                         sizeof(context->saHpiDomainMajorAlarms) );
        break;
    
        case COLUMN_SAHPIDOMAINMINORALARMS:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainMinorAlarms,
                         sizeof(context->saHpiDomainMinorAlarms) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMUSERLIMIT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (unsigned char*)&context->saHpiDomainAlarmUserLimit,
                         sizeof(context->saHpiDomainAlarmUserLimit) );
        break;
    
        case COLUMN_SAHPIDOMAINALARMOVERFLOW:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (unsigned char*)&context->saHpiDomainAlarmOverflow,
                         sizeof(context->saHpiDomainAlarmOverflow) );
        break;
    
        case COLUMN_SAHPIDOMAINGUID:
            /** SaHpiGuid = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (unsigned char*)&context->saHpiDomainGuid,
                         context->saHpiDomainGuid_len );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiDomainInfoTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiDomainInfoTable_get_by_idx
 */
const saHpiDomainInfoTable_context *
saHpiDomainInfoTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiDomainInfoTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


