/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiEventLogInfoTable.c 5452 2006-02-10 19:54:20Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include <saHpiEventLogInfoTable.h>
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceTable.h>
#include <saHpiResourceEventTable.h>
#include <saHpiDomainEventTable.h>
#include <saHpiEventLogTable.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiEventLogInfoTable_oid[] = { saHpiEventLogInfoTable_TABLE_OID };
size_t saHpiEventLogInfoTable_oid_len = OID_LENGTH(saHpiEventLogInfoTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

SaErrorT populate_saHpiEventLogInfo (SaHpiSessionIdT sessionid)
{

	SaErrorT rv;
	SaHpiEventLogInfoT event_log_info;

        SaHpiEntryIdT EntryId;
	SaHpiRptEntryT rpt_entry;

        SaHpiBoolT event_log_state;
	
	oid evt_log_info_oid[EVENT_LOG_INFO_INDEX_NR];
	netsnmp_index evt_log_info_index;
	saHpiEventLogInfoTable_context *evt_log_info_context;
	
	int isNewRow = MIB_TRUE;

	DEBUGMSGTL ((AGENT, "populate_saHpiEventLogInfo\n"));

        EntryId = SAHPI_FIRST_ENTRY;

        subagent_lock(&hpi_lock_data);	
	
        do {
                rv = saHpiRptEntryGet(sessionid, EntryId, &EntryId, &rpt_entry);

                if (rv != SA_OK) {
                        DEBUGMSGTL ((AGENT, 
                        "populate_saHpiEventLogInfo, saHpiRptEntryGet Failed: rv = %d\n",rv));
                        rv =  AGENT_ERR_INTERNAL_ERROR;
                        break;
                }
                rv = saHpiEventLogInfoGet (sessionid, 
                                           rpt_entry.ResourceId,
                                           &event_log_info);
                if (rv == SA_ERR_HPI_CAPABILITY) {
                        DEBUGMSGTL ((AGENT, "SA_ERR_HPI_CAPABILITY: populate_saHpiEventLogInfo Failed: rv = %d\n",rv));
                        rv = AGENT_ERR_INTERNAL_ERROR;
                        continue;
                } else if ( (rv != SA_OK) && (rv != SA_ERR_HPI_CAPABILITY) ) {
                        DEBUGMSGTL ((AGENT, "populate_saHpiEventLogInfo Failed: rv = %d\n",rv));
                        rv = AGENT_ERR_INTERNAL_ERROR;
                        break;
                }
	
                evt_log_info_index.len = EVENT_LOG_INFO_INDEX_NR;
                evt_log_info_oid[0] = get_domain_id(sessionid);
                evt_log_info_oid[1] = rpt_entry.ResourceId;
                evt_log_info_index.oids = (oid *) & evt_log_info_oid;
	
                /* See if it exists. */
                evt_log_info_context = NULL;
                evt_log_info_context = CONTAINER_FIND (cb.container, &evt_log_info_index);
		
                if (!evt_log_info_context) { 
                        // New entry. Add it
                        evt_log_info_context = 
                                saHpiEventLogInfoTable_create_row ( &evt_log_info_index);
                }
		else {
			isNewRow = MIB_FALSE;
		}	
                if (!evt_log_info_context) {
                        snmp_log (LOG_ERR, "Not enough memory for a EventLogInfo row!");
                        rv = AGENT_ERR_INTERNAL_ERROR;
                        break;
                }	


                /** UNSIGNED32 = ASN_UNSIGNED */
                evt_log_info_context->saHpiEventLogInfoEntries = event_log_info.Entries;

                /** UNSIGNED32 = ASN_UNSIGNED */
                evt_log_info_context->saHpiEventLogInfoSize = event_log_info.Size;

                /** UNSIGNED32 = ASN_UNSIGNED */
                evt_log_info_context->saHpiEventLogInfoUserEventMaxSize = 
                        event_log_info.UserEventMaxSize;

                /** SaHpiTime = ASN_OCTET_STR */
		hpitime_to_snmptime(event_log_info.UpdateTimestamp, 
		                 evt_log_info_context->saHpiEventLogInfoUpdateTimestamp);		
				 
		evt_log_info_context->saHpiEventLogInfoUpdateTimestamp_len = sizeof(SaHpiTimeT);		 

                /** SaHpiTime = ASN_OCTET_STR */
                evt_log_info_context->saHpiEventLogInfoTime_len = 
                       sizeof(SaHpiTimeT);
                memset(evt_log_info_context->saHpiEventLogInfoTime, 
                       0,  sizeof(SaHpiTimeT));
                hpitime_to_snmptime(event_log_info.CurrentTime, evt_log_info_context->saHpiEventLogInfoTime); ;

                /** TruthValue = ASN_INTEGER */
                evt_log_info_context->saHpiEventLogInfoIsEnabled =
                        (event_log_info.Enabled == SAHPI_TRUE) ? 
                                MIB_TRUE : MIB_FALSE;

                /** TruthValue = ASN_INTEGER */
                evt_log_info_context->saHpiEventLogInfoOverflowFlag =
                        (event_log_info.OverflowFlag == SAHPI_TRUE) ? 
                                MIB_TRUE : MIB_FALSE;

                /** TruthValue = ASN_INTEGER */
                evt_log_info_context->saHpiEventLogInfoOverflowResetable =
                        (event_log_info.OverflowResetable == SAHPI_TRUE) ? 
                                MIB_TRUE : MIB_FALSE; 

                /** INTEGER = ASN_INTEGER */
                evt_log_info_context->saHpiEventLogInfoOverflowAction = 
                        event_log_info.OverflowAction + 1;

                /** INTEGER = ASN_INTEGER */
                evt_log_info_context->saHpiEventLogInfoOverflowReset = 0;

                /** TruthValue = ASN_INTEGER */
                evt_log_info_context->saHpiEventLogClear = MIB_FALSE;

                /** TruthValue = ASN_INTEGER */
                rv = saHpiEventLogStateGet (sessionid, rpt_entry.ResourceId,
                                            &event_log_state);
                if (rv != SA_OK) {
                        DEBUGMSGTL ((AGENT, 
                        "populate_saHpiEventLogInfo, saHpiEventLogStateGet Failed: rv = %d\n",rv));
                        rv =  AGENT_ERR_INTERNAL_ERROR;
                        break;
                }
                evt_log_info_context->saHpiEventLogState =                         
		        (event_log_state == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE; 

		if (isNewRow == MIB_TRUE) {
                	CONTAINER_INSERT (cb.container, evt_log_info_context);
			populate_saHpiEventLog(sessionid, rpt_entry.ResourceId);
		}
		else { //reset it
			isNewRow = MIB_TRUE;
		}				


	} while (EntryId != SAHPI_LAST_ENTRY);

	
	isNewRow = MIB_TRUE;
        /******************************************************/
        /* now get the Domain Event Log Info for this session */
        /* this is accomplished by using                      */
        /* in saHpiEventLogInfoGet()                          */
        /* SAHPI_UNSPECIFIED_RESOURCE_ID                      */
        /******************************************************/
        rv = saHpiEventLogInfoGet (sessionid, 
                                   SAHPI_UNSPECIFIED_RESOURCE_ID,
                                   &event_log_info);
        if (rv != SA_OK) {
                DEBUGMSGTL ((AGENT, "getting Domain Event Log Failed: rv = %d\n",rv));
                
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
        }

        evt_log_info_index.len = EVENT_LOG_INFO_INDEX_NR;
        evt_log_info_oid[0] = get_domain_id(sessionid);
        evt_log_info_oid[1] = SAHPI_UNSPECIFIED_RESOURCE_ID;
        evt_log_info_index.oids = (oid *) & evt_log_info_oid;

        /* See if it exists. */
        evt_log_info_context = NULL;
        evt_log_info_context = CONTAINER_FIND (cb.container, &evt_log_info_index);

        if (!evt_log_info_context) { 
                // New entry. Add it
                evt_log_info_context = 
                        saHpiEventLogInfoTable_create_row ( &evt_log_info_index);
        }
	else {
		isNewRow = MIB_FALSE;
	}	
	
        if (!evt_log_info_context) {
                DEBUGMSGTL ((AGENT, "Not enough memory for a EventLogInfo row!"));
                snmp_log (LOG_ERR, "Not enough memory for a EventLogInfo row!");
                
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
        } 

        /** UNSIGNED32 = ASN_UNSIGNED */
        evt_log_info_context->saHpiEventLogInfoEntries = event_log_info.Entries;

        /** UNSIGNED32 = ASN_UNSIGNED */
        evt_log_info_context->saHpiEventLogInfoSize = event_log_info.Size;

        /** UNSIGNED32 = ASN_UNSIGNED */
        evt_log_info_context->saHpiEventLogInfoUserEventMaxSize = 
                event_log_info.UserEventMaxSize;

        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event_log_info.UpdateTimestamp, 
	                 evt_log_info_context->saHpiEventLogInfoUpdateTimestamp);
			 
        evt_log_info_context->saHpiEventLogInfoUpdateTimestamp_len = sizeof(SaHpiTimeT);			 

        /** SaHpiTime = ASN_OCTET_STR */
        evt_log_info_context->saHpiEventLogInfoTime_len = 
               sizeof(SaHpiTimeT);
        memset(evt_log_info_context->saHpiEventLogInfoTime, 
               0, sizeof(SaHpiTimeT));
        hpitime_to_snmptime(event_log_info.CurrentTime, evt_log_info_context->saHpiEventLogInfoTime);

        /** TruthValue = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogInfoIsEnabled =
                (event_log_info.Enabled == SAHPI_TRUE) ? 
                        MIB_TRUE : MIB_FALSE;

        /** TruthValue = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogInfoOverflowFlag =
                (event_log_info.OverflowFlag == SAHPI_TRUE) ? 
                        MIB_TRUE : MIB_FALSE;

        /** TruthValue = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogInfoOverflowResetable =
                (event_log_info.OverflowResetable == SAHPI_TRUE) ? 
                        MIB_TRUE : MIB_FALSE; 

        /** INTEGER = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogInfoOverflowAction = 
                event_log_info.OverflowAction + 1;

        /** INTEGER = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogInfoOverflowReset = 0;

        /** TruthValue = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogClear = MIB_FALSE;

        /** TruthValue = ASN_INTEGER */
        evt_log_info_context->saHpiEventLogState = 		
	          (event_log_info.Enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE; 
        
	if (isNewRow == MIB_TRUE) {
        	CONTAINER_INSERT (cb.container, evt_log_info_context);	
	}
	
	populate_saHpiEventLog(sessionid, SAHPI_UNSPECIFIED_RESOURCE_ID);
	
	subagent_unlock(&hpi_lock_data);
	
	return rv;

}

/*
 * @row_ctx - pointer to the row context
 *
 * return:
 */
int event_info_overflow_reset (saHpiEventLogInfoTable_context *row_ctx)
{

	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;

        DEBUGMSGTL ((AGENT, "event_info_overflow_reset, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];
	
        rc = saHpiEventLogOverflowReset(session_id, resource_id);
       
	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
		  	"Call to saHpiEventLogOverflowReset failed to reset overflow flag rc: %s.\n",
		 	 oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
		   	"Call to saHpiEventLogOverflowReset failed to reset overflow flag rc: %s.\n",
		   	oh_lookup_error(rc)));
		return get_snmp_error(rc);
 
	}		

	return SNMP_ERR_NOERROR; 
}	



/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int event_log_info_time_set (saHpiEventLogInfoTable_context *row_ctx)
{
	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;
        SaHpiTimeT              time;

        DEBUGMSGTL ((AGENT, "event_log_info_time_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];

        if (row_ctx->saHpiEventLogInfoTime_len > sizeof(SaHpiTimeoutT)) 
                return SNMP_ERR_TOOBIG;
        
        time = snmptime_to_hpitime(row_ctx->saHpiEventLogInfoTime);

        rc = saHpiEventLogTimeSet(session_id, resource_id, time);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to event_log_info_time_set"
                          " failed to set timeout [%lld] rc: %s.\n",
                          time, 
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			   "Call to event_log_info_time_set"
                           " failed to set timeout [%lld] rc: %s.\n",
                           time, 
			   oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/*
 * @row_ctx - pointer to the row context
 *
 * return:
 */
int event_log_info_clear (saHpiEventLogInfoTable_context *row_ctx)
{

	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;

        int                     rv = SNMP_ERR_NOERROR;

        DEBUGMSGTL ((AGENT, "event_log_info_clear, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];
	
        rc = saHpiEventLogClear(session_id, resource_id);

	if (rc != SA_OK) {
	
	        // Since the command wasn't successful, we reset the saHpiEventLogClear node.
	        row_ctx->saHpiEventLogClear = MIB_FALSE;
	
		snmp_log (LOG_ERR,
		  	"Call to saHpiEventLogClear failed to clear event logs rc: %s.\n",
		 	 oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
		   	"Call to saHpiEventLogClear failed to clear event logs rc: %s.\n",
		   	oh_lookup_error(rc)));
		return get_snmp_error(rc);
 
	}		

        /* it succeeded clear the underlying tables */
        rv = event_log_clear(session_id, resource_id, MIB_FALSE);

	return SNMP_ERR_NOERROR; 
	
}

/*
 * @row_ctx - pointer to the row context
 *
 * return:
 */
int event_log_info_state_set (saHpiEventLogInfoTable_context *row_ctx)
{

	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;

        DEBUGMSGTL ((AGENT, "event_log_info_state_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];
	
	if (row_ctx->saHpiEventLogState == MIB_TRUE) {
	    rc = saHpiEventLogStateSet(session_id, resource_id, SAHPI_TRUE);
	}
	else {
	    rc = saHpiEventLogStateSet(session_id, resource_id, SAHPI_FALSE);    
	}
       
	if (rc != SA_OK) {
	
		snmp_log (LOG_ERR,
		  	"Call to saHpiEventLogStateSet failed to enable/disable event logs rc: %s.\n",
		 	 oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
		   	"Call to saHpiEventLogStateSet failed to enable/disable event logs rc: %s.\n",
		   	oh_lookup_error(rc)));
		return get_snmp_error(rc);
 
	}		

	return SNMP_ERR_NOERROR; 
	
}


/*
 * @sessionid - the session id of the session
 *
 * return:
 */


SaErrorT event_log_info_update (SaHpiSessionIdT sessionid)
{
	SaErrorT rv;
	SaHpiEventLogInfoT event_log_info;

        SaHpiEntryIdT EntryId;
	SaHpiRptEntryT rpt_entry;

	int isNewRow = MIB_TRUE;
	
	oid evt_log_info_oid[EVENT_LOG_INFO_INDEX_NR];
	netsnmp_index evt_log_info_index;
	saHpiEventLogInfoTable_context *evt_log_info_context;

	DEBUGMSGTL ((AGENT, "event_log_info_update called\n"));

        EntryId = SAHPI_FIRST_ENTRY;
        do { //Get the plugin event log first
                rv = saHpiRptEntryGet(sessionid, EntryId, &EntryId, &rpt_entry);

                if (rv != SA_OK) {
                        DEBUGMSGTL ((AGENT, 
                        "event_log_info_update, saHpiRptEntryGet Failed: rv = %d\n",rv));
                        rv =  AGENT_ERR_INTERNAL_ERROR;
                        break;
                }
                
		rv = saHpiEventLogInfoGet (sessionid, 
                                           rpt_entry.ResourceId,
                                           &event_log_info);
					   
                if (rv == SA_ERR_HPI_CAPABILITY) {
                        DEBUGMSGTL ((AGENT, "SA_ERR_HPI_CAPABILITY: event_log_info_update Failed: rv = %d\n",rv));
                        rv = AGENT_ERR_INTERNAL_ERROR;
                        continue;
                } else if ( (rv != SA_OK) && (rv != SA_ERR_HPI_CAPABILITY) ) {
                        DEBUGMSGTL ((AGENT, "event_log_info_update Failed: rv = %d\n",rv));
                        rv = AGENT_ERR_INTERNAL_ERROR;
                        break;
                }
		
                evt_log_info_index.len = EVENT_LOG_INFO_INDEX_NR;
                evt_log_info_oid[0] = get_domain_id(sessionid);
                evt_log_info_oid[1] = rpt_entry.ResourceId;
                evt_log_info_index.oids = (oid *) & evt_log_info_oid;
	
                /* See if it exists. */
                evt_log_info_context = NULL;
                evt_log_info_context = CONTAINER_FIND (cb.container, &evt_log_info_index);
		

		if (!evt_log_info_context) { //Obviously a new entry
			evt_log_info_context = 
                                saHpiEventLogInfoTable_create_row ( &evt_log_info_index);
			
			if (!evt_log_info_context) {
                        	snmp_log (LOG_ERR, "Not enough memory for a EventLogInfo row!");
                        	rv = AGENT_ERR_INTERNAL_ERROR;
                        	break;
			}		
		
		}
		else {
			isNewRow = MIB_FALSE;
		}	
			
	
		if ((isNewRow == MIB_TRUE) || (compare_timestamp(event_log_info.UpdateTimestamp,
		                              evt_log_info_context->saHpiEventLogInfoUpdateTimestamp) != 0)) {	
                	
			/** UNSIGNED32 = ASN_UNSIGNED */
                	evt_log_info_context->saHpiEventLogInfoEntries = event_log_info.Entries;

                	/** UNSIGNED32 = ASN_UNSIGNED */
                	evt_log_info_context->saHpiEventLogInfoSize = event_log_info.Size;
	
                	/** UNSIGNED32 = ASN_UNSIGNED */
                	evt_log_info_context->saHpiEventLogInfoUserEventMaxSize = 
                        event_log_info.UserEventMaxSize;

                	/** SaHpiTime = ASN_OCTET_STR */
			hpitime_to_snmptime(event_log_info.UpdateTimestamp, 
			                 evt_log_info_context->saHpiEventLogInfoUpdateTimestamp);

                	evt_log_info_context->saHpiEventLogInfoUpdateTimestamp_len = sizeof(SaHpiTimeT);
			
			/** SaHpiTime = ASN_OCTET_STR */
                	evt_log_info_context->saHpiEventLogInfoTime_len = 
                	       sizeof(SaHpiTimeT);
                	memset(evt_log_info_context->saHpiEventLogInfoTime, 
                 	      0, sizeof(SaHpiTimeT));
			      
                	hpitime_to_snmptime(event_log_info.CurrentTime, evt_log_info_context->saHpiEventLogInfoTime); 

                	/** TruthValue = ASN_INTEGER */
                	evt_log_info_context->saHpiEventLogInfoIsEnabled =
                        	(event_log_info.Enabled == SAHPI_TRUE) ? 
                     	           MIB_TRUE : MIB_FALSE;

                	/** TruthValue = ASN_INTEGER */
                	evt_log_info_context->saHpiEventLogInfoOverflowFlag =
                       	 	(event_log_info.OverflowFlag == SAHPI_TRUE) ? 
                        	        MIB_TRUE : MIB_FALSE;

                	/** TruthValue = ASN_INTEGER */
                	evt_log_info_context->saHpiEventLogInfoOverflowResetable =
                        	(event_log_info.OverflowResetable == SAHPI_TRUE) ? 
                                	MIB_TRUE : MIB_FALSE; 

                	/** INTEGER = ASN_INTEGER */
                	evt_log_info_context->saHpiEventLogInfoOverflowAction = 
                        	event_log_info.OverflowAction + 1;

                	/** INTEGER = ASN_INTEGER */
                	evt_log_info_context->saHpiEventLogInfoOverflowReset = 0;

                	/** TruthValue = ASN_INTEGER */
                	evt_log_info_context->saHpiEventLogClear = MIB_FALSE;

                	/** TruthValue = ASN_INTEGER */                	
	                evt_log_info_context->saHpiEventLogState =                         
		        	(event_log_info.Enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE; 
			
			if (isNewRow == MIB_TRUE) {
				CONTAINER_INSERT (cb.container, evt_log_info_context);	
			}
			else {
				isNewRow = MIB_TRUE;
			}	
					
			populate_saHpiEventLog(sessionid, rpt_entry.ResourceId);			
			
                }	
		
	}while(EntryId != SAHPI_LAST_ENTRY);
	
        /******************************************************/
        /* now get the Domain Event Log Info for this session */
        /* this is accomplished by using                      */
        /* in saHpiEventLogInfoGet()                          */
        /* SAHPI_UNSPECIFIED_RESOURCE_ID                      */
        /******************************************************/	
        rv = saHpiEventLogInfoGet (sessionid, 
                                   SAHPI_UNSPECIFIED_RESOURCE_ID,
                                   &event_log_info);
        if (rv != SA_OK) {
                DEBUGMSGTL ((AGENT, "getting Domain Event Log Failed: rv = %d\n",rv));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        evt_log_info_index.len = EVENT_LOG_INFO_INDEX_NR;
        evt_log_info_oid[0] = get_domain_id(sessionid);
        evt_log_info_oid[1] = SAHPI_UNSPECIFIED_RESOURCE_ID;
        evt_log_info_index.oids = (oid *) & evt_log_info_oid;

        /* See if it exists. */
        evt_log_info_context = NULL;
        evt_log_info_context = CONTAINER_FIND (cb.container, &evt_log_info_index);

        if (!evt_log_info_context) { 
		evt_log_info_context = 
                	saHpiEventLogInfoTable_create_row ( &evt_log_info_index);
			
		if (!evt_log_info_context) {
                       	snmp_log (LOG_ERR, "Not enough memory for a EventLogInfo row!");
                       	rv = AGENT_ERR_INTERNAL_ERROR;
			return rv;
		}	
        }
	else {
		isNewRow = MIB_FALSE;
	}	
	
	if ((isNewRow == MIB_TRUE) || (compare_timestamp(event_log_info.UpdateTimestamp,
		                        evt_log_info_context->saHpiEventLogInfoUpdateTimestamp) != 0)) {
	        /** UNSIGNED32 = ASN_UNSIGNED */
        	evt_log_info_context->saHpiEventLogInfoEntries = event_log_info.Entries;

	        /** UNSIGNED32 = ASN_UNSIGNED */
        	evt_log_info_context->saHpiEventLogInfoSize = event_log_info.Size;

	        /** UNSIGNED32 = ASN_UNSIGNED */
        	evt_log_info_context->saHpiEventLogInfoUserEventMaxSize = 
                	event_log_info.UserEventMaxSize;

	        /** SaHpiTime = ASN_OCTET_STR */
		hpitime_to_snmptime(event_log_info.UpdateTimestamp, 
		                 evt_log_info_context->saHpiEventLogInfoUpdateTimestamp);
				 
		evt_log_info_context->saHpiEventLogInfoUpdateTimestamp_len = sizeof(SaHpiTimeT);		 

	        /** SaHpiTime = ASN_OCTET_STR */
        	evt_log_info_context->saHpiEventLogInfoTime_len = 
        	       	sizeof(SaHpiTimeT);
	        memset(evt_log_info_context->saHpiEventLogInfoTime, 
        	       0, sizeof(SaHpiTimeT));
	        hpitime_to_snmptime(event_log_info.CurrentTime, evt_log_info_context->saHpiEventLogInfoTime);;

	        /** TruthValue = ASN_INTEGER */
        	evt_log_info_context->saHpiEventLogInfoIsEnabled =
                	(event_log_info.Enabled == SAHPI_TRUE) ? 
                        	MIB_TRUE : MIB_FALSE;

	        /** TruthValue = ASN_INTEGER */
        	evt_log_info_context->saHpiEventLogInfoOverflowFlag =
                	(event_log_info.OverflowFlag == SAHPI_TRUE) ? 
                        	MIB_TRUE : MIB_FALSE;

	        /** TruthValue = ASN_INTEGER */
        	evt_log_info_context->saHpiEventLogInfoOverflowResetable =
                	(event_log_info.OverflowResetable == SAHPI_TRUE) ? 
                        	MIB_TRUE : MIB_FALSE; 

	        /** INTEGER = ASN_INTEGER */
        	evt_log_info_context->saHpiEventLogInfoOverflowAction = 
                	event_log_info.OverflowAction + 1;

	        /** INTEGER = ASN_INTEGER */
        	evt_log_info_context->saHpiEventLogInfoOverflowReset = 0;
	
        	/** TruthValue = ASN_INTEGER */
	        evt_log_info_context->saHpiEventLogClear = MIB_FALSE;

        	/** TruthValue = ASN_INTEGER */
	        evt_log_info_context->saHpiEventLogState = 		
		          (event_log_info.Enabled == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE; 
        
		if (isNewRow == MIB_TRUE) {
        		CONTAINER_INSERT (cb.container, evt_log_info_context);	
		}
	
		populate_saHpiEventLog(sessionid, SAHPI_UNSPECIFIED_RESOURCE_ID);
		     		
	}
	return SA_OK;	
}


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiEventLogInfoTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiEventLogInfoTable_cmp( const void *lhs, const void *rhs )
{
        saHpiEventLogInfoTable_context *context_l =
        (saHpiEventLogInfoTable_context *)lhs;
        saHpiEventLogInfoTable_context *context_r =
        (saHpiEventLogInfoTable_context *)rhs;
        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiEventLogInfoTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiEventLogInfoTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1])
		        return 0;
	}
	
	return 0;
}

/************************************************************
 * Initializes the saHpiEventLogInfoTable module
 */
void
init_saHpiEventLogInfoTable(void)
{
    initialize_table_saHpiEventLogInfoTable();

    /*
     * TODO: perform any startup stuff here, such as
     * populating the table with initial data.
     *
     * saHpiEventLogInfoTable_context * new_row = create_row(index);
     * CONTAINER_INSERT(cb.container,new_row);
     */
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiEventLogInfoTable_row_copy(saHpiEventLogInfoTable_context * dst,
                         saHpiEventLogInfoTable_context * src)
{
    if(!dst||!src)
        return 1;
        
   
   subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiEventLogInfoEntries = src->saHpiEventLogInfoEntries;

    dst->saHpiEventLogInfoSize = src->saHpiEventLogInfoSize;

    dst->saHpiEventLogInfoUserEventMaxSize = src->saHpiEventLogInfoUserEventMaxSize;

    memcpy( dst->saHpiEventLogInfoUpdateTimestamp, src->saHpiEventLogInfoUpdateTimestamp, src->saHpiEventLogInfoUpdateTimestamp_len );
    dst->saHpiEventLogInfoUpdateTimestamp_len = src->saHpiEventLogInfoUpdateTimestamp_len;;

    memcpy( dst->saHpiEventLogInfoTime, src->saHpiEventLogInfoTime, 
            src->saHpiEventLogInfoTime_len );
    dst->saHpiEventLogInfoTime_len = src->saHpiEventLogInfoTime_len;

    dst->saHpiEventLogInfoIsEnabled = src->saHpiEventLogInfoIsEnabled;

    dst->saHpiEventLogInfoOverflowFlag = src->saHpiEventLogInfoOverflowFlag;

    dst->saHpiEventLogInfoOverflowResetable = src->saHpiEventLogInfoOverflowResetable;

    dst->saHpiEventLogInfoOverflowAction = src->saHpiEventLogInfoOverflowAction;

    dst->saHpiEventLogInfoOverflowReset = src->saHpiEventLogInfoOverflowReset;

    dst->saHpiEventLogClear = src->saHpiEventLogClear;

    dst->saHpiEventLogState = src->saHpiEventLogState;

    subagent_unlock(&hpi_lock_data);
    
    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiEventLogInfoTable_extract_index( saHpiEventLogInfoTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        int err;
	
        DEBUGMSGTL ((AGENT, "saHpiEventLogInfoTable_extract_index, called\n"));

        
	subagent_lock(&hpi_lock_data);
	/*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
			return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
       memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
       var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiDomainId.next_variable = &var_saHpiResourceId;

       memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
       var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiResourceId.next_variable = NULL;

        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiResourceId */
		
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer); 		
        
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        
	subagent_unlock(&hpi_lock_data);
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiEventLogInfoTable_can_activate(saHpiEventLogInfoTable_context *undo_ctx,
                      saHpiEventLogInfoTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiEventLogInfoTable_can_deactivate(saHpiEventLogInfoTable_context *undo_ctx,
                        saHpiEventLogInfoTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiEventLogInfoTable_can_delete(saHpiEventLogInfoTable_context *undo_ctx,
                    saHpiEventLogInfoTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiEventLogInfoTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiEventLogInfoTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiEventLogInfoTable_context *
saHpiEventLogInfoTable_create_row( netsnmp_index* hdr)
{
    subagent_lock(&hpi_lock_data);
    
    
    saHpiEventLogInfoTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiEventLogInfoTable_context);
    if(!ctx) {
   	 subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiEventLogInfoTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiEventLogInfoTime = 0;
     ctx->saHpiEventLogInfoOverflowResetable = 0;
     ctx->saHpiEventLogInfoOverflowAction = 0;
     ctx->saHpiEventLogInfoOverflowReset = 0;
     ctx->saHpiEventLogClear = 0;
     ctx->saHpiEventLogState = 0;
    */

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiEventLogInfoTable_context *
saHpiEventLogInfoTable_duplicate_row( saHpiEventLogInfoTable_context * row_ctx)
{
    saHpiEventLogInfoTable_context * dup;

    if(!row_ctx)
        return NULL;

    
    subagent_lock(&hpi_lock_data);
    dup = SNMP_MALLOC_TYPEDEF(saHpiEventLogInfoTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    if(saHpiEventLogInfoTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiEventLogInfoTable_delete_row( saHpiEventLogInfoTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

    subagent_lock(&hpi_lock_data);
    
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    
    subagent_unlock(&hpi_lock_data);
	    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiEventLogInfoTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiEventLogInfoTable_context *row_ctx =
            (saHpiEventLogInfoTable_context *)rg->existing_row;
    //saHpiEventLogInfoTable_context *undo_ctx =
    //        (saHpiEventLogInfoTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {


        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
            rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);
            if (rc == SNMP_ERR_NOERROR ) {
                    if (var->val_len > SAF_UNSIGNED_64_LEN) {
                            rc = SNMP_ERR_WRONGLENGTH;
                            DEBUGMSGTL ((AGENT, 
                            "COLUMN_SAHPIEVENTLOGINFOTIME"
                            " SNMP_ERR_WRONGLENGTH\n"));
                    }
            }

        break;

        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiEventLogInfoOverflowReset));
        break;

        case COLUMN_SAHPIEVENTLOGCLEAR: // saHpiEventLogClear
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiEventLogClear));
        break;

        case COLUMN_SAHPIEVENTLOGSTATE: //saHpiEventLogStateSet
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiEventLogState));
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiEventLogInfoTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiEventLogInfoTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiEventLogInfoTable_context *row_ctx = (saHpiEventLogInfoTable_context *)rg->existing_row;
    //saHpiEventLogInfoTable_context *undo_ctx = (saHpiEventLogInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIEVENTLOGCLEAR:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        case COLUMN_SAHPIEVENTLOGSTATE:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiEventLogInfoTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiEventLogInfoTable_context *row_ctx = (saHpiEventLogInfoTable_context *)rg->existing_row;
//    saHpiEventLogInfoTable_context *undo_ctx = (saHpiEventLogInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;

    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiEventLogInfoTime, var->val.string, var->val_len);
            row_ctx->saHpiEventLogInfoTime_len = var->val_len;
            row_err = event_log_info_time_set(row_ctx);

        break;

        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
            //row_ctx->saHpiEventLogInfoOverflowReset = *var->val.integer;
            row_err = event_info_overflow_reset(row_ctx);
            if (row_err == SNMP_ERR_NOERROR) //we were successful; reset the flag.
	    	row_ctx->saHpiEventLogInfoOverflowFlag = MIB_FALSE;
        break;

        case COLUMN_SAHPIEVENTLOGCLEAR:
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiEventLogClear = *var->val.integer;
	    if (row_ctx->saHpiEventLogClear == MIB_TRUE)
                event_log_info_clear(row_ctx);
            if (row_err == SNMP_ERR_NOERROR) //we were successful; reset the flag.
	    	row_ctx->saHpiEventLogInfoOverflowFlag = MIB_FALSE;
 		
        break;

        case COLUMN_SAHPIEVENTLOGSTATE: 
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiEventLogState = *var->val.integer;
            event_log_info_state_set(row_ctx);

        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }
    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiEventLogInfoTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogInfoTable_context *row_ctx = (saHpiEventLogInfoTable_context *)rg->existing_row;
//    saHpiEventLogInfoTable_context *undo_ctx = (saHpiEventLogInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIEVENTLOGCLEAR:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIEVENTLOGSTATE:
            /** TruthValue = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiEventLogInfoTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogInfoTable_context *row_ctx = (saHpiEventLogInfoTable_context *)rg->existing_row;
//    saHpiEventLogInfoTable_context *undo_ctx = (saHpiEventLogInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIEVENTLOGCLEAR:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIEVENTLOGSTATE:
            /** TruthValue = ASN_INTEGER */
        break;

        default: 
                break;
                /** We shouldn't get here */
                /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiEventLogInfoTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogInfoTable_context *row_ctx = (saHpiEventLogInfoTable_context *)rg->existing_row;
//    saHpiEventLogInfoTable_context *undo_ctx = (saHpiEventLogInfoTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIEVENTLOGCLEAR:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIEVENTLOGSTATE:
            /** TruthValue = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 *
 * Initialize the saHpiEventLogInfoTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiEventLogInfoTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiEventLogInfoTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiEventLogInfoTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiEventLogInfoTable_oid,
                                             saHpiEventLogInfoTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiEventLogInfoTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiEventLogInfoTable_COL_MIN;
    table_info->max_column = saHpiEventLogInfoTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiEventLogInfoTable_get_value;
    cb.container = netsnmp_container_find("saHpiEventLogInfoTable_primary:"
                                          "saHpiEventLogInfoTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiEventLogInfoTable_secondary:"
                                                       "saHpiEventLogInfoTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiEventLogInfoTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiEventLogInfoTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiEventLogInfoTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiEventLogInfoTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiEventLogInfoTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiEventLogInfoTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiEventLogInfoTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiEventLogInfoTable_can_delete;

    cb.set_reserve1 = saHpiEventLogInfoTable_set_reserve1;
    cb.set_reserve2 = saHpiEventLogInfoTable_set_reserve2;
    cb.set_action = saHpiEventLogInfoTable_set_action;
    cb.set_commit = saHpiEventLogInfoTable_set_commit;
    cb.set_free = saHpiEventLogInfoTable_set_free;
    cb.set_undo = saHpiEventLogInfoTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiEventLogInfoTable",
                "Registering table saHpiEventLogInfoTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiEventLogInfoTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiEventLogInfoTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiEventLogInfoTable_context *context = (saHpiEventLogInfoTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIEVENTLOGINFOENTRIES:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiEventLogInfoEntries,
                         sizeof(context->saHpiEventLogInfoEntries) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOSIZE:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiEventLogInfoSize,
                         sizeof(context->saHpiEventLogInfoSize) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOUSEREVENTMAXSIZE:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiEventLogInfoUserEventMaxSize,
                         sizeof(context->saHpiEventLogInfoUserEventMaxSize) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOUPDATETIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiEventLogInfoUpdateTimestamp,
                         sizeof(context->saHpiEventLogInfoUpdateTimestamp) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOTIME:
            /** SafUnsigned64 = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiEventLogInfoTime,
                         context->saHpiEventLogInfoTime_len );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOISENABLED:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogInfoIsEnabled,
                         sizeof(context->saHpiEventLogInfoIsEnabled) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWFLAG:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogInfoOverflowFlag,
                         sizeof(context->saHpiEventLogInfoOverflowFlag) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESETABLE:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogInfoOverflowResetable,
                         sizeof(context->saHpiEventLogInfoOverflowResetable) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWACTION:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogInfoOverflowAction,
                         sizeof(context->saHpiEventLogInfoOverflowAction) );
        break;
    
        case COLUMN_SAHPIEVENTLOGINFOOVERFLOWRESET:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogInfoOverflowReset,
                         sizeof(context->saHpiEventLogInfoOverflowReset) );
        break;
    
        case COLUMN_SAHPIEVENTLOGCLEAR:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogClear,
                         sizeof(context->saHpiEventLogClear) );
        break;
    
        case COLUMN_SAHPIEVENTLOGSTATE:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogState,
                         sizeof(context->saHpiEventLogState) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiEventLogInfoTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiEventLogInfoTable_get_by_idx
 */
const saHpiEventLogInfoTable_context *
saHpiEventLogInfoTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiEventLogInfoTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


