/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiEventLogTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiEventLogTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceEventLogTable.h>
#include <saHpiSensorEventLogTable.h>
#include <saHpiHotSwapEventLogTable.h>
#include <saHpiWatchdogEventLogTable.h>
#include <saHpiSoftwareEventLogTable.h>
#include <saHpiOemEventLogTable.h>
#include <saHpiSensorEnableChangeEventLogTable.h>
#include <saHpiDomainEventLogTable.h>
#include <saHpiUserEventLogTable.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiEventLogTable_oid[] = { saHpiEventLogTable_TABLE_OID };
size_t saHpiEventLogTable_oid_len = OID_LENGTH(saHpiEventLogTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;		      
static GHashTable *dr_table;

SaErrorT populate_saHpiEventLog (SaHpiSessionIdT sessionid, SaHpiResourceIdT resourceid)
{

	SaErrorT rv;

        SaHpiEventLogEntryIdT      event_entry_id;
        SaHpiEventLogEntryIdT  pre_event_entry_id;
        SaHpiEventLogEntryT       event_log_entry;
       
	SaHpiRptEntryT  event_rpt_entry;
	SaHpiRdrT       event_rdr_entry;

	oid evt_log_oid[EVENT_LOG_INDEX_NR];
	netsnmp_index evt_log_index;
	saHpiEventLogTable_context *evt_log_context;

        DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

        oid child_oid[MAX_OID_LEN];
        size_t child_oid_len;
	
	int isNewRow = MIB_TRUE;

	printf( "populate_saHpiEventLog\n");
        printf(" ***************************************\n");
        printf(" ***************************************\n");
        printf(" EVENT LOG TABLES \n");
		
	event_log_clear(sessionid, resourceid, MIB_TRUE);
	
	printf("event_log_clear: done \n");

        subagent_lock(&hpi_lock_data);

	event_entry_id = SAHPI_OLDEST_ENTRY;
	
	if (resourceid != SAHPI_UNSPECIFIED_RESOURCE_ID) { 

       		
        	do {
                	rv = saHpiEventLogEntryGet (sessionid, 
                                            resourceid,
                                            event_entry_id,
                                            &pre_event_entry_id,
                                            &event_entry_id,
                                            &event_log_entry,
                                            &event_rdr_entry,
                                            &event_rpt_entry);

                	printf("        Entry ID [%d]\n", event_entry_id);
                	printf("        EventType [%d]\n", event_log_entry.Event.EventType);
                	printf("        rv [%d]\n", rv);


                	if (rv == SA_ERR_HPI_CAPABILITY) {
                     	        printf ("populate_saHpiEventLog, saHpiEventLogEntryGet() == SA_ERR_HPI_CAPABILITY\n");
                    	    	DEBUGMSGTL ((AGENT, "populate_saHpiEventLog, saHpiEventLogEntryGet() == SA_ERR_HPI_CAPABILITY\n"));
                  	    	rv = AGENT_ERR_NOERROR;
                           	break;
                	} else if (rv == SA_ERR_HPI_NOT_PRESENT) {
                      	  	printf ("populate_saHpiEventLog, saHpiEventLogEntryGet() == SA_ERR_HPI_NOT_PRESENT\n");
                      	 	DEBUGMSGTL ((AGENT, "populatesaHpiResourcesDiscover_saHpiEventLog, saHpiEventLogEntryGet() == SA_ERR_HPI_NOT_PRESENT\n"));
                       	 	rv = AGENT_ERR_NOERROR;
                        	break;
                	} else if (rv != SA_OK) {
                        	printf ("populate_saHpiEventLog Failed: rv = %d\n",rv);
                        	DEBUGMSGTL ((AGENT, "populate_saHpiEventLog Failed: rv = %d\n",rv));
                        	rv = AGENT_ERR_INTERNAL_ERROR;
                        	break;
                	}
							
                	switch (event_log_entry.Event.EventType) {
                	case SAHPI_ET_RESOURCE:
                        	printf("SAHPI_ET_RESOURCE: rv [%d]\n", rv);
                        	printf("        Event Type: [%s]\n", 
                               	oh_lookup_resourceeventtype(event_log_entry.Event.EventDataUnion.ResourceEvent.ResourceEventType));
                        	printf("        Resource: [%d]\n", event_log_entry.Event.Source);
                        	printf("        Severity: [%s]\n\n",oh_lookup_severity(event_log_entry.Event.Severity));
                        	populate_saHpiResourceEventLogTable(sessionid, &event_log_entry,                                           
                                                            child_oid, 
                                                            &child_oid_len);
                        	break;	
			case SAHPI_ET_DOMAIN:	
				printf("SAHPI_ET_DOMAIN: rv [%d]\n\n", rv);
				printf("        Event Type: [%s]\n", 
                               	oh_lookup_domaineventtype(event_log_entry.Event.EventDataUnion.DomainEvent.Type));
				printf("        Resource: [%d]\n", event_log_entry.Event.Source);
				printf("        Domain: [%d]\n\n",event_log_entry.Event.EventDataUnion.DomainEvent.DomainId);
        			populate_saHpiDomainEventLogTable(sessionid, 
	                                    &event_log_entry,					    
                                            child_oid, 
                                            &child_oid_len);
				break;							
                	case SAHPI_ET_SENSOR:
                        	printf("SAHPI_ET_SENSOR: rv [%d]\n", rv);
                        	printf("        Sensor Type: [%s]\n\n",
                                oh_lookup_sensortype(event_log_entry.Event.EventDataUnion.SensorEvent.SensorType));
                       		populate_saHpiSensorEventLogTable(sessionid, &event_log_entry,                                           
                                                          child_oid, 
                                                          &child_oid_len,
                                                          &event_rdr_entry);
                        	break;
                	case SAHPI_ET_SENSOR_ENABLE_CHANGE:
                        	printf("SAHPI_ET_SENSOR_ENABLE_CHANGE: rv [%d]\n\n", rv);
                        	populate_saHpiSensorEnableChangeEventLogTable(sessionid, &event_log_entry,                                           
                                                           child_oid, 
                                                           &child_oid_len);
                        	break;
                	case SAHPI_ET_HOTSWAP:
                        	printf("SAHPI_ET_HOTSWAP: rv [%d]\n\n", rv);
                        	populate_saHpiHotSwapEventLogTable(sessionid, &event_log_entry,  					 
                        				   child_oid, 
                        				   &child_oid_len);
							   

                        	break;
                	case SAHPI_ET_WATCHDOG:
                        	printf("SAHPI_ET_WATCHDOG: rv [%d]\n\n", rv);
                        	populate_saHpiWatchdogEventLogTable(sessionid, &event_log_entry, 
			                                  child_oid, 
                        				  &child_oid_len);
							 
                        	break;
                	case SAHPI_ET_HPI_SW:
                       		printf("SAHPI_ET_HPI_SW: rv [%d]\n\n", rv);
                        	populate_saHpiSoftwareEventLogTable(sessionid, &event_log_entry,                                           
                                                            child_oid, 
                                                            &child_oid_len);
                        	break;
                	case SAHPI_ET_OEM:
                       		printf("SAHPI_ET_OEM: rv [%d]\n\n", rv);
                        	populate_saHpiOemEventLogTable(sessionid, &event_log_entry,                                           
                                                        child_oid, 
                                                        &child_oid_len);
                        	break;
                	case SAHPI_ET_USER:
                        	printf("SAHPI_ET_USER: rv [%d]\n\n", rv);
                        	populate_saHpiUserEventLogTable(sessionid, &event_log_entry,                                           
                                                        child_oid, 
                                                        &child_oid_len);
                        	break;
                	default:
                        	printf("********* unknown event type *********\n");
                        	break;        
                	}
			
               		/* BUILD oid for new row */
                        /* assign the number of indices */
                	evt_log_index.len = EVENT_LOG_INDEX_NR;
                	/** Index saHpiDomainId is external */
                	evt_log_oid[0] = get_domain_id(sessionid);
                	/** Index saHpiResourceId is external */
                	evt_log_oid[1] = resourceid;
                	/** Index saHpiEventLogIndex is internal */
                	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
                	dr_pair.domainId_resourceId_arry[1] = resourceid;
                	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
                	if (dr_entry == NULL) {
                        	DEBUGMSGTL ((AGENT, 
                        	"ERROR: populate_saHpEventLogTable() domain_resource_pair_get returned NULL\n"));
                        	subagent_unlock(&hpi_lock_data);
				return AGENT_ERR_INTERNAL_ERROR;
                	}						
			evt_log_oid[2] = dr_entry->entry_id++;
                	/* assign the indices to the index */
                	evt_log_index.oids = (oid *) & evt_log_oid;
	
                	/* See if it exists. */
                	evt_log_context = NULL;
                	evt_log_context = CONTAINER_FIND (cb.container, &evt_log_index);
	
                	if (!evt_log_context) { 
                        // New entry. Add it
                        	evt_log_context = 
                                	saHpiEventLogTable_create_row ( &evt_log_index);
                	}
			else {
				isNewRow = MIB_FALSE;
			}
			
                	if (!evt_log_context) {
                        	snmp_log (LOG_ERR, "Not enough memory for a EventLog row!");
                        	rv = AGENT_ERR_INTERNAL_ERROR;
                        	break;
                	}

                	/** SaHpiEntryId = ASN_UNSIGNED */
                	evt_log_context->saHpiEventLogIndex = evt_log_oid[2];

                	/** INTEGER = ASN_INTEGER */
                	evt_log_context->saHpiEventLogType = 
                	        event_log_entry.Event.EventType + 1; 

                	/** SaHpiTime = ASN_OCTET_STR */
			hpitime_to_snmptime(event_log_entry.Timestamp, evt_log_context->saHpiEventLogAddedTimestamp);			                	
                        evt_log_context->saHpiEventLogAddedTimestamp_len = sizeof(SaHpiTimeT);
                	
			/** RowPointer = ASN_OBJECT_ID */
                	evt_log_context->saHpiEventLogRowPointer_len = 
                        	child_oid_len * sizeof (oid);
                	memcpy (evt_log_context->saHpiEventLogRowPointer, 
                        	child_oid, 
                        	evt_log_context->saHpiEventLogRowPointer_len);
		
			if (isNewRow == MIB_TRUE) {
                		CONTAINER_INSERT (cb.container, evt_log_context);
			}
			else { //reset it
				isNewRow = MIB_TRUE;
			}			

        	} while (event_entry_id != SAHPI_NO_MORE_ENTRIES);
	}	
	else {
        	
		do {	
		/******************************************************/
       	 	/* now get the Domain Event Log Info for this session */
       	 	/* this is accomplished by using                      */
        	/* in saHpiEventLogInfoGet()                          */
        	/* SAHPI_UNSPECIFIED_RESOURCE_ID                      */
        	/******************************************************/
        		rv = saHpiEventLogEntryGet (sessionid, 
                                    SAHPI_UNSPECIFIED_RESOURCE_ID,
                                    event_entry_id,
                                    &pre_event_entry_id,
                                    &event_entry_id,
                                    &event_log_entry,
                                    &event_rdr_entry,
                                    &event_rpt_entry);

                	printf("        Entry ID [%d]\n", event_entry_id);
                	printf("        EventType [%d]\n", event_log_entry.Event.EventType);
                	printf("        rv [%d]\n", rv);

        		if (rv == SA_ERR_HPI_NOT_PRESENT) {
                		DEBUGMSGTL ((AGENT, "Domain Event Log Empty\n"));
                		
				subagent_unlock(&hpi_lock_data);
				return AGENT_ERR_NOERROR;
        		}
        		if (rv != SA_OK) {
               		 	DEBUGMSGTL ((AGENT, "getting Domain Event Log Failed: rv = %d\n",rv));
                		
				subagent_unlock(&hpi_lock_data);
				return AGENT_ERR_INTERNAL_ERROR;
        		}
					    
                	switch (event_log_entry.Event.EventType) {
                	case SAHPI_ET_RESOURCE:
                        	printf("SAHPI_ET_RESOURCE: rv [%d]\n", rv);
                        	printf("        Event Type: [%s]\n", 
                               	oh_lookup_resourceeventtype(event_log_entry.Event.EventDataUnion.ResourceEvent.ResourceEventType));
                        	printf("        Resource: [%d]\n", event_log_entry.Event.Source);
                        	printf("        Severity: [%s]\n\n",oh_lookup_severity(event_log_entry.Event.Severity));
                        	populate_saHpiResourceEventLogTable(sessionid, &event_log_entry,                                           
                                                            child_oid, 
                                                            &child_oid_len);
                        	break;
       	 		case SAHPI_ET_DOMAIN:
				printf("SAHPI_ET_DOMAIN: rv [%d]\n\n", rv);
				printf("        Event Type: [%s]\n", 
                               	oh_lookup_domaineventtype(event_log_entry.Event.EventDataUnion.DomainEvent.Type));
				printf("        Resource: [%d]\n", event_log_entry.Event.Source);
				printf("        Domain: [%d]\n\n",event_log_entry.Event.EventDataUnion.DomainEvent.DomainId);
        			populate_saHpiDomainEventLogTable(sessionid, 
	                                    &event_log_entry,					    
                                            child_oid, 
                                            &child_oid_len);
				break;				
                	case SAHPI_ET_SENSOR:
                        	printf("SAHPI_ET_SENSOR: rv [%d]\n", rv);
                        	printf("        Sensor Type: [%s]\n\n",
                                oh_lookup_sensortype(event_log_entry.Event.EventDataUnion.SensorEvent.SensorType));
                       		populate_saHpiSensorEventLogTable(sessionid, &event_log_entry,                                           
                                                          child_oid, 
                                                          &child_oid_len,
                                                          &event_rdr_entry);
                        	break;
                	case SAHPI_ET_SENSOR_ENABLE_CHANGE:
                        	printf("SAHPI_ET_SENSOR_ENABLE_CHANGE: rv [%d]\n\n", rv);
                        	populate_saHpiSensorEnableChangeEventLogTable(sessionid, &event_log_entry,                                           
                                                           child_oid, 
                                                           &child_oid_len);
                        	break;
                	case SAHPI_ET_HOTSWAP:
                        	printf("SAHPI_ET_HOTSWAP: rv [%d]\n\n", rv);
                        	populate_saHpiHotSwapEventLogTable(sessionid, &event_log_entry,  					 
                        				   child_oid, 
                        				   &child_oid_len);
							   

                        	break;
                	case SAHPI_ET_WATCHDOG:
                        	printf("SAHPI_ET_WATCHDOG: rv [%d]\n\n", rv);
                        	populate_saHpiWatchdogEventLogTable(sessionid, &event_log_entry, 
			                                  child_oid, 
                        				  &child_oid_len);
							 
                        	break;
                	case SAHPI_ET_HPI_SW:
                       		printf("SAHPI_ET_HPI_SW: rv [%d]\n\n", rv);
                        	populate_saHpiSoftwareEventLogTable(sessionid, &event_log_entry,                                           
                                                            child_oid, 
                                                            &child_oid_len);
                        	break;
                	case SAHPI_ET_OEM:
                       		printf("SAHPI_ET_OEM: rv [%d]\n\n", rv);
                        	populate_saHpiOemEventLogTable(sessionid, &event_log_entry,                                           
                                                        child_oid, 
                                                        &child_oid_len);
                        	break;
                	case SAHPI_ET_USER:
                        	printf("SAHPI_ET_USER: rv [%d]\n\n", rv);
                        	populate_saHpiUserEventLogTable(sessionid, &event_log_entry,                                           
                                                        child_oid, 
                                                        &child_oid_len);
                        	break;
                	default:
                        	printf("********* unknown event type *********\n");
                        	break;        
                	}					    

        		/* BUILD oid for new row */
                	/* assign the number of indices */
	        	evt_log_index.len = EVENT_LOG_INDEX_NR;
	        	/** Index saHpiDomainId is external */
        		evt_log_oid[0] = get_domain_id(sessionid);
        		/** Index saHpiResourceId is external */
       	 		evt_log_oid[1] = SAHPI_UNSPECIFIED_RESOURCE_ID;
        		/** Index saHpiEventLogIndex is internal */
        		dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
        		dr_pair.domainId_resourceId_arry[1] = SAHPI_UNSPECIFIED_RESOURCE_ID;
        		dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
        		if (dr_entry == NULL) {
               	 		DEBUGMSGTL ((AGENT, 
                		"ERROR: populate_saHpEventLogTable() domain_resource_pair_get returned NULL\n"));
                		
				subagent_unlock(&hpi_lock_data);
				return AGENT_ERR_INTERNAL_ERROR;
        		}
        		evt_log_oid[2] = dr_entry->entry_id++;			
			
        		/* assign the indices to the index */
        		evt_log_index.oids = (oid *) & evt_log_oid;
	
        		/* See if it exists. */
	        	evt_log_context = NULL;
        		evt_log_context = CONTAINER_FIND (cb.container, &evt_log_index);
		
        		if (!evt_log_context) { 
	                // New entry. Add it
        	        	evt_log_context = 
                	        	saHpiEventLogTable_create_row ( &evt_log_index);
	        	}
			else {
				isNewRow = MIB_FALSE;
			}
				
        		if (!evt_log_context) {
                		snmp_log (LOG_ERR, "Not enough memory for a EventLog row!");
	                	rv = AGENT_ERR_INTERNAL_ERROR;
        		}
	
        		/** SaHpiEntryId = ASN_UNSIGNED */
        		evt_log_context->saHpiEventLogIndex = evt_log_oid[2];

	        	/** INTEGER = ASN_INTEGER */
        		evt_log_context->saHpiEventLogType = 
                		event_log_entry.Event.EventType + 1; 

	        	/** SaHpiTime = ASN_OCTET_STR */
			hpitime_to_snmptime(event_log_entry.Timestamp, evt_log_context->saHpiEventLogAddedTimestamp);
                        evt_log_context->saHpiEventLogAddedTimestamp_len = sizeof(SaHpiTimeT);
			
	        	/** RowPointer = ASN_OBJECT_ID */
        		evt_log_context->saHpiEventLogRowPointer_len = 
                		child_oid_len * sizeof (oid);
	        	memcpy (evt_log_context->saHpiEventLogRowPointer, 
        	        	child_oid, 
                		evt_log_context->saHpiEventLogRowPointer_len);
		
			if (isNewRow == MIB_TRUE) {
        			CONTAINER_INSERT (cb.container, evt_log_context);
			}
			else { //reset
				isNewRow = MIB_TRUE;	
			}	
		} while (event_entry_id != SAHPI_NO_MORE_ENTRIES);
	} 
	
	subagent_unlock(&hpi_lock_data);
	return rv;

}

/**
 * When a User Event Log Entry is successfully added in saHpiUserEventLogTable
 * a corresponding entry must be craeted here and the udpate count maintained
 * 
 * @session_id
 * @resource_id
 * @event
 * @child_oid
 * @child_oid_len
 * 
 * @return 
 */
int event_log_add(SaHpiSessionIdT session_id, 
                  SaHpiResourceIdT resource_id, 
                  SaHpiEventT *event,
                  oid *child_oid, 
                  size_t child_oid_len) 
{
        oid evt_log_oid[EVENT_LOG_INDEX_NR];
        netsnmp_index evt_log_index;
        saHpiEventLogTable_context *evt_log_context;

        DR_XREF *dr_entry;
        SaHpiDomainIdResourceIdArrayT dr_pair;

        /* BUILD oid for new row */
                /* assign the number of indices */
        evt_log_index.len = EVENT_LOG_INDEX_NR;
        /** Index saHpiDomainId is external */
        evt_log_oid[0] = get_domain_id(session_id);
        /** Index saHpiResourceId is external */
        evt_log_oid[1] = resource_id;
        /** Index saHpiEventLogIndex is internal */
        dr_pair.domainId_resourceId_arry[0] = get_domain_id(session_id);
        dr_pair.domainId_resourceId_arry[1] = resource_id;
        dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
        if (dr_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                "ERROR: populate_saHpEventLogTable() domain_resource_pair_get returned NULL\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        evt_log_oid[2] = dr_entry->entry_id++;
        /* assign the indices to the index */
        evt_log_index.oids = (oid *) & evt_log_oid;

        /* See if it exists. */
        evt_log_context = NULL;
        evt_log_context = CONTAINER_FIND (cb.container, &evt_log_index);

        if (!evt_log_context) { 
                // New entry. Add it
                evt_log_context = 
                        saHpiEventLogTable_create_row ( &evt_log_index);
        }
        if (!evt_log_context) {
                snmp_log (LOG_ERR, "Not enough memory for a EventLog row!");
                return  AGENT_ERR_INTERNAL_ERROR;
        }

        /** SaHpiEntryId = ASN_UNSIGNED */
        evt_log_context->saHpiEventLogIndex = evt_log_oid[2];

        /** INTEGER = ASN_INTEGER */
        evt_log_context->saHpiEventLogType = 
                event->EventType + 1; 

        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event->Timestamp, evt_log_context->saHpiEventLogAddedTimestamp);
	evt_log_context->saHpiEventLogAddedTimestamp_len = sizeof(SaHpiTimeT);

        /** RowPointer = ASN_OBJECT_ID */
        evt_log_context->saHpiEventLogRowPointer_len = 
                child_oid_len * sizeof (oid);
        memcpy (evt_log_context->saHpiEventLogRowPointer, 
                child_oid, 
                evt_log_context->saHpiEventLogRowPointer_len);

        CONTAINER_INSERT (cb.container, evt_log_context);

        return SNMP_ERR_NOERROR;
}


/**
 * 
 * @session_id
 * @resource_id
 * 
 * @return 
 */
SaErrorT event_log_clear (SaHpiSessionIdT session_id, SaHpiResourceIdT resource_id, int modifyTotal)
{
        netsnmp_index *row_idx;
        saHpiEventLogTable_context *event_log_ctx;
	
	DR_XREF *dr_entry;
        SaHpiDomainIdResourceIdArrayT dr_pair;

        DEBUGMSGTL ((AGENT, "event_log_clear, called\n"));

        dr_pair.domainId_resourceId_arry[0] = get_domain_id(session_id);
        dr_pair.domainId_resourceId_arry[1] = resource_id;
        dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
        if (dr_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                "ERROR: event_log_clear() domain_resource_pair_get returned NULL\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        
	dr_entry->entry_id = 0;

        row_idx = CONTAINER_FIRST(cb.container);                  
        if (row_idx) { //At least one entry was found.
               do {
                       event_log_ctx = CONTAINER_FIND(cb.container, row_idx);

                       if ((event_log_ctx->index.oids[saHpiEventLogDomainId_INDEX] ==
                            get_domain_id(session_id)) &&
                           (event_log_ctx->index.oids[saHpiEventLogResourceId_INDEX] ==
                            resource_id)) {

                               DEBUGMSGTL ((AGENT, "event_log_clear: deleting \n"));

                               /* all conditions met remove row */

                               switch (event_log_ctx->saHpiEventLogType - 1) {
                               case SAHPI_ET_RESOURCE:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_RESOURCE, "
                                                    "called\n"));
                                       resource_event_log_clear(session_id, 
                                                                resource_id,  
                                                                event_log_ctx->saHpiEventLogRowPointer,
                                                                event_log_ctx->saHpiEventLogRowPointer_len,
								modifyTotal);
                                       break;
                               case SAHPI_ET_DOMAIN:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_DOMAIN, "
                                                    "called\n"));
                                       domain_event_log_clear(session_id, 
                                                              resource_id,  
                                                              event_log_ctx->saHpiEventLogRowPointer,
                                                              event_log_ctx->saHpiEventLogRowPointer_len,
							      modifyTotal);
                                       break;
                               case SAHPI_ET_SENSOR:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_SENSOR, "
                                                    "called\n"));
                                       sensor_event_log_clear(session_id, 
                                                              resource_id,  
                                                              event_log_ctx->saHpiEventLogRowPointer,
                                                              event_log_ctx->saHpiEventLogRowPointer_len,
							      modifyTotal);
                                       break;
                               case SAHPI_ET_SENSOR_ENABLE_CHANGE:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_SENSOR_ENABLE_CHANGE, "
                                                    "called\n"));
                                       sen_en_change_event_log_clear(session_id, 
                                                                    resource_id,  
                                                                    event_log_ctx->saHpiEventLogRowPointer,
                                                                    event_log_ctx->saHpiEventLogRowPointer_len,
								    modifyTotal);
                                       break;
                               case SAHPI_ET_HOTSWAP:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_HOTSWAP, "
                                                    "called\n"));
                                       hotswap_event_log_clear(session_id, 
                                                               resource_id,  
                                                               event_log_ctx->saHpiEventLogRowPointer,
                                                               event_log_ctx->saHpiEventLogRowPointer_len,
							       modifyTotal);
                                       break;
                               case SAHPI_ET_WATCHDOG:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_WATCHDOG, "
                                                    "called\n"));
                                       watchdog_event_log_clear(session_id, 
                                                                resource_id,  
                                                                event_log_ctx->saHpiEventLogRowPointer,
                                                                event_log_ctx->saHpiEventLogRowPointer_len,
								modifyTotal);
                                       break;
                               case SAHPI_ET_HPI_SW:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_HPI_SW, "
                                                    "called\n"));
                                       software_event_log_clear(session_id, 
                                                                resource_id,  
                                                                event_log_ctx->saHpiEventLogRowPointer,
                                                                event_log_ctx->saHpiEventLogRowPointer_len,
								modifyTotal);
                                       break;
                               case SAHPI_ET_OEM:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_OEM, "
                                                    "called\n"));
                                       oem_event_log_clear(session_id, 
                                                           resource_id,  
                                                           event_log_ctx->saHpiEventLogRowPointer,
                                                           event_log_ctx->saHpiEventLogRowPointer_len,
							   modifyTotal);
                                       break;
                               case SAHPI_ET_USER:
                                       DEBUGMSGTL ((AGENT, "SAHPI_ET_USER, "
                                                    "called\n"));
                                       user_event_log_clear(session_id, 
                                                            resource_id,  
                                                            event_log_ctx->saHpiEventLogRowPointer,
                                                            event_log_ctx->saHpiEventLogRowPointer_len,
							    modifyTotal);
                                       break;                               
                               default:
                                       printf("********* unknown event type *********\n");
                                       break;        
                               }
			       
			       row_idx = CONTAINER_NEXT(cb.container, row_idx);
			       
			       CONTAINER_REMOVE (cb.container, event_log_ctx);
                               saHpiEventLogTable_delete_row (event_log_ctx);
                               DEBUGMSGTL ((AGENT, "event_log_clear: found row for "
                                                   "clearing/deletion\n"));	   
                       }
		       else {
		        	row_idx = CONTAINER_NEXT(cb.container, row_idx);	
		       }		
                       
               } while (row_idx);
       }

        return SA_OK;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiEventLogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiEventLogTable_cmp( const void *lhs, const void *rhs )
{
        saHpiEventLogTable_context *context_l =
        (saHpiEventLogTable_context *)lhs;
        saHpiEventLogTable_context *context_r =
        (saHpiEventLogTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiEventLogTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiEventLogTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiEventLogIndex */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2])
			        return 0;
		}
	}
	
	return 0;
}

/************************************************************
 * Initializes the saHpiEventLogTable module
 */
void
init_saHpiEventLogTable(void)
{
    initialize_table_saHpiEventLogTable();

    domain_resource_pair_initialize(&initialized, &dr_table);

    /*
     * TODO: perform any startup stuff here, such as
     * populating the table with initial data.
     *
     * saHpiEventLogTable_context * new_row = create_row(index);
     * CONTAINER_INSERT(cb.container,new_row);
     */
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiEventLogTable_row_copy(saHpiEventLogTable_context * dst,
                         saHpiEventLogTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    
    subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiEventLogIndex = src->saHpiEventLogIndex;

    dst->saHpiEventLogType = src->saHpiEventLogType;

    memcpy( dst->saHpiEventLogAddedTimestamp,
            src->saHpiEventLogAddedTimestamp,
	    src->saHpiEventLogAddedTimestamp_len);
	    
    dst->saHpiEventLogAddedTimestamp_len = src->saHpiEventLogAddedTimestamp_len;	    

    memcpy( dst->saHpiEventLogRowPointer, 
            src->saHpiEventLogRowPointer,  
            src->saHpiEventLogRowPointer_len );
    dst->saHpiEventLogRowPointer_len = src->saHpiEventLogRowPointer_len;

    subagent_unlock(&hpi_lock_data);
    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiEventLogTable_extract_index( saHpiEventLogTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiEventLogIndex;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiEventLogTable_extract_index, called\n"));

        
	subagent_lock(&hpi_lock_data);
	/*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
			return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
       memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
       var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiDomainId.next_variable = &var_saHpiResourceId;


       memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
       var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiResourceId.next_variable = &var_saHpiEventLogIndex;

       memset( &var_saHpiEventLogIndex, 0x00, sizeof(var_saHpiEventLogIndex) );
       var_saHpiEventLogIndex.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiEventLogIndex.next_variable = NULL;

       /* copy indexes into newly created context */
       memcpy(ctx->saHpiEventLogRowPointer, hdr->oids, hdr->len * sizeof(oid));
       ctx->saHpiEventLogRowPointer_len = hdr->len;
       err = SNMP_ERR_NOERROR;

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        
	subagent_unlock(&hpi_lock_data);
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiEventLogTable_can_activate(saHpiEventLogTable_context *undo_ctx,
                      saHpiEventLogTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiEventLogTable_can_deactivate(saHpiEventLogTable_context *undo_ctx,
                        saHpiEventLogTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiEventLogTable_can_delete(saHpiEventLogTable_context *undo_ctx,
                    saHpiEventLogTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiEventLogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiEventLogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiEventLogTable_context *
saHpiEventLogTable_create_row( netsnmp_index* hdr)
{
    
    subagent_lock(&hpi_lock_data);
    
    saHpiEventLogTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiEventLogTable_context);
    if(!ctx) { 
        subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiEventLogTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
    */

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiEventLogTable_context *
saHpiEventLogTable_duplicate_row( saHpiEventLogTable_context * row_ctx)
{
    saHpiEventLogTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
    
    dup = SNMP_MALLOC_TYPEDEF(saHpiEventLogTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    if(saHpiEventLogTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    
    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiEventLogTable_delete_row( saHpiEventLogTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

    
    subagent_lock(&hpi_lock_data);
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiEventLogTable_set_reserve1( netsnmp_request_group *rg )
{
//    saHpiEventLogTable_context *row_ctx =
//            (saHpiEventLogTable_context *)rg->existing_row;
//    saHpiEventLogTable_context *undo_ctx =
//            (saHpiEventLogTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiEventLogTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiEventLogTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiEventLogTable_context *row_ctx = (saHpiEventLogTable_context *)rg->existing_row;
//    saHpiEventLogTable_context *undo_ctx = (saHpiEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiEventLogTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogTable_context *row_ctx = (saHpiEventLogTable_context *)rg->existing_row;
//    saHpiEventLogTable_context *undo_ctx = (saHpiEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }
    subagent_unlock(&hpi_lock_data);
    return;

}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiEventLogTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogTable_context *row_ctx = (saHpiEventLogTable_context *)rg->existing_row;
//    saHpiEventLogTable_context *undo_ctx = (saHpiEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiEventLogTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogTable_context *row_ctx = (saHpiEventLogTable_context *)rg->existing_row;
//    saHpiEventLogTable_context *undo_ctx = (saHpiEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: 
                break;
                /** We shouldn't get here */
                /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiEventLogTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiEventLogTable_context *row_ctx = (saHpiEventLogTable_context *)rg->existing_row;
//    saHpiEventLogTable_context *undo_ctx = (saHpiEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 *
 * Initialize the saHpiEventLogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiEventLogTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiEventLogTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiEventLogTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiEventLogTable_oid,
                                             saHpiEventLogTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiEventLogTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiEventLogIndex */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiEventLogTable_COL_MIN;
    table_info->max_column = saHpiEventLogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiEventLogTable_get_value;
    cb.container = netsnmp_container_find("saHpiEventLogTable_primary:"
                                          "saHpiEventLogTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiEventLogTable_secondary:"
                                                       "saHpiEventLogTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiEventLogTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiEventLogTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiEventLogTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiEventLogTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiEventLogTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiEventLogTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiEventLogTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiEventLogTable_can_delete;

    cb.set_reserve1 = saHpiEventLogTable_set_reserve1;
    cb.set_reserve2 = saHpiEventLogTable_set_reserve2;
    cb.set_action = saHpiEventLogTable_set_action;
    cb.set_commit = saHpiEventLogTable_set_commit;
    cb.set_free = saHpiEventLogTable_set_free;
    cb.set_undo = saHpiEventLogTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiEventLogTable",
                "Registering table saHpiEventLogTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiEventLogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiEventLogTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiEventLogTable_context *context = (saHpiEventLogTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIEVENTLOGINDEX:
            /** SaHpiEntryId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiEventLogIndex,
                         sizeof(context->saHpiEventLogIndex) );
        break;
    
        case COLUMN_SAHPIEVENTLOGTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiEventLogType,
                         sizeof(context->saHpiEventLogType) );
        break;
    
        case COLUMN_SAHPIEVENTLOGADDEDTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiEventLogAddedTimestamp,
                         sizeof(context->saHpiEventLogAddedTimestamp) );
        break;
    
        case COLUMN_SAHPIEVENTLOGROWPOINTER:
            /** RowPointer = ASN_OBJECT_ID */
            snmp_set_var_typed_value(var, ASN_OBJECT_ID,
                         (u_char*)&context->saHpiEventLogRowPointer,
                         context->saHpiEventLogRowPointer_len );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiEventLogTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiEventLogTable_get_by_idx
 */
const saHpiEventLogTable_context *
saHpiEventLogTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiEventLogTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


