/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiFieldTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiFieldTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <saHpiResourceTable.h>
#include <saHpiAreaTable.h>
#include <session_info.h>
#include <hash_utils.h>
#include <oh_utils.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiFieldTable_oid[] = { saHpiFieldTable_TABLE_OID };
size_t saHpiFieldTable_oid_len = OID_LENGTH(saHpiFieldTable_oid);


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;               
static GHashTable *dria_table;

/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long field_entry_count = 0;
static oid saHpiFieldActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,4,8,5 };
int handle_saHpiFieldActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests);
int initialize_table_saHpiFieldActiveEntries(void);
int set_table_field_update (saHpiFieldTable_context *row_ctx);
int set_table_field_add (saHpiFieldTable_context *row_ctx);
int set_table_field_delete (saHpiFieldTable_context *row_ctx);


/**
 * 
 * @session_id:
 * @rdr_entry:
 * @rpt_entry:
 * @area_context:
 * 
 * @return: 
 */
SaErrorT populate_field (SaHpiSessionIdT session_id, 
                         SaHpiRdrT *rdr_entry,
                         SaHpiRptEntryT *rpt_entry,
                         saHpiAreaTable_context *area_context)
{

        DEBUGMSGTL ((AGENT, "populate_field, called\n"));

        SaErrorT rv = SA_OK;
        SaHpiEntryIdT field_id;
        SaHpiIdrFieldT field_entry;

        DRIA_XREF *dria_entry;
        SaHpiDomainIdResourceIdInventoryIdAreaIdArrayT dria_tuple;
        
        oid field_oid[FIELD_INDEX_NR];
        netsnmp_index field_index;
        saHpiFieldTable_context *field_context;

        SaHpiEntryIdT subagent_field_id;

        DEBUGMSGTL ((AGENT, "populate_field() called\n"));

        /* check for NULL pointers */
        if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_field() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_field() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

	subagent_lock(&hpi_lock_data);

        field_id = SAHPI_FIRST_ENTRY;
        do {

                rv = saHpiIdrFieldGet(session_id,
                                      rpt_entry->ResourceId,
                                      rdr_entry->RdrTypeUnion.InventoryRec.IdrId,
                                      area_context->saHpiAreaIdIndex,
                                      SAHPI_IDR_FIELDTYPE_UNSPECIFIED,
                                      field_id,
                                      &field_id,
                                      &field_entry);

                if (rv != SA_OK) {
                        DEBUGMSGTL ((AGENT, "saHpiIdrFieldGet Failed: rv = %d\n",rv));
                        rv =  AGENT_ERR_INTERNAL_ERROR;
                        break;
                }

                /* Create MIB AreaId value                              */
                /* this is not the same as the AreaId obtained from HPI */
                /* it is maintained in the subagent                     */
                dria_tuple.domainId_resourceId_idr_area_arry[0] = get_domain_id(session_id);
                dria_tuple.domainId_resourceId_idr_area_arry[1] = rpt_entry->ResourceId;
                dria_tuple.domainId_resourceId_idr_area_arry[2] = 
                        rdr_entry->RdrTypeUnion.InventoryRec.IdrId;
                dria_tuple.domainId_resourceId_idr_area_arry[3] = area_context->saHpiAreaId;

                /* domain_resource_idr_get() generates unique keys based on */
                /* domainid, resourceid, and IdrId tuples.                  */
                dria_entry = domain_resource_idr_area_get(&dria_tuple, &dria_table); 
                if (dria_entry == NULL) {
                        DEBUGMSGTL ((AGENT, 
                                     "ERROR: populate_area() domain_resource_idr_get returned NULL\n"));
			subagent_unlock(&hpi_lock_data);
                        return AGENT_ERR_INTERNAL_ERROR;
                }
                subagent_field_id = dria_entry->entry_id++;
                dria_entry->sa_hpi_area_id_index = area_context->saHpiAreaIdIndex;

                /* BUILD oid for new row */
                /* assign the number of indices */
                field_index.len = FIELD_INDEX_NR;
                /** Index saHpiDomainId is external */
                field_oid[0] = get_domain_id(session_id);
                /** Index saHpiResourceId is external */
                field_oid[1] = rpt_entry->ResourceId;
                /** Index saHpiResourceIsHistorical is external */
                field_oid[2] = MIB_FALSE;
                /** Index saHpiInventoryId */
                field_oid[3] = rdr_entry->RdrTypeUnion.InventoryRec.IdrId;
                /** Index saHpiAreaId */
                field_oid[4] = area_context->saHpiAreaId;
                /** Index saHpiAreaId */
                field_oid[5] = subagent_field_id;
                /* assign the indices to the index */
                field_index.oids = (oid *) & field_oid;

                /* See if Row exists. */
                field_context = NULL;
                field_context = CONTAINER_FIND(cb.container, &field_index);

                if (!field_context) {
                        // New entry. Add it
                        field_context = 
                        saHpiFieldTable_create_row(&field_index);
                }
                if (!field_context) {
                        snmp_log (LOG_ERR, "Not enough memory for a Area row!");
			subagent_unlock(&hpi_lock_data);
                        return AGENT_ERR_INTERNAL_ERROR;
                }

        /** SaHpiInstrumentId = ASN_UNSIGNED */
                field_context->saHpiFieldId = subagent_field_id;

        /** SaHpiInstrumentId = ASN_UNSIGNED */
                field_context->saHpiFieldIdIndex = field_entry.FieldId;

        /** INTEGER = ASN_INTEGER */
                field_context->saHpiFieldType = field_entry.Type + 1;
                field_context->field_type_set = MIB_TRUE;

        /** TruthValue = ASN_INTEGER */
                field_context->saHpiFieldIsReadOnly = 
                        (field_entry.ReadOnly == SAHPI_TRUE) ? 
                                MIB_TRUE : MIB_FALSE;

        /** SaHpiTextType = ASN_INTEGER */
                field_context->saHpiFieldTextType = 
                        field_entry.Field.DataType + 1;
            field_context->text_type_set = MIB_TRUE;

        /** SaHpiTextLanguage = ASN_INTEGER */
            field_context->saHpiFieldTextLanguage = 
                    field_entry.Field.Language + 1;
            field_context->text_language_set = MIB_TRUE;

        /** SaHpiText = ASN_OCTET_STR */
            memset(field_context->saHpiFieldText, 0, 
                   sizeof(field_context->saHpiFieldText));
            memcpy(field_context->saHpiFieldText, 
                   field_entry.Field.Data, 
                   field_entry.Field.DataLength); 
            field_context->saHpiFieldText_len = field_entry.Field.DataLength;
            field_context->text_set = MIB_TRUE;

        /** RowStatus = ASN_INTEGER */
            field_context->saHpiFieldStatus = SNMP_ROW_ACTIVE;

            CONTAINER_INSERT (cb.container, field_context);

        } while (field_id !=  SAHPI_LAST_ENTRY );

        field_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);

        return rv;
}


SaErrorT clear_field(SaHpiDomainIdT   domainId, 
                     SaHpiResourceIdT resourceId,
                     SaHpiIdrIdT      idrId,
                     SaHpiEntryIdT    areaId)
{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiFieldTable_context *ctx;

        DRIA_XREF *dria_entry;
        SaHpiDomainIdResourceIdInventoryIdAreaIdArrayT dria_tuple;

	DEBUGMSGTL ((AGENT, "clear_field, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domainId));	
	DEBUGMSGTL ((AGENT, "           resourceId [%d]\n", resourceId));
	DEBUGMSGTL ((AGENT, "           idrId      [%d]\n", idrId));
	DEBUGMSGTL ((AGENT, "           areaId     [%d]\n", areaId));

        /***************************/
        /* build the key to lookup */
        /***************************/
        dria_tuple.domainId_resourceId_idr_area_arry[0] = domainId;
        dria_tuple.domainId_resourceId_idr_area_arry[1] = resourceId;
        dria_tuple.domainId_resourceId_idr_area_arry[2] = idrId;
        dria_tuple.domainId_resourceId_idr_area_arry[3] = areaId;

        /* domain_resource_idr_get() generates unique keys based on */
        /* domainid, resourceid, and IdrId tuples.                  */
        dria_entry = domain_resoruce_idr_area_lookup(&dria_tuple, &dria_table);

        if (dria_entry == NULL) {
                DEBUGMSGTL ((AGENT, 
                             "INFO: clear_field() "
                             "domain_resoruce_idr_area_lookup  returned "
                             "NULL\n"));
                rv = SA_ERR_HPI_NOT_PRESENT;
        } else {
                dria_entry->entry_id = 0;
        }

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        /* based on the found row_idx get the pointer   */
                        /* to its context (row data)                    */
                        ctx = CONTAINER_FIND(cb.container, row_idx);

                        /* before we delete the context we should get the  */
                        /* next row (context) if any before we delete this */ 
                        /* one.                                            */
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if ((ctx->index.oids[saHpiDomainId_INDEX] ==
                             domainId) &&

                            (ctx->index.oids[saHpiResourceEntryId_INDEX] ==
                             resourceId)) {                     
                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctx);
                                saHpiFieldTable_delete_row (ctx);
                                DEBUGMSGTL ((AGENT, "clear_field: "
                                                    "found row: removing\n"));
                        }

                } while (row_idx);
        } 

        return rv;
}

/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int set_table_field_update (saHpiFieldTable_context *row_ctx)
{
        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiIdrIdT         idr_id;    
        SaHpiIdrFieldT      field_entry;

        DRIA_XREF *dria_entry;
        SaHpiDomainIdResourceIdInventoryIdAreaIdArrayT dria_tuple;


        DEBUGMSGTL ((AGENT, "set_table_field_update() called\n"));

        if ((row_ctx->field_type_set == MIB_TRUE) &&
            (row_ctx->text_type_set == MIB_TRUE) &&
            (row_ctx->text_language_set == MIB_TRUE) &&
            (row_ctx->text_set == MIB_TRUE)) {

                if (!row_ctx)
                        return AGENT_ERR_NULL_DATA;

                session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_field_INDEX]);
                resource_id = row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                idr_id = row_ctx->index.oids[saHpiInventoryId_field_INDEX];
#if 0
printf("*** set_table_field_update *******************\n");
printf("**********************************************\n");
printf("session_id [%d]\n", session_id);
printf("resource_id [%d]\n", resource_id);
printf("idr_id [%d]\n", idr_id);
printf("**********************************************\n");
#endif
                /* Get AreaIdIndex                                      */
                /* this is the same as the AreaId obtained from HPI     */
                /* it is maintained in the subagent                     */
                dria_tuple.domainId_resourceId_idr_area_arry[0] = 
                        row_ctx->index.oids[saHpiDomainId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[1] = 
                        row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[2] = 
                        row_ctx->index.oids[saHpiInventoryId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[3] = 
                        row_ctx->index.oids[saHpiAreaId_field_INDEX];

                dria_entry = domain_resoruce_idr_area_lookup(&dria_tuple, &dria_table);
                if (!dria_entry) {
                        DEBUGMSGTL ((AGENT, "domain_resoruce_idr_area_lookup() FIALED!!!!!!!\n"));
                        return SNMP_ERR_NOERROR; 
                } 
#if 0
printf("**********************************************\n");
printf("row_ctx->index.oids[saHpiDomainId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiDomainId_field_INDEX]);
printf("row_ctx->index.oids[saHpiResourceEntryId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiResourceEntryId_field_INDEX]);
printf("row_ctx->index.oids[saHpiInventoryId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiInventoryId_field_INDEX]);
printf("row_ctx->index.oids[saHpiAreaId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiAreaId_field_INDEX]);
printf("dria_entry->sa_hpi_area_id_index [%d]\n", 
       (int)dria_entry->sa_hpi_area_id_index);
printf("**********************************************\n");
#endif
                field_entry.AreaId = dria_entry->sa_hpi_area_id_index;

                memcpy(field_entry.Field.Data, 
                       row_ctx->saHpiFieldText, 
                       row_ctx->saHpiFieldText_len);
                field_entry.Field.DataLength = row_ctx->saHpiFieldText_len;
                field_entry.Field.DataType = row_ctx->saHpiFieldTextType - 1;
                field_entry.Field.Language = row_ctx->saHpiFieldTextLanguage - 1;
                /* field_entry.FieldId   not valid going in */
                /* field_entry.ReadOnly  not valid going in */
                field_entry.Type = row_ctx->saHpiFieldType - 1;

                rc = saHpiIdrFieldSet (session_id, resource_id,
                                      idr_id, &field_entry);

                if (rc == SA_OK) {
                        DEBUGMSGTL ((AGENT, "saHpiIdrFieldSet() SUCCEEDED!!!!!!!\n"));
                        return SNMP_ERR_NOERROR; 
                } else if (rc != SA_OK) {
                        snmp_log (LOG_ERR,
                                  "set_table_field_update: Call to saHpiIdrFieldSet() failed to set Mode rc: %s.\n",
                                  oh_lookup_error(rc));
                        DEBUGMSGTL ((AGENT,
                                     "set_table_field_update: Call to saHpiIdrFieldSet() failed to set Mode rc: %s.\n",
                                     oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }
        }
        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int set_table_field_add (saHpiFieldTable_context *row_ctx)
{
        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiIdrIdT         idr_id;    
        SaHpiIdrInfoT       idr_info; 
        SaHpiIdrFieldT      field_entry;

        DRIA_XREF *dria_entry;
        SaHpiDomainIdResourceIdInventoryIdAreaIdArrayT dria_tuple;

        DEBUGMSGTL ((AGENT, "set_table_field_add() called\n"));

        if ((row_ctx->field_type_set == MIB_TRUE) &&
            (row_ctx->text_type_set == MIB_TRUE) &&
            (row_ctx->text_language_set == MIB_TRUE) &&
            (row_ctx->text_set == MIB_TRUE)) {


                if (!row_ctx)
                        return AGENT_ERR_NULL_DATA;


                session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_field_INDEX]);
                resource_id = row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                idr_id = row_ctx->index.oids[saHpiInventoryId_field_INDEX];
#if 0
printf("*** set_table_field_add *******************\n");
printf("**********************************************\n");
printf("session_id [%d]\n", session_id);
printf("resource_id [%d]\n", resource_id);
printf("idr_id [%d]\n", idr_id);
printf("**********************************************\n");
#endif
                /* Get AreaIdIndex                                      */
                /* this is the same as the AreaId obtained from HPI     */
                /* it is maintained in the subagent                     */
                dria_tuple.domainId_resourceId_idr_area_arry[0] = 
                        row_ctx->index.oids[saHpiDomainId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[1] = 
                        row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[2] = 
                        row_ctx->index.oids[saHpiInventoryId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[3] = 
                        row_ctx->index.oids[saHpiAreaId_field_INDEX];

                dria_entry = domain_resoruce_idr_area_lookup(&dria_tuple, &dria_table);
                if (!dria_entry) {
                        DEBUGMSGTL ((AGENT, "domain_resoruce_idr_area_lookup() FIALED!!!!!!!\n"));
                        return SNMP_ERR_NOERROR; 
                } 
#if 0
printf("**********************************************\n");
printf("row_ctx->index.oids[saHpiDomainId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiDomainId_field_INDEX]);
printf("row_ctx->index.oids[saHpiResourceEntryId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiResourceEntryId_field_INDEX]);
printf("row_ctx->index.oids[saHpiInventoryId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiInventoryId_field_INDEX]);
printf("row_ctx->index.oids[saHpiAreaId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiAreaId_field_INDEX]);
printf("dria_entry->sa_hpi_area_id_index [%d]\n", 
       (int)dria_entry->sa_hpi_area_id_index);
printf("**********************************************\n");
#endif
                field_entry.AreaId = dria_entry->sa_hpi_area_id_index;

                memcpy(field_entry.Field.Data, 
                       row_ctx->saHpiFieldText, 
                       row_ctx->saHpiFieldText_len);
                field_entry.Field.DataLength = row_ctx->saHpiFieldText_len;
                field_entry.Field.DataType = row_ctx->saHpiFieldTextType - 1;
                field_entry.Field.Language = row_ctx->saHpiFieldTextLanguage - 1;
                /* field_entry.FieldId   not valid going in */
                /* field_entry.ReadOnly  not valid going in */
                field_entry.Type = row_ctx->saHpiFieldType - 1;

                rc = saHpiIdrFieldAdd(session_id, resource_id, 
                                      idr_id, &field_entry);
                if (rc != SA_OK) {
                        snmp_log (LOG_ERR,
                                  "set_table_field_add: Call to saHpiIdrFieldAdd() failed to set Mode rc: %s.\n",
                                  oh_lookup_error(rc));
                        DEBUGMSGTL ((AGENT,
                                     "set_table_field_add: Call to saHpiIdrFieldAdd() failed to set Mode rc: %s.\n",
                                     oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }

                rc = saHpiIdrInfoGet(session_id, resource_id, idr_id, &idr_info);
                if (rc != SA_OK) {
                                snmp_log (LOG_ERR,
                                "set_table_field_add: Call to saHpiIdrInfoGet() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc));
                                DEBUGMSGTL ((AGENT,
                                "set_table_field_add: Call to saHpiIdrInfoGet() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc)));
                                return get_snmp_error(rc);
                }

                SaHpiEntryIdT area_area_id;
                SaHpiIdrAreaHeaderT header;
                area_area_id = dria_entry->sa_hpi_area_id_index;
                rc = saHpiIdrAreaHeaderGet(session_id, resource_id, idr_id, 
                                           SAHPI_IDR_AREATYPE_UNSPECIFIED,
                                           area_area_id, &area_area_id, 
                                           &header);
                if (rc != SA_OK) {
                                snmp_log (LOG_ERR,
                                "set_table_field_add: Call to saHpiIdrAreaHeaderGet() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc));
                                DEBUGMSGTL ((AGENT,
                                "set_table_field_add: Call to saHpiIdrAreaHeaderGet() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc)));
                                return get_snmp_error(rc);
                }

                rc = update_num_data_fields (row_ctx->index.oids, header.NumFields);
                if (rc != SA_OK) {
                                snmp_log (LOG_ERR,
                                "set_table_field_add: Call to update_num_data_fields() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc));
                                DEBUGMSGTL ((AGENT,
                                "set_table_field_add: Call to update_num_data_fields() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc)));
                                return get_snmp_error(rc);
                }

                row_ctx->saHpiFieldIdIndex = field_entry.FieldId;  /* from saHpiIdrFieldAdd() */
                row_ctx->saHpiFieldIsReadOnly =                    /* from saHpiIdrInfoGet()  */
                        (idr_info.ReadOnly == SAHPI_TRUE) ? 
                                MIB_TRUE : MIB_FALSE;
                row_ctx->saHpiFieldStatus = SNMP_ROW_ACTIVE;      
                DEBUGMSGTL ((AGENT, "saHpiIdrFieldAdd() + saHpiIdrInfoGet() + saHpiIdrAreaHeaderGet() SUCCEEDED!!!!!!!\n"));
        } 

        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int set_table_field_delete (saHpiFieldTable_context *row_ctx)
{
        SaErrorT            rc = SA_OK;
        SaHpiSessionIdT     session_id;
        SaHpiResourceIdT    resource_id;
        SaHpiIdrIdT         idr_id;    
        SaHpiEntryIdT       area_id;
        SaHpiEntryIdT       field_id;

        SaHpiEntryIdT       next_area_id;
        SaHpiIdrAreaHeaderT header;

        DRIA_XREF *dria_entry;
        SaHpiDomainIdResourceIdInventoryIdAreaIdArrayT dria_tuple;

        DEBUGMSGTL ((AGENT, "set_table_field_delete() called\n"));

        if ((row_ctx->field_type_set == MIB_TRUE) &&
            (row_ctx->text_type_set == MIB_TRUE) &&
            (row_ctx->text_language_set == MIB_TRUE) &&
            (row_ctx->text_set == MIB_TRUE)) {

                if (!row_ctx)
                        return AGENT_ERR_NULL_DATA;

                session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_field_INDEX]);
                resource_id = row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                idr_id = row_ctx->index.oids[saHpiInventoryId_field_INDEX];
#if 0
printf("*** set_table_field_delete *******************\n");
printf("**********************************************\n");
printf("session_id [%d]\n", session_id);
printf("resource_id [%d]\n", resource_id);
printf("idr_id [%d]\n", idr_id);
printf("**********************************************\n");
#endif
                /* Get AreaIdIndex                                      */
                /* this is the same as the AreaId obtained from HPI     */
                /* it is maintained in the subagent                     */
                dria_tuple.domainId_resourceId_idr_area_arry[0] = 
                        row_ctx->index.oids[saHpiDomainId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[1] = 
                        row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[2] = 
                        row_ctx->index.oids[saHpiInventoryId_field_INDEX];
                dria_tuple.domainId_resourceId_idr_area_arry[3] = 
                        row_ctx->index.oids[saHpiAreaId_field_INDEX];

                dria_entry = domain_resoruce_idr_area_lookup(&dria_tuple, &dria_table);
                if (!dria_entry) {
                        DEBUGMSGTL ((AGENT, "set_table_field_delete: domain_resoruce_idr_area_lookup() FIALED!!!!!!!\n"));
                        return SNMP_ERR_NOERROR; 
                } 
#if 0
printf("**********************************************\n");
printf("row_ctx->index.oids[saHpiDomainId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiDomainId_field_INDEX]);
printf("row_ctx->index.oids[saHpiResourceEntryId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiResourceEntryId_field_INDEX]);
printf("row_ctx->index.oids[saHpiInventoryId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiInventoryId_field_INDEX]);
printf("row_ctx->index.oids[saHpiAreaId_INDEX] [%d]\n", 
       (int)row_ctx->index.oids[saHpiAreaId_field_INDEX]);
printf("dria_entry->sa_hpi_area_id_index [%d]\n", 
       (int)dria_entry->sa_hpi_area_id_index);
printf("**********************************************\n");
#endif
                area_id = dria_entry->sa_hpi_area_id_index;
                field_id = row_ctx->saHpiFieldIdIndex;

                rc = saHpiIdrFieldDelete(session_id, resource_id,
                                         idr_id, area_id, field_id); 
                if (rc != SA_OK) {
                        snmp_log (LOG_ERR,
                                  "set_table_field_delete: Call to saHpiIdrFieldDelete() failed to set Mode rc: %s.\n",
                                  oh_lookup_error(rc));
                        DEBUGMSGTL ((AGENT,
                                     "set_table_field_delete: Call to saHpiIdrFieldDelete() failed to set Mode rc: %s.\n",
                                     oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }

                rc = saHpiIdrAreaHeaderGet(session_id, resource_id, idr_id, 
                                           SAHPI_IDR_AREATYPE_UNSPECIFIED,
                                           area_id, &next_area_id, 
                                           &header);
                if (rc != SA_OK) {
                                snmp_log (LOG_ERR,
                                "set_table_field_add: Call to saHpiIdrAreaHeaderGet() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc));
                                DEBUGMSGTL ((AGENT,
                                "set_table_field_add: Call to saHpiIdrAreaHeaderGet() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc)));
                                return get_snmp_error(rc);
                }

                rc = update_num_data_fields (row_ctx->index.oids, header.NumFields);
                if (rc != SA_OK) {
                                snmp_log (LOG_ERR,
                                "set_table_field_add: Call to update_num_data_fields() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc));
                                DEBUGMSGTL ((AGENT,
                                "set_table_field_add: Call to update_num_data_fields() failed to set Mode rc: %s.\n",
                                oh_lookup_error(rc)));
                                return get_snmp_error(rc);
                }

                DEBUGMSGTL ((AGENT, 
                "saHpiIdrFieldDelete() + saHpiIdrAreaHeaderGet() + update_num_data_fields() SUCCEEDED!!!!!!!\n"));
        } 

        return SNMP_ERR_NOERROR; 
}


/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int delete_fields(SaHpiSessionIdT  session_id,
				       SaHpiResourceIdT resource_id,
				       SaHpiIdrIdT	idr_id,
				       SaHpiEntryIdT    area_id)
{
        SaErrorT                 rc = SA_OK;
	saHpiFieldTable_context *field_ctx;
	netsnmp_index           *field_index;			

        DEBUGMSGTL ((AGENT, "delete_fields, called\n"));

	field_index = CONTAINER_FIRST(cb.container);
	
	if (field_index) {

		do {	
			field_ctx = CONTAINER_FIND(cb.container, field_index);		
			
			if (  (field_ctx->index.oids[saHpiDomainId_field_INDEX] == 
		                                                      get_domain_id(session_id))
		 	 	 && (field_ctx->index.oids[saHpiResourceEntryId_field_INDEX] == resource_id) 
	     	  	  	 && (field_ctx->index.oids[saHpiInventoryId_field_INDEX] == idr_id) 
	          	  	 && (field_ctx->index.oids[saHpiAreaId_field_INDEX] == area_id) )
			{	
				
				field_index = CONTAINER_NEXT(cb.container, field_index);	   	       		
				
				if (CONTAINER_REMOVE(cb.container, field_ctx) != 0) 
				{
				     DEBUGMSGTL ((AGENT,"delete_fields: FAILURE IN REMOVE FIELDS\n"));
				     return SA_ERR_HPI_UNKNOWN;
				}
					
				saHpiFieldTable_delete_row(field_ctx);
				field_entry_count = CONTAINER_SIZE (cb.container);
                                DEBUGMSGTL ((AGENT, "delete_fields: found row for "
                                                     "deletion\n"));						
									
			}
			else {
				field_index = CONTAINER_NEXT(cb.container, field_index);
			}
								
		}while(field_index);	

	}	

	return rc;
}	

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int handle_saHpiFieldActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests)
{
    /* We are never called for a GETNEXT if it's registered as a
       "instance", as it's "magically" handled for us.  */

    /* a instance handler also only hands us one request at a time, so
       we don't need to loop over a list of requests; we'll only get one. */
    
        DEBUGMSGTL ((AGENT, "handle_saHpiFieldActiveEntries, called\n"));

    field_entry_count = CONTAINER_SIZE (cb.container);

    switch(reqinfo->mode) {

        case MODE_GET:
            snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
                                     (u_char *) &field_entry_count,
                                     sizeof(field_entry_count));
            break;


        default:
            /* we should never get here, so this is a really bad error */
            return SNMP_ERR_GENERR;
    }

    return SNMP_ERR_NOERROR;
} 

/**
 * 
 * @return: 
 */
int initialize_table_saHpiFieldActiveEntries(void)
{
        DEBUGMSGTL ((AGENT, "initialize_table_saHpiFieldActiveEntries, called\n"));

        netsnmp_register_scalar(
                netsnmp_create_handler_registration(
                        "saHpiFieldActiveEntries", 
                        handle_saHpiFieldActiveEntries,
                        saHpiFieldActiveEntries_oid, 
                        OID_LENGTH(saHpiFieldActiveEntries_oid),
                        HANDLER_CAN_RONLY
        ));
        return 0;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiFieldTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiFieldTable_cmp( const void *lhs, const void *rhs )
{
    saHpiFieldTable_context *context_l =
        (saHpiFieldTable_context *)lhs;
    saHpiFieldTable_context *context_r =
        (saHpiFieldTable_context *)rhs;

    /*
     * check primary key, then secondary. Add your own code if
     * there are more than 2 indexes
     */
    DEBUGMSGTL ((AGENT, "saHpiFieldTable_cmp, called\n"));

    /* check for NULL pointers */
    if (lhs == NULL || rhs == NULL ) {
            DEBUGMSGTL((AGENT,"saHpiFieldTable_cmp() NULL pointer ERROR\n" ));
            return 0;
    }
    /* CHECK FIRST INDEX,  saHpiDomainId */
    if ( context_l->index.oids[0] < context_r->index.oids[0])
            return -1;

    if ( context_l->index.oids[0] > context_r->index.oids[0])
            return 1;

    if ( context_l->index.oids[0] == context_r->index.oids[0]) {
            /* If saHpiDomainId index is equal sort by second index */
            /* CHECK SECOND INDEX,  saHpiResourceEntryId */
            if ( context_l->index.oids[1] < context_r->index.oids[1])
                    return -1;

            if ( context_l->index.oids[1] > context_r->index.oids[1])
                    return 1;

            if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                    /* If saHpiResourceEntryId index is equal sort by third index */
                    /* CHECK THIRD INDEX,  saHpiResourceIsHistorical */
                    if ( context_l->index.oids[2] < context_r->index.oids[2])
                            return -1;

                    if ( context_l->index.oids[2] > context_r->index.oids[2])
                            return 1;

                    if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                            /* If saHpiResourceIsHistorical index is equal sort by forth index */
                            /* CHECK FORTH INDEX,  saHpiInventoryId */
                            if ( context_l->index.oids[3] < context_r->index.oids[3])
                                    return -1;

                            if ( context_l->index.oids[3] > context_r->index.oids[3])
                                    return 1;

                            if ( context_l->index.oids[3] == context_r->index.oids[3]) {
                                    /* If saHpiInventoryId index is equal sort by forth index */
                                    /* CHECK FIFTH INDEX,  saHpiAreaId */
                                    if ( context_l->index.oids[4] < context_r->index.oids[4])
                                            return -1;

                                    if ( context_l->index.oids[4] > context_r->index.oids[4])
                                            return 1;

                                    if ( context_l->index.oids[4] == context_r->index.oids[4]) {
                                            /* If saHpiInventoryId index is equal sort by forth index */
                                            /* CHECK SIXTh INDEX,  saHpiFieldId */
                                            if ( context_l->index.oids[5] < context_r->index.oids[5])
                                                    return -1;
                                            
                                            if ( context_l->index.oids[5] > context_r->index.oids[5])
                                                    return 1;
                                            
                                            if ( context_l->index.oids[5] == context_r->index.oids[5])
                                                    return 0;
                                    }
                            }
                    }
            }
    }

    return 0;

}

/************************************************************
 * Initializes the saHpiFieldTable module
 */
void
init_saHpiFieldTable(void)
{
        DEBUGMSGTL ((AGENT, "init_saHpiFieldTable, called\n"));

    initialize_table_saHpiFieldTable();

    initialize_table_saHpiFieldActiveEntries();

    domain_resource_idr_initialize(&initialized, &dria_table);
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiFieldTable_row_copy(saHpiFieldTable_context * dst,
                         saHpiFieldTable_context * src)
{
    if(!dst||!src)
        return 1;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_row_copy, called\n"));
        
   
    subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiFieldId = src->saHpiFieldId;

    dst->saHpiFieldIdIndex = src->saHpiFieldIdIndex;

    dst->saHpiFieldType = src->saHpiFieldType;

    dst->saHpiFieldIsReadOnly = src->saHpiFieldIsReadOnly;

    dst->saHpiFieldTextType = src->saHpiFieldTextType;

    dst->saHpiFieldTextLanguage = src->saHpiFieldTextLanguage;

    memcpy( dst->saHpiFieldText, src->saHpiFieldText, src->saHpiFieldText_len );
    dst->saHpiFieldText_len = src->saHpiFieldText_len;

    dst->saHpiFieldStatus = src->saHpiFieldStatus;

    
    subagent_unlock(&hpi_lock_data);
    return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiFieldTable_extract_index( saHpiFieldTable_context * ctx, netsnmp_index * hdr )
{
    /*
     * temporary local storage for extracting oid index
     *
     * extract index uses varbinds (netsnmp_variable_list) to parse
     * the index OID into the individual components for each index part.
     */
    /** TODO: add storage for external index(s)! */
    netsnmp_variable_list var_saHpiDomainId;
    netsnmp_variable_list var_saHpiResourceId;
    netsnmp_variable_list var_saHpiResourceIsHistorical;
    netsnmp_variable_list var_saHpiInventoryId;
    netsnmp_variable_list var_saHpiAreaId;
    netsnmp_variable_list var_saHpiFieldId;
    int err;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_extract_index, called\n"));

    
    subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(hdr) {
        netsnmp_assert(ctx->index.oids == NULL);
        if(snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                           hdr->len * sizeof(oid) )) {
            subagent_unlock(&hpi_lock_data);
	    return -1;
        }
        ctx->index.len = hdr->len;
    }

    /*
     * initialize variable that will hold each component of the index.
     * If there are multiple indexes for the table, the variable_lists
     * need to be linked together, in order.
     */
       /** TODO: add code for external index(s)! */
       memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
       var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiDomainId.next_variable = &var_saHpiResourceId;

       memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
       var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiResourceId.next_variable = &var_saHpiResourceIsHistorical;

       memset( &var_saHpiResourceIsHistorical, 0x00, sizeof(var_saHpiResourceIsHistorical) );
       var_saHpiResourceIsHistorical.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiResourceIsHistorical.next_variable = &var_saHpiInventoryId;

       memset( &var_saHpiInventoryId, 0x00, sizeof(var_saHpiInventoryId) );
       var_saHpiInventoryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiInventoryId.next_variable = &var_saHpiAreaId;

       memset( &var_saHpiAreaId, 0x00, sizeof(var_saHpiAreaId) );
       var_saHpiAreaId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiAreaId.next_variable = &var_saHpiFieldId;

       memset( &var_saHpiFieldId, 0x00, sizeof(var_saHpiFieldId) );
       var_saHpiFieldId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiFieldId.next_variable = NULL;


    /*
     * parse the oid into the individual index components
     */
    err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
    if (err == SNMP_ERR_NOERROR) {
       /*
        * copy index components into the context structure
        */
              /** skipping external index saHpiDomainId */
   
              /** skipping external index saHpiResourceId */
   
              /** skipping external index saHpiResourceIsHistorical */
   
              /** skipping external index saHpiInventoryId */
   
              /** skipping external index saHpiAreaId */
   
                ctx->saHpiFieldId = *var_saHpiFieldId.val.integer;
   
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
                err = saHpiResourceIsHistorical_check_index(*var_saHpiResourceIsHistorical.val.integer);
                err = saHpiInventoryId_check_index(*var_saHpiInventoryId.val.integer);
                err = saHpiAreaId_check_index(*var_saHpiAreaId.val.integer);
                err = saHpiFieldId_check_index(*var_saHpiFieldId.val.integer);
    }

    /*
     * parsing may have allocated memory. free it.
     */
    snmp_reset_var_buffers( &var_saHpiDomainId );

    subagent_unlock(&hpi_lock_data);
    return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiFieldTable_can_activate(saHpiFieldTable_context *undo_ctx,
                      saHpiFieldTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiFieldTable_can_activate, called\n"));
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiFieldTable_can_deactivate(saHpiFieldTable_context *undo_ctx,
                        saHpiFieldTable_context *row_ctx,
                        netsnmp_request_group * rg)
{

        DEBUGMSGTL ((AGENT, "saHpiFieldTable_can_deactivate, called\n"));
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiFieldTable_can_delete(saHpiFieldTable_context *undo_ctx,
                    saHpiFieldTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiFieldTable_can_delete, called\n"));
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiFieldTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiFieldTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiFieldTable_context *
saHpiFieldTable_create_row( netsnmp_index* hdr)
{
    
    subagent_lock(&hpi_lock_data);
    
    saHpiFieldTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiFieldTable_context);
    if(!ctx) {
        subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    DEBUGMSGTL ((AGENT, "saHpiFieldTable_create_row, called\n"));

    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiFieldTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiFieldType = 0;
     ctx->saHpiFieldTextType = 0;
     ctx->saHpiFieldTextLanguage = 0;
     ctx->saHpiFieldText = 0;
     ctx->saHpiFieldStatus = 0;
    */

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiFieldTable_context *
saHpiFieldTable_duplicate_row( saHpiFieldTable_context * row_ctx)
{
    saHpiFieldTable_context * dup;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_duplicate_row, called\n"));

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
    
    dup = SNMP_MALLOC_TYPEDEF(saHpiFieldTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    if(saHpiFieldTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiFieldTable_delete_row( saHpiFieldTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        DEBUGMSGTL ((AGENT, "saHpiFieldTable_delete_row, called\n"));

    subagent_lock(&hpi_lock_data);
    
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiFieldTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiFieldTable_context *row_ctx =
            (saHpiFieldTable_context *)rg->existing_row;
//    saHpiFieldTable_context *undo_ctx =
//            (saHpiFieldTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;

    DRIA_XREF *dria_entry;
    SaHpiDomainIdResourceIdInventoryIdAreaIdArrayT dria_tuple;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_set_reserve1, called\n"));

    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiFieldType));
        break;

        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiFieldTextType));
        break;

        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiFieldTextLanguage));
        break;

        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
                rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
                if (rc == SNMP_ERR_NOERROR ) {
                        if (var->val_len > sizeof(row_ctx->saHpiFieldText)) {
                                rc = SNMP_ERR_WRONGLENGTH;
                        }
                }
        break;

        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiFieldStatus));
            /* check for valid status and state transitions*/
            if ((rc == SNMP_ERR_NOERROR ) &&  (rg->row_created) && 
                (*var->val.integer != SNMP_ROW_CREATEANDWAIT)) {
                    DEBUGMSGTL ((AGENT, 
                    "AREA_RS: COLUMN_SAHPIAREAROWSTATUS, should be SNMP_ROW_CREATEANDWAIT on new row\n"));
                    rc = SNMP_ERR_WRONGVALUE;
            } else if ((rc == SNMP_ERR_NOERROR ) && (!rg->row_created)) {
                    if ((row_ctx->saHpiFieldStatus != SNMP_ROW_DESTROY) &&
                        (*var->val.integer == SNMP_ROW_DESTROY)) {
                            rc = SNMP_ERR_NOERROR;
                    } else {
                            DEBUGMSGTL ((AGENT, 
                            "COLUMN_SA_HPI_AREA_ROW_STATUS, SNMP_ERR_INCONSISTENTVALUE\n"));
                            rc = SNMP_ERR_INCONSISTENTVALUE;
                    }
            }

            /* if this is a new row check the validity of the user specified AreaId */
            if (((rg->row_created)) && (rc == SNMP_ERR_NOERROR)) {
                    /* check for valid areaId specified */
                    /* this is not the same as the AreaId obtained from HPI */
                    /* it is maintained in the subagent                     */
                    dria_tuple.domainId_resourceId_idr_area_arry[0] = 
                            row_ctx->index.oids[saHpiDomainId_field_INDEX];
                    dria_tuple.domainId_resourceId_idr_area_arry[1] = 
                            row_ctx->index.oids[saHpiResourceEntryId_field_INDEX];
                    dria_tuple.domainId_resourceId_idr_area_arry[2] = 
                            row_ctx->index.oids[saHpiInventoryId_field_INDEX];
                    dria_tuple.domainId_resourceId_idr_area_arry[3] = 
                            row_ctx->index.oids[saHpiAreaId_field_INDEX];

                    /* domain_resource_idr_area_get() generates unique keys based on */
                    /* domainid, resourceid, and IdrId tuples.                  */
                    dria_entry = domain_resoruce_idr_area_lookup(&dria_tuple, &dria_table); 

                    if (dria_entry == NULL) {
                            DEBUGMSGTL ((AGENT, 
                                         "ERROR: saHpiFieldTable_set_reserve1() domain_resource_idr_area_get returned NULL\n"));
                            rc = SNMP_ERR_GENERR;
                    }

                    if ( row_ctx->index.oids[saHpiFieldId_field_INDEX] >= dria_entry->entry_id ) {
                            dria_entry->entry_id = row_ctx->index.oids[saHpiFieldId_field_INDEX];
                            dria_entry->entry_id++;
                    } else {
                            DEBUGMSGTL ((AGENT,"***********************************************\n"));
                            DEBUGMSGTL ((AGENT,"dria_entry->entry_id [%d]\n", dria_entry->entry_id));
                            DEBUGMSGTL ((AGENT,"row_ctx->index.oids[saHpiFieldId_INDEX] [%ld]\n", 
                                         row_ctx->index.oids[saHpiFieldId_field_INDEX]));
                            DEBUGMSGTL ((AGENT, 
                                         "ERROR: saHpiFieldTable_set_reserve1() User specified FieldId invalid!!!\n"));
                            DEBUGMSGTL ((AGENT,"***********************************************\n"));
                            rc = SNMP_ERR_WRONGVALUE;
                    }
            }
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiFieldTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiFieldTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiFieldTable_context *row_ctx = (saHpiFieldTable_context *)rg->existing_row;
//    saHpiFieldTable_context *undo_ctx = (saHpiFieldTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_set_reserve2, called\n"));

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
                /*
                 * TODO: routine to check valid values
                 *
                 * EXAMPLE:
                 *
                * if ( XXX_check_value( var->val.integer, XXX ) ) {
            *    rc = SNMP_ERR_INCONSISTENTVALUE;
            *    rc = SNMP_ERR_BADVALUE;
            * }
            */

        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiFieldTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiFieldTable_context *row_ctx = (saHpiFieldTable_context *)rg->existing_row;
//    saHpiFieldTable_context *undo_ctx = (saHpiFieldTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;

    subagent_lock(&hpi_lock_data);

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_set_action, called\n"));

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiFieldType = *var->val.integer;
            row_ctx->field_type_set = MIB_TRUE;
            if (row_ctx->saHpiFieldStatus == SNMP_ROW_ACTIVE) {
                    set_table_field_update (row_ctx);
            } else if (row_ctx->saHpiFieldStatus == SNMP_ROW_CREATEANDWAIT) {
                    set_table_field_add (row_ctx);
            }
        break;

        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            row_ctx->saHpiFieldTextType = *var->val.integer;
            row_ctx->text_type_set = MIB_TRUE;
            if (row_ctx->saHpiFieldStatus == SNMP_ROW_ACTIVE) {
                    set_table_field_update (row_ctx);
            } else if (row_ctx->saHpiFieldStatus == SNMP_ROW_CREATEANDWAIT) {
                    set_table_field_add (row_ctx);
            }
        break;

        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            row_ctx->saHpiFieldTextLanguage = *var->val.integer;
            row_ctx->text_language_set = MIB_TRUE;
            if (row_ctx->saHpiFieldStatus == SNMP_ROW_ACTIVE) {
                    set_table_field_update (row_ctx);
            } else if (row_ctx->saHpiFieldStatus == SNMP_ROW_CREATEANDWAIT) {
                    set_table_field_add (row_ctx);
            }
        break;

        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiFieldText,var->val.string,var->val_len);
            row_ctx->saHpiFieldText_len = var->val_len;
            row_ctx->text_set = MIB_TRUE;
            if (row_ctx->saHpiFieldStatus == SNMP_ROW_ACTIVE) {
                    set_table_field_update (row_ctx);
            } else if (row_ctx->saHpiFieldStatus == SNMP_ROW_CREATEANDWAIT) {
                    set_table_field_add (row_ctx);
            }
        break;

        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
            if (*var->val.integer == SNMP_ROW_DESTROY) {
                    if (row_ctx->saHpiFieldStatus == SNMP_ROW_ACTIVE) {
                            row_err = set_table_field_delete (row_ctx);
                    }
                    rg->row_deleted = 1;
            }
            row_ctx->saHpiFieldStatus = *var->val.integer;
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }
    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiFieldTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiFieldTable_context *row_ctx = (saHpiFieldTable_context *)rg->existing_row;
//    saHpiFieldTable_context *undo_ctx = (saHpiFieldTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_set_commit, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiFieldTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiFieldTable_context *row_ctx = (saHpiFieldTable_context *)rg->existing_row;
//    saHpiFieldTable_context *undo_ctx = (saHpiFieldTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_set_free, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: 
        break;
        /** We shouldn't get here */
        /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiFieldTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiFieldTable_context *row_ctx = (saHpiFieldTable_context *)rg->existing_row;
//    saHpiFieldTable_context *undo_ctx = (saHpiFieldTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_set_undo, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 *
 * Initialize the saHpiFieldTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiFieldTable(void)
{
    netsnmp_table_registration_info *table_info;

    DEBUGMSGTL ((AGENT, "initialize_table_saHpiFieldTable, called\n"));

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiFieldTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiFieldTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiFieldTable_oid,
                                             saHpiFieldTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiFieldTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceIsHistorical */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiInventoryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiAreaId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiFieldId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiFieldTable_COL_MIN;
    table_info->max_column = saHpiFieldTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiFieldTable_get_value;
    cb.container = netsnmp_container_find("saHpiFieldTable_primary:"
                                          "saHpiFieldTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiFieldTable_secondary:"
                                                       "saHpiFieldTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiFieldTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiFieldTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiFieldTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiFieldTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiFieldTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiFieldTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiFieldTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiFieldTable_can_delete;

    cb.set_reserve1 = saHpiFieldTable_set_reserve1;
    cb.set_reserve2 = saHpiFieldTable_set_reserve2;
    cb.set_action = saHpiFieldTable_set_action;
    cb.set_commit = saHpiFieldTable_set_commit;
    cb.set_free = saHpiFieldTable_set_free;
    cb.set_undo = saHpiFieldTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiFieldTable",
                "Registering table saHpiFieldTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiFieldTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiFieldTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiFieldTable_context *context = (saHpiFieldTable_context *)item;

    DEBUGMSGTL ((AGENT, "saHpiFieldTable_get_value, called\n"));

    switch(table_info->colnum) {

        case COLUMN_SAHPIFIELDID:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiFieldId,
                         sizeof(context->saHpiFieldId) );
        break;
    
        case COLUMN_SAHPIFIELDIDINDEX:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiFieldIdIndex,
                         sizeof(context->saHpiFieldIdIndex) );
        break;
    
        case COLUMN_SAHPIFIELDTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiFieldType,
                         sizeof(context->saHpiFieldType) );
        break;
    
        case COLUMN_SAHPIFIELDISREADONLY:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiFieldIsReadOnly,
                         sizeof(context->saHpiFieldIsReadOnly) );
        break;
    
        case COLUMN_SAHPIFIELDTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiFieldTextType,
                         sizeof(context->saHpiFieldTextType) );
        break;
    
        case COLUMN_SAHPIFIELDTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiFieldTextLanguage,
                         sizeof(context->saHpiFieldTextLanguage) );
        break;
    
        case COLUMN_SAHPIFIELDTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiFieldText,
                         context->saHpiFieldText_len );
        break;
    
        case COLUMN_SAHPIFIELDSTATUS:
            /** RowStatus = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiFieldStatus,
                         sizeof(context->saHpiFieldStatus) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiFieldTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiFieldTable_get_by_idx
 */
const saHpiFieldTable_context *
saHpiFieldTable_get_by_idx(netsnmp_index * hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiFieldTable_get_by_idx, called\n"));

    return (const saHpiFieldTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


