/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiHotSwapTable.c 5429 2005-11-14 22:23:12Z ddearauj $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiHotSwapTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <saHpiAutoInsertTimeoutTable.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiHotSwapTable_oid[] = { saHpiHotSwapTable_TABLE_OID };
size_t saHpiHotSwapTable_oid_len = OID_LENGTH(saHpiHotSwapTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * oid and fucntion declarations scalars
 */
 
static u_long hotswap_entry_count = 0;
static oid saHpiHotSwapActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,2,11,2 };

int handle_saHpiHotSwapActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests);
				
int initialize_table_saHpiHotSwapActiveEntries(void);


/**
 * 
 * @sessionid:
 * @rpt_entry:
 * 
 * @return 
 */
SaErrorT populate_hotswap(SaHpiSessionIdT sessionid,
                          SaHpiRptEntryT *rpt_entry)
{

        DEBUGMSGTL ((AGENT, "populate_hotswap, called\n"));

        SaErrorT rv = SA_OK;
        int new_row = MIB_FALSE;

        oid hotswap_oid[HOTSWAP_INDEX_NR];
        netsnmp_index hotswap_idx;
        saHpiHotSwapTable_context *hotswap_ctx;
	
	SaHpiHsIndicatorStateT indicator_state;
	SaHpiHsStateT hs_state;
	SaHpiTimeoutT extract_timeout;

        DEBUGMSGTL ((AGENT, "SAHPI_HOTSWAP_RDR populate_hotswap() called\n"));

        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_hotswap() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

	subagent_lock(&hpi_lock_data);

        /* BUILD oid for new row */
        /* assign the number of indices */
        hotswap_idx.len = HOTSWAP_INDEX_NR;
        /** Index saHpiDomainId is external */
        hotswap_oid[0] = get_domain_id(sessionid);
        /** Index saHpiResourceId is external */
        hotswap_oid[1] = rpt_entry->ResourceId;
        /** Index saHpiResourceIsHistorical is external */
        hotswap_oid[2] = MIB_FALSE;
        /* assign the indices to the index */
        hotswap_idx.oids = (oid *) & hotswap_oid;

        /* See if Row exists. */
        hotswap_ctx = NULL;
        hotswap_ctx = CONTAINER_FIND(cb.container, &hotswap_idx);

        if (!hotswap_ctx) {
                // New entry. Add it
                hotswap_ctx = 
                saHpiHotSwapTable_create_row(&hotswap_idx);
                new_row = MIB_TRUE;
        }
        if (!hotswap_ctx) {
                snmp_log (LOG_ERR, "Not enough memory for a HotSwap row!");
                
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
        }

        /** INTEGER = ASN_INTEGER */	
	if (SA_OK == saHpiHotSwapIndicatorStateGet (sessionid, 
				        rpt_entry->ResourceId, 
					&indicator_state) ) {
		hotswap_ctx->saHpiHotSwapIndicator = indicator_state + 1;
			
	}
	
        /** SaHpiHotSwapState = ASN_INTEGER */
	if (SA_OK == saHpiHotSwapStateGet(sessionid,
					rpt_entry->ResourceId,
					&hs_state
					  ) ){
		hotswap_ctx->saHpiHotSwapState = hs_state + 1; 
	}
	
        /** SaHpiTime = ASN_OCTET_STR */                     
	if ( SA_OK == saHpiAutoExtractTimeoutGet(sessionid,
			                rpt_entry->ResourceId,
					&extract_timeout) ) {
                hotswap_ctx->saHpiHotSwapExtractTimeout_len = 
                        sizeof(extract_timeout);
                memcpy(hotswap_ctx->saHpiHotSwapExtractTimeout,
                       &extract_timeout, 
                       hotswap_ctx->saHpiHotSwapExtractTimeout_len);
	}

        /** INTEGER = ASN_INTEGER */
        hotswap_ctx->saHpiHotSwapActionRequest = 0;              

        /** INTEGER = ASN_INTEGER */
        hotswap_ctx->saHpiHotSwapPolicyCancel = 0;

        /** INTEGER = ASN_INTEGER */
        hotswap_ctx->saHpiHotSwapResourceRequest = 0;

        if (new_row == MIB_TRUE)
                CONTAINER_INSERT (cb.container, hotswap_ctx);

        hotswap_entry_count = CONTAINER_SIZE (cb.container);

        populate_saHpiAutoInsertTimeoutTable(sessionid);

	subagent_unlock(&hpi_lock_data);

	return rv;	
}


/**
 * 
 * @sessionid
 * @event
 * @rdr_entry
 * @rpt_entry
 * 
 * @return 
 */
SaErrorT async_hotswap_add(SaHpiSessionIdT sessionid,
                           SaHpiEventT *event, 
                           SaHpiRdrT *rdr_entry, 
                           SaHpiRptEntryT *rpt_entry)
{
        SaErrorT rv = SA_OK;

        oid hotswap_oid[HOTSWAP_INDEX_NR];
        netsnmp_index hotswap_idx;
        saHpiHotSwapTable_context *hotswap_ctx;
	
	SaHpiHsIndicatorStateT indicator_state;
	SaHpiTimeoutT extract_timeout;

        DEBUGMSGTL ((AGENT, "async_hotswap_add(), called\n"));

        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: async_hotswap_add() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        /* BUILD oid for new row */
        /* assign the number of indices */
        hotswap_idx.len = HOTSWAP_INDEX_NR;
        /** Index saHpiDomainId is external */
        hotswap_oid[0] = get_domain_id(sessionid);
        /** Index saHpiResourceId is external */
        hotswap_oid[1] = rpt_entry->ResourceId;
        /** Index saHpiResourceIsHistorical is external */
        hotswap_oid[2] = MIB_FALSE;
        /* assign the indices to the index */
        hotswap_idx.oids = (oid *) & hotswap_oid;

        /* See if Row exists. */
        hotswap_ctx = NULL;
        hotswap_ctx = CONTAINER_FIND(cb.container, &hotswap_idx);

        if (!hotswap_ctx) {
                DEBUGMSGTL ((AGENT, "async_hotswap_add() row NOT found!\n"));
                return AGENT_ERR_NOT_FOUND;
        } else {
                DEBUGMSGTL ((AGENT, "async_hotswap_add() row FOUND!\n"));
        }

        /** INTEGER = ASN_INTEGER */	
	if (SA_OK == saHpiHotSwapIndicatorStateGet (sessionid, 
				        rpt_entry->ResourceId, 
					&indicator_state) ) {
		hotswap_ctx->saHpiHotSwapIndicator = indicator_state + 1;
			
	}       
	
        /** SaHpiHotSwapState = ASN_INTEGER */
        hotswap_ctx->saHpiHotSwapState = 
                event->EventDataUnion.HotSwapEvent.HotSwapState + 1; 
	
        /** SaHpiTime = ASN_OCTET_STR */                     
	if ( SA_OK == saHpiAutoExtractTimeoutGet(sessionid,
			                rpt_entry->ResourceId,
					&extract_timeout) ) {
                hotswap_ctx->saHpiHotSwapExtractTimeout_len = 
                        sizeof(extract_timeout);
                memcpy(hotswap_ctx->saHpiHotSwapExtractTimeout,
                       &extract_timeout, 
                       hotswap_ctx->saHpiHotSwapExtractTimeout_len);
	}

        populate_saHpiAutoInsertTimeoutTable(sessionid);

        return rv;	
}


/*
 * int hot_swap_indicator_set (saHpiHotSwapTable_context * ctx)
 */
int hot_swap_indicator_set (saHpiHotSwapTable_context *row_ctx)
{
	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;
        SaHpiHsIndicatorStateT  state;

        DEBUGMSGTL ((AGENT, "hot_swap_indicator_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];
        state = row_ctx->saHpiHotSwapIndicator - 1;

        rc = saHpiHotSwapIndicatorStateSet(session_id, resource_id, state);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to saHpiHotSwapIndicatorStateSet failed to set IndicatorState rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			   "Call to saHpiHotSwapIndicatorStateSet failed to set IndicatorState rc: %s.\n",
			   oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/*
 * int hot_swap_state_set (saHpiHotSwapTable_context * ctx)
 */
int hot_swap_state_set (saHpiHotSwapTable_context *row_ctx)
{
	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;

        DEBUGMSGTL ((AGENT, "hot_swap_state_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];

        if ( (row_ctx->saHpiHotSwapState - 1) == SAHPI_HS_STATE_ACTIVE ) {
                rc = saHpiResourceActiveSet(session_id, resource_id);
        } else if ( (row_ctx->saHpiHotSwapState - 1) == SAHPI_HS_STATE_INACTIVE ) {
                rc = saHpiResourceInactiveSet(session_id, resource_id);
        }

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to hot_swap_state_set failed to set ResourceState [%s] rc: %s.\n",
                          oh_lookup_hsstate(row_ctx->saHpiHotSwapState - 1), 
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			   "Call to hot_swap_state_set failed to set ResourceState [%s] rc: %s.\n",
                           oh_lookup_hsstate(row_ctx->saHpiHotSwapState - 1), 
			   oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/*
 * int hot_swap_request_action_set (saHpiHotSwapTable_context *row_ctx)
 */
int hot_swap_action_request_set (saHpiHotSwapTable_context *row_ctx)
{
	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;
	SaHpiHsActionT          action;

        DEBUGMSGTL ((AGENT, "hot_swap_action_request_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];
        action = row_ctx->saHpiHotSwapActionRequest - 1;

        rc = saHpiHotSwapActionRequest(session_id, resource_id, action);


	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to hot_swap_action_request_set failed to set action [%s] rc: %s.\n",
                          oh_lookup_hsaction(row_ctx->saHpiHotSwapActionRequest - 1), 
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			   "Call to hot_swap_action_request_set failed to set action [%s] rc: %s.\n",
                           oh_lookup_hsaction(row_ctx->saHpiHotSwapActionRequest - 1), 
			   oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 

}    

/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int hot_swap_auto_extract_timeout_set (saHpiHotSwapTable_context *row_ctx)
{
	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;
	SaHpiTimeoutT           timeout;

        DEBUGMSGTL ((AGENT, "hot_swap_extract_timeout_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];

        if (row_ctx->saHpiHotSwapExtractTimeout_len > sizeof(SaHpiTimeoutT)) 
                return SNMP_ERR_TOOBIG;
        
        timeout = snmptime_to_hpitime (row_ctx->saHpiHotSwapExtractTimeout);

        rc = saHpiAutoExtractTimeoutSet(session_id, resource_id, timeout);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to hot_swap_extract_timeout_set"
                          " failed to set timeout [%lld] rc: %s.\n",
                          timeout, 
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			   "Call to hot_swap_extract_timeout_set"
                           " failed to set timeout [%lld] rc: %s.\n",
                           timeout, 
			   oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/*
 * int hot_swap_policy_set (saHpiHotSwapTable_context * ctx)
 */
int hot_swap_policy_set (saHpiHotSwapTable_context *row_ctx)
{
	SaErrorT                rc = SA_OK;
	SaHpiSessionIdT         session_id;
	SaHpiResourceIdT        resource_id;

        DEBUGMSGTL ((AGENT, "hot_swap_policy_set, called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->index.oids[saHpiResourceId_INDEX];

        rc = saHpiHotSwapPolicyCancel(session_id, resource_id);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to hot_swap_policy_set "
                          "saHpiHotSwapPolicyCancel() failed. rc: [%s].\n",
			  oh_lookup_error(rc));
                DEBUGMSGTL ((AGENT,
			  "Call to hot_swap_policy_set "
                          "saHpiHotSwapPolicyCancel() failed. rc: [%s].\n",
			  oh_lookup_error(rc)));
		return get_snmp_error(rc);
	} 

	return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiHotSwapActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiHotSwapActiveEntries, called\n"));

        hotswap_entry_count = CONTAINER_SIZE (cb.container);       
        
	switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			        (u_char *) &hotswap_entry_count,
        			        sizeof(hotswap_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}


/**
 * 
 * @return: 
 */
int initialize_table_saHpiHotSwapActiveEntries(void)
{
        
	DEBUGMSGTL ((AGENT, "initialize_table_saHpiHotSwapActiveEntries, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiHotSwapActiveEntries", 
					handle_saHpiHotSwapActiveEntries,
                                        saHpiHotSwapActiveEntries_oid, 
					OID_LENGTH(saHpiHotSwapActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiHotSwapTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiHotSwapTable_cmp( const void *lhs, const void *rhs )
{
        saHpiHotSwapTable_context *context_l =
        (saHpiHotSwapTable_context *)lhs;
        saHpiHotSwapTable_context *context_r =
        (saHpiHotSwapTable_context *)rhs;
        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiHotSwapTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiResourceIsHistorical */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) 
                                 return 0;
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiHotSwapTable module
 */
void
init_saHpiHotSwapTable(void)
{
        DEBUGMSGTL ((AGENT, "init_saHpiHotSwapTable, called\n"));

        initialize_table_saHpiHotSwapTable();

        initialize_table_saHpiHotSwapActiveEntries();
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiHotSwapTable_row_copy(saHpiHotSwapTable_context * dst,
                         saHpiHotSwapTable_context * src)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_row_copy, called\n"));

    if(!dst||!src)
        return 1;
        
    subagent_lock(&hpi_lock_data);
    
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
	return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiHotSwapIndicator = src->saHpiHotSwapIndicator;

    dst->saHpiHotSwapState = src->saHpiHotSwapState;

    memcpy( dst->saHpiHotSwapExtractTimeout, 
            src->saHpiHotSwapExtractTimeout, 
            src->saHpiHotSwapExtractTimeout_len );
    dst->saHpiHotSwapExtractTimeout_len = src->saHpiHotSwapExtractTimeout_len;

    dst->saHpiHotSwapActionRequest = src->saHpiHotSwapActionRequest;

    dst->saHpiHotSwapPolicyCancel = src->saHpiHotSwapPolicyCancel;

    dst->saHpiHotSwapResourceRequest = src->saHpiHotSwapResourceRequest;

    subagent_unlock(&hpi_lock_data);
    return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiHotSwapTable_extract_index( saHpiHotSwapTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiResourceIsHistorical;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_extract_index, called\n"));

        subagent_lock(&hpi_lock_data);
	
	/*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
			return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiResourceIsHistorical;

        memset( &var_saHpiResourceIsHistorical, 0x00, sizeof(var_saHpiResourceIsHistorical) );
        var_saHpiResourceIsHistorical.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceIsHistorical.next_variable = NULL;

        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */

                /** skipping external index saHpiResourceId */

                /** skipping external index saHpiResourceIsHistorical */

                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
                err = saHpiResourceIsHistorical_check_index(*var_saHpiResourceIsHistorical.val.integer);
                
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiHotSwapTable_can_activate(saHpiHotSwapTable_context *undo_ctx,
                      saHpiHotSwapTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_can_activate, called\n"));

    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiHotSwapTable_can_deactivate(saHpiHotSwapTable_context *undo_ctx,
                        saHpiHotSwapTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_can_deactivate, called\n"));
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiHotSwapTable_can_delete(saHpiHotSwapTable_context *undo_ctx,
                    saHpiHotSwapTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_can_delete, called\n"));
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiHotSwapTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiHotSwapTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiHotSwapTable_context *
saHpiHotSwapTable_create_row( netsnmp_index* hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_create_row, called\n"));

   subagent_lock(&hpi_lock_data);
   
    saHpiHotSwapTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiHotSwapTable_context);
    if(!ctx) {
    	subagent_unlock(&hpi_lock_data);
        return NULL; }
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiHotSwapTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
	return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiHotSwapIndicator = 0;
     ctx->saHpiHotSwapState = 0;
     ctx->saHpiHotSwapExtractTimeout = 0;
     ctx->saHpiHotSwapActionRequest = 0;
     ctx->saHpiHotSwapPolicyCancel = 0;
     ctx->saHpiHotSwapResourceRequest = 0;
    */

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiHotSwapTable_context *
saHpiHotSwapTable_duplicate_row( saHpiHotSwapTable_context * row_ctx)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_duplicate_row, called\n"));

    saHpiHotSwapTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
    
    dup = SNMP_MALLOC_TYPEDEF(saHpiHotSwapTable_context);
    if(!dup) {
    	subagent_unlock(&hpi_lock_data);
        return NULL;}
        
    if(saHpiHotSwapTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiHotSwapTable_delete_row( saHpiHotSwapTable_context * ctx )
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_delete_row, called\n"));

  /* netsnmp_mutex_destroy(ctx->lock); */

    subagent_lock(&hpi_lock_data);
    
    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiHotSwapTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiHotSwapTable_context *row_ctx =
            (saHpiHotSwapTable_context *)rg->existing_row;
//    saHpiHotSwapTable_context *undo_ctx =
//            (saHpiHotSwapTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_set_reserve1, called\n"));

    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiHotSwapIndicator));
        break;

        case COLUMN_SAHPIHOTSWAPSTATE:
            /** SaHpiHotSwapState = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiHotSwapState));
        break;

        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
            rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);
            if (rc == SNMP_ERR_NOERROR ) {
                    if (var->val_len > SAF_UNSIGNED_64_LEN) {
                            rc = SNMP_ERR_WRONGLENGTH;
                            DEBUGMSGTL ((AGENT, 
                            "COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT"
                            " SNMP_ERR_WRONGLENGTH\n"));
                    }
            }
            break;

        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiHotSwapActionRequest));
        break;

        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiHotSwapPolicyCancel));
        break;

        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiHotSwapResourceRequest));
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiHotSwapTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiHotSwapTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiHotSwapTable_context *row_ctx = (saHpiHotSwapTable_context *)rg->existing_row;
//    saHpiHotSwapTable_context *undo_ctx = (saHpiHotSwapTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_set_reserve2, called\n"));

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
                if (!oh_lookup_hsindicatorstate(*var->val.integer - 1))
                        rc = SNMP_ERR_WRONGVALUE;
        break;


        case COLUMN_SAHPIHOTSWAPSTATE:
                if ( ((*var->val.integer - 1) != SAHPI_HS_STATE_INACTIVE) &&
                     ((*var->val.integer - 1) != SAHPI_HS_STATE_ACTIVE) )   {
                        DEBUGMSGTL ((AGENT, "COLUMN_SAHPIHOTSWAPSTATE [%s]"
                                     " saHpiHotSwapTable_set_reserve2"
                                     " Only transistions to ACTIVE or INACITVE"
                                     " are allowed\n",
                                     oh_lookup_hsstate(*var->val.integer - 1)
                                     ));
                        rc = SNMP_ERR_BADVALUE;
                }
        break;

        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
                if ( ((*var->val.integer - 1) != SAHPI_HS_ACTION_INSERTION) &&
                     ((*var->val.integer - 1) != SAHPI_HS_ACTION_EXTRACTION) )   {
                        DEBUGMSGTL ((AGENT, "COLUMN_SAHPIHOTSWAPACTIONREQUEST"
                                     " saHpiHotSwapTable_set_reserve2"
                                     " Only HotSwap action requests of"
				     " SAHPI_HS_ACTION_INSERTION and" 
				     " SAHPI_HS_ACTION_EXTRACTION"
                                     " are allowed\n"));
                        rc = SNMP_ERR_BADVALUE;
                }            
        break;

        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
                if ( ((*var->val.integer - 1) != SAHPI_HS_STATE_INACTIVE) ||
                     ((*var->val.integer - 1) != SAHPI_HS_STATE_ACTIVE) )   {
                        DEBUGMSGTL ((AGENT, "COLUMN_SAHPIHOTSWAPRESOURCEREQUEST"
                                     " saHpiHotSwapTable_set_reserve2"
                                     " Only transistions to INACTIVE or ACITVE"
                                     " are allowed\n"));
                        rc = SNMP_ERR_BADVALUE;
                }
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiHotSwapTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiHotSwapTable_context *row_ctx = (saHpiHotSwapTable_context *)rg->existing_row;
//    saHpiHotSwapTable_context *undo_ctx = (saHpiHotSwapTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    subagent_lock(&hpi_lock_data);

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_set_action, called\n"));
    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiHotSwapIndicator = *var->val.integer;
            row_err = hot_swap_indicator_set(row_ctx);
        break;

        case COLUMN_SAHPIHOTSWAPSTATE:
            /** SaHpiHotSwapState = ASN_INTEGER */
            row_ctx->saHpiHotSwapState = *var->val.integer;
            row_err = hot_swap_state_set(row_ctx);
        break;

        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiHotSwapExtractTimeout, var->val.string, var->val_len );
            row_ctx->saHpiHotSwapExtractTimeout_len = var->val_len;
            row_err = hot_swap_auto_extract_timeout_set (row_ctx);
        break;
                                                       
        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiHotSwapActionRequest = *var->val.integer;
	    row_err = hot_swap_action_request_set(row_ctx);
        break;

        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiHotSwapPolicyCancel = *var->val.integer;
            row_err = hot_swap_policy_set(row_ctx);
        break;

        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiHotSwapResourceRequest = *var->val.integer;
	    row_err = hot_swap_state_set(row_ctx);
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */

    /*
     * check activation/deactivation
     */
    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }

    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiHotSwapTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiHotSwapTable_context *row_ctx = (saHpiHotSwapTable_context *)rg->existing_row;
//    saHpiHotSwapTable_context *undo_ctx = (saHpiHotSwapTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_set_commit, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPSTATE:
            /** SaHpiHotSwapState = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
            /** INTEGER = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiHotSwapTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiHotSwapTable_context *row_ctx = (saHpiHotSwapTable_context *)rg->existing_row;
//    saHpiHotSwapTable_context *undo_ctx = (saHpiHotSwapTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_set_free, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPSTATE:
            /** SaHpiHotSwapState = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
            /** INTEGER = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
	        break;
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiHotSwapTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiHotSwapTable_context *row_ctx = (saHpiHotSwapTable_context *)rg->existing_row;
//    saHpiHotSwapTable_context *undo_ctx = (saHpiHotSwapTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_set_undo, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPSTATE:
            /** SaHpiHotSwapState = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
            /** INTEGER = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}


/************************************************************
 *
 * Initialize the saHpiHotSwapTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiHotSwapTable(void)
{
    netsnmp_table_registration_info *table_info;

    DEBUGMSGTL ((AGENT, "initialize_table_saHpiHotSwapTable, called\n"));

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiHotSwapTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiHotSwapTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiHotSwapTable_oid,
                                             saHpiHotSwapTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiHotSwapTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceIsHistorical */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);

    table_info->min_column = saHpiHotSwapTable_COL_MIN;
    table_info->max_column = saHpiHotSwapTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiHotSwapTable_get_value;
    cb.container = netsnmp_container_find("saHpiHotSwapTable_primary:"
                                          "saHpiHotSwapTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiHotSwapTable_secondary:"
                                                       "saHpiHotSwapTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiHotSwapTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiHotSwapTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiHotSwapTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiHotSwapTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiHotSwapTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiHotSwapTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiHotSwapTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiHotSwapTable_can_delete;

    cb.set_reserve1 = saHpiHotSwapTable_set_reserve1;
    cb.set_reserve2 = saHpiHotSwapTable_set_reserve2;
    cb.set_action = saHpiHotSwapTable_set_action;
    cb.set_commit = saHpiHotSwapTable_set_commit;
    cb.set_free = saHpiHotSwapTable_set_free;
    cb.set_undo = saHpiHotSwapTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiHotSwapTable",
                "Registering table saHpiHotSwapTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiHotSwapTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiHotSwapTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiHotSwapTable_context *context = (saHpiHotSwapTable_context *)item;

    DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_get_value, called\n"));

    switch(table_info->colnum) {

        case COLUMN_SAHPIHOTSWAPINDICATOR:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiHotSwapIndicator,
                         sizeof(context->saHpiHotSwapIndicator) );
        break;
    
        case COLUMN_SAHPIHOTSWAPSTATE:
            /** SaHpiHotSwapState = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiHotSwapState,
                         sizeof(context->saHpiHotSwapState) );
        break;
    
        case COLUMN_SAHPIHOTSWAPEXTRACTTIMEOUT:
            /** SafUnsigned64 = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiHotSwapExtractTimeout,
                         context->saHpiHotSwapExtractTimeout_len );
        break;
    
        case COLUMN_SAHPIHOTSWAPACTIONREQUEST:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiHotSwapActionRequest,
                         sizeof(context->saHpiHotSwapActionRequest) );
        break;
    
        case COLUMN_SAHPIHOTSWAPPOLICYCANCEL:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiHotSwapPolicyCancel,
                         sizeof(context->saHpiHotSwapPolicyCancel) );
        break;
    
        case COLUMN_SAHPIHOTSWAPRESOURCEREQUEST:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiHotSwapResourceRequest,
                         sizeof(context->saHpiHotSwapResourceRequest) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiHotSwapTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiHotSwapTable_get_by_idx
 */
const saHpiHotSwapTable_context *
saHpiHotSwapTable_get_by_idx(netsnmp_index * hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiHotSwapTable_get_by_idx, called\n"));

    return (const saHpiHotSwapTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


