/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiResourceEventLogTable.c 5420 2005-11-14 20:46:04Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiResourceEventLogTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiResourceEventLogTable_oid[] = { saHpiResourceEventLogTable_TABLE_OID };
size_t saHpiResourceEventLogTable_oid_len = OID_LENGTH(saHpiResourceEventLogTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;		      
static GHashTable *dr_table;


/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long resource_event_log_entry_count_total = 0;
static u_long resource_event_log_entry_count = 0;

static oid saHpiResourceEventLogLifetimeChanges_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,3 };
static oid saHpiResourceEventLogActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,4 }; 

int handle_saHpiResourceEventLogLifetimeChanges(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);

int handle_saHpiResourceEventLogActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);
					
int initialize_table_saHpiResourceEventLogLifetimeChanges(void);
int initialize_table_saHpiResourceEventLogActiveEntries(void); 

/**
 * 
 * @sessionid:
 * @event:
 * @this_child_oid:
 * @this_child_oid_len:
 * 
 * @return:
 */
SaErrorT populate_saHpiResourceEventLogTable(SaHpiSessionIdT sessionid, 
                                             SaHpiEventLogEntryT *event,
                                             oid * this_child_oid, 
                                             size_t *this_child_oid_len)
{
	SaErrorT rv = SA_OK;

	oid res_evt_oid[RESOURCE_EVENT_LOG_INDEX_NR];
	netsnmp_index res_evt_idx;
	saHpiResourceEventLogTable_context *res_evt_ctx;

	oid column[2];
	int column_len = 2;

        int isNewRow = MIB_TRUE;

        DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

        DEBUGMSGTL ((AGENT, "populate_saHpiResourceEventLogTable, called\n"));

	/* check for NULL pointers */
	if (!event) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiResourceEventLogTable() passed NULL event pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}    

	/* BUILD oid for new row */
		/* assign the number of indices */
	res_evt_idx.len = RESOURCE_EVENT_LOG_INDEX_NR;
		/** Index saHpiDomainId is external */
	res_evt_oid[0] = get_domain_id(sessionid);
		/** Index saHpiResourceId is external */
	res_evt_oid[1] = event->Event.Source;
		/** Index saHpiEventSeverity is external */
	res_evt_oid[2] = event->Event.Severity + 1;
                /** Index saHpiResourceEventEntryId is internal */
	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
	dr_pair.domainId_resourceId_arry[1] = event->Event.Source;
	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
	if (dr_entry == NULL) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiResourceEventLogTable() domain_resource_pair_get returned NULL\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}
	res_evt_oid[3] = dr_entry->entry_id++;
		/* assign the indices to the index */
	res_evt_idx.oids = (oid *) & res_evt_oid;

	/* create full oid on This row for parent RowPointer */
	column[0] = 1;
	column[1] = COLUMN_SAHPIRESOURCEEVENTLOGTIMESTAMP;
	memset(this_child_oid, 0, sizeof(this_child_oid));
	build_full_oid(saHpiResourceEventLogTable_oid, saHpiResourceEventLogTable_oid_len,
			column, column_len,
			&res_evt_idx,
			this_child_oid, MAX_OID_LEN, this_child_oid_len);

	/* See if Row exists. */
	res_evt_ctx = NULL;
	res_evt_ctx = CONTAINER_FIND(cb.container, &res_evt_idx);

	if (!res_evt_ctx) { 
		// New entry. Add it
		res_evt_ctx = 
			saHpiResourceEventLogTable_create_row(&res_evt_idx);
	}
	else {
		isNewRow = MIB_FALSE;
	}	
	if (!res_evt_ctx) {
		snmp_log (LOG_ERR, "Not enough memory for a Resource Event Log row!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}

        /** SaHpiEntryId = ASN_UNSIGNED */
        res_evt_ctx->saHpiResourceEventLogEntryId = res_evt_oid[3];

        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event->Timestamp, res_evt_ctx->saHpiResourceEventLogTimestamp);
        res_evt_ctx->saHpiResourceEventLogTimestamp_len = sizeof(SaHpiTimeT);
	
        /** INTEGER = ASN_INTEGER */
        res_evt_ctx->saHpiResourceEventLogType = 
               event->Event.EventDataUnion.ResourceEvent.ResourceEventType + 1;
	
	if (isNewRow == MIB_TRUE) {
	
		CONTAINER_INSERT (cb.container, res_evt_ctx);
	}
		
	resource_event_log_entry_count = CONTAINER_SIZE (cb.container);

        return SA_OK;
}


/**
 * 
 * @session_id
 * @resource_id
 * @saHpiEventLogRowPointer
 * @aHpiEventLogRowPointer_len
 * 
 * @return 
 */
SaErrorT resource_event_log_clear(SaHpiSessionIdT session_id, 
                                  SaHpiResourceIdT resource_id,  
                                  oid *saHpiEventLogRowPointer, 
                                  size_t saHpiEventLogRowPointer_len,
				  int modifyTotal)
{
	oid res_evt_oid[RESOURCE_EVENT_LOG_INDEX_NR];
	netsnmp_index res_evt_idx;
	netsnmp_index *res_index;
	saHpiResourceEventLogTable_context *res_evt_ctx;
	
	DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;	

	int column_len = 2;

        DEBUGMSGTL ((AGENT, "resource_event_log_clear, called\n"));
	DEBUGMSGTL ((AGENT, "Attempting to delete resource el row with the following indexes:\n"));
        DEBUGMSGTL ((AGENT,"     Domain   [%ld]\n", saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len]));
        DEBUGMSGTL ((AGENT,"     Resource [%ld]\n", saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 1]));
	DEBUGMSGTL ((AGENT,"     Severity [%s]\n",  oh_lookup_severity(
						saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 2]-1)));
	DEBUGMSGTL ((AGENT,"     Entry id [%ld]\n", saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 3]));

	/* BUILD oid to look for row */
	/* assign the number of indices */
	res_evt_idx.len = RESOURCE_EVENT_LOG_INDEX_NR;
	/** Index saHpiDomainId is external */
	res_evt_oid[0] = saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len];
	/** Index saHpiResourceId is external */
	res_evt_oid[1] = saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 1];
	/** Index saHpiEventSeverity is external */
	res_evt_oid[2] = saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 2];
        /** Index saHpiResourceEventEntryId is internal */
	res_evt_oid[3] = saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 3];
	/* assign the indices to the index */
	res_evt_idx.oids = (oid *) & res_evt_oid;
	
	res_index = CONTAINER_FIRST(cb.container);
	res_evt_ctx = CONTAINER_FIND(cb.container, &res_evt_idx);
	
	if (!res_evt_ctx) {
		DEBUGMSGTL ((AGENT, "resource_event_log_clear did not find a row to delete\n"));
	}
	else {
		
		DEBUGMSGTL ((AGENT, "resource_event_log_clear found row to delete\n"));
		CONTAINER_REMOVE(cb.container, res_evt_ctx);
		saHpiResourceEventLogTable_delete_row(res_evt_ctx);
		
		/* Reset the entry id for this domain/resource pair */
		dr_pair.domainId_resourceId_arry[0] = saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len];
		dr_pair.domainId_resourceId_arry[1] = saHpiEventLogRowPointer[saHpiResourceEventLogTable_oid_len + column_len + 1];
		dr_entry = domain_resoruce_pair_lookup(&dr_pair, &dr_table); 
		
		if (dr_entry == NULL) {
			DEBUGMSGTL ((AGENT, 
			"ERROR: resource_event_log_clear() domain_resoruce_pair_lookup returned NULL\n"));
			return AGENT_ERR_INTERNAL_ERROR;
		}
	
		DEBUGMSGTL ((AGENT, 
			"resource_event_log_clear() resetting entry_id to 0\n"));
		dr_entry->entry_id = 0;		
		
		resource_event_log_entry_count = CONTAINER_SIZE (cb.container);	

		if (modifyTotal == MIB_TRUE) {
			resource_event_log_entry_count_total--;
		}	
	}	
		
        return SA_OK;

}


					
/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiResourceEventLogLifetimeChanges(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */
 
        DEBUGMSGTL ((AGENT, "handle_saHpiResourceEventLogLifetimeChanges, called\n"));
       
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_COUNTER,
        			(u_char *) &resource_event_log_entry_count_total,
        	                sizeof(resource_event_log_entry_count_total));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
} 
 
/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiResourceEventLogActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiResourceEventLogActiveEntries, called\n"));

        resource_event_log_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
        	snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			(u_char *) &resource_event_log_entry_count,
        			sizeof(resource_event_log_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @return: 
 */
int initialize_table_saHpiResourceEventLogLifetimeChanges(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiResourceEventLogLifetimeChanges, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiResourceEventLogLifetimeChanges", 
				        handle_saHpiResourceEventLogLifetimeChanges,
                                        saHpiResourceEventLogLifetimeChanges_oid,
					OID_LENGTH(saHpiResourceEventLogLifetimeChanges_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
} 
 
/**
 * 
 * @return: 
 */
int initialize_table_saHpiResourceEventLogActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiResourceEventLogActiveEntries, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiResourceEventLogActiveEntries",
					 handle_saHpiResourceEventLogActiveEntries,
                                         saHpiResourceEventLogActiveEntries_oid, 
					 OID_LENGTH(saHpiResourceEventLogActiveEntries_oid),
                                         HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiResourceEventLogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiResourceEventLogTable_cmp( const void *lhs, const void *rhs )
{
        saHpiResourceEventLogTable_context *context_l =
        (saHpiResourceEventLogTable_context *)lhs;
        saHpiResourceEventLogTable_context *context_r =
        (saHpiResourceEventLogTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiResourceEventLogTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiEventSeverity */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                /* If saHpiEventSeverity index is equal sort by forth index */
                                /* CHECK FORTH INDEX,  saHpiResourceEventLogEntryId */
                                if ( context_l->index.oids[3] < context_r->index.oids[3])
                                        return -1;

                                if ( context_l->index.oids[3] > context_r->index.oids[3])
                                        return 1;

                                if ( context_l->index.oids[3] == context_r->index.oids[3])
                                        return 0;
                        }
                }
        }

        return 0;
}


/************************************************************
 * Initializes the saHpiResourceEventLogTable module
 */
void
init_saHpiResourceEventLogTable(void)
{
        
	DEBUGMSGTL ((AGENT, "init_saHpiResourceEventLogTable, called\n"));

        initialize_table_saHpiResourceEventLogTable();

        initialize_table_saHpiResourceEventLogLifetimeChanges();

        initialize_table_saHpiResourceEventLogActiveEntries();     

        domain_resource_pair_initialize(&initialized, &dr_table);
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiResourceEventLogTable_row_copy(saHpiResourceEventLogTable_context * dst,
                         saHpiResourceEventLogTable_context * src)
{
        DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_row_copy, called\n"));

    if(!dst||!src)
        return 1;
        
    subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
        return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiResourceEventLogEntryId = src->saHpiResourceEventLogEntryId;

    memcpy( dst->saHpiResourceEventLogTimestamp,
            src->saHpiResourceEventLogTimestamp,
	    src->saHpiResourceEventLogTimestamp_len);
    
    dst->saHpiResourceEventLogTimestamp_len = src->saHpiResourceEventLogTimestamp_len;
    
    dst->saHpiResourceEventLogType = src->saHpiResourceEventLogType;

    subagent_unlock(&hpi_lock_data);
    return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiResourceEventLogTable_extract_index( saHpiResourceEventLogTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiEventSeverity;
        netsnmp_variable_list var_saHpiResourceEventLogEntryId;
        int err;
        
	DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_extract_index, called\n"));
        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
                        return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
       memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
       var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiDomainId.next_variable = &var_saHpiResourceId;

       memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
       var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiResourceId.next_variable = &var_saHpiEventSeverity;

       memset( &var_saHpiEventSeverity, 0x00, sizeof(var_saHpiEventSeverity) );
       var_saHpiEventSeverity.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiEventSeverity.next_variable = &var_saHpiResourceEventLogEntryId;

       memset( &var_saHpiResourceEventLogEntryId, 0x00, sizeof(var_saHpiResourceEventLogEntryId) );
       var_saHpiResourceEventLogEntryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
       /** TODO: link this index to the next, or NULL for the last one */
       var_saHpiResourceEventLogEntryId.next_variable = NULL;

        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiResourceId */
   
                /** skipping external index saHpiEventSeverity */
    
                ctx->saHpiResourceEventLogEntryId = *var_saHpiResourceEventLogEntryId.val.integer;
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);     
                err = saHpiEventSeverity_check_index(*var_saHpiEventSeverity.val.integer);
		err = saHpiResourceEventLogEntryId_check_index(*var_saHpiResourceEventLogEntryId.val.integer);
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiResourceEventLogTable_can_activate(saHpiResourceEventLogTable_context *undo_ctx,
                      saHpiResourceEventLogTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
	DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_can_activate, called\n"));

    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiResourceEventLogTable_can_deactivate(saHpiResourceEventLogTable_context *undo_ctx,
                        saHpiResourceEventLogTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
	DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_can_deactivate, called\n"));
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiResourceEventLogTable_can_delete(saHpiResourceEventLogTable_context *undo_ctx,
                    saHpiResourceEventLogTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
	DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_can_delete, called\n"));
        subagent_lock(&hpi_lock_data);
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiResourceEventLogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1) {
            subagent_unlock(&hpi_lock_data);
            return 0;

    }
    
    /*
     * TODO: check for other deletion requirements here
     */
    subagent_unlock(&hpi_lock_data);
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiResourceEventLogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiResourceEventLogTable_context *
saHpiResourceEventLogTable_create_row( netsnmp_index* hdr)
{
    saHpiResourceEventLogTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiResourceEventLogTable_context);

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_create_row, called\n"));

    if(!ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiResourceEventLogTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
        return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
    */
    resource_event_log_entry_count_total++;

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiResourceEventLogTable_context *
saHpiResourceEventLogTable_duplicate_row( saHpiResourceEventLogTable_context * row_ctx)
{
    saHpiResourceEventLogTable_context * dup;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_duplicate_row, called\n"));

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);

    dup = SNMP_MALLOC_TYPEDEF(saHpiResourceEventLogTable_context);
    if(!dup) {
            subagent_unlock(&hpi_lock_data);
            return NULL;
    }
        
    if(saHpiResourceEventLogTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiResourceEventLogTable_delete_row( saHpiResourceEventLogTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_delete_row, called\n"));

    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );
    
    resource_event_log_entry_count_total++;
    
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiResourceEventLogTable_set_reserve1( netsnmp_request_group *rg )
{
//    saHpiResourceEventLogTable_context *row_ctx =
//            (saHpiResourceEventLogTable_context *)rg->existing_row;
//    saHpiResourceEventLogTable_context *undo_ctx =
//            (saHpiResourceEventLogTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_set_reserve1, called\n"));

    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiResourceEventLogTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiResourceEventLogTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiResourceEventLogTable_context *row_ctx = (saHpiResourceEventLogTable_context *)rg->existing_row;
//    saHpiResourceEventLogTable_context *undo_ctx = (saHpiResourceEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_set_reserve2, called\n"));

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiResourceEventLogTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiResourceEventLogTable_context *row_ctx = (saHpiResourceEventLogTable_context *)rg->existing_row;
//    saHpiResourceEventLogTable_context *undo_ctx = (saHpiResourceEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    subagent_lock(&hpi_lock_data);

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_set_action, called\n"));

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }
    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiResourceEventLogTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiResourceEventLogTable_context *row_ctx = (saHpiResourceEventLogTable_context *)rg->existing_row;
//    saHpiResourceEventLogTable_context *undo_ctx = (saHpiResourceEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_set_commit, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiResourceEventLogTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiResourceEventLogTable_context *row_ctx = (saHpiResourceEventLogTable_context *)rg->existing_row;
//    saHpiResourceEventLogTable_context *undo_ctx = (saHpiResourceEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_set_free, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: 
                break;
                /** We shouldn't get here */
                /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiResourceEventLogTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiResourceEventLogTable_context *row_ctx = (saHpiResourceEventLogTable_context *)rg->existing_row;
//    saHpiResourceEventLogTable_context *undo_ctx = (saHpiResourceEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_set_undo, called\n"));

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}


/************************************************************
 *
 * Initialize the saHpiResourceEventLogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiResourceEventLogTable(void)
{
    netsnmp_table_registration_info *table_info;

    DEBUGMSGTL ((AGENT, "initialize_table_saHpiResourceEventLogTable, called\n"));

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiResourceEventLogTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiResourceEventLogTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiResourceEventLogTable_oid,
                                             saHpiResourceEventLogTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiResourceEventLogTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiEventSeverity */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiResourceEventLogEntryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiResourceEventLogTable_COL_MIN;
    table_info->max_column = saHpiResourceEventLogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiResourceEventLogTable_get_value;
    cb.container = netsnmp_container_find("saHpiResourceEventLogTable_primary:"
                                          "saHpiResourceEventLogTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiResourceEventLogTable_secondary:"
                                                       "saHpiResourceEventLogTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiResourceEventLogTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiResourceEventLogTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiResourceEventLogTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiResourceEventLogTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiResourceEventLogTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiResourceEventLogTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiResourceEventLogTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiResourceEventLogTable_can_delete;

    cb.set_reserve1 = saHpiResourceEventLogTable_set_reserve1;
    cb.set_reserve2 = saHpiResourceEventLogTable_set_reserve2;
    cb.set_action = saHpiResourceEventLogTable_set_action;
    cb.set_commit = saHpiResourceEventLogTable_set_commit;
    cb.set_free = saHpiResourceEventLogTable_set_free;
    cb.set_undo = saHpiResourceEventLogTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiResourceEventLogTable",
                "Registering table saHpiResourceEventLogTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiResourceEventLogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiResourceEventLogTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiResourceEventLogTable_context *context = (saHpiResourceEventLogTable_context *)item;

    DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_get_value, called\n"));

    switch(table_info->colnum) {

        case COLUMN_SAHPIRESOURCEEVENTLOGENTRYID:
            /** SaHpiEntryId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiResourceEventLogEntryId,
                         sizeof(context->saHpiResourceEventLogEntryId) );
        break;
    
        case COLUMN_SAHPIRESOURCEEVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiResourceEventLogTimestamp,
                         sizeof(context->saHpiResourceEventLogTimestamp) );
        break;
    
        case COLUMN_SAHPIRESOURCEEVENTLOGTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiResourceEventLogType,
                         sizeof(context->saHpiResourceEventLogType) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiResourceEventLogTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiResourceEventLogTable_get_by_idx
 */
const saHpiResourceEventLogTable_context *
saHpiResourceEventLogTable_get_by_idx(netsnmp_index * hdr)
{
        DEBUGMSGTL ((AGENT, "saHpiResourceEventLogTable_get_by_idx, called\n"));

    return (const saHpiResourceEventLogTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


