/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiResourceTable.c 5421 2005-11-14 21:02:01Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiResourceTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <saHpiRdrTable.h>
#include <saHpiHotSwapTable.h>
#include <saHpiDomainInfoTable.h>
#include <saHpiDomainAlarmTable.h>

#include <oh_utils.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiResourceTable_oid[] = { saHpiResourceTable_TABLE_OID};
size_t saHpiResourceTable_oid_len = OID_LENGTH(saHpiResourceTable_oid);

/*************************************************************
 * oid declarations scalars
 */
static oid saHpiResourceActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,2,7 };
static u_long resource_entry_count = 0;

/*************************************************************
 * function declarations: OpenHpi
 */
int populate_saHpiResourceTable(SaHpiSessionIdT sessionid)
{       
	SaErrorT                rv;
        int                     new_row = MIB_FALSE;
	SaHpiDomainInfoT        domain_info;    
	SaHpiEntryIdT           EntryId;
	SaHpiRptEntryT          RptEntry;
	oh_big_textbuffer       bigbuf;
	SaHpiResetActionT       ResetAction;
	SaHpiPowerStateT        State;

	SaHpiUint16T    rs_cap;
	SaHpiUint8T     hs_cap;

	oid resource_oid[RESOURCE_INDEX_NR];
	netsnmp_index resource_index;
	saHpiResourceTable_context *resource_context;

	DEBUGMSGTL ((AGENT, "populate_saHpiResourceTable() Called\n"));

	
	subagent_lock(&hpi_lock_data);
	/* Get the DomainInfo structur,  This is how we get theDomainId for this Session */
	rv = saHpiDomainInfoGet(sessionid, &domain_info);
	if (rv != SA_OK) {
		DEBUGMSGTL ((AGENT, 
			     "populate_saHpiResourceTable: saHpiDomainInfoGet() Failed: rv = %d\n",
			     rv));
		
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}

	EntryId = SAHPI_FIRST_ENTRY;
	do {
		rv = saHpiRptEntryGet(sessionid, EntryId, &EntryId, &RptEntry);

		if (rv != SA_OK) {
			DEBUGMSGTL ((AGENT, "saHpiRptEntryGet Failed: rv = %d\n",rv));
			rv =  AGENT_ERR_INTERNAL_ERROR;
			break;
		}

		resource_index.len = RESOURCE_INDEX_NR;
		resource_oid[0] = domain_info.DomainId;
		resource_oid[1] = RptEntry.ResourceId;
		resource_oid[2] = MIB_FALSE;
		resource_index.oids = (oid *) & resource_oid;

		/* See if it exists. */
		resource_context = NULL;
		resource_context = CONTAINER_FIND (cb.container, &resource_index);

		if (!resource_context) {
			// New entry. Add it
			resource_context = 
			saHpiResourceTable_create_row ( &resource_index);
                        new_row = MIB_TRUE;
		}
		if (!resource_context) {
			snmp_log (LOG_ERR, "Not enough memory for a Resource row!");
			rv = AGENT_ERR_INTERNAL_ERROR;
			break;
		}

		/** UNSIGNED32 = ASN_UNSIGNED */
		resource_context->saHpiResourceId = RptEntry.EntryId;

		/** SaHpiEntryId = ASN_UNSIGNED */
		resource_context->saHpiResourceEntryId = RptEntry.ResourceId;

		/** SaHpiEntityPath = ASN_OCTET_STR */
		memset( resource_context->saHpiResourceEntityPath, 
			0, sizeof(oh_big_textbuffer));
		memset(&bigbuf, 0, sizeof(oh_big_textbuffer));        
		rv = oh_decode_entitypath(&RptEntry.ResourceEntity, &bigbuf );
		if (rv != SA_OK) {
			DEBUGMSGTL ((AGENT, 
				     "ERROR: populate_saHpiResourceTable RPT, oh_decode_entitypath() rv = %d\n",rv));
			rv =  AGENT_ERR_INTERNAL_ERROR;
			saHpiResourceTable_delete_row( resource_context );
			break;
		}
		memcpy( resource_context->saHpiResourceEntityPath, 
			bigbuf.Data, 
			bigbuf.DataLength );            
		if ((bigbuf.Data[bigbuf.DataLength - 1] == 0x00) && 
		    (bigbuf.DataType == SAHPI_TL_TYPE_TEXT))
			resource_context->saHpiResourceEntityPath_len = 
			bigbuf.DataLength - 1;
		else
			resource_context->saHpiResourceEntityPath_len = 
			bigbuf.DataLength;

		/** BITS = ASN_OCTET_STR */
                res_cap_map(&rs_cap, &RptEntry);
		memcpy( resource_context->saHpiResourceCapabilities, 
			&rs_cap, sizeof(rs_cap) ); 
		resource_context->saHpiResourceCapabilities_len = sizeof(rs_cap);

		/** BITS = ASN_OCTET_STR */
                hotswap_cap_map(&hs_cap, &RptEntry);
		*resource_context->saHpiResourceHotSwapCapabilities = hs_cap;                                   
		resource_context->saHpiResourceHotSwapCapabilities_len = sizeof(hs_cap);    

		/** SaHpiSeverity = ASN_INTEGER */
		resource_context->saHpiResourceSeverity = 
		RptEntry.ResourceSeverity + 1;

		/** TruthValue = ASN_INTEGER */
		resource_context->saHpiResourceFailed = 
		(RptEntry.ResourceFailed == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

		/** Unsigned8 = ASN_INTEGER */
		resource_context->saHpiResourceInfoResourceRev = 
		RptEntry.ResourceInfo.ResourceRev;

		/** Unsigned8 = ASN_INTEGER */
		resource_context->saHpiResourceInfoSpecificVer =
		RptEntry.ResourceInfo.SpecificVer;

		/** Unsigned8 = ASN_INTEGER */
		resource_context->saHpiResourceInfoDeviceSupport =
		RptEntry.ResourceInfo.DeviceSupport;

		/** SaHpiManufacturerId = ASN_UNSIGNED */
		resource_context->saHpiResourceInfoManufacturerId =
		RptEntry.ResourceInfo.ManufacturerId;

		/** Unsigned16 = ASN_INTEGER */
		resource_context->saHpiResourceInfoProductId =
		RptEntry.ResourceInfo.ProductId;            

		/** Unsigned8 = ASN_INTEGER */
		resource_context->saHpiResourceInfoFirmwareMajorRev =
		RptEntry.ResourceInfo.FirmwareMajorRev;

		/** Unsigned8 = ASN_INTEGER */
		resource_context->saHpiResourceInfoFirmwareMinorRev =
		RptEntry.ResourceInfo.FirmwareMinorRev;

		/** Unsigned8 = ASN_INTEGER */
		resource_context->saHpiResourceInfoAuxFirmwareRev =
		RptEntry.ResourceInfo.AuxFirmwareRev;

		/** SaHpiGuid = ASN_OCTET_STR */
		memset(resource_context->saHpiResourceInfoGuid, 0, sizeof(SaHpiGuidT));
		memcpy( resource_context->saHpiResourceInfoGuid, 
			RptEntry.ResourceInfo.Guid, 
			sizeof(SaHpiGuidT) );
		resource_context->saHpiResourceInfoGuid_len = sizeof(SaHpiGuidT);

		/** SaHpiTextType = ASN_INTEGER */
		resource_context->saHpiResourceTagTextType = 
		RptEntry.ResourceTag.DataType + 1;

		/** SaHpiTextLanguage = ASN_INTEGER */
		resource_context->saHpiResourceTagTextLanguage = 
		RptEntry.ResourceTag.Language + 1;

		/** SaHpiText = ASN_OCTET_STR */
		memset( resource_context->saHpiResourceTag, 
			0, SAHPI_MAX_TEXT_BUFFER_LENGTH);
		memcpy( resource_context->saHpiResourceTag, 
			RptEntry.ResourceTag.Data, 
			RptEntry.ResourceTag.DataLength );                      

		if ((RptEntry.ResourceTag.Data[RptEntry.ResourceTag.DataLength-1] == 0x00) 
		    && (RptEntry.ResourceTag.DataType == SAHPI_TL_TYPE_TEXT))
			resource_context->saHpiResourceTag_len = 
			RptEntry.ResourceTag.DataLength - 1;
		else
			resource_context->saHpiResourceTag_len = 
			RptEntry.ResourceTag.DataLength;   

		DEBUGMSGTL ((AGENT, "populate_saHpiResourceTable RptEntry.ResourceTag.DataLength rv = %d\n", 
                             RptEntry.ResourceTag.DataLength));
		DEBUGMSGTL ((AGENT, "populate_saHpiResourceTable strlen(RptEntry.ResourceTag.Data) rv = %ld\n", 
                             (long)strlen((const char *)RptEntry.ResourceTag.Data))); 


		/** INTEGER = ASN_INTEGER */
		resource_context->saHpiResourceParmControl = 0;	/* undefined */

		/** INTEGER = ASN_INTEGER */
		rv = saHpiResourceResetStateGet(
					       sessionid, 
					       RptEntry.ResourceId,
					       &ResetAction );
		if (rv == SA_ERR_HPI_INVALID_CMD) {
			DEBUGMSGTL ((AGENT, 
				     "ERROR: populate_saHpiResourceTable saHpiResourceResetStateGet() rv = %d\n",
				     rv));
			rv =  AGENT_ERR_INTERNAL_ERROR;
			saHpiResourceTable_delete_row( resource_context );
			break;
		}
		if (rv == SA_ERR_HPI_CAPABILITY ) {
			resource_context->saHpiResourceResetAction = 0;
		} else {
			resource_context->saHpiResourceResetAction = ResetAction + 1;
		}

		/** INTEGER = ASN_INTEGER */
		rv = saHpiResourcePowerStateGet(
					       sessionid, 
					       RptEntry.ResourceId, 
					       &State);                
		if ( rv == SA_ERR_HPI_INVALID_CMD ) {
			DEBUGMSGTL ((AGENT, 
				     "ERROR: populate_saHpiResourceTable saHpiResourcePowerStateGet() rv = %d\n",
				     rv));
			rv =  AGENT_ERR_INTERNAL_ERROR;
			saHpiResourceTable_delete_row( resource_context );
			break;
		}
		if (rv == SA_ERR_HPI_CAPABILITY) {
			resource_context->saHpiResourcePowerAction = 0;         
		} else {
			resource_context->saHpiResourcePowerAction = State + 1;
		}

		/** TruthValue = ASN_INTEGER */
		resource_context->saHpiResourceIsHistorical = MIB_FALSE;                

                if (new_row == MIB_TRUE) 
                        CONTAINER_INSERT (cb.container, resource_context);

		if (RptEntry.ResourceCapabilities & 
                    SAHPI_CAPABILITY_MANAGED_HOTSWAP) {
			 rv = populate_hotswap (sessionid, &RptEntry);
                }
		if (RptEntry.ResourceCapabilities & SAHPI_CAPABILITY_RDR) {
			rv = populate_saHpiRdrTable(sessionid, 
						    &RptEntry,
						    resource_index.oids, 
						    resource_index.len);
		}
		
		new_row = MIB_FALSE;

	} while (EntryId != SAHPI_LAST_ENTRY);

	resource_entry_count = CONTAINER_SIZE (cb.container);

	DEBUGMSGTL ((AGENT, "resource_entry_count = %ld\n", resource_entry_count));

        subagent_unlock(&hpi_lock_data);

	return rv;              
}

SaErrorT async_resource_add(SaHpiSessionIdT sessionid, SaHpiEventT *event, 
                            SaHpiRdrT *rdr_event, SaHpiRptEntryT *rpt_entry_event)
{       
	SaErrorT                rv;
	SaHpiDomainInfoT        domain_info;    
	oh_big_textbuffer       bigbuf;
	SaHpiResetActionT       ResetAction;
	SaHpiPowerStateT        State;
        SaHpiRptEntryT          rpt_entry;

	SaHpiUint16T    rs_cap;
	SaHpiUint8T     hs_cap;

	oid resource_oid[RESOURCE_INDEX_NR];
	netsnmp_index resource_index;
	saHpiResourceTable_context *resource_context;

        int new_row = MIB_FALSE;
        int remove_rpt_entry = MIB_FALSE;

DEBUGMSGTL ((AGENT, "*****************************\n"));
DEBUGMSGTL ((AGENT, "async_resource_add() Called\n"));
DEBUGMSGTL ((AGENT, "*****************************\n"));

	/* Get the DomainInfo structure,  This is how we get theDomainId for this Session */
	rv = saHpiDomainInfoGet(sessionid, &domain_info);
	if (rv != SA_OK) {
		DEBUGMSGTL ((AGENT, 
			     "async_event_resource: saHpiDomainInfoGet() Failed: rv = %d\n",
			     rv));
		return AGENT_ERR_INTERNAL_ERROR;
	}

        /* Get the RPT entry structure,  If the RPT entry does not exist */
        /* within HPI we need to remove it from our NET-SNMP container */
        rv = saHpiRptEntryGetByResourceId (sessionid, event->Source, &rpt_entry);    
        if (rv == SA_OK) {
                remove_rpt_entry = MIB_FALSE;
                DEBUGMSGTL ((AGENT, 
                             "async_event_resource: saHpiRptEntryGetByResourceId Present\n"));
        } else if (rv != SA_OK) {
                remove_rpt_entry = MIB_TRUE;
                DEBUGMSGTL ((AGENT, 
                             "async_event_resource: saHpiRptEntryGetByResourceId NOT Present\n"));
        }

DEBUGMSGTL ((AGENT, "**** sessionid [%d]\n", sessionid));
DEBUGMSGTL ((AGENT, "**** DomainId  [%d]\n", domain_info.DomainId));
DEBUGMSGTL ((AGENT, "**** event->Source [%d]\n", event->Source));
DEBUGMSGTL ((AGENT, "**** MIB_FALSE [%d]\n", MIB_FALSE));

        /* build index */
	resource_index.len = RESOURCE_INDEX_NR;
	resource_oid[0] = domain_info.DomainId;
	resource_oid[1] = rpt_entry.ResourceId;
	resource_oid[2] = MIB_FALSE;
	resource_index.oids = (oid *) & resource_oid;

	/* See if it exists. */
	resource_context = NULL;
	resource_context = CONTAINER_FIND (cb.container, &resource_index);

        if ((!resource_context) && (remove_rpt_entry == MIB_FALSE)) {

		// New entry. Add it
		resource_context = saHpiResourceTable_create_row ( &resource_index);
                new_row = MIB_TRUE;

	        if (!resource_context) {
                        snmp_log (LOG_ERR, "async_event_resource: Not enough memory for a Resource row!");
                        return AGENT_ERR_INTERNAL_ERROR;
                }

                DEBUGMSGTL ((AGENT, "!!!!!!!!!! async_resource_add: NEW ROW !!!!!!!!!!!\n"));

        } else if ((!resource_context) && (remove_rpt_entry == MIB_TRUE)) {

                DEBUGMSGTL ((AGENT, "!!!!!!!!!! async_resource_add: NOT FOUND ROW !!!!!!!!!!!\n"));

                return SA_OK; /* was present and we didn't want it present */

        } else if ((resource_context) && (remove_rpt_entry == MIB_TRUE)) {

                DEBUGMSGTL ((AGENT, "!!!!!!!!!! async_resource_add: FOUND ROW DELETING!!!!!!!!!!!\n"));

                saHpiResourceTable_delete_row( resource_context );

                return SA_OK;

        } else if ((resource_context) && (remove_rpt_entry == MIB_FALSE)) {

                DEBUGMSGTL ((AGENT, "!!!!!!!!!! async_resource_add: FOUND ROW UPDATING!!!!!!!!!!!\n"));

        }

        /* If you want to use more timely data use the rpt data returned */
        /* by saHpiEventGet() */
        if (rpt_entry_event->ResourceCapabilities != 0x00) {
                memset(&rpt_entry, 0, sizeof(SaHpiRptEntryT));
                memcpy(&rpt_entry, rpt_entry_event, sizeof(SaHpiRptEntryT));
        }

	/** UNSIGNED32 = ASN_UNSIGNED */
	resource_context->saHpiResourceId = rpt_entry.EntryId;

	/** SaHpiEntryId = ASN_UNSIGNED */
	resource_context->saHpiResourceEntryId = rpt_entry.ResourceId;

	/** SaHpiEntityPath = ASN_OCTET_STR */
	memset( resource_context->saHpiResourceEntityPath, 
		0, sizeof(oh_big_textbuffer));
	memset(&bigbuf, 0, sizeof(oh_big_textbuffer));        
	rv = oh_decode_entitypath(&rpt_entry.ResourceEntity, &bigbuf);
	if (rv != SA_OK) {
		DEBUGMSGTL ((AGENT, 
			     "ERROR: async_resource_add: RPT, oh_decode_entitypath() rv = %d\n",rv));
		saHpiResourceTable_delete_row( resource_context );
		return AGENT_ERR_INTERNAL_ERROR;
	}
	memcpy( resource_context->saHpiResourceEntityPath, 
		bigbuf.Data, 
		bigbuf.DataLength );            
	if ((bigbuf.Data[bigbuf.DataLength - 1] == 0x00) && 
	    (bigbuf.DataType == SAHPI_TL_TYPE_TEXT))
		resource_context->saHpiResourceEntityPath_len = 
		bigbuf.DataLength - 1;
	else
		resource_context->saHpiResourceEntityPath_len = 
		bigbuf.DataLength;

	/** BITS = ASN_OCTET_STR */
        res_cap_map(&rs_cap, &rpt_entry);
	memcpy( resource_context->saHpiResourceCapabilities, 
		&rs_cap, sizeof(rs_cap) ); 
	resource_context->saHpiResourceCapabilities_len = sizeof(rs_cap);

	/** BITS = ASN_OCTET_STR */     
        hotswap_cap_map(&hs_cap, &rpt_entry);
	*resource_context->saHpiResourceHotSwapCapabilities = hs_cap;                                   
	resource_context->saHpiResourceHotSwapCapabilities_len = sizeof(hs_cap);    

	/** SaHpiSeverity = ASN_INTEGER */
	resource_context->saHpiResourceSeverity = 
	rpt_entry.ResourceSeverity + 1;

	/** TruthValue = ASN_INTEGER */
        switch (event->EventDataUnion.ResourceEvent.ResourceEventType) {
        case SAHPI_RESE_RESOURCE_RESTORED:
                resource_context->saHpiResourceFailed = MIB_FALSE;
                break;
        case SAHPI_RESE_RESOURCE_FAILURE:
                resource_context->saHpiResourceFailed = MIB_TRUE;
                break;
        default:
                break;
        }

	/** Unsigned8 = ASN_INTEGER */
	resource_context->saHpiResourceInfoResourceRev = 
	rpt_entry.ResourceInfo.ResourceRev;

	/** Unsigned8 = ASN_INTEGER */
	resource_context->saHpiResourceInfoSpecificVer =
	rpt_entry.ResourceInfo.SpecificVer;

	/** Unsigned8 = ASN_INTEGER */
	resource_context->saHpiResourceInfoDeviceSupport =
	rpt_entry.ResourceInfo.DeviceSupport;

	/** SaHpiManufacturerId = ASN_UNSIGNED */
	resource_context->saHpiResourceInfoManufacturerId =
	rpt_entry.ResourceInfo.ManufacturerId;

	/** Unsigned16 = ASN_INTEGER */
	resource_context->saHpiResourceInfoProductId =
	rpt_entry.ResourceInfo.ProductId;            

	/** Unsigned8 = ASN_INTEGER */
	resource_context->saHpiResourceInfoFirmwareMajorRev =
	rpt_entry.ResourceInfo.FirmwareMajorRev;

	/** Unsigned8 = ASN_INTEGER */
	resource_context->saHpiResourceInfoFirmwareMinorRev =
	rpt_entry.ResourceInfo.FirmwareMinorRev;

	/** Unsigned8 = ASN_INTEGER */
	resource_context->saHpiResourceInfoAuxFirmwareRev =
	rpt_entry.ResourceInfo.AuxFirmwareRev;

	/** SaHpiGuid = ASN_OCTET_STR */
	memset(resource_context->saHpiResourceInfoGuid, 0, sizeof(SaHpiGuidT));
	memcpy( resource_context->saHpiResourceInfoGuid, 
		rpt_entry.ResourceInfo.Guid, 
		sizeof(SaHpiGuidT) );
	resource_context->saHpiResourceInfoGuid_len = sizeof(SaHpiGuidT);

	/** SaHpiTextType = ASN_INTEGER */
	resource_context->saHpiResourceTagTextType = 
	rpt_entry.ResourceTag.DataType + 1;

	/** SaHpiTextLanguage = ASN_INTEGER */
	resource_context->saHpiResourceTagTextLanguage = 
	rpt_entry.ResourceTag.Language + 1;

	/** SaHpiText = ASN_OCTET_STR */
	memset( resource_context->saHpiResourceTag, 
		0, SAHPI_MAX_TEXT_BUFFER_LENGTH);
	memcpy( resource_context->saHpiResourceTag, 
		rpt_entry.ResourceTag.Data, 
		rpt_entry.ResourceTag.DataLength );                      

	if ((rpt_entry.ResourceTag.Data[rpt_entry.ResourceTag.DataLength-1] == 0x00) 
	    && (rpt_entry.ResourceTag.DataType == SAHPI_TL_TYPE_TEXT))
		resource_context->saHpiResourceTag_len = 
		rpt_entry.ResourceTag.DataLength - 1;
	else
		resource_context->saHpiResourceTag_len = 
		rpt_entry.ResourceTag.DataLength;   

	/** INTEGER = ASN_INTEGER */
	resource_context->saHpiResourceParmControl = 0;	/* undefined */

	/** INTEGER = ASN_INTEGER */
	rv = saHpiResourceResetStateGet(sessionid, 
                                        rpt_entry.ResourceId,
                                        &ResetAction );
	if (rv == SA_ERR_HPI_INVALID_CMD) {
		DEBUGMSGTL ((AGENT, 
			     "ERROR: async_resource_add:"
                             " saHpiResourceResetStateGet() rv = %d\n",
			     rv));
		saHpiResourceTable_delete_row( resource_context );
		return AGENT_ERR_INTERNAL_ERROR;
	}
	if (rv == SA_ERR_HPI_CAPABILITY ) {
		resource_context->saHpiResourceResetAction = 0;
	} else {
		resource_context->saHpiResourceResetAction = ResetAction + 1;
	}

	/** INTEGER = ASN_INTEGER */
	rv = saHpiResourcePowerStateGet(sessionid, 
                                        rpt_entry.ResourceId, 
                                        &State);                
	if ( rv == SA_ERR_HPI_INVALID_CMD ) {
		DEBUGMSGTL ((AGENT, 
                             "ERROR: async_resource_add: "
                             "saHpiResourcePowerStateGet() rv = %d\n",
			     rv));
		saHpiResourceTable_delete_row( resource_context );
		return AGENT_ERR_INTERNAL_ERROR;
	}
	if (rv == SA_ERR_HPI_CAPABILITY) {
		resource_context->saHpiResourcePowerAction = 0;         
	} else {
		resource_context->saHpiResourcePowerAction = State + 1;
	}

	/** TruthValue = ASN_INTEGER */
	resource_context->saHpiResourceIsHistorical = MIB_FALSE;  

        /***************************/
        /* update applicable RDR's */
        /***************************/
	if ((rpt_entry.ResourceCapabilities & SAHPI_CAPABILITY_RDR) && 
            (event->EventDataUnion.ResourceEvent.ResourceEventType != 
             SAHPI_RESE_RESOURCE_FAILURE)) {

                rv = populate_saHpiRdrTable(sessionid, 
                                            rpt_entry_event,
                                            resource_index.oids, 
                                            resource_index.len);

	}
	
	if ((rpt_entry.ResourceCapabilities & SAHPI_CAPABILITY_MANAGED_HOTSWAP) &&
	    (new_row == MIB_TRUE)) {
		
		rv = populate_hotswap (sessionid, &rpt_entry);
        }

        /*********************************************************/
        /* need to update DomainInfoTable,  RPT data has changed */
        /*********************************************************/
        rv = populate_saHpiDomainInfoTable(sessionid);

        /*********************************************************/
        /* need to update DomainAlarmTable,  RPT data has changed*/
        /*********************************************************/
        //TODO DMJ: this could be cleaner, calling this fucntion will 
        // loop thru the entire table for the given sessionid, we could
        // check and call only when a new entry is expected instead
        rv = populate_saHpiDomainAlarmTable(sessionid);

        if (new_row == MIB_TRUE) 
                CONTAINER_INSERT (cb.container, resource_context);
        

	resource_entry_count = CONTAINER_SIZE (cb.container);


	DEBUGMSGTL ((AGENT, "async_event_resource: resource_entry_count = %ld\n", 
                     resource_entry_count));

	return rv;              
}

/*
 * int set_table_severity (saHpiTable_context * ctx)
 */
int set_table_severity (saHpiResourceTable_context *row_ctx)
{
	SaErrorT                        rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiSeverityT      severity;

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->saHpiResourceEntryId;
	severity = row_ctx->saHpiResourceSeverity - 1;

	rc = saHpiResourceSeveritySet(session_id, resource_id, severity);            

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "Call to saHpiResourceSeverity failed with return code: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "Call to saHpiResourceSeveritySet failed with rc: %s.\n",
			     oh_lookup_error(rc)));
		return AGENT_ERR_OPERATION;
	}

	return AGENT_ERR_NOERROR; 

}

/*
 * int set_table_resource_parm_control (saHpiTable_context * ctx)
 */
int set_table_resource_parm_control (saHpiResourceTable_context *row_ctx)
{
	SaErrorT                        rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiParmActionT    action;

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->saHpiResourceEntryId;
	action = row_ctx->saHpiResourceParmControl - 1;

	rc = saHpiParmControl (session_id, resource_id, action);

	if ((rc != SA_OK) && (rc != SA_ERR_HPI_CAPABILITY)) {
		snmp_log (LOG_ERR,
			  "Call to saHpiParmControl failed with return code: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "Call to saHpiParmControl failed with rc: %s.\n",
			     oh_lookup_error(rc)));
		return AGENT_ERR_OPERATION;
	} else if (rc == SA_ERR_HPI_CAPABILITY) {
		row_ctx->saHpiResourceParmControl = 4;
	}

	return AGENT_ERR_NOERROR; 

}

/*
 * int set_table_resource_reset_action (saHpiTable_context * ctx)
 */
int set_table_resource_reset_action (saHpiResourceTable_context *row_ctx)
{
	SaErrorT                        rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiResetActionT   reset_action;

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->saHpiResourceEntryId;
	reset_action = row_ctx->saHpiResourceResetAction - 1;

	rc = saHpiResourceResetStateSet (session_id, resource_id, reset_action);

	if ((rc != SA_OK) && (rc != SA_ERR_HPI_CAPABILITY)) {
		snmp_log (LOG_ERR,
			  "Call to saHpiResourceResetStateSet failed with return code: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "Call to saHpiResourceResetStateSet failed with rc: %s.\n",
			     oh_lookup_error(rc)));          
		return AGENT_ERR_OPERATION;
	} else if (rc == SA_ERR_HPI_CAPABILITY) {
		row_ctx->saHpiResourceResetAction = 5;
	}

	return AGENT_ERR_NOERROR; 

}


/*
 * int set_table_resource_power_action (saHpiTable_context * ctx)
 */
int set_table_resource_power_action (saHpiResourceTable_context *row_ctx)
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiPowerStateT    power_action;

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->saHpiResourceEntryId;
	power_action = row_ctx->saHpiResourcePowerAction - 1;   

	rc = saHpiResourcePowerStateSet (session_id, resource_id, power_action); 

	if ((rc != SA_OK) && (rc != SA_ERR_HPI_CAPABILITY)) {
		snmp_log (LOG_ERR,
			  "Call to saHpiResourcePowerStateSet failed with return code: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "Call to saHpiResourcePowerStateSet failed with rc: %s.\n",
			     oh_lookup_error(rc)));
		return AGENT_ERR_OPERATION;
	} else if (rc == SA_ERR_HPI_CAPABILITY) {
		row_ctx->saHpiResourcePowerAction = 4;
	}

	return AGENT_ERR_NOERROR; 

}


/*
 * int set_table_resource_tag (saHpiResourceTable_context *row_ctx)
 */
int set_table_resource_tag (saHpiResourceTable_context *row_ctx)
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
	SaHpiTextBufferT    buffer;

        DEBUGMSGTL ((AGENT, "saHpiResourceTable.c -> set_table_resource_tag_text_type called\n"));

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

	    
	session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_INDEX]);
	resource_id = row_ctx->saHpiResourceEntryId; 
	
	// copy the data into a TextBufferT structure
	buffer.DataType   = row_ctx->saHpiResourceTagTextType - 1;
	buffer.Language   = row_ctx->saHpiResourceTagTextLanguage - 1;
	buffer.DataLength = row_ctx->saHpiResourceTag_len;
	memcpy(buffer.Data, row_ctx->saHpiResourceTag, row_ctx->saHpiResourceTag_len);

	rc = saHpiResourceTagSet (session_id, resource_id, &buffer); 

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			    "Call to saHpiResourceTagSet failed with return code: %s.\n",
			    oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "Call to saHpiResourceTagSet failed with rc: %s.\n",
			     oh_lookup_error(rc)));
		return AGENT_ERR_OPERATION;

        }
		
	return AGENT_ERR_NOERROR; 
}	


/*
 *  int handle_saHpiResourceActiveEntries()
 */
int handle_saHpiResourceActiveEntries(netsnmp_mib_handler *handler,
				   netsnmp_handler_registration *reginfo,
				   netsnmp_agent_request_info   *reqinfo,
				   netsnmp_request_info         *requests)
{

	DEBUGMSGTL ((AGENT,     
		     "handle_saHpiResourceActiveEntries: Entry Count is %ld\n", 
		     resource_entry_count));  

	/* We are never called for a GETNEXT if it's registered as a
	   "instance", as it's "magically" handled for us.  */

	/* a instance handler also only hands us one request at a time, so
	   we don't need to loop over a list of requests; we'll only get one. */

        resource_entry_count = CONTAINER_SIZE (cb.container);

	switch (reqinfo->mode) {
	
	case MODE_GET:
		snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
					 (u_char *) &resource_entry_count,
					 sizeof(resource_entry_count));
		break;

	default:
		/* we should never get here, so this is a really bad error */
		return SNMP_ERR_GENERR;
	}

	return SNMP_ERR_NOERROR;
}

/*
 * int initialize_table_saHpiResourceActiveEntries()
 */
int initialize_table_saHpiResourceActiveEntries(void)
{
	DEBUGMSGTL ((AGENT, "initialize_table_saHpiResourceActiveEntries, called\n"));     
	netsnmp_register_scalar(
			       netsnmp_create_handler_registration("saHpiResourceActiveEntries", 
								   handle_saHpiResourceActiveEntries,
								   saHpiResourceActiveEntries_oid, 
								   OID_LENGTH(saHpiResourceActiveEntries_oid),
								   HANDLER_CAN_RONLY) );
	return 0;
}


/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiResourceTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiResourceTable_cmp( const void *lhs, const void *rhs )
{
	saHpiResourceTable_context *context_l =
	(saHpiResourceTable_context *)lhs;
	saHpiResourceTable_context *context_r =
	(saHpiResourceTable_context *)rhs;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_cmp, called\n"));        

	/*
	 * check primary key, then secondary. Add your own code if
	 * there are more than 2 indexes
	 */

	/* check for NULL pointers */
	if (lhs == NULL || rhs == NULL ) {
		DEBUGMSGTL((AGENT,"saHpiResourceTable_cmp() NULL pointer ERROR\n" ));
		return 0;
	}
	/* CHECK FIRST INDEX,  saHpiDomainId */
	if ( context_l->index.oids[0] < context_r->index.oids[0])
		return -1;

	if ( context_l->index.oids[0] > context_r->index.oids[0])
		return 1;

	if ( context_l->index.oids[0] == context_r->index.oids[0]) {
		/* If saHpiDomainId index is equal sort by second index */
		/* CHECK FIRST INDEX,  saHpiResourceEntryId */
		if ( context_l->index.oids[1] < context_r->index.oids[1])
			return -1;

		if ( context_l->index.oids[1] > context_r->index.oids[1])
			return 1;

		if ( context_l->index.oids[1] == context_r->index.oids[1]) {
			/* If saHpiResourceEntryId index is equal sort by second index */
			/* CHECK FIRST INDEX,  saHpiResourceIsHistorical */
			if ( context_l->index.oids[2] < context_r->index.oids[2])
				return -1;

			if ( context_l->index.oids[2] > context_r->index.oids[2])
				return 1;

			if ( context_l->index.oids[2] == context_r->index.oids[2])
				return 0;
		}
	}

	return 0;
}

/************************************************************
 * Initializes the saHpiResourceTable module
 */
void
init_saHpiResourceTable(void)
{
	DEBUGMSGTL ((AGENT, "init_saHpiResourceTable, called\n"));

	initialize_table_saHpiResourceTable();

	initialize_table_saHpiResourceActiveEntries();
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiResourceTable_row_copy(saHpiResourceTable_context * dst,
				       saHpiResourceTable_context * src)
{

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_row_copy, called\n"));

	if (!dst||!src)
		return 1;

        subagent_lock(&hpi_lock_data);
	/*
	 * copy index, if provided
	 */
	if (dst->index.oids)
		free(dst->index.oids);
	if (snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
			    src->index.len * sizeof(oid) )) {
		dst->index.oids = NULL;
                subagent_unlock(&hpi_lock_data);
		return 1;
	}
	dst->index.len = src->index.len;


	/*
	 * copy components into the context structure
	 */
	/** TODO: add code for external index(s)! */
	dst->saHpiResourceId = src->saHpiResourceId;

	dst->saHpiResourceEntryId = src->saHpiResourceEntryId;

	memcpy( dst->saHpiResourceEntityPath, src->saHpiResourceEntityPath, src->saHpiResourceEntityPath_len );
	dst->saHpiResourceEntityPath_len = src->saHpiResourceEntityPath_len;

	memcpy( dst->saHpiResourceCapabilities, src->saHpiResourceCapabilities, src->saHpiResourceCapabilities_len );
	dst->saHpiResourceCapabilities_len = src->saHpiResourceCapabilities_len;

	memcpy( dst->saHpiResourceHotSwapCapabilities, src->saHpiResourceHotSwapCapabilities, src->saHpiResourceHotSwapCapabilities_len );
	dst->saHpiResourceHotSwapCapabilities_len = src->saHpiResourceHotSwapCapabilities_len;

	dst->saHpiResourceSeverity = src->saHpiResourceSeverity;

	dst->saHpiResourceFailed = src->saHpiResourceFailed;

	dst->saHpiResourceInfoResourceRev = src->saHpiResourceInfoResourceRev;

	dst->saHpiResourceInfoSpecificVer = src->saHpiResourceInfoSpecificVer;

	dst->saHpiResourceInfoDeviceSupport = src->saHpiResourceInfoDeviceSupport;

	dst->saHpiResourceInfoManufacturerId = src->saHpiResourceInfoManufacturerId;

	dst->saHpiResourceInfoProductId = src->saHpiResourceInfoProductId;

	dst->saHpiResourceInfoFirmwareMajorRev = src->saHpiResourceInfoFirmwareMajorRev;

	dst->saHpiResourceInfoFirmwareMinorRev = src->saHpiResourceInfoFirmwareMinorRev;

	dst->saHpiResourceInfoAuxFirmwareRev = src->saHpiResourceInfoAuxFirmwareRev;

	memcpy( dst->saHpiResourceInfoGuid, src->saHpiResourceInfoGuid, src->saHpiResourceInfoGuid_len );
	dst->saHpiResourceInfoGuid_len = src->saHpiResourceInfoGuid_len;

	dst->saHpiResourceTagTextType = src->saHpiResourceTagTextType;

	dst->saHpiResourceTagTextLanguage = src->saHpiResourceTagTextLanguage;

	memcpy( dst->saHpiResourceTag, src->saHpiResourceTag, src->saHpiResourceTag_len );
	dst->saHpiResourceTag_len = src->saHpiResourceTag_len;

	dst->saHpiResourceParmControl = src->saHpiResourceParmControl;

	dst->saHpiResourceResetAction = src->saHpiResourceResetAction;

	dst->saHpiResourcePowerAction = src->saHpiResourcePowerAction;

	dst->saHpiResourceIsHistorical = src->saHpiResourceIsHistorical;

        subagent_unlock(&hpi_lock_data);
	return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiResourceTable_extract_index( saHpiResourceTable_context * ctx, netsnmp_index * hdr )
{
	/*
	 * temporary local storage for extracting oid index
	 *
	 * extract index uses varbinds (netsnmp_variable_list) to parse
	 * the index OID into the individual components for each index part.
	 */
	netsnmp_variable_list var_saHpiDomainId;
	netsnmp_variable_list var_saHpiResourceEntryId;
	netsnmp_variable_list var_saHpiResourceIsHistorical;
	int err;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_extract_index, called\n"));    
        subagent_lock(&hpi_lock_data);

	/*
	 * copy index, if provided
	 */
	if (hdr) {
		netsnmp_assert(ctx->index.oids == NULL);
		if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
				    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
			return -1;
		}
		ctx->index.len = hdr->len;
	}

	/*
	 * initialize variable that will hold each component of the index.
	 * If there are multiple indexes for the table, the variable_lists
	 * need to be linked together, in order.
	 */
	memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
	var_saHpiDomainId.type = ASN_UNSIGNED; 
	var_saHpiDomainId.next_variable = &var_saHpiResourceEntryId;


	memset( &var_saHpiResourceEntryId, 0x00, sizeof(var_saHpiResourceEntryId) );
	var_saHpiResourceEntryId.type = ASN_UNSIGNED; 
	var_saHpiResourceEntryId.next_variable = &var_saHpiResourceIsHistorical;


	memset( &var_saHpiResourceIsHistorical, 0x00, sizeof(var_saHpiResourceIsHistorical) );
	var_saHpiResourceIsHistorical.type = ASN_INTEGER; 
	var_saHpiResourceIsHistorical.next_variable = NULL;

	/*
	 * parse the oid into the individual index components
	 */
	err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
	if (err == SNMP_ERR_NOERROR) {

		/*
		 * copy index components into the context structure
		 */
		/** skipping external index saHpiDomainId */

		ctx->saHpiResourceEntryId = 
		*var_saHpiResourceEntryId.val.integer;

		ctx->saHpiResourceIsHistorical = 
		*var_saHpiResourceIsHistorical.val.integer;

		/*
		 *Check Index 
	 */
		err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
		err = saHpiResourceEntryId_check_index(*var_saHpiResourceEntryId.val.integer);  
		err = saHpiResourceIsHistorical_check_index(*var_saHpiResourceIsHistorical.val.integer);
	}

	/*
	 * parsing may have allocated memory. free it.
	 */
	snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
	return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiResourceTable_can_activate(saHpiResourceTable_context *undo_ctx,
				    saHpiResourceTable_context *row_ctx,
				    netsnmp_request_group * rg)
{

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_can_activate, called\n"));

	/*
	 * TODO: check for activation requirements here
	 */


	/*
	 * be optimistic.
	 */
	return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiResourceTable_can_deactivate(saHpiResourceTable_context *undo_ctx,
				      saHpiResourceTable_context *row_ctx,
				      netsnmp_request_group * rg)
{
	DEBUGMSGTL ((AGENT, "saHpiResourceTable_can_deactivate, called\n"));
	/*
	 * TODO: check for deactivation requirements here
	 */
	return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiResourceTable_can_delete(saHpiResourceTable_context *undo_ctx,
				  saHpiResourceTable_context *row_ctx,
				  netsnmp_request_group * rg)
{

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_can_delete, called\n"));
        subagent_lock(&hpi_lock_data);

	/*
	 * probably shouldn't delete a row that we can't
	 * deactivate.
	 */
	if (saHpiResourceTable_can_deactivate(undo_ctx,row_ctx,rg) != 1) {
                subagent_unlock(&hpi_lock_data);
		return 0;
        }

	/*
	 * TODO: check for other deletion requirements here
	 */
        subagent_unlock(&hpi_lock_data);
	return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiResourceTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiResourceTable_context *
saHpiResourceTable_create_row( netsnmp_index* hdr)
{

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_create_row, called\n"));
        subagent_lock(&hpi_lock_data);

	saHpiResourceTable_context * ctx =
	SNMP_MALLOC_TYPEDEF(saHpiResourceTable_context);
	if (!ctx) {
                subagent_unlock(&hpi_lock_data);
		return NULL;
        }

	/*
	 * TODO: check indexes, if necessary.
	 */
	if (saHpiResourceTable_extract_index( ctx, hdr )) {
		free(ctx->index.oids);
		free(ctx);
                subagent_unlock(&hpi_lock_data);
		return NULL;
	}

	/* netsnmp_mutex_init(ctx->lock);
	   netsnmp_mutex_lock(ctx->lock); */

	/*
	 * TODO: initialize any default values here. This is also
	 * first place you really should allocate any memory for
	 * yourself to use.  If you allocated memory earlier,
	 * make sure you free it for earlier error cases!
	 */
	/**
	 ctx->saHpiResourceSeverity = 0;
	 ctx->saHpiResourceTagTextType = 0;
	 ctx->saHpiResourceTagTextLanguage = 0;
	 ctx->saHpiResourceTag = 0;
	 ctx->saHpiResourceParmControl = 0;
	 ctx->saHpiResourceResetAction = 0;
	 ctx->saHpiResourcePowerAction = 0;
	*/

        subagent_unlock(&hpi_lock_data);
	return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiResourceTable_context *
saHpiResourceTable_duplicate_row( saHpiResourceTable_context * row_ctx)
{
	saHpiResourceTable_context * dup;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_duplicate_row, called\n"));

	if (!row_ctx)
		return NULL;

        subagent_lock(&hpi_lock_data);
	dup = SNMP_MALLOC_TYPEDEF(saHpiResourceTable_context);
	if (!dup) {
                subagent_unlock(&hpi_lock_data);
		return NULL;
        }

	if (saHpiResourceTable_row_copy(dup,row_ctx)) {
		free(dup);
		dup = NULL;
	}

        subagent_unlock(&hpi_lock_data);
	return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiResourceTable_delete_row( saHpiResourceTable_context * ctx )
{

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_delete_row, called\n"));
        subagent_lock(&hpi_lock_data);

	/* netsnmp_mutex_destroy(ctx->lock); */

	if (ctx->index.oids)
		free(ctx->index.oids);

	/*
	 * TODO: release any memory you allocated here...
	 */

	/*
	 * release header
	 */
	free( ctx );
        subagent_unlock(&hpi_lock_data);

	return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiResourceTable_set_reserve1( netsnmp_request_group *rg )
{
	saHpiResourceTable_context *row_ctx =
	(saHpiResourceTable_context *)rg->existing_row;
//	saHpiResourceTable_context *undo_ctx =
//	(saHpiResourceTable_context *)rg->undo_info;
	netsnmp_variable_list *var;
	netsnmp_request_group_item *current;
	int rc;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_set_reserve1, called\n"));

	/*
	 * TODO: loop through columns, check syntax and lengths. For
	 * columns which have no dependencies, you could also move
	 * the value/range checking here to attempt to catch error
	 * cases as early as possible.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;
		rc = SNMP_ERR_NOERROR;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPIRESOURCESEVERITY:
			/** SaHpiSeverity = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiResourceSeverity));
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
			/** SaHpiTextType = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiResourceTagTextType));
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
			/** SaHpiTextLanguage = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiResourceTagTextLanguage));
			break;

		case COLUMN_SAHPIRESOURCETAG:          
			/** SaHpiText = ASN_OCTET_STR */  
			rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
			if (rc == SNMP_ERR_NOERROR ) {
				if (var->val_len > SAHPI_MAX_TEXT_BUFFER_LENGTH) {
					rc = SNMP_ERR_WRONGLENGTH;
				}
			}
			if (rc == SNMP_ERR_NOERROR)
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCETAG NO ERROR: %d\n", rc));
			else
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCETAG ERROR: %d\n", rc));                          
			break;

		case COLUMN_SAHPIRESOURCEPARMCONTROL:
			/** INTEGER = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiResourceParmControl));
			break;

		case COLUMN_SAHPIRESOURCERESETACTION:
			/** INTEGER = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiResourceResetAction));
			break;

		case COLUMN_SAHPIRESOURCEPOWERACTION:
			/** INTEGER = ASN_INTEGER */
			rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
							    sizeof(row_ctx->saHpiResourcePowerAction));
			break;

		default: /** We shouldn't get here */
			rc = SNMP_ERR_GENERR;
			snmp_log(LOG_ERR, "unknown column in "
				 "saHpiResourceTable_set_reserve1\n");
		}

		if (rc)
			netsnmp_request_set_error( current->ri, rc );
		rg->status = SNMP_MAX( rg->status, current->ri->status );
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

void saHpiResourceTable_set_reserve2( netsnmp_request_group *rg )
{
//	saHpiResourceTable_context *row_ctx = (saHpiResourceTable_context *)rg->existing_row;
//	saHpiResourceTable_context *undo_ctx = (saHpiResourceTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;
	netsnmp_variable_list *var;
	int rc;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_set_reserve2, called\n"));

	rg->rg_void = rg->list->ri;

	/*
	 * TODO: loop through columns, check for valid
	 * values and any range constraints.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;
		rc = SNMP_ERR_NOERROR;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPIRESOURCESEVERITY:
			if (oh_lookup_severity(*var->val.integer - 1) == NULL) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCESEVERITY: SNMP_ERR_BADVALUE\n"));                  
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
			/** SaHpiTextType = ASN_INTEGER */
			if (oh_lookup_texttype(*var->val.integer - 1) == NULL) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCETAGTEXTTYPE: SNMP_ERR_BADVALUE\n"));               
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
			/** SaHpiTextLanguage = ASN_INTEGER */
			if ( oh_lookup_language(*var->val.integer - 1) == NULL ) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE: SNMP_ERR_BADVALUE\n"));                                
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPIRESOURCETAG:
			/** SaHpiText = ASN_OCTET_STR */
			if ( var->val_len > SAHPI_MAX_TEXT_BUFFER_LENGTH  ) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCETAG: SNMP_ERR_BADVALUE\n"));                               
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPIRESOURCEPARMCONTROL:
			if (oh_lookup_parmaction(*var->val.integer - 1) == NULL) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCEPARMCONTROL: SNMP_ERR_BADVALUE\n"));               
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPIRESOURCERESETACTION:
			if (oh_lookup_resetaction(*var->val.integer - 1) == NULL) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCERESETACTION: SNMP_ERR_BADVALUE\n"));               
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		case COLUMN_SAHPIRESOURCEPOWERACTION:
			if (oh_lookup_powerstate(*var->val.integer - 1) == NULL) {
				DEBUGMSGTL ((AGENT, 
					     "COLUMN_SAHPIRESOURCEPOWERACTION: SNMP_ERR_BADVALUE\n"));               
				rc = SNMP_ERR_BADVALUE;
			}
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}

		if (rc)
			netsnmp_request_set_error( current->ri, rc);
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiResourceTable_set_action( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
	saHpiResourceTable_context *row_ctx = (saHpiResourceTable_context *)rg->existing_row;
//	saHpiResourceTable_context *undo_ctx = (saHpiResourceTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	int row_err = 0;
        subagent_lock(&hpi_lock_data);

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_set_action, called\n"));

	/*
	 * TODO: loop through columns, copy varbind values
	 * to context structure for the row.
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPIRESOURCESEVERITY:
			/** SaHpiSeverity = ASN_INTEGER */ 
			row_ctx->saHpiResourceSeverity = *var->val.integer;
			if (set_table_severity (row_ctx) != AGENT_ERR_NOERROR)
				row_err = SNMP_ERR_GENERR;
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
			/** SaHpiTextType = ASN_INTEGER */
			row_ctx->saHpiResourceTagTextType = *var->val.integer;			
			row_err = set_table_resource_tag (row_ctx);
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
			/** SaHpiTextLanguage = ASN_INTEGER */
			row_ctx->saHpiResourceTagTextLanguage = *var->val.integer;
			row_err = set_table_resource_tag (row_ctx);
			break;

		case COLUMN_SAHPIRESOURCETAG:
			/** SaHpiText = ASN_OCTET_STR */
			memcpy(row_ctx->saHpiResourceTag,var->val.string,var->val_len);
			row_ctx->saHpiResourceTag_len = var->val_len;
			row_err = set_table_resource_tag (row_ctx);
			break;

		case COLUMN_SAHPIRESOURCEPARMCONTROL:
			/** INTEGER = ASN_INTEGER */
			row_ctx->saHpiResourceParmControl = *var->val.integer;
			if (set_table_resource_parm_control (row_ctx) != AGENT_ERR_NOERROR)
				row_err = SNMP_ERR_GENERR;
			break;

		case COLUMN_SAHPIRESOURCERESETACTION:
			/** INTEGER = ASN_INTEGER */
			row_ctx->saHpiResourceResetAction = *var->val.integer;
			if (set_table_resource_reset_action (row_ctx) != AGENT_ERR_NOERROR)
				row_err = SNMP_ERR_GENERR;
			break;

		case COLUMN_SAHPIRESOURCEPOWERACTION:
			/** INTEGER = ASN_INTEGER */
			row_ctx->saHpiResourcePowerAction = *var->val.integer;
			if (set_table_resource_power_action (row_ctx) != AGENT_ERR_NOERROR)
				row_err = SNMP_ERR_GENERR;
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * check activation/deactivation
	 */
	if (row_err) {
		netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
					       row_err);
	}

        subagent_unlock(&hpi_lock_data);
        return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiResourceTable_set_commit( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiResourceTable_context *row_ctx = (saHpiResourceTable_context *)rg->existing_row;
//	saHpiResourceTable_context *undo_ctx = (saHpiResourceTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_set_commit, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPIRESOURCESEVERITY:
			/** SaHpiSeverity = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
			/** SaHpiTextType = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
			/** SaHpiTextLanguage = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAG:
			/** SaHpiText = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPIRESOURCEPARMCONTROL:
			/** INTEGER = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCERESETACTION:
			/** INTEGER = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCEPOWERACTION:
			/** INTEGER = ASN_INTEGER */
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiResourceTable_set_free( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiResourceTable_context *row_ctx = (saHpiResourceTable_context *)rg->existing_row;
//	saHpiResourceTable_context *undo_ctx = (saHpiResourceTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_set_free, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPIRESOURCESEVERITY:
			/** SaHpiSeverity = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
			/** SaHpiTextType = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
			/** SaHpiTextLanguage = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAG:
			/** SaHpiText = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPIRESOURCEPARMCONTROL:
			/** INTEGER = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCERESETACTION:
			/** INTEGER = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCEPOWERACTION:
			/** INTEGER = ASN_INTEGER */
			break;

		default: 
			break;	/** We shouldn't get here */
				/** should have been logged in reserve1 */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiResourceTable_set_undo( netsnmp_request_group *rg )
{
	netsnmp_variable_list *var;
//	saHpiResourceTable_context *row_ctx = (saHpiResourceTable_context *)rg->existing_row;
//	saHpiResourceTable_context *undo_ctx = (saHpiResourceTable_context *)rg->undo_info;
	netsnmp_request_group_item *current;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_set_undo, called\n"));

	/*
	 * loop through columns
	 */
	for ( current = rg->list; current; current = current->next ) {

		var = current->ri->requestvb;

		switch (current->tri->colnum) {
		
		case COLUMN_SAHPIRESOURCESEVERITY:
			/** SaHpiSeverity = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
			/** SaHpiTextType = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
			/** SaHpiTextLanguage = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCETAG:
			/** SaHpiText = ASN_OCTET_STR */
			break;

		case COLUMN_SAHPIRESOURCEPARMCONTROL:
			/** INTEGER = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCERESETACTION:
			/** INTEGER = ASN_INTEGER */
			break;

		case COLUMN_SAHPIRESOURCEPOWERACTION:
			/** INTEGER = ASN_INTEGER */
			break;

		default: /** We shouldn't get here */
			netsnmp_assert(0); /** why wasn't this caught in reserve1? */
		}
	}

	/*
	 * done with all the columns. Could check row related
	 * requirements here.
	 */
}


/************************************************************
 *
 * Initialize the saHpiResourceTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiResourceTable(void)
{
	netsnmp_table_registration_info *table_info;

	DEBUGMSGTL ((AGENT, "initialize_table_saHpiResourceTable, called\n"));

	if (my_handler) {
		snmp_log(LOG_ERR, "initialize_table_saHpiResourceTable_handler called again\n");
		return;
	}

	memset(&cb, 0x00, sizeof(cb));

	/** create the table structure itself */
	table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

	/* if your table is read only, it's easiest to change the
	   HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
	my_handler = netsnmp_create_handler_registration("saHpiResourceTable",
							 netsnmp_table_array_helper_handler,
							 saHpiResourceTable_oid,
							 saHpiResourceTable_oid_len,
							 HANDLER_CAN_RWRITE);

	if (!my_handler || !table_info) {
		snmp_log(LOG_ERR, "malloc failed in "
			 "initialize_table_saHpiResourceTable_handler\n");
		return;	/** mallocs failed */
	}

	/***************************************************
	 * Setting up the table's definition
	 */
	/*
	 * TODO: add any external indexes here.
	 */
	/** TODO: add code for external index(s)! */

	/*
	 * internal indexes
	 */
	/** index: saHpiDomainId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
	/** index: saHpiResourceEntryId */
	netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
	/** index: saHpiResourceIsHistorical */
	netsnmp_table_helper_add_index(table_info, ASN_INTEGER);

	table_info->min_column = saHpiResourceTable_COL_MIN;
	table_info->max_column = saHpiResourceTable_COL_MAX;

	/***************************************************
	 * registering the table with the master agent
	 */
	cb.get_value = saHpiResourceTable_get_value;
	cb.container = netsnmp_container_find("saHpiResourceTable_primary:"
					      "saHpiResourceTable:"
					      "table_container");

	netsnmp_container_add_index(cb.container,
				    netsnmp_container_find("saHpiResourceTable_secondary:"
							   "saHpiResourceTable:"
							   "table_container"));
	cb.container->next->compare = saHpiResourceTable_cmp;


	cb.can_set = 1;

	cb.create_row = (UserRowMethod*)saHpiResourceTable_create_row;

	cb.duplicate_row = (UserRowMethod*)saHpiResourceTable_duplicate_row;
	cb.delete_row = (UserRowMethod*)saHpiResourceTable_delete_row;
	cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiResourceTable_row_copy;

	cb.can_activate = (Netsnmp_User_Row_Action *)saHpiResourceTable_can_activate;
	cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiResourceTable_can_deactivate;
	cb.can_delete = (Netsnmp_User_Row_Action *)saHpiResourceTable_can_delete;

	cb.set_reserve1 = saHpiResourceTable_set_reserve1;
	cb.set_reserve2 = saHpiResourceTable_set_reserve2;
	cb.set_action = saHpiResourceTable_set_action;
	cb.set_commit = saHpiResourceTable_set_commit;
	cb.set_free = saHpiResourceTable_set_free;
	cb.set_undo = saHpiResourceTable_set_undo;

	DEBUGMSGTL(("initialize_table_saHpiResourceTable",
		    "Registering table saHpiResourceTable "
		    "as a table array\n"));
	netsnmp_table_container_register(my_handler, table_info, &cb,
					 cb.container, 1);
}

/************************************************************
 * saHpiResourceTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiResourceTable_get_value(
				netsnmp_request_info *request,
				netsnmp_index *item,
				netsnmp_table_request_info *table_info )
{
	netsnmp_variable_list *var = request->requestvb;
	saHpiResourceTable_context *context = (saHpiResourceTable_context *)item;

	DEBUGMSGTL ((AGENT, "saHpiResourceTable_get_value, called\n"));

	switch (table_info->colnum) {
	
	case COLUMN_SAHPIRESOURCEID:
		/** UNSIGNED32 = ASN_UNSIGNED */
		snmp_set_var_typed_value(var, ASN_UNSIGNED,
					 (unsigned char*)&context->saHpiResourceId,
					 sizeof(context->saHpiResourceId) );
		break;

	case COLUMN_SAHPIRESOURCEENTRYID:
		/** SaHpiEntryId = ASN_UNSIGNED */
		snmp_set_var_typed_value(var, ASN_UNSIGNED,
					 (unsigned char*)&context->saHpiResourceEntryId,
					 sizeof(context->saHpiResourceEntryId) );
		break;

	case COLUMN_SAHPIRESOURCEENTITYPATH:
		/** SaHpiEntityPath = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (unsigned char*)&context->saHpiResourceEntityPath,
					 context->saHpiResourceEntityPath_len );
		break;

	case COLUMN_SAHPIRESOURCECAPABILITIES:
		/** BITS = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (unsigned char*)&context->saHpiResourceCapabilities,
					 context->saHpiResourceCapabilities_len );
		break;

	case COLUMN_SAHPIRESOURCEHOTSWAPCAPABILITIES:
		/** BITS = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (unsigned char*)&context->saHpiResourceHotSwapCapabilities,
					 context->saHpiResourceHotSwapCapabilities_len );
		break;

	case COLUMN_SAHPIRESOURCESEVERITY:
		/** SaHpiSeverity = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceSeverity,
					 sizeof(context->saHpiResourceSeverity) );
		break;

	case COLUMN_SAHPIRESOURCEFAILED:
		/** TruthValue = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceFailed,
					 sizeof(context->saHpiResourceFailed) );
		break;

	case COLUMN_SAHPIRESOURCEINFORESOURCEREV:
		/** Unsigned8 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoResourceRev,
					 sizeof(context->saHpiResourceInfoResourceRev) );
		break;

	case COLUMN_SAHPIRESOURCEINFOSPECIFICVER:
		/** Unsigned8 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoSpecificVer,
					 sizeof(context->saHpiResourceInfoSpecificVer) );
		break;

	case COLUMN_SAHPIRESOURCEINFODEVICESUPPORT:
		/** Unsigned8 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoDeviceSupport,
					 sizeof(context->saHpiResourceInfoDeviceSupport) );
		break;

	case COLUMN_SAHPIRESOURCEINFOMANUFACTURERID:
		/** SaHpiManufacturerId = ASN_UNSIGNED */
		snmp_set_var_typed_value(var, ASN_UNSIGNED,
					 (unsigned char*)&context->saHpiResourceInfoManufacturerId,
					 sizeof(context->saHpiResourceInfoManufacturerId) );
		break;

	case COLUMN_SAHPIRESOURCEINFOPRODUCTID:
		/** Unsigned16 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoProductId,
					 sizeof(context->saHpiResourceInfoProductId) );
		break;

	case COLUMN_SAHPIRESOURCEINFOFIRMWAREMAJORREV:
		/** Unsigned8 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoFirmwareMajorRev,
					 sizeof(context->saHpiResourceInfoFirmwareMajorRev) );
		break;

	case COLUMN_SAHPIRESOURCEINFOFIRMWAREMINORREV:
		/** Unsigned8 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoFirmwareMinorRev,
					 sizeof(context->saHpiResourceInfoFirmwareMinorRev) );
		break;

	case COLUMN_SAHPIRESOURCEINFOAUXFIRMWAREREV:
		/** Unsigned8 = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceInfoAuxFirmwareRev,
					 sizeof(context->saHpiResourceInfoAuxFirmwareRev) );
		break;

	case COLUMN_SAHPIRESOURCEINFOGUID:
		/** SaHpiGuid = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (unsigned char*)&context->saHpiResourceInfoGuid,
					 context->saHpiResourceInfoGuid_len );
		break;

	case COLUMN_SAHPIRESOURCETAGTEXTTYPE:
		/** SaHpiTextType = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceTagTextType,
					 sizeof(context->saHpiResourceTagTextType) );
		break;

	case COLUMN_SAHPIRESOURCETAGTEXTLANGUAGE:
		/** SaHpiTextLanguage = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceTagTextLanguage,
					 sizeof(context->saHpiResourceTagTextLanguage) );
		break;

	case COLUMN_SAHPIRESOURCETAG:
		/** SaHpiText = ASN_OCTET_STR */
		snmp_set_var_typed_value(var, ASN_OCTET_STR,
					 (unsigned char*)&context->saHpiResourceTag,
					 context->saHpiResourceTag_len );
		break;

	case COLUMN_SAHPIRESOURCEPARMCONTROL:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceParmControl,
					 sizeof(context->saHpiResourceParmControl) );
		break;

	case COLUMN_SAHPIRESOURCERESETACTION:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceResetAction,
					 sizeof(context->saHpiResourceResetAction) );
		break;

	case COLUMN_SAHPIRESOURCEPOWERACTION:
		/** INTEGER = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourcePowerAction,
					 sizeof(context->saHpiResourcePowerAction) );
		break;

	case COLUMN_SAHPIRESOURCEISHISTORICAL:
		/** TruthValue = ASN_INTEGER */
		snmp_set_var_typed_value(var, ASN_INTEGER,
					 (unsigned char*)&context->saHpiResourceIsHistorical,
					 sizeof(context->saHpiResourceIsHistorical) );
		break;

	default: /** We shouldn't get here */
		snmp_log(LOG_ERR, "unknown column in "
			 "saHpiResourceTable_get_value\n");
		return SNMP_ERR_GENERR;
	}
	return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiResourceTable_get_by_idx
 */
const saHpiResourceTable_context *
saHpiResourceTable_get_by_idx(netsnmp_index * hdr)
{
	DEBUGMSGTL ((AGENT, "saHpiResourceTable_get_by_idx, called\n"));

	return(const saHpiResourceTable_context *)
	CONTAINER_FIND(cb.container, hdr );
}




