/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiSensorEventLogTable.c 5423 2005-11-14 21:12:55Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiSensorEventLogTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiSensorEventLogTable_oid[] = { saHpiSensorEventLogTable_TABLE_OID };
size_t saHpiSensorEventLogTable_oid_len = OID_LENGTH(saHpiSensorEventLogTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;		      
static GHashTable *dr_table;

/*************************************************************
 * oid and function declarations scalars
 */
static u_long sensor_event_log_entry_count_total = 0;
static u_long sensor_event_log_entry_count = 0; 
static oid saHpiSensorEventLogLifetimeChanges_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,9 }; 
static oid saHpiSensorEventLogActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,10 };

int handle_saHpiSensorEventEntryLogCountTotal(netsnmp_mib_handler *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests);
			  
int handle_saHpiSensorEventLogActiveEntries(netsnmp_mib_handler *handler,
                                         netsnmp_handler_registration *reginfo,
                                         netsnmp_agent_request_info   *reqinfo,
                                         netsnmp_request_info         *requests);
					 
int initialize_table_saHpiSensorEventEntryLogCountTotal(void);
int initialize_table_saHpiSensorEventLogActiveEntries(void);


/**
 * 
 * @sessionid:
 * @event:
 * @this_child_oid:
 * @this_child_oid_len:
 * 
 * @return:
 */
SaErrorT populate_saHpiSensorEventLogTable(SaHpiSessionIdT sessionid, 
                                             SaHpiEventLogEntryT *event,
                                             oid * this_child_oid, 
                                             size_t *this_child_oid_len,
                                             SaHpiRdrT *event_rdr_entry)
{
	SaErrorT rv = SA_OK;

	oid sen_evt_oid[SENSOR_EVENT_LOG_INDEX_NR];
	netsnmp_index sen_evt_idx;
	saHpiSensorEventLogTable_context *sen_evt_ctx;

	oid column[2];
	int column_len = 2;
	
	int isNewRow = MIB_TRUE;

        DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

        SaHpiTextBufferT buff;

        DEBUGMSGTL ((AGENT, "populate_saHpiSensorEventLogTable, called\n"));

	/* check for NULL pointers */
	if (!event) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiSensorEventLogTable() passed NULL event pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}    

	
	subagent_lock(&hpi_lock_data);
	/* BUILD oid for new row */
		/* assign the number of indices */
	sen_evt_idx.len = SENSOR_EVENT_LOG_INDEX_NR;
		/** Index saHpiDomainId is external */
	sen_evt_oid[0] = get_domain_id(sessionid);
		/** Index saHpiResourceId is external */
	sen_evt_oid[1] = event->Event.Source;
		/** Index saHpiSensorNum is external */
	sen_evt_oid[2] = event->Event.EventDataUnion.SensorEvent.SensorNum;
		/** Index saHpiEventSeverity is external */
	sen_evt_oid[3] = event->Event.Severity + 1;

                /** Index saHpiSensorEventEntryId is internal */
	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
	dr_pair.domainId_resourceId_arry[1] = event->Event.Source;
	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
	if (dr_entry == NULL) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiSensorEventLogTable() domain_resource_pair_get returned NULL\n"));
		
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}
	sen_evt_oid[4] = dr_entry->entry_id++;
		/* assign the indices to the index */
	sen_evt_idx.oids = (oid *) & sen_evt_oid;

	/* create full oid on This row for parent RowPointer */
	column[0] = 1;
	column[1] = COLUMN_SAHPISENSOREVENTLOGTIMESTAMP;
	memset(this_child_oid, 0, MAX_OID_LEN);
	build_full_oid(saHpiSensorEventLogTable_oid, saHpiSensorEventLogTable_oid_len,
			column, column_len,
			&sen_evt_idx,
			this_child_oid, MAX_OID_LEN, this_child_oid_len);

	/* See if Row exists. */
	sen_evt_ctx = NULL;
	sen_evt_ctx = CONTAINER_FIND(cb.container, &sen_evt_idx);

	if (!sen_evt_ctx) { 
		// New entry. Add it
		sen_evt_ctx = 
			saHpiSensorEventLogTable_create_row(&sen_evt_idx);
	}
	else {
		isNewRow = MIB_FALSE;
	}
	
	if (!sen_evt_ctx) {
		snmp_log (LOG_ERR, "Not enough memory for a Sensor Event Log row!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}


        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event->Timestamp, sen_evt_ctx->saHpiSensorEventLogTimestamp);
        sen_evt_ctx->saHpiSensorEventLogTimestamp_len = sizeof(SaHpiTimeT);
        /** SaHpiSensorType = ASN_INTEGER */
        sen_evt_ctx->saHpiSensorEventLogType = event->Event.EventType + 1;

        /** SaHpiEventCategory = ASN_INTEGER */
        sen_evt_ctx->saHpiSensorEventLogCategory = 
                event->Event.EventDataUnion.SensorEvent.EventCategory;

        /** TruthValue = ASN_INTEGER */
        sen_evt_ctx->saHpiSensorEventLogAssertion = 
            (event->Event.EventDataUnion.SensorEvent.Assertion == SAHPI_TRUE)
		? MIB_TRUE : MIB_FALSE;

        /** SaHpiEventState = ASN_OCTET_STR */
        rv = oh_decode_eventstate(event->Event.EventDataUnion.SensorEvent.EventState,
			      event->Event.EventDataUnion.SensorEvent.EventCategory,
			      &buff);
	if (rv != SA_OK) {
		snmp_log (LOG_ERR, 
                "populate_saHpiResourceEventLogTable: oh_decode_eventstate error!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}
        memcpy(sen_evt_ctx->saHpiSensorEventLogState,
               buff.Data, buff.DataLength);
        sen_evt_ctx->saHpiSensorEventLogState_len = buff.DataLength;

        /** SaHpiOptionalData = ASN_OCTET_STR */
        rv = oh_decode_sensoroptionaldata(
                event->Event.EventDataUnion.SensorEvent.OptionalDataPresent,
		&buff);
	if (rv != SA_OK) {
		snmp_log (LOG_ERR, 
                "populate_saHpiResourceEventLogTable: oh_decode_sensoroptionaldata error!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}
        memcpy(sen_evt_ctx->saHpiSensorEventLogOptionalData, buff.Data,
               buff.DataLength);
        sen_evt_ctx->saHpiSensorEventLogOptionalData_len = buff.DataLength;

        /** SaHpiSensorReadingType = ASN_INTEGER */
        sen_evt_ctx->saHpiSensorEventLogTriggerReadingType =
                event->Event.EventDataUnion.SensorEvent.TriggerReading.Type + 1;

        /** SaHpiSensorReadingValue = ASN_OCTET_STR */
        if (event_rdr_entry) {
                rv =  oh_decode_sensorreading(
                        event->Event.EventDataUnion.SensorEvent.TriggerReading,
                        event_rdr_entry->RdrTypeUnion.SensorRec.DataFormat,
                        &buff);
                if (rv != SA_OK) {
                        snmp_log (LOG_ERR, 
                                  "populate_saHpiResourceEventLogTable: oh_decode_sensorreading, TriggerReading error!");
                        rv = AGENT_ERR_INTERNAL_ERROR;
                }
                memcpy(sen_evt_ctx->saHpiSensorEventLogTriggerReading, buff.Data, buff.DataLength);
                sen_evt_ctx->saHpiSensorEventLogTriggerReading_len = buff.DataLength;
        }

        /** SaHpiSensorReadingType = ASN_INTEGER */
        sen_evt_ctx->saHpiSensorEventLogTriggerThresholdType =
                event->Event.EventDataUnion.SensorEvent.TriggerThreshold.Type + 1;

        /** SaHpiSensorReadingValue = ASN_OCTET_STR */
        if (event_rdr_entry) {
                rv =  oh_decode_sensorreading(
                        event->Event.EventDataUnion.SensorEvent.TriggerThreshold,
                        event_rdr_entry->RdrTypeUnion.SensorRec.DataFormat,
                        &buff);
                if (rv != SA_OK) {
                        snmp_log (LOG_ERR, 
                                  "populate_saHpiResourceEventLogTable: oh_decode_sensorreading TriggerReading error!");
                        rv = AGENT_ERR_INTERNAL_ERROR;
                }
                memcpy(sen_evt_ctx->saHpiSensorEventLogTriggerThreshold, buff.Data, buff.DataLength);
                sen_evt_ctx->saHpiSensorEventLogTriggerThreshold_len = buff.DataLength;
        }

        /** SaHpiEventState = ASN_OCTET_STR */
        rv = oh_decode_eventstate(event->Event.EventDataUnion.SensorEvent.PreviousState,
                                  event->Event.EventDataUnion.SensorEvent.EventCategory,
                                  &buff);
	if (rv != SA_OK) {
		snmp_log (LOG_ERR, 
                "populate_saHpiResourceEventLogTable: oh_decode_eventstate PreviousState error!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}
        memcpy(sen_evt_ctx->saHpiSensorEventLogPreviousState,
               buff.Data, buff.DataLength);
        sen_evt_ctx->saHpiSensorEventLogPreviousState_len = buff.DataLength;

        /** SaHpiEventState = ASN_OCTET_STR */
        rv = oh_decode_eventstate(event->Event.EventDataUnion.SensorEvent.CurrentState,
                                  event->Event.EventDataUnion.SensorEvent.EventCategory,
                                  &buff);
	if (rv != SA_OK) {
		snmp_log (LOG_ERR, 
                "populate_saHpiResourceEventLogTable: oh_decode_eventstate CurrentState error!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}
        memcpy(sen_evt_ctx->saHpiSensorEventLogCurrentState,
               buff.Data, buff.DataLength);
        sen_evt_ctx->saHpiSensorEventLogCurrentState_len = buff.DataLength;

        /** UNSIGNED32 = ASN_UNSIGNED */
        sen_evt_ctx->saHpiSensorEventLogOem = 
                event->Event.EventDataUnion.SensorEvent.Oem;

        /** UNSIGNED32 = ASN_UNSIGNED */
        sen_evt_ctx->saHpiSensorEventLogSpecific = 
                event->Event.EventDataUnion.SensorEvent.SensorSpecific;

	if (isNewRow == MIB_TRUE) {
	
		CONTAINER_INSERT (cb.container, sen_evt_ctx);
	}
		
	sensor_event_log_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);
	
	return SA_OK;
}

/**
 * 
 * @session_id
 * @resource_id
 * @saHpiEventLogRowPointer
 * @aHpiEventLogRowPointer_len
 * 
 * @return 
 */
SaErrorT sensor_event_log_clear(SaHpiSessionIdT session_id, 
                                SaHpiResourceIdT resource_id,  
                                oid *saHpiEventLogRowPointer, 
                                size_t saHpiEventLogRowPointer_len,
				int modifyTotal)
{

	oid sen_evt_oid[SENSOR_EVENT_LOG_INDEX_NR];
	netsnmp_index sen_evt_idx;
	netsnmp_index *sensor_index;
	saHpiSensorEventLogTable_context *sen_evt_ctx;
	
	DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;	

	int column_len = 2;
	
        DEBUGMSGTL ((AGENT, "sensor_event_log_clear, called\n"));
	DEBUGMSGTL ((AGENT, "Attempting to delete sensor el row with the following indexes:\n"));
        DEBUGMSGTL ((AGENT,"     Domain   [%ld]\n", saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len]));
        DEBUGMSGTL ((AGENT,"     Resource [%ld]\n", saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 1]));
	DEBUGMSGTL ((AGENT,"     Sensor # [%ld]\n", saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 2])); 
	DEBUGMSGTL ((AGENT,"     Severity [%s]\n",  oh_lookup_severity(
						saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 3]-1)));
	DEBUGMSGTL ((AGENT,"     Entry Id # [%ld]\n", saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 4]));								
   
	/* BUILD oid to lookup row */
		/* assign the number of indices */
	sen_evt_idx.len = SENSOR_EVENT_LOG_INDEX_NR;
		/** Index saHpiDomainId is external */
	sen_evt_oid[0] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len];
		/** Index saHpiResourceId is external */
	sen_evt_oid[1] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 1];
		/** Index saHpiSensorNum is external */
	sen_evt_oid[2] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 2];
		/** Index saHpiEventSeverity is external */
	sen_evt_oid[3] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 3];
                /** Index saHpiSensorEventEntryId is internal */	
	sen_evt_oid[4] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 4];	
		/* assign the indices to the index */	
	sen_evt_idx.oids = (oid *) & sen_evt_oid;

	sensor_index = CONTAINER_FIRST(cb.container);
	sen_evt_ctx = CONTAINER_FIND(cb.container, &sen_evt_idx);
	
	if (!sen_evt_ctx) {
		DEBUGMSGTL ((AGENT, "sensor_event_log_clear did not find a row to delete\n"));
	}
	else {
		
		DEBUGMSGTL ((AGENT, "sensor_event_log_clear found row to delete\n"));
		CONTAINER_REMOVE(cb.container, sen_evt_ctx);
		saHpiSensorEventLogTable_delete_row(sen_evt_ctx);
			
		/* Reset the entry id for this domain/resource pair */
		dr_pair.domainId_resourceId_arry[0] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len];
		dr_pair.domainId_resourceId_arry[1] = saHpiEventLogRowPointer[saHpiSensorEventLogTable_oid_len + column_len + 1];		
		dr_entry = domain_resoruce_pair_lookup(&dr_pair, &dr_table); 
		
		if (dr_entry == NULL) {
			DEBUGMSGTL ((AGENT, 
			"ERROR: sensor_event_log_clear() domain_resoruce_pair_lookup returned NULL\n"));
			return AGENT_ERR_INTERNAL_ERROR;
		}
	
		DEBUGMSGTL ((AGENT, 
			"sensor_event_log_clear() resetting entry_id to 0\n"));
		dr_entry->entry_id = 0;		
		
		sensor_event_log_entry_count = CONTAINER_SIZE (cb.container);
		if (modifyTotal == MIB_TRUE) {
			sensor_event_log_entry_count_total--;	
		}	
	}	

        return SA_OK;
}


/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int handle_saHpiSensorEventEntryLogCountTotal(netsnmp_mib_handler *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */
           
        DEBUGMSGTL ((AGENT, "handle_saHpiSensorEventEntryLogCountTotal, called\n"));
            
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_COUNTER,
        			        (u_char *) &sensor_event_log_entry_count_total,
        			        sizeof(sensor_event_log_entry_count_total));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int handle_saHpiSensorEventLogActiveEntries(netsnmp_mib_handler *handler,
                                     netsnmp_handler_registration *reginfo,
                                     netsnmp_agent_request_info   *reqinfo,
                                     netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */
           
        DEBUGMSGTL ((AGENT, "handle_saHpiSensorEventLogActiveEntries, called\n"));

        sensor_event_log_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
        	snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        				(u_char *) &sensor_event_log_entry_count,
        				sizeof(sensor_event_log_entry_count));
        	break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @return: 
 */
int initialize_table_saHpiSensorEventEntryLogCountTotal(void)
{
        DEBUGMSGTL ((AGENT, "initialize_table_saHpiSensorEventEntryLogCountTotal, called\n"));   

        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
			               "saHpiSensorEventEntryLogCountTotal", 
				       handle_saHpiSensorEventEntryLogCountTotal,
                                       saHpiSensorEventLogLifetimeChanges_oid, 
				       OID_LENGTH(saHpiSensorEventLogLifetimeChanges_oid),
                                       HANDLER_CAN_RONLY ));
				       
        return SNMP_ERR_NOERROR;
}
/**
 * 
 * @return: 
 */
int initialize_table_saHpiSensorEventLogActiveEntries(void)
{
        DEBUGMSGTL ((AGENT, "initialize_table_saHpiSensorEventLogActiveEntries, called\n"));   

        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
			               "saHpiSensorEventLogActiveEntries", 
				        handle_saHpiSensorEventLogActiveEntries,
                                        saHpiSensorEventLogActiveEntries_oid, 
					OID_LENGTH(saHpiSensorEventLogActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));
        return SNMP_ERR_NOERROR;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
	

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiSensorEventLogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiSensorEventLogTable_cmp( const void *lhs, const void *rhs )
{
        saHpiSensorEventLogTable_context *context_l =
        (saHpiSensorEventLogTable_context *)lhs;
        saHpiSensorEventLogTable_context *context_r =
        (saHpiSensorEventLogTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiSensorEventLogTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiSensorEventLogTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiSensorNum */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                /* If saHpiResourceIsHistorical index is equal sort by forth index */
                                /* CHECK FORTH INDEX,  saHpiEventSeverity */
                                if ( context_l->index.oids[3] < context_r->index.oids[3])
                                        return -1;

                                if ( context_l->index.oids[3] > context_r->index.oids[3])
                                        return 1;

                                if ( context_l->index.oids[3] == context_r->index.oids[3]) {
				        /* If saHpiEventSeverity index is equal sort by forth index */
                                        /* CHECK FORTH INDEX,  saHpiSensorEventEntryId */
                                        if ( context_l->index.oids[4] < context_r->index.oids[4])
                                                return -1;

                                        if ( context_l->index.oids[4] > context_r->index.oids[4])
                                                return 1;

                                        if ( context_l->index.oids[4] == context_r->index.oids[4])
					        return 0;
				}	
                        }
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiSensorEventLogTable module
 */
void
init_saHpiSensorEventLogTable(void)
{
        DEBUGMSGTL ((AGENT, "init_saHpiSensorEventLogTable, called\n"));
        
        initialize_table_saHpiSensorEventLogTable();
        
        initialize_table_saHpiSensorEventEntryLogCountTotal();
        
        initialize_table_saHpiSensorEventLogActiveEntries();

        domain_resource_pair_initialize(&initialized, &dr_table);
     
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiSensorEventLogTable_row_copy(saHpiSensorEventLogTable_context * dst,
                         saHpiSensorEventLogTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    subagent_lock(&hpi_lock_data);
    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
        return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    memcpy( dst->saHpiSensorEventLogTimestamp,
            src->saHpiSensorEventLogTimestamp,
	    src->saHpiSensorEventLogTimestamp_len);
    
    dst->saHpiSensorEventLogTimestamp_len = src->saHpiSensorEventLogTimestamp_len;
    
    dst->saHpiSensorEventLogType = src->saHpiSensorEventLogType;

    dst->saHpiSensorEventLogCategory = src->saHpiSensorEventLogCategory;

    dst->saHpiSensorEventLogAssertion = src->saHpiSensorEventLogAssertion;

    memcpy( dst->saHpiSensorEventLogState, src->saHpiSensorEventLogState, src->saHpiSensorEventLogState_len );
    dst->saHpiSensorEventLogState_len = src->saHpiSensorEventLogState_len;

    memcpy( dst->saHpiSensorEventLogOptionalData, src->saHpiSensorEventLogOptionalData, src->saHpiSensorEventLogOptionalData_len );
    dst->saHpiSensorEventLogOptionalData_len = src->saHpiSensorEventLogOptionalData_len;

    dst->saHpiSensorEventLogTriggerReadingType = src->saHpiSensorEventLogTriggerReadingType;

    memcpy( dst->saHpiSensorEventLogTriggerReading, src->saHpiSensorEventLogTriggerReading, src->saHpiSensorEventLogTriggerReading_len );
    dst->saHpiSensorEventLogTriggerReading_len = src->saHpiSensorEventLogTriggerReading_len;

    dst->saHpiSensorEventLogTriggerThresholdType = src->saHpiSensorEventLogTriggerThresholdType;

    memcpy( dst->saHpiSensorEventLogTriggerThreshold, src->saHpiSensorEventLogTriggerThreshold, src->saHpiSensorEventLogTriggerThreshold_len );
    dst->saHpiSensorEventLogTriggerThreshold_len = src->saHpiSensorEventLogTriggerThreshold_len;

    memcpy( dst->saHpiSensorEventLogPreviousState, src->saHpiSensorEventLogPreviousState, src->saHpiSensorEventLogPreviousState_len );
    dst->saHpiSensorEventLogPreviousState_len = src->saHpiSensorEventLogPreviousState_len;

    memcpy( dst->saHpiSensorEventLogCurrentState, src->saHpiSensorEventLogCurrentState, src->saHpiSensorEventLogCurrentState_len );
    dst->saHpiSensorEventLogCurrentState_len = src->saHpiSensorEventLogCurrentState_len;

    dst->saHpiSensorEventLogOem = src->saHpiSensorEventLogOem;

    dst->saHpiSensorEventLogSpecific = src->saHpiSensorEventLogSpecific;

    subagent_unlock(&hpi_lock_data);
    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiSensorEventLogTable_extract_index( saHpiSensorEventLogTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiSensorNum;
        netsnmp_variable_list var_saHpiEventSeverity;
        netsnmp_variable_list var_saHpiSensorEventEntryId;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiSensorEventLogTable_extract_index, called\n"));
        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
                        return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiSensorNum;

        memset( &var_saHpiSensorNum, 0x00, sizeof(var_saHpiSensorNum) );
        var_saHpiSensorNum.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiSensorNum.next_variable = &var_saHpiEventSeverity;

        memset( &var_saHpiEventSeverity, 0x00, sizeof(var_saHpiEventSeverity) );
        var_saHpiEventSeverity.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiEventSeverity.next_variable = &var_saHpiSensorEventEntryId;

        memset( &var_saHpiSensorEventEntryId, 0x00, sizeof(var_saHpiSensorEventEntryId) );
        var_saHpiSensorEventEntryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiSensorEventEntryId.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiResourceId */
   
                /** skipping external index saHpiSensorNum */
   
                /** skipping external index saHpiEventSeverity */
   
                /** skipping external index saHpiSensorEventEntryId */
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);
		err = saHpiSensorNum_check_index(*var_saHpiSensorNum.val.integer);
		err = saHpiEventSeverity_check_index(*var_saHpiEventSeverity.val.integer);
		err = saHpiSensorEventEntryId_check_index(*var_saHpiSensorEventEntryId.val.integer); 
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiSensorEventLogTable_can_activate(saHpiSensorEventLogTable_context *undo_ctx,
                      saHpiSensorEventLogTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiSensorEventLogTable_can_deactivate(saHpiSensorEventLogTable_context *undo_ctx,
                        saHpiSensorEventLogTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiSensorEventLogTable_can_delete(saHpiSensorEventLogTable_context *undo_ctx,
                    saHpiSensorEventLogTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiSensorEventLogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiSensorEventLogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiSensorEventLogTable_context *
saHpiSensorEventLogTable_create_row( netsnmp_index* hdr)
{
    saHpiSensorEventLogTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiSensorEventLogTable_context);
    if(!ctx)
        return NULL;

        
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiSensorEventLogTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
        return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
    */
    sensor_event_log_entry_count_total++;

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiSensorEventLogTable_context *
saHpiSensorEventLogTable_duplicate_row( saHpiSensorEventLogTable_context * row_ctx)
{
    saHpiSensorEventLogTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);

    dup = SNMP_MALLOC_TYPEDEF(saHpiSensorEventLogTable_context);
    if(!dup) {
            subagent_unlock(&hpi_lock_data);
            return NULL;
    }
        
    if(saHpiSensorEventLogTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiSensorEventLogTable_delete_row( saHpiSensorEventLogTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        subagent_lock(&hpi_lock_data);

    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    sensor_event_log_entry_count_total++;

    subagent_unlock(&hpi_lock_data);
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiSensorEventLogTable_set_reserve1( netsnmp_request_group *rg )
{
//    saHpiSensorEventLogTable_context *row_ctx =
//            (saHpiSensorEventLogTable_context *)rg->existing_row;
//    saHpiSensorEventLogTable_context *undo_ctx =
//            (saHpiSensorEventLogTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiSensorEventLogTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiSensorEventLogTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiSensorEventLogTable_context *row_ctx = (saHpiSensorEventLogTable_context *)rg->existing_row;
//    saHpiSensorEventLogTable_context *undo_ctx = (saHpiSensorEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiSensorEventLogTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSensorEventLogTable_context *row_ctx = (saHpiSensorEventLogTable_context *)rg->existing_row;
//    saHpiSensorEventLogTable_context *undo_ctx = (saHpiSensorEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }
    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiSensorEventLogTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSensorEventLogTable_context *row_ctx = (saHpiSensorEventLogTable_context *)rg->existing_row;
//    saHpiSensorEventLogTable_context *undo_ctx = (saHpiSensorEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiSensorEventLogTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSensorEventLogTable_context *row_ctx = (saHpiSensorEventLogTable_context *)rg->existing_row;
//    saHpiSensorEventLogTable_context *undo_ctx = (saHpiSensorEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: 
                break;
                /** We shouldn't get here */
                /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiSensorEventLogTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSensorEventLogTable_context *row_ctx = (saHpiSensorEventLogTable_context *)rg->existing_row;
//    saHpiSensorEventLogTable_context *undo_ctx = (saHpiSensorEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 *
 * Initialize the saHpiSensorEventLogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiSensorEventLogTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiSensorEventLogTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiSensorEventLogTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiSensorEventLogTable_oid,
                                             saHpiSensorEventLogTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiSensorEventLogTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiSensorNum */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiEventSeverity */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiSensorEventEntryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiSensorEventLogTable_COL_MIN;
    table_info->max_column = saHpiSensorEventLogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiSensorEventLogTable_get_value;
    cb.container = netsnmp_container_find("saHpiSensorEventLogTable_primary:"
                                          "saHpiSensorEventLogTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiSensorEventLogTable_secondary:"
                                                       "saHpiSensorEventLogTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiSensorEventLogTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiSensorEventLogTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiSensorEventLogTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiSensorEventLogTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiSensorEventLogTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiSensorEventLogTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiSensorEventLogTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiSensorEventLogTable_can_delete;

    cb.set_reserve1 = saHpiSensorEventLogTable_set_reserve1;
    cb.set_reserve2 = saHpiSensorEventLogTable_set_reserve2;
    cb.set_action = saHpiSensorEventLogTable_set_action;
    cb.set_commit = saHpiSensorEventLogTable_set_commit;
    cb.set_free = saHpiSensorEventLogTable_set_free;
    cb.set_undo = saHpiSensorEventLogTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiSensorEventLogTable",
                "Registering table saHpiSensorEventLogTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiSensorEventLogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiSensorEventLogTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiSensorEventLogTable_context *context = (saHpiSensorEventLogTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPISENSOREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogTimestamp,
                         sizeof(context->saHpiSensorEventLogTimestamp) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGTYPE:
            /** SaHpiSensorType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSensorEventLogType,
                         sizeof(context->saHpiSensorEventLogType) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGCATEGORY:
            /** SaHpiEventCategory = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSensorEventLogCategory,
                         sizeof(context->saHpiSensorEventLogCategory) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGASSERTION:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSensorEventLogAssertion,
                         sizeof(context->saHpiSensorEventLogAssertion) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogState,
                         context->saHpiSensorEventLogState_len );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGOPTIONALDATA:
            /** SaHpiOptionalData = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogOptionalData,
                         context->saHpiSensorEventLogOptionalData_len );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGTRIGGERREADINGTYPE:
            /** SaHpiSensorReadingType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSensorEventLogTriggerReadingType,
                         sizeof(context->saHpiSensorEventLogTriggerReadingType) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGTRIGGERREADING:
            /** SaHpiSensorReadingValue = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogTriggerReading,
                         context->saHpiSensorEventLogTriggerReading_len );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGTRIGGERTHRESHOLDTYPE:
            /** SaHpiSensorReadingType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSensorEventLogTriggerThresholdType,
                         sizeof(context->saHpiSensorEventLogTriggerThresholdType) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGTRIGGERTHRESHOLD:
            /** SaHpiSensorReadingValue = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogTriggerThreshold,
                         context->saHpiSensorEventLogTriggerThreshold_len );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGPREVIOUSSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogPreviousState,
                         context->saHpiSensorEventLogPreviousState_len );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGCURRENTSTATE:
            /** SaHpiEventState = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSensorEventLogCurrentState,
                         context->saHpiSensorEventLogCurrentState_len );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGOem:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiSensorEventLogOem,
                         sizeof(context->saHpiSensorEventLogOem) );
        break;
    
        case COLUMN_SAHPISENSOREVENTLOGSPECIFIC:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiSensorEventLogSpecific,
                         sizeof(context->saHpiSensorEventLogSpecific) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiSensorEventLogTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiSensorEventLogTable_get_by_idx
 */
const saHpiSensorEventLogTable_context *
saHpiSensorEventLogTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiSensorEventLogTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


