/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiSoftwareEventTable.c 5427 2005-11-14 22:10:52Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiSoftwareEventTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiSoftwareEventTable_oid[] = { saHpiSoftwareEventTable_TABLE_OID };
size_t saHpiSoftwareEventTable_oid_len = OID_LENGTH(saHpiSoftwareEventTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;		      
static GHashTable *dr_table;

/*************************************************************
 * oid and fucntion declarations scalars
 */

static u_long software_event_entry_count_total = 0;
static u_long software_event_entry_count = 0;

static oid saHpiSoftwareEventLifetimeChanges_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,1,22 };
static oid saHpiSoftwareEventActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,1,23 }; 

int handle_saHpiSoftwareEventLifetimeChanges(netsnmp_mib_handler *handler,
                                             netsnmp_handler_registration *reginfo,
                                             netsnmp_agent_request_info   *reqinfo,
                                             netsnmp_request_info         *requests);
					     
int handle_saHpiSoftwareEventActiveEntries(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);					     

int initialize_table_saHpiSoftwareEventLifetimeChanges(void);
int initialize_table_saHpiSoftwareEventActiveEntries(void);


SaErrorT populate_saHpiSoftwareEventTable(SaHpiSessionIdT sessionid,
                                          SaHpiEventT *event,
                                          oid * this_child_oid, 
                                          size_t *this_child_oid_len)
{
	SaErrorT rv = SA_OK;
        int new_row = MIB_FALSE;

	oid software_evt_oid[SOFTWARE_EVENT_INDEX_NR];
	netsnmp_index software_evt_idx;
	saHpiSoftwareEventTable_context *software_evt_ctx;

	oid column[2];
	int column_len = 2;

        DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

        DEBUGMSGTL ((AGENT, "populate_saHpiSoftwareEventTable, called\n"));

	/* check for NULL pointers */
	if (!event) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiSoftwareEventTable() passed NULL event pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	} 
	
	
	subagent_lock(&hpi_lock_data);
	/* BUILD oid for new row */
		/* assign the number of indices */
	software_evt_idx.len = SOFTWARE_EVENT_INDEX_NR;
		/** Index saHpiDomainId is external */
	software_evt_oid[0] = get_domain_id(sessionid);
		/** Index saHpiResourceId is external */
	software_evt_oid[1] = event->Source;
	        /** Index saHpiEventSeverity is external */
	software_evt_oid[2] = event->Severity + 1;
                /** Index saHpiSoftwareEventEntryId is internal */
	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
	dr_pair.domainId_resourceId_arry[1] = event->Source;
	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
	if (dr_entry == NULL) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiSoftwareEventTable() domain_resource_pair_get returned NULL\n"));
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}
	software_evt_oid[3] = dr_entry->entry_id++;	
	software_evt_idx.oids = (oid *) & software_evt_oid;
	   
	/* See if Row exists. */
	software_evt_ctx = NULL;
	software_evt_ctx = CONTAINER_FIND(cb.container, &software_evt_idx);

	if (!software_evt_ctx) { 
		// New entry. Add it
		software_evt_ctx = 
			saHpiSoftwareEventTable_create_row(&software_evt_idx);
                new_row = MIB_TRUE;
	}
	if (!software_evt_ctx) {
		snmp_log (LOG_ERR, "Not enough memory for a Software Event row!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}

        /** SaHpiEntryId = ASN_UNSIGNED */
        software_evt_ctx->saHpiSoftwareEventEntryId = software_evt_oid[3];

        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event->Timestamp, software_evt_ctx->saHpiSoftwareEventTimestamp);
        software_evt_ctx->saHpiSoftwareEventTimestamp_len = sizeof(SaHpiTimeT);
        /** SaHpiManufacturerId = ASN_UNSIGNED */
        software_evt_ctx->saHpiSoftwareEventManufacturerIdT = 
	                                event->EventDataUnion.HpiSwEvent.MId;
        
        /** INTEGER = ASN_INTEGER */
        software_evt_ctx->saHpiSoftwareEventType = 
	                                event->EventDataUnion.HpiSwEvent.Type + 1;

        /** SaHpiTextType = ASN_INTEGER */
        software_evt_ctx->saHpiSoftwareEventTextType = 
	                        event->EventDataUnion.HpiSwEvent.EventData.DataType + 1;
			     
        /** SaHpiTextLanguage = ASN_INTEGER */
        software_evt_ctx->saHpiSoftwareEventTextLanguage = 
	                        event->EventDataUnion.HpiSwEvent.EventData.Language + 1;

        /** SaHpiText = ASN_OCTET_STR */
        software_evt_ctx->saHpiSoftwareEventText_len = 
	                        event->EventDataUnion.HpiSwEvent.EventData.DataLength;
				
	memcpy(software_evt_ctx->saHpiSoftwareEventText, 
	          	        event->EventDataUnion.HpiSwEvent.EventData.Data,
				        SAHPI_MAX_TEXT_BUFFER_LENGTH);
              			
        if (new_row == MIB_TRUE) 
                CONTAINER_INSERT (cb.container, software_evt_ctx);
		
	software_event_entry_count = CONTAINER_SIZE (cb.container);

	/* create full oid on This row for parent RowPointer */
	column[0] = 1;
	column[1] = COLUMN_SAHPISOFTWAREEVENTTIMESTAMP;
	memset(this_child_oid, 0, MAX_OID_LEN);
	build_full_oid(saHpiSoftwareEventTable_oid, saHpiSoftwareEventTable_oid_len,
			column, column_len,
			&software_evt_idx,
			this_child_oid, MAX_OID_LEN, this_child_oid_len);

        
	subagent_unlock(&hpi_lock_data);
	return SA_OK;   					


}

SaErrorT async_software_event_add(SaHpiSessionIdT sessionid, 
                                  SaHpiEventT *event,
                                  SaHpiRdrT *rdr,
                                  SaHpiRptEntryT *rpt_entry,				  
                                  oid * this_child_oid, 
                                  size_t *this_child_oid_len)
{
	SaErrorT rv = SA_OK;
        int new_row = MIB_FALSE;

	oid software_evt_oid[SOFTWARE_EVENT_INDEX_NR];
	netsnmp_index software_evt_idx;
	saHpiSoftwareEventTable_context *software_evt_ctx;

	oid column[2];
	int column_len = 2;

        DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

        DEBUGMSGTL ((AGENT, "async_software_event_add, called\n"));

	/* check for NULL pointers */
	if (!event) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: async_software_event_add() passed NULL event pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	} 
	
	if (!rdr) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: async_software_event_add() passed NULL rdr pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}
	
	if (!rpt_entry) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: async_software_event_add() passed NULL rpt_entry pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}	
	
	/* BUILD oid for new row */
		/* assign the number of indices */
	software_evt_idx.len = SOFTWARE_EVENT_INDEX_NR;
		/** Index saHpiDomainId is external */
	software_evt_oid[0] = get_domain_id(sessionid);
		/** Index saHpiResourceId is external */
	software_evt_oid[1] = event->Source;
	        /** Index saHpiEventSeverity is external */
	software_evt_oid[2] = event->Severity + 1;
                /** Index saHpiSoftwareEventEntryId is internal */
	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
	dr_pair.domainId_resourceId_arry[1] = event->Source;
	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
	if (dr_entry == NULL) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: async_software_event_add() domain_resource_pair_get returned NULL\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}
	software_evt_oid[3] = dr_entry->entry_id++;	
	software_evt_idx.oids = (oid *) & software_evt_oid;
	   
	/* See if Row exists. */
	software_evt_ctx = NULL;
	software_evt_ctx = CONTAINER_FIND(cb.container, &software_evt_idx);

	if (!software_evt_ctx) { 
		// New entry. Add it
		software_evt_ctx = 
			saHpiSoftwareEventTable_create_row(&software_evt_idx);
                new_row = MIB_TRUE;
	}
	if (!software_evt_ctx) {
		snmp_log (LOG_ERR, "Not enough memory for a Software Event row!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}

        /** SaHpiEntryId = ASN_UNSIGNED */
        software_evt_ctx->saHpiSoftwareEventEntryId = software_evt_oid[3];

        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event->Timestamp, software_evt_ctx->saHpiSoftwareEventTimestamp);
        software_evt_ctx->saHpiSoftwareEventTimestamp_len = sizeof(SaHpiTimeT);
        /** SaHpiManufacturerId = ASN_UNSIGNED */
        software_evt_ctx->saHpiSoftwareEventManufacturerIdT = 
	                                event->EventDataUnion.HpiSwEvent.MId;
        
        /** INTEGER = ASN_INTEGER */
        software_evt_ctx->saHpiSoftwareEventType = 
	                                event->EventDataUnion.HpiSwEvent.Type + 1;

        /** SaHpiTextType = ASN_INTEGER */
        software_evt_ctx->saHpiSoftwareEventTextType = 
	                        event->EventDataUnion.HpiSwEvent.EventData.DataType + 1;
			     
        /** SaHpiTextLanguage = ASN_INTEGER */
        software_evt_ctx->saHpiSoftwareEventTextLanguage = 
	                        event->EventDataUnion.HpiSwEvent.EventData.Language + 1;

        /** SaHpiText = ASN_OCTET_STR */
        software_evt_ctx->saHpiSoftwareEventText_len = 
	                        event->EventDataUnion.HpiSwEvent.EventData.DataLength;
				
	memcpy(software_evt_ctx->saHpiSoftwareEventText, 
	          	        event->EventDataUnion.HpiSwEvent.EventData.Data,
				        SAHPI_MAX_TEXT_BUFFER_LENGTH);
        if (new_row == MIB_TRUE) 
                CONTAINER_INSERT (cb.container, software_evt_ctx);
		
	software_event_entry_count = CONTAINER_SIZE (cb.container);
	
	/* create full oid on This row for parent RowPointer */
	column[0] = 1;
	column[1] = COLUMN_SAHPISOFTWAREEVENTTIMESTAMP;
	memset(this_child_oid, 0, MAX_OID_LEN);
	build_full_oid(saHpiSoftwareEventTable_oid, saHpiSoftwareEventTable_oid_len,
			column, column_len,
			&software_evt_idx,
			this_child_oid, MAX_OID_LEN, this_child_oid_len);

        return SA_OK;   					

}

					     
/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiSoftwareEventLifetimeChanges(netsnmp_mib_handler *handler,
                                         netsnmp_handler_registration *reginfo,
                                         netsnmp_agent_request_info   *reqinfo,
                                         netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiSoftwareEventLifetimeChanges, called\n"));

        
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_COUNTER,
         			        (u_char *) &software_event_entry_count_total,
         			        sizeof(software_event_entry_count_total));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiSoftwareEventActiveEntries(netsnmp_mib_handler *handler,
                                    netsnmp_handler_registration *reginfo,
                                    netsnmp_agent_request_info   *reqinfo,
                                    netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiSoftwareEventActiveEntries, called\n"));

        software_event_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			        (u_char *) &software_event_entry_count,
        			        sizeof(software_event_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @return: 
 */
int 
initialize_table_saHpiSoftwareEventLifetimeChanges(void)
{
        DEBUGMSGTL ((AGENT, "initialize_table_saHpiSoftwareEventLifetimeChanges, called\n"));
        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
			               "saHpiSoftwareEventLifetimeChanges", 
				       handle_saHpiSoftwareEventLifetimeChanges,
                                       saHpiSoftwareEventLifetimeChanges_oid, 
				       OID_LENGTH(saHpiSoftwareEventLifetimeChanges_oid),
                                       HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;

}

 /**
 * 
 * @return: 
 */
int 
initialize_table_saHpiSoftwareEventActiveEntries(void)
{
        DEBUGMSGTL ((AGENT, "initialize_table_saHpiSoftwareEventActiveEntries, called\n"));
    
        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
			       "saHpiSoftwareEventActiveEntries", 
			       handle_saHpiSoftwareEventActiveEntries,
                               saHpiSoftwareEventActiveEntries_oid, 
			       OID_LENGTH(saHpiSoftwareEventActiveEntries_oid),
                               HANDLER_CAN_RONLY ));
			       
        return SNMP_ERR_NOERROR;			       
}
 
 
/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiSoftwareEventTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiSoftwareEventTable_cmp( const void *lhs, const void *rhs )
{
        saHpiSoftwareEventTable_context *context_l =
        (saHpiSoftwareEventTable_context *)lhs;
        saHpiSoftwareEventTable_context *context_r =
        (saHpiSoftwareEventTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiSoftwareEventTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiSoftwareEventTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiEventSeverity */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                /* If saHpiEventSeverity index is equal sort by forth index */
                                /* CHECK FORTH INDEX,  saHpiSoftwareEventEntryId */
                                if ( context_l->index.oids[3] < context_r->index.oids[3])
                                        return -1;

                                if ( context_l->index.oids[3] > context_r->index.oids[3])
                                        return 1;

                                if ( context_l->index.oids[3] == context_r->index.oids[3])
                                        return 0;
                        }
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiSoftwareEventTable module
 */
void
init_saHpiSoftwareEventTable(void)
{
        DEBUGMSGTL ((AGENT, "init_saHpiSoftwareEventTable, called\n"));

        initialize_table_saHpiSoftwareEventTable();

        initialize_table_saHpiSoftwareEventLifetimeChanges();
        initialize_table_saHpiSoftwareEventActiveEntries();
	
        domain_resource_pair_initialize(&initialized, &dr_table);	
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiSoftwareEventTable_row_copy(saHpiSoftwareEventTable_context * dst,
                         saHpiSoftwareEventTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    subagent_lock(&hpi_lock_data);

    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
        return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiSoftwareEventEntryId = src->saHpiSoftwareEventEntryId;

    memcpy( dst->saHpiSoftwareEventTimestamp, 
            src->saHpiSoftwareEventTimestamp,
	    src->saHpiSoftwareEventTimestamp_len);
	    
    dst->saHpiSoftwareEventTimestamp_len = src->saHpiSoftwareEventTimestamp_len;
    
    dst->saHpiSoftwareEventManufacturerIdT = src->saHpiSoftwareEventManufacturerIdT;

    dst->saHpiSoftwareEventType = src->saHpiSoftwareEventType;

    dst->saHpiSoftwareEventTextType = src->saHpiSoftwareEventTextType;

    dst->saHpiSoftwareEventTextLanguage = src->saHpiSoftwareEventTextLanguage;

    memcpy( dst->saHpiSoftwareEventText, src->saHpiSoftwareEventText, src->saHpiSoftwareEventText_len );
    dst->saHpiSoftwareEventText_len = src->saHpiSoftwareEventText_len;

    subagent_unlock(&hpi_lock_data);
    return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiSoftwareEventTable_extract_index( saHpiSoftwareEventTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiEventSeverity;
        netsnmp_variable_list var_saHpiSoftwareEventEntryId;
        int err;
       
        DEBUGMSGTL ((AGENT, "saHpiSoftwareEventTable_extract_index, called\n"));

        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
                        return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiEventSeverity;

        memset( &var_saHpiEventSeverity, 0x00, sizeof(var_saHpiEventSeverity) );
        var_saHpiEventSeverity.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiEventSeverity.next_variable = &var_saHpiSoftwareEventEntryId;

        memset( &var_saHpiSoftwareEventEntryId, 0x00, sizeof(var_saHpiSoftwareEventEntryId) );
        var_saHpiSoftwareEventEntryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiSoftwareEventEntryId.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiResourceId */
   
                /** skipping external index saHpiEventSeverity */
   
                ctx->saHpiSoftwareEventEntryId = *var_saHpiSoftwareEventEntryId.val.integer;
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
                err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
                err = saHpiEventSeverity_check_index(*var_saHpiEventSeverity.val.integer);
                err = saHpiSoftwareEventEntryId_check_index(*var_saHpiSoftwareEventEntryId.val.integer);
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiSoftwareEventTable_can_activate(saHpiSoftwareEventTable_context *undo_ctx,
                      saHpiSoftwareEventTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiSoftwareEventTable_can_deactivate(saHpiSoftwareEventTable_context *undo_ctx,
                        saHpiSoftwareEventTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiSoftwareEventTable_can_delete(saHpiSoftwareEventTable_context *undo_ctx,
                    saHpiSoftwareEventTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiSoftwareEventTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiSoftwareEventTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiSoftwareEventTable_context *
saHpiSoftwareEventTable_create_row( netsnmp_index* hdr)
{
    saHpiSoftwareEventTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiSoftwareEventTable_context);
    if(!ctx)
        return NULL;


    subagent_lock(&hpi_lock_data);
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiSoftwareEventTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
        return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
    */

    software_event_entry_count_total++;
    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiSoftwareEventTable_context *
saHpiSoftwareEventTable_duplicate_row( saHpiSoftwareEventTable_context * row_ctx)
{
    saHpiSoftwareEventTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);

    dup = SNMP_MALLOC_TYPEDEF(saHpiSoftwareEventTable_context);
    if(!dup) {
            subagent_unlock(&hpi_lock_data);
            return NULL;
    }
        
    if(saHpiSoftwareEventTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiSoftwareEventTable_delete_row( saHpiSoftwareEventTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        subagent_lock(&hpi_lock_data);  

    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    software_event_entry_count_total++;

    subagent_unlock(&hpi_lock_data);
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiSoftwareEventTable_set_reserve1( netsnmp_request_group *rg )
{
//    saHpiSoftwareEventTable_context *row_ctx =
//            (saHpiSoftwareEventTable_context *)rg->existing_row;
//    saHpiSoftwareEventTable_context *undo_ctx =
//            (saHpiSoftwareEventTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiSoftwareEventTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiSoftwareEventTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiSoftwareEventTable_context *row_ctx = (saHpiSoftwareEventTable_context *)rg->existing_row;
//    saHpiSoftwareEventTable_context *undo_ctx = (saHpiSoftwareEventTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiSoftwareEventTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSoftwareEventTable_context *row_ctx = (saHpiSoftwareEventTable_context *)rg->existing_row;
//    saHpiSoftwareEventTable_context *undo_ctx = (saHpiSoftwareEventTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */

    /*
     * check activation/deactivation
     */
    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }

        subagent_unlock(&hpi_lock_data);
        return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiSoftwareEventTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSoftwareEventTable_context *row_ctx = (saHpiSoftwareEventTable_context *)rg->existing_row;
//    saHpiSoftwareEventTable_context *undo_ctx = (saHpiSoftwareEventTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiSoftwareEventTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSoftwareEventTable_context *row_ctx = (saHpiSoftwareEventTable_context *)rg->existing_row;
//    saHpiSoftwareEventTable_context *undo_ctx = (saHpiSoftwareEventTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
	        break;
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiSoftwareEventTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiSoftwareEventTable_context *row_ctx = (saHpiSoftwareEventTable_context *)rg->existing_row;
//    saHpiSoftwareEventTable_context *undo_ctx = (saHpiSoftwareEventTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 *
 * Initialize the saHpiSoftwareEventTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiSoftwareEventTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiSoftwareEventTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiSoftwareEventTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiSoftwareEventTable_oid,
                                             saHpiSoftwareEventTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiSoftwareEventTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiEventSeverity */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiSoftwareEventEntryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiSoftwareEventTable_COL_MIN;
    table_info->max_column = saHpiSoftwareEventTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiSoftwareEventTable_get_value;
    cb.container = netsnmp_container_find("saHpiSoftwareEventTable_primary:"
                                          "saHpiSoftwareEventTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiSoftwareEventTable_secondary:"
                                                       "saHpiSoftwareEventTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiSoftwareEventTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiSoftwareEventTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiSoftwareEventTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiSoftwareEventTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiSoftwareEventTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiSoftwareEventTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiSoftwareEventTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiSoftwareEventTable_can_delete;

    cb.set_reserve1 = saHpiSoftwareEventTable_set_reserve1;
    cb.set_reserve2 = saHpiSoftwareEventTable_set_reserve2;
    cb.set_action = saHpiSoftwareEventTable_set_action;
    cb.set_commit = saHpiSoftwareEventTable_set_commit;
    cb.set_free = saHpiSoftwareEventTable_set_free;
    cb.set_undo = saHpiSoftwareEventTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiSoftwareEventTable",
                "Registering table saHpiSoftwareEventTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiSoftwareEventTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiSoftwareEventTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiSoftwareEventTable_context *context = (saHpiSoftwareEventTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPISOFTWAREEVENTENTRYID:
            /** SaHpiEntryId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiSoftwareEventEntryId,
                         sizeof(context->saHpiSoftwareEventEntryId) );
        break;
    
        case COLUMN_SAHPISOFTWAREEVENTTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSoftwareEventTimestamp,
                         sizeof(context->saHpiSoftwareEventTimestamp) );
        break;
    
        case COLUMN_SAHPISOFTWAREEVENTMANUFACTURERIDT:
            /** SaHpiManufacturerId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiSoftwareEventManufacturerIdT,
                         sizeof(context->saHpiSoftwareEventManufacturerIdT) );
        break;
    
        case COLUMN_SAHPISOFTWAREEVENTTYPE:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSoftwareEventType,
                         sizeof(context->saHpiSoftwareEventType) );
        break;
    
        case COLUMN_SAHPISOFTWAREEVENTTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSoftwareEventTextType,
                         sizeof(context->saHpiSoftwareEventTextType) );
        break;
    
        case COLUMN_SAHPISOFTWAREEVENTTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiSoftwareEventTextLanguage,
                         sizeof(context->saHpiSoftwareEventTextLanguage) );
        break;
    
        case COLUMN_SAHPISOFTWAREEVENTTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiSoftwareEventText,
                         context->saHpiSoftwareEventText_len );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiSoftwareEventTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiSoftwareEventTable_get_by_idx
 */
const saHpiSoftwareEventTable_context *
saHpiSoftwareEventTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiSoftwareEventTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


