/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiUserEventLogTable.c 5427 2005-11-14 22:10:52Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiUserEventLogTable.h"
#include <saHpiEventLogTable.h>
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiEventThread.h>
#include <hpiLock.h>

static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiUserEventLogTable_oid[] = { saHpiUserEventLogTable_TABLE_OID };
size_t saHpiUserEventLogTable_oid_len = OID_LENGTH(saHpiUserEventLogTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * objects for hash table
 */
static int initialized = FALSE;		      
static GHashTable *dr_table;

/*************************************************************
 * oid and fucntion declarations scalars
 */
static u_long user_event_log_entry_count_total = 0;
static u_long user_event_log_entry_count = 0;

static oid saHpiUserEventLogLifetimeChanges_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,27 };
static oid saHpiUserEventLogActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,3,2,28 };

int handle_saHpiUserEventLogLifetimeChanges(netsnmp_mib_handler *handler,
                                        netsnmp_handler_registration *reginfo,
                                        netsnmp_agent_request_info   *reqinfo,
                                        netsnmp_request_info         *requests);

int handle_saHpiUserEventLogActiveEntries(netsnmp_mib_handler *handler,
                                netsnmp_handler_registration *reginfo,
                                netsnmp_agent_request_info   *reqinfo,
                                netsnmp_request_info         *requests);

int initialize_table_saHpiUserEventLogLifetimeChanges(void);
int initialize_table_saHpiUserEventLogActiveEntries(void);


SaErrorT populate_saHpiUserEventLogTable(SaHpiSessionIdT sessionid, 
                                         SaHpiEventLogEntryT *event,
                                         oid * this_child_oid, 
                                         size_t *this_child_oid_len)
{
	SaErrorT rv = SA_OK;

	oid user_evt_oid[USER_EVENT_LOG_INDEX_NR];
	netsnmp_index user_evt_idx;
	saHpiUserEventLogTable_context *user_evt_ctx;

	oid column[2];
	int column_len = 2;
	
	int isNewRow = MIB_TRUE;

        DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;

        DEBUGMSGTL ((AGENT, "populate_saHpiUserEventLogTable, called\n"));

	/* check for NULL pointers */
	if (!event) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiUserEventLogTable() passed NULL event pointer\n"));
		return AGENT_ERR_INTERNAL_ERROR;
	}    


        printf("POPULATING USER EVENT WITH RESOURCE OF %d\n", event->Event.Source);

	
	subagent_lock(&hpi_lock_data);
	/* BUILD oid for new row */
		/* assign the number of indices */
	user_evt_idx.len = USER_EVENT_LOG_INDEX_NR;
		/** Index saHpiDomainId is external */
	user_evt_oid[0] = get_domain_id(sessionid);
		/** Index saHpiEventSeverity is external */
	user_evt_oid[1] = event->Event.Severity + 1;
                /** Index saHpiUserEventEntryId is external */
	dr_pair.domainId_resourceId_arry[0] = get_domain_id(sessionid);
	dr_pair.domainId_resourceId_arry[1] = event->Event.Source;
	dr_entry = domain_resource_pair_get(&dr_pair, &dr_table); 
	if (dr_entry == NULL) {
		DEBUGMSGTL ((AGENT, 
		"ERROR: populate_saHpiUserEventLogTable() domain_resource_pair_get returned NULL\n"));
		subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
	}
	user_evt_oid[2] = dr_entry->entry_id++;
		/* assign the indices to the index */
	user_evt_idx.oids = (oid *) & user_evt_oid;
	
	/* create full oid on This row for parent RowPointer */
	column[0] = 1;
	column[1] = COLUMN_SAHPIUSEREVENTLOGTIMESTAMP;
	memset(this_child_oid, 0, MAX_OID_LEN);
	build_full_oid(saHpiUserEventLogTable_oid, saHpiUserEventLogTable_oid_len,
			column, column_len,
			&user_evt_idx,
			this_child_oid, MAX_OID_LEN, this_child_oid_len);	

	/* See if Row exists. */
	user_evt_ctx = NULL;
	user_evt_ctx = CONTAINER_FIND(cb.container, &user_evt_idx);

	if (!user_evt_ctx) { 
		// New entry. Add it
		user_evt_ctx = 
			saHpiUserEventLogTable_create_row(&user_evt_idx);
	}
	else {
		isNewRow = MIB_FALSE;
	}	
	
	if (!user_evt_ctx) {
		snmp_log (LOG_ERR, "Not enough memory for a User Event Log row!");
		rv = AGENT_ERR_INTERNAL_ERROR;
	}


        /** SaHpiTime = ASN_OCTET_STR */
	hpitime_to_snmptime(event->Timestamp, user_evt_ctx->saHpiUserEventLogTimestamp);
	user_evt_ctx->saHpiUserEventLogTimestamp_len = sizeof(SaHpiTimeT);
		
        /** SaHpiTextType = ASN_INTEGER */
        user_evt_ctx->saHpiUserEventLogTextType = 
                event->Event.EventDataUnion.UserEvent.UserEventData.DataType + 1;
		
        /** SaHpiTextLanguage = ASN_INTEGER */
        user_evt_ctx->saHpiUserEventLogTextLanguage = 
                event->Event.EventDataUnion.UserEvent.UserEventData.Language + 1;	
		
        /** SaHpiText = ASN_OCTET_STR */
        memcpy(user_evt_ctx->saHpiUserEventLogText, 
	        event->Event.EventDataUnion.UserEvent.UserEventData.Data,
		event->Event.EventDataUnion.UserEvent.UserEventData.DataLength);
		
	user_evt_ctx->saHpiUserEventLogText_len = 
	        event->Event.EventDataUnion.UserEvent.UserEventData.DataLength;				

        user_evt_ctx->saHpiUserEventLogRowStatus = SAHPIUSEREVENTLOGROWSTATUS_ACTIVE;

	if (isNewRow == MIB_TRUE) {
		CONTAINER_INSERT (cb.container, user_evt_ctx);
	}
		
	user_event_log_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);
	
	return SA_OK;
}

/**
 * 
 * @session_id
 * @resource_id
 * @saHpiEventLogRowPointer
 * @aHpiEventLogRowPointer_len
 * 
 * @return 
 */
SaErrorT user_event_log_clear(SaHpiSessionIdT session_id, 
                              SaHpiResourceIdT resource_id,  
                              oid *saHpiEventLogRowPointer, 
                              size_t saHpiEventLogRowPointer_len,
			      int modifyTotal)
{
	oid user_evt_oid[USER_EVENT_LOG_INDEX_NR];
	netsnmp_index user_evt_idx;
	netsnmp_index *user_index;
	saHpiUserEventLogTable_context *user_evt_ctx;
	
	DR_XREF *dr_entry;
	SaHpiDomainIdResourceIdArrayT dr_pair;	

	int column_len = 2;

        DEBUGMSGTL ((AGENT, "user_event_log_clear, called\n"));
	DEBUGMSGTL ((AGENT, "Attempting to delete user el row with the following indexes:\n"));
        DEBUGMSGTL ((AGENT,"     Domain    [%ld]\n", saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len]));
	DEBUGMSGTL ((AGENT,"     Severity  [%s]\n",  oh_lookup_severity(
						saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len + 1]-1)));
	DEBUGMSGTL ((AGENT,"     Entry id  [%ld]\n", saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len + 2]));

	/* BUILD oid to lookup row */
		/* assign the number of indices */
	user_evt_idx.len = USER_EVENT_LOG_INDEX_NR;
		/** Index saHpiDomainId is external */
	user_evt_oid[0] = saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len];
		/** Index saHpiEventSeverity is external */
	user_evt_oid[1] = saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len + 1];
                /** Index saHpiUserEventEntryId is external */
	user_evt_oid[2] = saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len + 2];
		/* assign the indices to the index */
	user_evt_idx.oids = (oid *) & user_evt_oid;

	user_index = CONTAINER_FIRST(cb.container);
	user_evt_ctx = CONTAINER_FIND(cb.container, &user_evt_idx);
	
	if (!user_evt_ctx) {
		DEBUGMSGTL ((AGENT, "user_event_log_clear did not find a row to delete\n"));
	}
	else {
		
		DEBUGMSGTL ((AGENT, "user_event_log_clear found row to delete\n"));
		CONTAINER_REMOVE(cb.container, user_evt_ctx);
		saHpiUserEventLogTable_delete_row(user_evt_ctx);
		
		/* Reset the entry id for this domain/resource pair */
		dr_pair.domainId_resourceId_arry[0] = saHpiEventLogRowPointer[saHpiUserEventLogTable_oid_len + column_len];
		dr_pair.domainId_resourceId_arry[1] = resource_id;
		dr_entry = domain_resoruce_pair_lookup(&dr_pair, &dr_table); 
		
		if (dr_entry == NULL) {
			DEBUGMSGTL ((AGENT, 
			"ERROR: user_event_log_clear() domain_resoruce_pair_lookup returned NULL\n"));
			return AGENT_ERR_INTERNAL_ERROR;
		}
	
		DEBUGMSGTL ((AGENT, 
			"user_event_log_clear() resetting entry_id to 0\n"));
		dr_entry->entry_id = 0;			
		
		user_event_log_entry_count = CONTAINER_SIZE (cb.container);
		
		if (modifyTotal == MIB_TRUE) {	
			user_event_log_entry_count_total--;
		}	
	}			
	
	
        return SA_OK;
}


/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int user_event_log_add (saHpiUserEventLogTable_context *row_ctx)
{
        SaErrorT                rc = SA_OK;
        SaHpiSessionIdT         session_id;
        SaHpiResourceIdT        resource_id;
        SaHpiEventT             event;

	netsnmp_index user_evt_idx;

        oid this_child_oid[MAX_OID_LEN];
        size_t this_child_oid_len;

	oid column[2];
	int column_len = 2;

        DEBUGMSGTL ((AGENT, "user_event_log_add() called\n"));

          if ((row_ctx->text_type_set     == MIB_TRUE) &&
              (row_ctx->text_language_set == MIB_TRUE) &&
              (row_ctx->text_set          == MIB_TRUE)) {

                if (!row_ctx)
                        return AGENT_ERR_NULL_DATA;

                session_id = get_session_id(row_ctx->index.oids[saHpiDomainId_event_log_INDEX]);
                resource_id = SAHPI_UNSPECIFIED_RESOURCE_ID;

                event.EventType = SAHPI_ET_USER;
                event.Severity  = row_ctx->index.oids[saHpiEventSeverity_event_log_INDEX] - 1;
                event.Source    = SAHPI_UNSPECIFIED_RESOURCE_ID;
                //event.Timestamp = 0; shall be set by hpi implementation 
                memcpy(event.EventDataUnion.UserEvent.UserEventData.Data,
                       row_ctx->saHpiUserEventLogText,
                       row_ctx->saHpiUserEventLogText_len);
                event.EventDataUnion.UserEvent.UserEventData.DataLength =
                        row_ctx->saHpiUserEventLogText_len;
                event.EventDataUnion.UserEvent.UserEventData.DataType =
                        row_ctx->saHpiUserEventLogTextType - 1;
                event.EventDataUnion.UserEvent.UserEventData.Language = 
                        row_ctx->saHpiUserEventLogTextLanguage - 1;


                /* the premise here is we won't set the row to active until */
                /* the event is received from the event handling thread */
                /* at which point we will need to set this tables RowStatus */
                /* to ACTIVE(1) and update the EventTable as well */
                rc = saHpiEventLogEntryAdd (session_id, 
                                            resource_id,
                                            &event);
                if (rc != SA_OK) {
                        snmp_log (LOG_ERR,
                                  "user_event_add: Call to "
                                  "saHpiEventLogentryAdd() failed "
                                  "to set Mode rc: %s.\n",
                                  oh_lookup_error(rc));
                        DEBUGMSGTL ((AGENT,
                                     "user_event_log_add: Call to "
                                     "saHpiEventLogentryAdd() failed to set "
                                     "Mode rc: %s.\n",
                                     oh_lookup_error(rc)));
                        return get_snmp_error(rc);
                }

                row_ctx->saHpiUserEventLogRowStatus = SAHPIUSEREVENTLOGROWSTATUS_ACTIVE;
		
		hpitime_to_snmptime(event.Timestamp, row_ctx->saHpiUserEventLogTimestamp);
		row_ctx->saHpiUserEventLogTimestamp_len = sizeof(SaHpiTimeT);
		
                row_ctx->saHpiEventAdd_called = MIB_TRUE;

                /* create full oid on This row for parent RowPointer */
                column[0] = 1;
                column[1] = COLUMN_SAHPIUSEREVENTLOGTIMESTAMP;
                memset(this_child_oid, 0, MAX_OID_LEN);
                build_full_oid(saHpiUserEventLogTable_oid, 
                               saHpiUserEventLogTable_oid_len,
                               column, column_len,
                               &user_evt_idx,
                               this_child_oid, MAX_OID_LEN, &this_child_oid_len);

                /* Now Push newly craeted RowPointer and data up to */
                /* saHpiEventLogTable */
                rc = event_log_add(session_id, resource_id, &event, 
                                   this_child_oid, this_child_oid_len);


        }  


        return SNMP_ERR_NOERROR; 
}

/**
 * 
 * @row_ctx:
 * 
 * @return: 
 */
int user_event_log_delete (saHpiUserEventLogTable_context *row_ctx)
{
        DEBUGMSGTL ((AGENT, "user_event_add() called\n"));
                    
        return SNMP_ERR_NOERROR;
}


/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiUserEventLogLifetimeChanges(netsnmp_mib_handler *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */
        
        DEBUGMSGTL ((AGENT, "handle_saHpiUserEvenLogtLifetimeChanges, called\n"));

        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_COUNTER,
        			        (u_char *) &user_event_log_entry_count_total,
        			        sizeof(user_event_log_entry_count_total));
            break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiUserEventLogActiveEntries(netsnmp_mib_handler *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiUserEventLogActiveEntries, called\n"));

        user_event_log_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			        (u_char *) &user_event_log_entry_count,
        			        sizeof(user_event_log_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
}

/**
 * 
 * @return: 
 */
int initialize_table_saHpiUserEventLogLifetimeChanges(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiUserEvenLogtLifetimeChanges, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiUserEventLogLifetimeChanges", 
				        handle_saHpiUserEventLogLifetimeChanges,
                                        saHpiUserEventLogLifetimeChanges_oid,
			                OID_LENGTH(saHpiUserEventLogLifetimeChanges_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}


/**
 * 
 * @return: 
 */
int initialize_table_saHpiUserEventLogActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiUserEventLogActiveEntries, called\n"));

        netsnmp_register_scalar(
                                netsnmp_create_handler_registration(
				        "saHpiUserEventLogActiveEntries", 
				        handle_saHpiUserEventLogActiveEntries,
                                        saHpiUserEventLogActiveEntries_oid, 
					OID_LENGTH(saHpiUserEventLogActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));

        return SNMP_ERR_NOERROR;
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiUserEventLogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiUserEventLogTable_cmp( const void *lhs, const void *rhs )
{
        saHpiUserEventLogTable_context *context_l =
        (saHpiUserEventLogTable_context *)lhs;
        saHpiUserEventLogTable_context *context_r =
        (saHpiUserEventLogTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiUserEventLogTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiUserEventLogTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiEventSeverity */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiEventSeverity index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiUserEventEntryId */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2])
			        return 0;
		}
	}
	
	return 0;
}

/************************************************************
 * Initializes the saHpiUserEventLogTable module
 */
void
init_saHpiUserEventLogTable(void)
{
        
        DEBUGMSGTL ((AGENT, "init_saHpiUserEventLogTable, called\n"));
		
	initialize_table_saHpiUserEventLogTable();

        initialize_table_saHpiUserEventLogLifetimeChanges();
        initialize_table_saHpiUserEventLogActiveEntries();
	
        domain_resource_pair_initialize(&initialized, &dr_table);
	
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiUserEventLogTable_row_copy(saHpiUserEventLogTable_context * dst,
                         saHpiUserEventLogTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    subagent_lock(&hpi_lock_data);

    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
        return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    memcpy( dst->saHpiUserEventLogTimestamp, src->saHpiUserEventLogTimestamp, src->saHpiUserEventLogTimestamp_len );
    dst->saHpiUserEventLogTimestamp_len = src->saHpiUserEventLogTimestamp_len;

    dst->saHpiUserEventLogTextType = src->saHpiUserEventLogTextType;

    dst->saHpiUserEventLogTextLanguage = src->saHpiUserEventLogTextLanguage;

    memcpy( dst->saHpiUserEventLogText, src->saHpiUserEventLogText, src->saHpiUserEventLogText_len );
    dst->saHpiUserEventLogText_len = src->saHpiUserEventLogText_len;

    dst->saHpiUserEventLogRowStatus = src->saHpiUserEventLogRowStatus;

    subagent_unlock(&hpi_lock_data);
    return 0;
}


/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiUserEventLogTable_extract_index( saHpiUserEventLogTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiEventSeverity;
        netsnmp_variable_list var_saHpiUserEventEntryId;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiUserEventLogTable_extract_index, called\n"));

        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
                if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
                                    hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
                        return -1;
                }
                ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiEventSeverity;

        memset( &var_saHpiEventSeverity, 0x00, sizeof(var_saHpiEventSeverity) );
        var_saHpiEventSeverity.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiEventSeverity.next_variable = &var_saHpiUserEventEntryId;

        memset( &var_saHpiUserEventEntryId, 0x00, sizeof(var_saHpiUserEventEntryId) );
        var_saHpiUserEventEntryId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiUserEventEntryId.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
                /*
                 * copy index components into the context structure
                 */
                /** skipping external index saHpiDomainId */
   
                /** skipping external index saHpiEventSeverity */
   
                /** skipping external index saHpiUserEventEntryId */
   
                err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
		err = saHpiEventSeverity_check_index(*var_saHpiEventSeverity.val.integer);
		err = saHpiUserEventEntryId_check_index(*var_saHpiUserEventEntryId.val.integer);

        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiUserEventLogTable_can_activate(saHpiUserEventLogTable_context *undo_ctx,
                      saHpiUserEventLogTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiUserEventLogTable_can_deactivate(saHpiUserEventLogTable_context *undo_ctx,
                        saHpiUserEventLogTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiUserEventLogTable_can_delete(saHpiUserEventLogTable_context *undo_ctx,
                    saHpiUserEventLogTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiUserEventLogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

#ifdef saHpiUserEventLogTable_ROW_CREATION
/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiUserEventLogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiUserEventLogTable_context *
saHpiUserEventLogTable_create_row( netsnmp_index* hdr)
{
    saHpiUserEventLogTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiUserEventLogTable_context);
    if(!ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);
        
    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiUserEventLogTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
        return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiUserEventLogTimestamp = 0;
     ctx->saHpiUserEventLogTextType = 0;
     ctx->saHpiUserEventLogTextLanguage = 0;
     ctx->saHpiUserEventLogText = 0;
     ctx->saHpiUserEventLogRowStatus = 0;
    */
    user_event_log_entry_count_total++;

    subagent_unlock(&hpi_lock_data);
    return ctx;
}
#endif

/************************************************************
 * the *_duplicate row routine
 */
saHpiUserEventLogTable_context *
saHpiUserEventLogTable_duplicate_row( saHpiUserEventLogTable_context * row_ctx)
{
    saHpiUserEventLogTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);

    dup = SNMP_MALLOC_TYPEDEF(saHpiUserEventLogTable_context);
    if(!dup) {
            subagent_unlock(&hpi_lock_data);
            return NULL;
    }
        
    if(saHpiUserEventLogTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiUserEventLogTable_delete_row( saHpiUserEventLogTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        subagent_lock(&hpi_lock_data);

    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    user_event_log_entry_count_total++;

    subagent_unlock(&hpi_lock_data);
    return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiUserEventLogTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiUserEventLogTable_context *row_ctx =
            (saHpiUserEventLogTable_context *)rg->existing_row;
//    saHpiUserEventLogTable_context *undo_ctx =
//            (saHpiUserEventLogTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;


    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            rc = netsnmp_check_vb_type_and_size(var, ASN_OCTET_STR,
                                                sizeof(row_ctx->saHpiUserEventLogTimestamp));

            /* check rowstatus is in correct state, createANDwait */
            if ( rc != SNMP_ERR_NOERROR )  {
                    DEBUGMSGTL ((AGENT, "saHpiUserEventLogTable_set_reserve1, "
                                        "netsnmp_check_vb_type_and_size failed\n"));
            }

            /* check rowstatus is in correct state, createANDwait */
            if ((rc == SNMP_ERR_NOERROR) && 
                (row_ctx->saHpiUserEventLogRowStatus != SNMP_ROW_CREATEANDWAIT)) {
                    rc = SNMP_ERR_NOCREATION;
                    DEBUGMSGTL ((AGENT, "saHpiUserEventLogTable_set_reserve1, "
                                        "SNMP_ERR_NOCREATION\n"));
            }
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiUserEventLogTextType));
            /* check rowstatus is in correct state, createANDwait */
            if ((rc == SNMP_ERR_NOERROR) && 
                (row_ctx->saHpiUserEventLogRowStatus != SNMP_ROW_CREATEANDWAIT))
                    rc = SNMP_ERR_NOCREATION;

        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiUserEventLogTextLanguage));
            /* check rowstatus is in correct state, createANDwait */
            if ((rc == SNMP_ERR_NOERROR) && 
                (row_ctx->saHpiUserEventLogRowStatus != SNMP_ROW_CREATEANDWAIT))
                    rc = SNMP_ERR_NOCREATION;

        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
	    rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
            if (rc == SNMP_ERR_NOERROR ) {
            	    if (var->val_len > sizeof(row_ctx->saHpiUserEventLogText)) {
            		    rc = SNMP_ERR_WRONGLENGTH;
                    }
            }	    	    
            /* check rowstatus is in correct state, createANDwait */
            if ((rc == SNMP_ERR_NOERROR) && 
                (row_ctx->saHpiUserEventLogRowStatus != SNMP_ROW_CREATEANDWAIT))
                    rc = SNMP_ERR_NOCREATION;
        break;

        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiUserEventLogRowStatus));
           /* check for valid status and state transitions*/
            if ((rc == SNMP_ERR_NOERROR ) &&  (rg->row_created) && 
                (*var->val.integer != SNMP_ROW_CREATEANDWAIT)) {
                    DEBUGMSGTL ((AGENT, 
                    "COLUMN_SAHPIUSEREVENTLOGROWSTATUS, should be SNMP_ROW_CREATEANDWAIT on new row\n"));
                    rc = SNMP_ERR_WRONGVALUE;
            } else if ((rc == SNMP_ERR_NOERROR ) && (!rg->row_created)) {
                    if ((row_ctx->saHpiUserEventLogRowStatus != SNMP_ROW_DESTROY) &&
                        (*var->val.integer == SNMP_ROW_DESTROY)) {
                            rc = SNMP_ERR_NOERROR;
                    } else {
                            DEBUGMSGTL ((AGENT, 
                            "COLUMN_SAHPIUSEREVENTLOGROWSTATUS, SNMP_ERR_INCONSISTENTVALUE\n"));
                            rc = SNMP_ERR_INCONSISTENTVALUE;
                    }
            }
            /* if this is a new row check the validity of the user specified EntryId */
            if (((rg->row_created)) && (rc == SNMP_ERR_NOERROR)) {
                    /* check for valid saHpiUserEventLogEntryId specified */
                    /* this is not the same as the EntryId obtained from HPI */
                    /* it is maintained in the subagent                     */
                    DR_XREF *dr_entry;
                    SaHpiDomainIdResourceIdArrayT dr_pair;

                    dr_pair.domainId_resourceId_arry[0] = 
                            row_ctx->index.oids[saHpiDomainId_event_log_INDEX];
                    dr_pair.domainId_resourceId_arry[1] = 
                            SAHPI_UNSPECIFIED_RESOURCE_ID;

                    /* domainid, resourceid, */
                    dr_entry = domain_resource_pair_get(&dr_pair, &dr_table);
                    if (dr_entry == NULL) {
                            DEBUGMSGTL ((AGENT, 
                                         "ERROR: saHpiUserEventLogTable_set_reserve1()"
                                         " domain_resource_pair_get returned NULL\n"));
                            rc = SNMP_ERR_GENERR;
                    }

                    if ( row_ctx->index.oids[saHpiUserEventEntryId_event_log_INDEX] >= dr_entry->entry_id ) {
                            dr_entry->entry_id = row_ctx->index.oids[saHpiUserEventEntryId_event_log_INDEX];
                            dr_entry->entry_id++;
                    } else {
                            DEBUGMSGTL ((AGENT,"***********************************************\n"));
                            DEBUGMSGTL ((AGENT,"dr_entry->entry_id [%d]\n", dr_entry->entry_id));
                            DEBUGMSGTL ((AGENT,"row_ctx->index.oids[saHpiUserEventEntryId_event_log_INDEX] [%ld]\n", 
                                         row_ctx->index.oids[saHpiUserEventEntryId_event_log_INDEX]));
                            DEBUGMSGTL ((AGENT, 
                                         "ERROR: saHpiUserEventLogTable_set_reserve1() "
                                         "User specified UserEventLogEntryId invalid!!!\n"));
                            DEBUGMSGTL ((AGENT,"***********************************************\n"));
                            rc = SNMP_ERR_WRONGVALUE;
                    }
            }
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiUserEventLogTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiUserEventLogTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiUserEventLogTable_context *row_ctx = (saHpiUserEventLogTable_context *)rg->existing_row;
//    saHpiUserEventLogTable_context *undo_ctx = (saHpiUserEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
                if (oh_lookup_texttype(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
                if (oh_lookup_language(*var->val.integer - 1) == NULL)
                        rc = SNMP_ERR_BADVALUE;
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiUserEventLogTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiUserEventLogTable_context *row_ctx = (saHpiUserEventLogTable_context *)rg->existing_row;
//    saHpiUserEventLogTable_context *undo_ctx = (saHpiUserEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int            row_err = 0;
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
	    hpitime_to_snmptime((SaHpiTimeT)*var->val.integer, row_ctx->saHpiUserEventLogTimestamp);
	    row_ctx->saHpiUserEventLogTimestamp_len = sizeof(SaHpiTimeT);
            row_ctx->timestamp_set = MIB_TRUE;
            if (row_ctx->saHpiUserEventLogRowStatus == SNMP_ROW_CREATEANDWAIT) {
                    row_err = user_event_log_add (row_ctx);
            } 
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            row_ctx->saHpiUserEventLogTextType = *var->val.integer;
            row_ctx->text_type_set = MIB_TRUE;
            if (row_ctx->saHpiUserEventLogRowStatus == SNMP_ROW_CREATEANDWAIT) {
                    row_err = user_event_log_add (row_ctx);
            } 

        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            row_ctx->saHpiUserEventLogTextLanguage = *var->val.integer;
            row_ctx->text_language_set = MIB_TRUE;
            if (row_ctx->saHpiUserEventLogRowStatus == SNMP_ROW_CREATEANDWAIT) {
                    row_err = user_event_log_add (row_ctx);
            } 
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiUserEventLogText,var->val.string,var->val_len);
            row_ctx->saHpiUserEventLogText_len = var->val_len;
            row_ctx->text_set = MIB_TRUE;
            if (row_ctx->saHpiUserEventLogRowStatus == SNMP_ROW_CREATEANDWAIT) {
                    row_err = user_event_log_add (row_ctx);
            } 

        break;

        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
            if (*var->val.integer == SNMP_ROW_DESTROY) {
                    if (row_ctx->saHpiUserEventLogRowStatus == SNMP_ROW_ACTIVE) {
                            row_err = user_event_log_delete (row_ctx);
                    }
                    rg->row_deleted = 1;
            }
            row_ctx->saHpiUserEventLogRowStatus = *var->val.integer;

        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }

    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiUserEventLogTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiUserEventLogTable_context *row_ctx = (saHpiUserEventLogTable_context *)rg->existing_row;
//    saHpiUserEventLogTable_context *undo_ctx = (saHpiUserEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiUserEventLogTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiUserEventLogTable_context *row_ctx = (saHpiUserEventLogTable_context *)rg->existing_row;
//    saHpiUserEventLogTable_context *undo_ctx = (saHpiUserEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
	        break;
            /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiUserEventLogTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiUserEventLogTable_context *row_ctx = (saHpiUserEventLogTable_context *)rg->existing_row;
//    saHpiUserEventLogTable_context *undo_ctx = (saHpiUserEventLogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
        break;

        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}



/************************************************************
 *
 * Initialize the saHpiUserEventLogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiUserEventLogTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiUserEventLogTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiUserEventLogTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiUserEventLogTable_oid,
                                             saHpiUserEventLogTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiUserEventLogTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiEventSeverity */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiUserEventEntryId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiUserEventLogTable_COL_MIN;
    table_info->max_column = saHpiUserEventLogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiUserEventLogTable_get_value;
    cb.container = netsnmp_container_find("saHpiUserEventLogTable_primary:"
                                          "saHpiUserEventLogTable:"
                                          "table_container");
#ifdef saHpiUserEventLogTable_IDX2
    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiUserEventLogTable_secondary:"
                                                       "saHpiUserEventLogTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiUserEventLogTable_cmp;
#endif
#ifdef saHpiUserEventLogTable_SET_HANDLING
    cb.can_set = 1;
#ifdef saHpiUserEventLogTable_ROW_CREATION
    cb.create_row = (UserRowMethod*)saHpiUserEventLogTable_create_row;
#endif
    cb.duplicate_row = (UserRowMethod*)saHpiUserEventLogTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiUserEventLogTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiUserEventLogTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiUserEventLogTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiUserEventLogTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiUserEventLogTable_can_delete;

    cb.set_reserve1 = saHpiUserEventLogTable_set_reserve1;
    cb.set_reserve2 = saHpiUserEventLogTable_set_reserve2;
    cb.set_action = saHpiUserEventLogTable_set_action;
    cb.set_commit = saHpiUserEventLogTable_set_commit;
    cb.set_free = saHpiUserEventLogTable_set_free;
    cb.set_undo = saHpiUserEventLogTable_set_undo;
#endif
    DEBUGMSGTL(("initialize_table_saHpiUserEventLogTable",
                "Registering table saHpiUserEventLogTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiUserEventLogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiUserEventLogTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiUserEventLogTable_context *context = (saHpiUserEventLogTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIUSEREVENTLOGTIMESTAMP:
            /** SaHpiTime = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiUserEventLogTimestamp,
                         sizeof(context->saHpiUserEventLogTimestamp) );
        break;
    
        case COLUMN_SAHPIUSEREVENTLOGTEXTTYPE:
            /** SaHpiTextType = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiUserEventLogTextType,
                         sizeof(context->saHpiUserEventLogTextType) );
        break;
    
        case COLUMN_SAHPIUSEREVENTLOGTEXTLANGUAGE:
            /** SaHpiTextLanguage = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiUserEventLogTextLanguage,
                         sizeof(context->saHpiUserEventLogTextLanguage) );
        break;
    
        case COLUMN_SAHPIUSEREVENTLOGTEXT:
            /** SaHpiText = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiUserEventLogText,
                         context->saHpiUserEventLogText_len );
        break;
    
        case COLUMN_SAHPIUSEREVENTLOGROWSTATUS:
            /** RowStatus = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiUserEventLogRowStatus,
                         sizeof(context->saHpiUserEventLogRowStatus) );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiUserEventLogTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiUserEventLogTable_get_by_idx
 */
const saHpiUserEventLogTable_context *
saHpiUserEventLogTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiUserEventLogTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


