/*
 * (C) Copyright IBM Corp. 2005
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   David Judkovics  <djudkovi@us.ibm.com>
 *   Daniel de Araujo <ddearauj@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18.2.2 2004/02/09 18:21:47 rstory Exp $
 *
 * $Id: saHpiWatchdogTable.c 5428 2005-11-14 22:19:59Z djudkovi $
 *
 *
 * For help understanding NET-SNMP in general, please check the 
 *     documentation and FAQ at:
 *
 *     http://www.net-snmp.org/
 *
 *
 * For help understanding this code, the agent and how it processes
 *     requests, please check the following references.
 *
 *     http://www.net-snmp.org/tutorial-5/
 *
 *
 * You can also join the #net-snmp channel on irc.freenode.net
 *     and ask for help there.
 *
 *
 * And if all else fails, send a detailed message to the developers
 *     describing the problem you are having to:
 *
 *    net-snmp-coders@lists.sourceforge.net
 *
 *
 * Yes, there is lots of code here that you might not use. But it is much
 * easier to remove code than to add it!
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <SaHpi.h>
#include "saHpiWatchdogTable.h"
#include <hpiSubagent.h>
#include <hpiCheckIndice.h>
#include <session_info.h>
#include <oh_utils.h>
#include <hpiLock.h>


static     netsnmp_handler_registration *my_handler = NULL;
static     netsnmp_table_array_callbacks cb;

oid saHpiWatchdogTable_oid[] = { saHpiWatchdogTable_TABLE_OID };
size_t saHpiWatchdogTable_oid_len = OID_LENGTH(saHpiWatchdogTable_oid);

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/

/*************************************************************
 * oid and fucntion declarations scalars
 */
static int set_table_watchdog_set (saHpiWatchdogTable_context *row_ctx); 
static int set_table_watchdog_reset (saHpiWatchdogTable_context *row_ctx); 

static u_long watchdog_entry_count = 0; 
static oid saHpiWatchdogActiveEntries_oid[] = { 1,3,6,1,4,1,18568,2,1,1,4,3 };

int handle_saHpiWatchdogActiveEntries(netsnmp_mib_handler *handler,
                                   netsnmp_handler_registration *reginfo,
                                   netsnmp_agent_request_info   *reqinfo,
                                   netsnmp_request_info         *requests);

int initialize_table_saHpiWatchdogActiveEntries(void);



/**
 * 
 * @sessionid:
 * @rdr_entry:
 * @rpt_entry:
 * @full_oid:
 * @full_oid_len:
 * @child_oid:
 * @child_oid_len:
 * 
 * @return 
 */
SaErrorT populate_watchdog(SaHpiSessionIdT sessionid, 
                           SaHpiRdrT *rdr_entry,
                           SaHpiRptEntryT *rpt_entry,
                           oid *full_oid, size_t full_oid_len,
                           oid *child_oid, size_t *child_oid_len)
{

        DEBUGMSGTL ((AGENT, "populate_watchdog, called\n"));

        SaErrorT rv = SA_OK;
        int new_row = MIB_FALSE;
        SaHpiWatchdogT watchdog;
        SaHpiTextBufferT buffer;

        oid watchdog_oid[WATCHDOG_INDEX_NR];
        netsnmp_index watchdog_idx;
        saHpiWatchdogTable_context *watchdog_ctx;

        oid column[2];
        int column_len = 2;

        DEBUGMSGTL ((AGENT, "SAHPI_WATCHDOG_RDR populate_watchdog() called\n"));

        /* check for NULL pointers */
        if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_watchdog() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: populate_watchdog() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        
	subagent_lock(&hpi_lock_data);
	/* BUILD oid for new row */
        /* assign the number of indices */
        watchdog_idx.len = WATCHDOG_INDEX_NR;
        /** Index saHpiDomainId is external */
        watchdog_oid[0] = get_domain_id(sessionid);
        /** Index saHpiResourceId is external */
        watchdog_oid[1] = rpt_entry->ResourceId;
        /** Index saHpiResourceIsHistorical is external */
        watchdog_oid[2] = MIB_FALSE;
        /** Index saHpiSensorNum */
        watchdog_oid[3] = rdr_entry->RdrTypeUnion.WatchdogRec.WatchdogNum;
        /* assign the indices to the index */
        watchdog_idx.oids = (oid *) & watchdog_oid;

        /* create full oid on This row for parent RowPointer */
        column[0] = 1;
        column[1] = COLUMN_SAHPIWATCHDOGLOG;
        memset(child_oid, 0, MAX_OID_LEN);
        build_full_oid(saHpiWatchdogTable_oid, saHpiWatchdogTable_oid_len,
                       column, column_len,
                       &watchdog_idx,
                       child_oid, MAX_OID_LEN, child_oid_len);

        /* See if Row exists. */
        watchdog_ctx = NULL;
        watchdog_ctx = CONTAINER_FIND(cb.container, &watchdog_idx);

        if (!watchdog_ctx) {
                // New entry. Add it
                watchdog_ctx = 
                saHpiWatchdogTable_create_row(&watchdog_idx);
                new_row = MIB_TRUE;
        }
        if (!watchdog_ctx) {
                snmp_log (LOG_ERR, "Not enough memory for a Annunciator row!");
                subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
        }

        rv = saHpiWatchdogTimerGet (sessionid,
                                    rpt_entry->ResourceId,
                                    rdr_entry->RdrTypeUnion.WatchdogRec.WatchdogNum,
                                    &watchdog);
        if (rv != SA_OK) {
                subagent_unlock(&hpi_lock_data);
		return AGENT_ERR_INTERNAL_ERROR;
        }


        /** SaHpiInstrumentId = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogNum = 
                rdr_entry->RdrTypeUnion.WatchdogRec.WatchdogNum;

        /** TruthValue = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogLog = 
                (watchdog.Log == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

        /** TruthValue = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogRunning = 
                (watchdog.Running == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

        /** SaHpiWatchdogTimerUse = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogTimerUseState = 
                watchdog.TimerUse + 1;

        /** INTEGER = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogTimerAction =
                watchdog.TimerAction + 1;

        /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogPretimerInterrupt =
                watchdog.PretimerInterrupt + 1;

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogPreTimeoutInterval =
                watchdog.PreTimeoutInterval;

        /** OCTETSTR = ASN_OCTET_STR */
        rv = oh_decode_exp_flags(watchdog.TimerUseExpFlags, &buffer);
        watchdog_ctx->saHpiWatchdogTimerUseExpFlags_len = buffer.DataLength;
        memcpy(watchdog_ctx->saHpiWatchdogTimerUseExpFlags,
               buffer.Data,
               buffer.DataLength);

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogTimerInitialCount = watchdog.InitialCount;

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogTimerPresentCount = watchdog.PresentCount;

        /** TruthValue = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogTimerReset = MIB_FALSE;  

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogOem = 
                rdr_entry->RdrTypeUnion.WatchdogRec.Oem;

        /** RowPointer = ASN_OBJECT_ID */
        memset(watchdog_ctx->saHpiWatchdogRDR, 
               0, 
               sizeof(watchdog_ctx->saHpiWatchdogRDR));
        watchdog_ctx->saHpiWatchdogRDR_len = 
        full_oid_len * sizeof(oid);
        memcpy(watchdog_ctx->saHpiWatchdogRDR, 
               full_oid, 
               watchdog_ctx->saHpiWatchdogRDR_len);

        if (new_row == MIB_TRUE) 
                CONTAINER_INSERT (cb.container, watchdog_ctx);

        watchdog_entry_count = CONTAINER_SIZE (cb.container);

        subagent_unlock(&hpi_lock_data);
	
	return rv;
}

/**
 * 
 * @sessionid
 * @event
 * @rdr_entry
 * @rpt_entry
 * 
 * @return 
 */
SaErrorT async_watchdog_add(SaHpiSessionIdT sessionid, SaHpiEventT *event, 
                            SaHpiRdrT *rdr_entry, SaHpiRptEntryT *rpt_entry)
{

        SaErrorT rv = SA_OK;
        SaHpiWatchdogT watchdog;
        SaHpiTextBufferT buffer;

        oid watchdog_oid[WATCHDOG_INDEX_NR];
        netsnmp_index watchdog_idx;
        saHpiWatchdogTable_context *watchdog_ctx;

        DEBUGMSGTL ((AGENT, "async_watchdog_add(), called\n"));

        /* check for NULL pointers */
        if (!rdr_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: async_watchdog_add() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }
        if (!rpt_entry) {
                DEBUGMSGTL ((AGENT, 
                             "ERROR: async_watchdog_add() passed NULL rdr_entry pointer\n"));
                return AGENT_ERR_INTERNAL_ERROR;
        }

        /* BUILD oid for new row */
        /* assign the number of indices */
        watchdog_idx.len = WATCHDOG_INDEX_NR;
        /** Index saHpiDomainId is external */
        watchdog_oid[0] = get_domain_id(sessionid);
        /** Index saHpiResourceId is external */
        watchdog_oid[1] = rpt_entry->ResourceId;
        /** Index saHpiResourceIsHistorical is external */
        watchdog_oid[2] = MIB_FALSE;
        /** Index saHpiSensorNum */
        watchdog_oid[3] = rdr_entry->RdrTypeUnion.WatchdogRec.WatchdogNum;
        /* assign the indices to the index */
        watchdog_idx.oids = (oid *) & watchdog_oid;

        /* See if Row exists. */
        watchdog_ctx = NULL;
        watchdog_ctx = CONTAINER_FIND(cb.container, &watchdog_idx);

        if (!watchdog_ctx) {
                DEBUGMSGTL ((AGENT, "async_watchdog_add(): row not found\n"));
                return AGENT_ERR_NOT_FOUND;
        }

        rv = saHpiWatchdogTimerGet (sessionid,
                                    rpt_entry->ResourceId,
                                    rdr_entry->RdrTypeUnion.WatchdogRec.WatchdogNum,
                                    &watchdog);
        if (rv != SA_OK) {
                return AGENT_ERR_INTERNAL_ERROR;
        }


        /** SaHpiInstrumentId = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogNum = 
                rdr_entry->RdrTypeUnion.WatchdogRec.WatchdogNum;

        /** TruthValue = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogLog = 
                (watchdog.Log == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

        /** TruthValue = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogRunning = 
                (watchdog.Running == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

        /** SaHpiWatchdogTimerUse = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogTimerUseState = 
                watchdog.TimerUse + 1;

        /** INTEGER = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogTimerAction =
                watchdog.TimerAction + 1;

        /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogPretimerInterrupt =
                watchdog.PretimerInterrupt + 1;

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogPreTimeoutInterval =
                watchdog.PreTimeoutInterval;

        /** OCTETSTR = ASN_OCTET_STR */
        rv = oh_decode_exp_flags(watchdog.TimerUseExpFlags, &buffer);
        watchdog_ctx->saHpiWatchdogTimerUseExpFlags_len = buffer.DataLength;
        memcpy(watchdog_ctx->saHpiWatchdogTimerUseExpFlags,
               buffer.Data,
               buffer.DataLength);

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogTimerInitialCount = watchdog.InitialCount;

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogTimerPresentCount = watchdog.PresentCount;

        /** TruthValue = ASN_INTEGER */
        watchdog_ctx->saHpiWatchdogTimerReset = MIB_FALSE;  

        /** UNSIGNED32 = ASN_UNSIGNED */
        watchdog_ctx->saHpiWatchdogOem = 
                rdr_entry->RdrTypeUnion.WatchdogRec.Oem;

        /** RowPointer = ASN_OBJECT_ID */
        /* don't have  all the necessary data AND if we found the Row   */
        /* this should not change                                       */
        /*
        memset(watchdog_ctx->saHpiWatchdogRDR, 
               0, 
               sizeof(watchdog_ctx->saHpiWatchdogRDR));
        watchdog_ctx->saHpiWatchdogRDR_len = 
        full_oid_len * sizeof(oid);
        memcpy(watchdog_ctx->saHpiWatchdogRDR, 
               full_oid, 
               watchdog_ctx->saHpiWatchdogRDR_len);
        */

        return rv;
}


/**
 * 
 * @domainId
 * @resourceId
 * 
 * @return 
 */
SaErrorT clear_watchdog(SaHpiDomainIdT domainId, 
                        SaHpiResourceIdT resourceId)

{
        SaErrorT rv = SA_OK;
        netsnmp_index *row_idx;
        saHpiWatchdogTable_context *ctx;

	DEBUGMSGTL ((AGENT, "clear_watchdog, called\n"));	
	DEBUGMSGTL ((AGENT, "           domainId   [%d]\n", domainId));	
	DEBUGMSGTL ((AGENT, "           resourceId [%d]\n", resourceId));

        row_idx = CONTAINER_FIRST(cb.container);
        if (row_idx) //At least one entry was found.
        {
                do {
                        /* based on the found row_idx get the pointer   */
                        /* to its context (row data)                    */
                        ctx = CONTAINER_FIND(cb.container, row_idx);

                        /* before we delete the context we should get the  */
                        /* next row (context) if any before we delete this */ 
                        /* one.                                            */
                        row_idx = CONTAINER_NEXT(cb.container, row_idx);

                        if ((ctx->index.oids[watchdog_saHpiDomainId_INDEX] ==
                             domainId) &&

                            (ctx->index.oids[watchdog_saHpiResourceEntryId_INDEX] ==
                             resourceId)) {

                                /* all conditions met remove row */
                                CONTAINER_REMOVE (cb.container, ctx);
                                saHpiWatchdogTable_delete_row (ctx);
                                watchdog_entry_count = 
                                        CONTAINER_SIZE (cb.container);
                                DEBUGMSGTL ((AGENT, "clear_watchdog: "
                                                    "found row: removing\n"));

                        }

                } while (row_idx);
        } 

        return rv;
}

static int set_table_watchdog_set (saHpiWatchdogTable_context *row_ctx) 
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
        SaHpiWatchdogNumT   watchdog_num;
	SaHpiWatchdogT      watchdog; 

        SaHpiTextBufferT buffer;

	DEBUGMSGTL ((AGENT, "set_table_watchdog_set, called\n"));  

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

        /* set the session */
        session_id = get_session_id(row_ctx->index.oids[watchdog_saHpiDomainId_INDEX]);

        /* set the session */
        resource_id = row_ctx->index.oids[watchdog_saHpiResourceEntryId_INDEX];

        /* set the WatchdogNum */
        watchdog_num = row_ctx->saHpiWatchdogNum;

        /* set the InitialCount */
        watchdog.InitialCount = row_ctx->saHpiWatchdogTimerInitialCount;

        /* set the Log */
        watchdog.Log = row_ctx->saHpiWatchdogLog;

        /* set the PresentCount */
        watchdog.PresentCount = row_ctx->saHpiWatchdogTimerPresentCount;

        /* set the PreTimeoutInterval */
        watchdog.PreTimeoutInterval = row_ctx->saHpiWatchdogPreTimeoutInterval;

        /* set the PretimerInterrupt */
        watchdog.PretimerInterrupt = row_ctx->saHpiWatchdogPretimerInterrupt - 1;

        /* set the Running */
        watchdog.Running = row_ctx->saHpiWatchdogRunning;

        /* set the TimerAction */
        watchdog.TimerAction = row_ctx->saHpiWatchdogTimerAction - 1;

        /* set the TimerUse */
        watchdog.TimerUse = row_ctx->saHpiWatchdogTimerUseState - 1;

        /* set the exp flags */
        buffer.DataLength = row_ctx->saHpiWatchdogTimerUseExpFlags_len;
        memcpy(buffer.Data, row_ctx->saHpiWatchdogTimerUseExpFlags, row_ctx->saHpiWatchdogTimerUseExpFlags_len);
        watchdog.TimerUseExpFlags = 0;

        if (oh_encode_exp_flags(&buffer, &watchdog.TimerUseExpFlags) != SNMP_ERR_NOERROR) {
                return SNMP_ERR_GENERR;
        }

	rc = saHpiWatchdogTimerSet (session_id, resource_id, 
                                    watchdog_num, &watchdog);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "set_table_watchdog_set: Call to saHpiWatchdogTimerSet failed to set Mode rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "set_table_watchdog_set: Call to saHpiWatchdogTimerSet failed to set Mode rc: %s.\n",
			     oh_lookup_error(rc)));
		return get_snmp_error(rc);
	}  else {
                DEBUGMSGTL ((AGENT, "set_table_watchdog_set: Success!!!\n"));

        }

	return SNMP_ERR_NOERROR;
}


static int set_table_watchdog_reset (saHpiWatchdogTable_context *row_ctx) 
{
	SaErrorT            rc = SA_OK;
	SaHpiSessionIdT     session_id;
	SaHpiResourceIdT    resource_id;
        SaHpiWatchdogNumT   watchdog_num;

        //SaHpiTextBufferT buffer;

	DEBUGMSGTL ((AGENT, "set_table_watchdog_set, called\n"));  

	if (!row_ctx)
		return AGENT_ERR_NULL_DATA;

        /* set the session */
        session_id = get_session_id(row_ctx->index.oids[watchdog_saHpiDomainId_INDEX]);

        /* set the session */
        resource_id = row_ctx->index.oids[watchdog_saHpiResourceEntryId_INDEX];

        /* set the WatchdogNum */
        watchdog_num = row_ctx->saHpiWatchdogNum;

	rc = saHpiWatchdogTimerReset (session_id, resource_id, watchdog_num);

	if (rc != SA_OK) {
		snmp_log (LOG_ERR,
			  "set_table_watchdog_reset: Call to saHpiWatchdogTimerReset failed to set Mode rc: %s.\n",
			  oh_lookup_error(rc));
		DEBUGMSGTL ((AGENT,
			     "set_table_watchdog_reset: Call to saHpiWatchdogTimerReset failed to set Mode rc: %s.\n",
			     oh_lookup_error(rc)));
		return get_snmp_error(rc);
	}  else {
                DEBUGMSGTL ((AGENT, "set_table_watchdog_reset: Success!!!\n"));

        }

	return SNMP_ERR_NOERROR;
}


/**
 * 
 * @handler:
 * @reginfo:
 * @reqinfo:
 * @requests:
 * 
 * @return:
 */
int
handle_saHpiWatchdogActiveEntries(netsnmp_mib_handler *handler,
                          netsnmp_handler_registration *reginfo,
                          netsnmp_agent_request_info   *reqinfo,
                          netsnmp_request_info         *requests)
{
        /* We are never called for a GETNEXT if it's registered as a
           "instance", as it's "magically" handled for us.  */
        /* a instance handler also only hands us one request at a time, so
           we don't need to loop over a list of requests; we'll only get one. */

        DEBUGMSGTL ((AGENT, "handle_saHpiWatchdogActiveEntries, called\n"));

        watchdog_entry_count = CONTAINER_SIZE (cb.container);
        
        switch(reqinfo->mode) {

        case MODE_GET:
                snmp_set_var_typed_value(requests->requestvb, ASN_GAUGE,
        			        (u_char *) &watchdog_entry_count,
        			        sizeof(watchdog_entry_count));
                break;


        default:
                /* we should never get here, so this is a really bad error */
                return SNMP_ERR_GENERR;
        }

        return SNMP_ERR_NOERROR;
} 

/**
 * 
 * @return: 
 */
int initialize_table_saHpiWatchdogActiveEntries(void)
{

        DEBUGMSGTL ((AGENT, "initialize_table_saHpiWatchdogActiveEntries, called\n"));

        netsnmp_register_scalar(
                               netsnmp_create_handler_registration(
			                "saHpiWatchdogActiveEntries", 
			                handle_saHpiWatchdogActiveEntries,
                                        saHpiWatchdogActiveEntries_oid, 
			                OID_LENGTH(saHpiWatchdogActiveEntries_oid),
                                        HANDLER_CAN_RONLY ));
			       
	return SNMP_ERR_NOERROR;

}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************
 * keep binary tree to find context by name
 */
static int saHpiWatchdogTable_cmp( const void *lhs, const void *rhs );

/************************************************************
 * compare two context pointers here. Return -1 if lhs < rhs,
 * 0 if lhs == rhs, and 1 if lhs > rhs.
 */
static int
saHpiWatchdogTable_cmp( const void *lhs, const void *rhs )
{
        saHpiWatchdogTable_context *context_l =
        (saHpiWatchdogTable_context *)lhs;
        saHpiWatchdogTable_context *context_r =
        (saHpiWatchdogTable_context *)rhs;

        /*
         * check primary key, then secondary. Add your own code if
         * there are more than 2 indexes
         */
        DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_cmp, called\n"));

        /* check for NULL pointers */
        if (lhs == NULL || rhs == NULL ) {
                DEBUGMSGTL((AGENT,"saHpiWatchdogTable_cmp() NULL pointer ERROR\n" ));
                return 0;
        }
        /* CHECK FIRST INDEX,  saHpiDomainId */
        if ( context_l->index.oids[0] < context_r->index.oids[0])
                return -1;

        if ( context_l->index.oids[0] > context_r->index.oids[0])
                return 1;

        if ( context_l->index.oids[0] == context_r->index.oids[0]) {
                /* If saHpiDomainId index is equal sort by second index */
                /* CHECK SECOND INDEX,  saHpiResourceEntryId */
                if ( context_l->index.oids[1] < context_r->index.oids[1])
                        return -1;

                if ( context_l->index.oids[1] > context_r->index.oids[1])
                        return 1;

                if ( context_l->index.oids[1] == context_r->index.oids[1]) {
                        /* If saHpiResourceEntryId index is equal sort by third index */
                        /* CHECK THIRD INDEX,  saHpiResourceIsHistorical */
                        if ( context_l->index.oids[2] < context_r->index.oids[2])
                                return -1;

                        if ( context_l->index.oids[2] > context_r->index.oids[2])
                                return 1;

                        if ( context_l->index.oids[2] == context_r->index.oids[2]) {
                                /* If saHpiResourceIsHistorical index is equal sort by forth index */
                                /* CHECK FORTH INDEX,  saHpiWatchdogNum */
                                if ( context_l->index.oids[3] < context_r->index.oids[3])
                                        return -1;

                                if ( context_l->index.oids[3] > context_r->index.oids[3])
                                        return 1;

                                if ( context_l->index.oids[3] == context_r->index.oids[3])
                                        return 0;
                        }
                }
        }

        return 0;
}

/************************************************************
 * Initializes the saHpiWatchdogTable module
 */
void
init_saHpiWatchdogTable(void)
{

        DEBUGMSGTL ((AGENT, "init_saHpiWatchdogTable, called\n"));
     
        initialize_table_saHpiWatchdogTable();

        initialize_table_saHpiWatchdogActiveEntries();
}

/************************************************************
 * the *_row_copy routine
 */
static int saHpiWatchdogTable_row_copy(saHpiWatchdogTable_context * dst,
                         saHpiWatchdogTable_context * src)
{
    if(!dst||!src)
        return 1;
        
    subagent_lock(&hpi_lock_data);

    /*
     * copy index, if provided
     */
    if(dst->index.oids)
        free(dst->index.oids);
    if(snmp_clone_mem( (void*)&dst->index.oids, src->index.oids,
                           src->index.len * sizeof(oid) )) {
        dst->index.oids = NULL;
        subagent_unlock(&hpi_lock_data);
        return 1;
    }
    dst->index.len = src->index.len;
    

    /*
     * copy components into the context structure
     */
    /** TODO: add code for external index(s)! */
    dst->saHpiWatchdogNum = src->saHpiWatchdogNum;

    dst->saHpiWatchdogLog = src->saHpiWatchdogLog;

    dst->saHpiWatchdogRunning = src->saHpiWatchdogRunning;

    dst->saHpiWatchdogTimerUseState = src->saHpiWatchdogTimerUseState;

    dst->saHpiWatchdogTimerAction = src->saHpiWatchdogTimerAction;

    dst->saHpiWatchdogPretimerInterrupt = src->saHpiWatchdogPretimerInterrupt;

    dst->saHpiWatchdogPreTimeoutInterval = src->saHpiWatchdogPreTimeoutInterval;

    memcpy( dst->saHpiWatchdogTimerUseExpFlags, src->saHpiWatchdogTimerUseExpFlags, src->saHpiWatchdogTimerUseExpFlags_len );
    dst->saHpiWatchdogTimerUseExpFlags_len = src->saHpiWatchdogTimerUseExpFlags_len;

    dst->saHpiWatchdogTimerInitialCount = src->saHpiWatchdogTimerInitialCount;

    dst->saHpiWatchdogTimerPresentCount = src->saHpiWatchdogTimerPresentCount;

    dst->saHpiWatchdogTimerReset = src->saHpiWatchdogTimerReset;

    dst->saHpiWatchdogOem = src->saHpiWatchdogOem;

    memcpy( src->saHpiWatchdogRDR, dst->saHpiWatchdogRDR, src->saHpiWatchdogRDR_len );
    dst->saHpiWatchdogRDR_len = src->saHpiWatchdogRDR_len;

    subagent_unlock(&hpi_lock_data);
    return 0;
}

/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
saHpiWatchdogTable_extract_index( saHpiWatchdogTable_context * ctx, netsnmp_index * hdr )
{
        /*
         * temporary local storage for extracting oid index
         *
         * extract index uses varbinds (netsnmp_variable_list) to parse
         * the index OID into the individual components for each index part.
         */
        /** TODO: add storage for external index(s)! */
        netsnmp_variable_list var_saHpiDomainId;
        netsnmp_variable_list var_saHpiResourceId;
        netsnmp_variable_list var_saHpiResourceIsHistorical;
        netsnmp_variable_list var_saHpiWatchdogNum;
        int err;

        DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_extract_index, called\n"));

        subagent_lock(&hpi_lock_data);

        /*
         * copy index, if provided
         */
        if (hdr) {
                netsnmp_assert(ctx->index.oids == NULL);
        	if (snmp_clone_mem( (void*)&ctx->index.oids, hdr->oids,
        	         		hdr->len * sizeof(oid) )) {
                        subagent_unlock(&hpi_lock_data);
        		return -1;
        	}
        	ctx->index.len = hdr->len;
        }

        /*
         * initialize variable that will hold each component of the index.
         * If there are multiple indexes for the table, the variable_lists
         * need to be linked together, in order.
         */
        /** TODO: add code for external index(s)! */
        memset( &var_saHpiDomainId, 0x00, sizeof(var_saHpiDomainId) );
        var_saHpiDomainId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiDomainId.next_variable = &var_saHpiResourceId;

        memset( &var_saHpiResourceId, 0x00, sizeof(var_saHpiResourceId) );
        var_saHpiResourceId.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceId.next_variable = &var_saHpiResourceIsHistorical;

        memset( &var_saHpiResourceIsHistorical, 0x00, sizeof(var_saHpiResourceIsHistorical) );
        var_saHpiResourceIsHistorical.type = ASN_INTEGER; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiResourceIsHistorical.next_variable = &var_saHpiWatchdogNum;

        memset( &var_saHpiWatchdogNum, 0x00, sizeof(var_saHpiWatchdogNum) );
        var_saHpiWatchdogNum.type = ASN_UNSIGNED; /* type hint for parse_oid_indexes */
        /** TODO: link this index to the next, or NULL for the last one */
        var_saHpiWatchdogNum.next_variable = NULL;


        /*
         * parse the oid into the individual index components
         */
        err = parse_oid_indexes( hdr->oids, hdr->len, &var_saHpiDomainId );
        if (err == SNMP_ERR_NOERROR) {
        	    /*
        	     * copy index components into the context structure
        	     */
        	    /** skipping external index saHpiDomainId */

        	    /** skipping external index saHpiResourceId */

        	    /** skipping external index saHpiResourceIsHistorical */

        	    ctx->saHpiWatchdogNum = *var_saHpiWatchdogNum.val.integer;


        	    err = saHpiDomainId_check_index(*var_saHpiDomainId.val.integer);
        	    err = saHpiResourceEntryId_check_index(*var_saHpiResourceId.val.integer);  
        	    err = saHpiResourceIsHistorical_check_index(*var_saHpiResourceIsHistorical.val.integer);
        	    err = saHpiWatchdogNum_check_index(*var_saHpiWatchdogNum.val.integer);    
        }

        /*
         * parsing may have allocated memory. free it.
         */
        snmp_reset_var_buffers( &var_saHpiDomainId );

        subagent_unlock(&hpi_lock_data);
        return err;
}

/************************************************************
 * the *_can_activate routine is called
 * when a row is changed to determine if all the values
 * set are consistent with the row's rules for a row status
 * of ACTIVE.
 *
 * return 1 if the row could be ACTIVE
 * return 0 if the row is not ready for the ACTIVE state
 */
int saHpiWatchdogTable_can_activate(saHpiWatchdogTable_context *undo_ctx,
                      saHpiWatchdogTable_context *row_ctx,
                      netsnmp_request_group * rg)
{
    /*
     * TODO: check for activation requirements here
     */


    /*
     * be optimistic.
     */
    return 1;
}

/************************************************************
 * the *_can_deactivate routine is called when a row that is
 * currently ACTIVE is set to a state other than ACTIVE. If
 * there are conditions in which a row should not be allowed
 * to transition out of the ACTIVE state (such as the row being
 * referred to by another row or table), check for them here.
 *
 * return 1 if the row can be set to a non-ACTIVE state
 * return 0 if the row must remain in the ACTIVE state
 */
int saHpiWatchdogTable_can_deactivate(saHpiWatchdogTable_context *undo_ctx,
                        saHpiWatchdogTable_context *row_ctx,
                        netsnmp_request_group * rg)
{
    /*
     * TODO: check for deactivation requirements here
     */
    return 1;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int saHpiWatchdogTable_can_delete(saHpiWatchdogTable_context *undo_ctx,
                    saHpiWatchdogTable_context *row_ctx,
                    netsnmp_request_group * rg)
{
    /*
     * probably shouldn't delete a row that we can't
     * deactivate.
     */
    if(saHpiWatchdogTable_can_deactivate(undo_ctx,row_ctx,rg) != 1)
        return 0;
    
    /*
     * TODO: check for other deletion requirements here
     */
    return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiWatchdogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiWatchdogTable_context *
saHpiWatchdogTable_create_row( netsnmp_index* hdr)
{
    saHpiWatchdogTable_context * ctx =
        SNMP_MALLOC_TYPEDEF(saHpiWatchdogTable_context);
    if(!ctx)
        return NULL;
        
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: check indexes, if necessary.
     */
    if(saHpiWatchdogTable_extract_index( ctx, hdr )) {
        free(ctx->index.oids);
        free(ctx);
        subagent_unlock(&hpi_lock_data);
        return NULL;
    }

    /* netsnmp_mutex_init(ctx->lock);
       netsnmp_mutex_lock(ctx->lock); */

    /*
     * TODO: initialize any default values here. This is also
     * first place you really should allocate any memory for
     * yourself to use.  If you allocated memory earlier,
     * make sure you free it for earlier error cases!
     */
    /**
     ctx->saHpiWatchdogLog = 0;
     ctx->saHpiWatchdogRunning = 0;
     ctx->saHpiWatchdogTimerUseState = 0;
     ctx->saHpiWatchdogTimerAction = 0;
     ctx->saHpiWatchdogPretimerInterrupt = 0;
     ctx->saHpiWatchdogPreTimeoutInterval = 0;
     ctx->saHpiWatchdogTimerUseExpFlags = 0;
     ctx->saHpiWatchdogTimerInitialCount = 0;
     ctx->saHpiWatchdogTimerPresentCount = 0;
     ctx->saHpiWatchdogTimerReset = 0;
    */

    subagent_unlock(&hpi_lock_data);
    return ctx;
}

/************************************************************
 * the *_duplicate row routine
 */
saHpiWatchdogTable_context *
saHpiWatchdogTable_duplicate_row( saHpiWatchdogTable_context * row_ctx)
{
    saHpiWatchdogTable_context * dup;

    if(!row_ctx)
        return NULL;

    subagent_lock(&hpi_lock_data);

    dup = SNMP_MALLOC_TYPEDEF(saHpiWatchdogTable_context);
    if(!dup) {
            subagent_unlock(&hpi_lock_data);
            return NULL;
    }
        
    if(saHpiWatchdogTable_row_copy(dup,row_ctx)) {
        free(dup);
        dup = NULL;
    }

    subagent_unlock(&hpi_lock_data);
    return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index * saHpiWatchdogTable_delete_row( saHpiWatchdogTable_context * ctx )
{
  /* netsnmp_mutex_destroy(ctx->lock); */

        subagent_lock(&hpi_lock_data);

    if(ctx->index.oids)
        free(ctx->index.oids);

    /*
     * TODO: release any memory you allocated here...
     */

    /*
     * release header
     */
    free( ctx );

    subagent_unlock(&hpi_lock_data);
    return NULL;
}

/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void saHpiWatchdogTable_set_reserve1( netsnmp_request_group *rg )
{
    saHpiWatchdogTable_context *row_ctx =
            (saHpiWatchdogTable_context *)rg->existing_row;
//    saHpiWatchdogTable_context *undo_ctx =
//            (saHpiWatchdogTable_context *)rg->undo_info;
    netsnmp_variable_list *var;
    netsnmp_request_group_item *current;
    int rc;

    SaHpiTextBufferT buffer;
    SaHpiWatchdogExpFlagsT exp_flags = 0;

    /*
     * TODO: loop through columns, check syntax and lengths. For
     * columns which have no dependencies, you could also move
     * the value/range checking here to attempt to catch error
     * cases as early as possible.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiWatchdogLog));
        break;

        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiWatchdogRunning));
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiWatchdogTimerUseState));
            if (rc == SNMP_ERR_NOERROR) {
                    if(!oh_lookup_watchdogtimeruse(row_ctx->saHpiWatchdogTimerUseState - 1))
                            rc = SNMP_ERR_BADVALUE;
            }
        break;

        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiWatchdogTimerAction));
            if (rc == SNMP_ERR_NOERROR) {
                    if (!oh_lookup_watchdogactionevent(row_ctx->saHpiWatchdogTimerAction - 1))
                            rc = SNMP_ERR_BADVALUE;
            }
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiWatchdogPretimerInterrupt));
            if (rc == SNMP_ERR_NOERROR) {
                    if (!oh_lookup_watchdogpretimerinterrupt(row_ctx->saHpiWatchdogPretimerInterrupt - 1))
                            rc = SNMP_ERR_BADVALUE;
            }
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                                                sizeof(row_ctx->saHpiWatchdogPreTimeoutInterval));
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
            rc = netsnmp_check_vb_type(var, ASN_OCTET_STR);                 
                if (rc == SNMP_ERR_NOERROR ) {
                        if (var->val_len > sizeof(row_ctx->saHpiWatchdogTimerUseExpFlags)) {
                                rc = SNMP_ERR_WRONGLENGTH;
			}
		}

            if (rc == SNMP_ERR_NOERROR) {
                    buffer.DataLength = var->val_len;
                    memcpy(buffer.Data, var->val.string, var->val_len);
                    rc = oh_encode_exp_flags(&buffer, &exp_flags);
            }

        break;

        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                                                sizeof(row_ctx->saHpiWatchdogTimerInitialCount));
        break;

        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            rc = netsnmp_check_vb_type_and_size(var, ASN_UNSIGNED,
                                                sizeof(row_ctx->saHpiWatchdogTimerPresentCount));
        break;

        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
            rc = netsnmp_check_vb_type_and_size(var, ASN_INTEGER,
                                                sizeof(row_ctx->saHpiWatchdogTimerReset));
        break;

        default: /** We shouldn't get here */
            rc = SNMP_ERR_GENERR;
            snmp_log(LOG_ERR, "unknown column in "
                     "saHpiWatchdogTable_set_reserve1\n");
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc );
        rg->status = SNMP_MAX( rg->status, current->ri->status );
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

void saHpiWatchdogTable_set_reserve2( netsnmp_request_group *rg )
{
//    saHpiWatchdogTable_context *row_ctx = (saHpiWatchdogTable_context *)rg->existing_row;
//    saHpiWatchdogTable_context *undo_ctx = (saHpiWatchdogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;
    netsnmp_variable_list *var;
    int rc;

    rg->rg_void = rg->list->ri;

    /*
     * TODO: loop through columns, check for valid
     * values and any range constraints.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;
        rc = SNMP_ERR_NOERROR;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( XXX_check_value( var->val.string, XXX ) ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
                    /*
                     * TODO: routine to check valid values
                     *
                     * EXAMPLE:
                     *
                    * if ( *var->val.integer != XXX ) {
                *    rc = SNMP_ERR_INCONSISTENTVALUE;
                *    rc = SNMP_ERR_BADVALUE;
                * }
                */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
                rc = netsnmp_check_vb_truthvalue(current->ri->requestvb);
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }

        if (rc)
           netsnmp_request_set_error( current->ri, rc);
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void saHpiWatchdogTable_set_action( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
    saHpiWatchdogTable_context *row_ctx = (saHpiWatchdogTable_context *)rg->existing_row;
//    saHpiWatchdogTable_context *undo_ctx = (saHpiWatchdogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    int row_err = 0;
    subagent_lock(&hpi_lock_data);

    /*
     * TODO: loop through columns, copy varbind values
     * to context structure for the row.
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiWatchdogLog = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
            row_ctx->saHpiWatchdogRunning = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
            row_ctx->saHpiWatchdogTimerUseState = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
            row_ctx->saHpiWatchdogTimerAction = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
            row_ctx->saHpiWatchdogPretimerInterrupt = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
            row_ctx->saHpiWatchdogPreTimeoutInterval = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
            memcpy(row_ctx->saHpiWatchdogTimerUseExpFlags,var->val.string,var->val_len);
            row_ctx->saHpiWatchdogTimerUseExpFlags_len = var->val_len;
            row_err = set_table_watchdog_set (row_ctx); 

        break;

        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            row_ctx->saHpiWatchdogTimerInitialCount = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            row_ctx->saHpiWatchdogTimerPresentCount = *var->val.integer;
            row_err = set_table_watchdog_set (row_ctx); 
        break;

        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
            //row_ctx->saHpiWatchdogTimerReset = *var->val.integer;
            row_err = set_table_watchdog_reset (row_ctx); 
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    if(row_err) {
        netsnmp_request_set_error((netsnmp_request_info*)rg->rg_void,
                                       row_err);
    }

    subagent_unlock(&hpi_lock_data);
    return;
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void saHpiWatchdogTable_set_commit( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiWatchdogTable_context *row_ctx = (saHpiWatchdogTable_context *)rg->existing_row;
//    saHpiWatchdogTable_context *undo_ctx = (saHpiWatchdogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void saHpiWatchdogTable_set_free( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiWatchdogTable_context *row_ctx = (saHpiWatchdogTable_context *)rg->existing_row;
//    saHpiWatchdogTable_context *undo_ctx = (saHpiWatchdogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
        break;

        default: 
                break;
                /** We shouldn't get here */
                /** should have been logged in reserve1 */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void saHpiWatchdogTable_set_undo( netsnmp_request_group *rg )
{
    netsnmp_variable_list *var;
//    saHpiWatchdogTable_context *row_ctx = (saHpiWatchdogTable_context *)rg->existing_row;
//    saHpiWatchdogTable_context *undo_ctx = (saHpiWatchdogTable_context *)rg->undo_info;
    netsnmp_request_group_item *current;

    /*
     * loop through columns
     */
    for( current = rg->list; current; current = current->next ) {

        var = current->ri->requestvb;

        switch(current->tri->colnum) {

        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
        break;

        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
        break;

        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
        break;

        default: /** We shouldn't get here */
            netsnmp_assert(0); /** why wasn't this caught in reserve1? */
        }
    }

    /*
     * done with all the columns. Could check row related
     * requirements here.
     */
}

/************************************************************
 *
 * Initialize the saHpiWatchdogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiWatchdogTable(void)
{
    netsnmp_table_registration_info *table_info;

    if(my_handler) {
        snmp_log(LOG_ERR, "initialize_table_saHpiWatchdogTable_handler called again\n");
        return;
    }

    memset(&cb, 0x00, sizeof(cb));

    /** create the table structure itself */
    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);

    /* if your table is read only, it's easiest to change the
       HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY */
    my_handler = netsnmp_create_handler_registration("saHpiWatchdogTable",
                                             netsnmp_table_array_helper_handler,
                                             saHpiWatchdogTable_oid,
                                             saHpiWatchdogTable_oid_len,
                                             HANDLER_CAN_RWRITE);
            
    if (!my_handler || !table_info) {
        snmp_log(LOG_ERR, "malloc failed in "
                 "initialize_table_saHpiWatchdogTable_handler\n");
        return; /** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */
    /*
     * TODO: add any external indexes here.
     */
        /** TODO: add code for external index(s)! */

    /*
     * internal indexes
     */
        /** index: saHpiDomainId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceId */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);
        /** index: saHpiResourceIsHistorical */
        netsnmp_table_helper_add_index(table_info, ASN_INTEGER);
        /** index: saHpiWatchdogNum */
        netsnmp_table_helper_add_index(table_info, ASN_UNSIGNED);

    table_info->min_column = saHpiWatchdogTable_COL_MIN;
    table_info->max_column = saHpiWatchdogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
    cb.get_value = saHpiWatchdogTable_get_value;
    cb.container = netsnmp_container_find("saHpiWatchdogTable_primary:"
                                          "saHpiWatchdogTable:"
                                          "table_container");

    netsnmp_container_add_index(cb.container,
                                netsnmp_container_find("saHpiWatchdogTable_secondary:"
                                                       "saHpiWatchdogTable:"
                                                       "table_container"));
    cb.container->next->compare = saHpiWatchdogTable_cmp;


    cb.can_set = 1;

    cb.create_row = (UserRowMethod*)saHpiWatchdogTable_create_row;

    cb.duplicate_row = (UserRowMethod*)saHpiWatchdogTable_duplicate_row;
    cb.delete_row = (UserRowMethod*)saHpiWatchdogTable_delete_row;
    cb.row_copy = (Netsnmp_User_Row_Operation *)saHpiWatchdogTable_row_copy;

    cb.can_activate = (Netsnmp_User_Row_Action *)saHpiWatchdogTable_can_activate;
    cb.can_deactivate = (Netsnmp_User_Row_Action *)saHpiWatchdogTable_can_deactivate;
    cb.can_delete = (Netsnmp_User_Row_Action *)saHpiWatchdogTable_can_delete;

    cb.set_reserve1 = saHpiWatchdogTable_set_reserve1;
    cb.set_reserve2 = saHpiWatchdogTable_set_reserve2;
    cb.set_action = saHpiWatchdogTable_set_action;
    cb.set_commit = saHpiWatchdogTable_set_commit;
    cb.set_free = saHpiWatchdogTable_set_free;
    cb.set_undo = saHpiWatchdogTable_set_undo;

    DEBUGMSGTL(("initialize_table_saHpiWatchdogTable",
                "Registering table saHpiWatchdogTable "
                "as a table array\n"));
    netsnmp_table_container_register(my_handler, table_info, &cb,
                                     cb.container, 1);
}

/************************************************************
 * saHpiWatchdogTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int saHpiWatchdogTable_get_value(
            netsnmp_request_info *request,
            netsnmp_index *item,
            netsnmp_table_request_info *table_info )
{
    netsnmp_variable_list *var = request->requestvb;
    saHpiWatchdogTable_context *context = (saHpiWatchdogTable_context *)item;

    switch(table_info->colnum) {

        case COLUMN_SAHPIWATCHDOGNUM:
            /** SaHpiInstrumentId = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiWatchdogNum,
                         sizeof(context->saHpiWatchdogNum) );
        break;
    
        case COLUMN_SAHPIWATCHDOGLOG:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiWatchdogLog,
                         sizeof(context->saHpiWatchdogLog) );
        break;
    
        case COLUMN_SAHPIWATCHDOGRUNNING:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiWatchdogRunning,
                         sizeof(context->saHpiWatchdogRunning) );
        break;
    
        case COLUMN_SAHPIWATCHDOGTIMERUSESTATE:
            /** SaHpiWatchdogTimerUse = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiWatchdogTimerUseState,
                         sizeof(context->saHpiWatchdogTimerUseState) );
        break;
    
        case COLUMN_SAHPIWATCHDOGTIMERACTION:
            /** INTEGER = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiWatchdogTimerAction,
                         sizeof(context->saHpiWatchdogTimerAction) );
        break;
    
        case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
            /** SaHpiWatchdogPreTimerAction = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiWatchdogPretimerInterrupt,
                         sizeof(context->saHpiWatchdogPretimerInterrupt) );
        break;
    
        case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiWatchdogPreTimeoutInterval,
                         sizeof(context->saHpiWatchdogPreTimeoutInterval) );
        break;
    
        case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
            /** OCTETSTR = ASN_OCTET_STR */
            snmp_set_var_typed_value(var, ASN_OCTET_STR,
                         (u_char*)&context->saHpiWatchdogTimerUseExpFlags,
                         context->saHpiWatchdogTimerUseExpFlags_len );
        break;
    
        case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiWatchdogTimerInitialCount,
                         sizeof(context->saHpiWatchdogTimerInitialCount) );
        break;
    
        case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiWatchdogTimerPresentCount,
                         sizeof(context->saHpiWatchdogTimerPresentCount) );
        break;
    
        case COLUMN_SAHPIWATCHDOGTIMERRESET:
            /** TruthValue = ASN_INTEGER */
            snmp_set_var_typed_value(var, ASN_INTEGER,
                         (u_char*)&context->saHpiWatchdogTimerReset,
                         sizeof(context->saHpiWatchdogTimerReset) );
        break;
    
        case COLUMN_SAHPIWATCHDOGOem:
            /** UNSIGNED32 = ASN_UNSIGNED */
            snmp_set_var_typed_value(var, ASN_UNSIGNED,
                         (u_char*)&context->saHpiWatchdogOem,
                         sizeof(context->saHpiWatchdogOem) );
        break;
    
        case COLUMN_SAHPIWATCHDOGRDR:
            /** RowPointer = ASN_OBJECT_ID */
            snmp_set_var_typed_value(var, ASN_OBJECT_ID,
                         (u_char*)&context->saHpiWatchdogRDR,
                         context->saHpiWatchdogRDR_len );
        break;
    
    default: /** We shouldn't get here */
        snmp_log(LOG_ERR, "unknown column in "
                 "saHpiWatchdogTable_get_value\n");
        return SNMP_ERR_GENERR;
    }
    return SNMP_ERR_NOERROR;
}

/************************************************************
 * saHpiWatchdogTable_get_by_idx
 */
const saHpiWatchdogTable_context *
saHpiWatchdogTable_get_by_idx(netsnmp_index * hdr)
{
    return (const saHpiWatchdogTable_context *)
        CONTAINER_FIND(cb.container, hdr );
}


